/*********************************************************************

	Color File.c
	
	This file contains functions for file I/O handling in the
	"Adventures in Color Printing" color/grayscale sample,
	"Color Adventures."
	
	Dave Hersey, Apple Developer Technical Support.
	
	 Edit Trail 
	
	creation/reorg/rebirth:						 6/28/92  - dmh
	sync'd with other samples:					 11/7/92  - dmh
	fixed SpoolInstall so that it
	works with the printer port:				 11/8/92  - dmh
	final clean up and sync with other samples:	11/28/92  - dmh
	final-final clean up.  Added error recovery
	to spoolPicProc so it doesn't read forever
	on a file error:							11/29/92  - dmh
	removed unnec. parameter passed to
	TearDownCursor call that the compiler
	didn't pick up on:							12/30/92  - dmh
	
*********************************************************************/

#pragma segment File

#include "Color Headers.h"

static OSErr			gError;					// Error variable for SpoolPicProc.
static short			gSpoolRefNum;			// RefNum to PICT file we're spooling in.
static CQDProcs			gCQDBottlenecks;		// Our CQD and QD bottleneck procs for
static QDProcs			gQDBottlenecks;			// spooling.
static Ptr				gSavedBottlenecks;		// Ptr to bottlenecks we replaced.
static Boolean			gSpoolFirstTimeThru;	// True first time through spool rtn.
static long				gSpoolCacheSize;		// Size of cache we use during spooling.
static Handle			gSpoolCache = nil;		// Handle to cache-buffer for spooling.
static long				gSpoolCacheIdx;			// Current byte offset in cache-buffer.

static Boolean			gUseColorCursors;		// If true, spinning cursors are
												// crsrs, not CURSs.
static ACurHdl			gACur;					// Spinning cursors' 'acur' resource.
static Handle			gCursor[kMaxNumCursors];// Handles of cursors to spin.
static short			gFinalCursorID;			// Cursor ID to use when done spinning.
static unsigned long	gTicksLastCursor;		// TickCount when we last spun cursors.


/************************************************************
  SetUpCursors - This routine does initialization for the
  animated cursor routines BumpCursor and TearDownCursors.
  
  The caller passes in the resource ID of an 'acur' reource,
  and this routine loads the related cursors.  Finally,
  it calls BumpCursor to start using the first cursor.

  We attempt to load 'crsr' resources if they're available.
  If not, we use 'CURS' ones.  That's so we can use color
  cursors if they're around.

  finalCursorID is the ID of the cursor to use when we're
  all done (in TearDownCursors).  It may be:
  
  	 kArrowCursor - for the arrow cursor.

	 or a valid ID of a 'crsr' or 'CURS' resource.
	  (TearDownCursors tries for a 'crsr' first.)

*************************************************************/

void SetUpCursors(short acurID, short finalCursorID)
{
	short	oldRes, cu;
	
// Save the current resource file, then switch to our
// application's, get the requested 'acur' resource and store
// the resource ID of the cursor we should use in
// TearDownCursors.

	oldRes = CurResFile();
	UseResFile(gAppResRefNum);
	
	gACur = (ACurHdl) GetResource('acur', acurID);
	gFinalCursorID = finalCursorID;


// If we loaded the resource, detach it and get, load and
// detach each cursor it points to.  We store the cursors
// in global handles so that we can easily access them.  If
// the acur resource references more cursors than we've
// allocated space for, then we just use as many as we can
// handle.  Note that we try to load 'crsr' resources first,
// and if that fails we try 'CURS' resoruces.  That's so we
// can use color cursors if they're available.
	
	if (gACur != nil)
	{
		gUseColorCursors = true;
		
		DetachResource((Handle) gACur);
		(*gACur)->numCursors = MIN((*gACur)->numCursors, kMaxNumCursors);

		for (cu = 0; cu < (*gACur)->numCursors; cu++)
		{
			if (gUseColorCursors)
			{
				gCursor[cu] = (Handle) GetCCursor((*gACur)->cursID[cu] >>16);

				if ((cu == 0) && (gCursor[cu] == nil))
					gUseColorCursors = false;
			}
			
			if (!gUseColorCursors)
			{
				gCursor[cu] = (Handle) GetCursor((*gACur)->cursID[cu] >>16);


// We need to lock the non-color cursors because we have to dereference their
// handle for SetCursor.

				if (gCursor[cu])
				{
					MoveHHi(gCursor[cu]);
					HLock(gCursor[cu]);
				}
			}
			

			if (gCursor[cu] != nil)
				DetachResource((Handle) gCursor[cu]);
		}


// Set the "next frame" counter and the "tick count when
// we last changed cursors" to 0.  That's so that we'll
// immediately display the first cursor, which we do by
// calling BumpCursor.  Finally, restore the old
// resource file.

		(*gACur)->nextFrame = gTicksLastCursor = 0;
		BumpCursor();
	}

	UseResFile(oldRes);
}


/************************************************************
  BumpCursor - This routine displays the next cursor in our
  animated cursor list (from the 'acur' resource whose ID
  we passed to SetUpCursors.

*************************************************************/

void BumpCursor()
{
	register short			curFrame;
	register unsigned long	numTicks;

// If there's an 'acur' resource loaded and we're in the foreground,
// see if we should change cursors.  If so, (if the time that's passed
// since the last change is at least kMinCursorTicks) store the new tick
// count, grab the next cursor to use and bump the frame counter.

	if (gACur != nil && !gInBackground)
	{
		numTicks = TickCount();
		if ((numTicks - gTicksLastCursor) < kMinCursorTicks)
			return;
	
		gTicksLastCursor = numTicks;
		curFrame = (*gACur)->nextFrame;
		(*gACur)->nextFrame = ++(*gACur)->nextFrame % (*gACur)->numCursors;


// Call the appropriate SetCursor" call, based on whether we have color
// or black and white cursors loaded.  Non-color cursors will be locked.

		if (gCursor[curFrame] != nil)
		{
			if (gUseColorCursors)
				SetCCursor((CCrsrHandle) gCursor[curFrame]);
			else
				SetCursor(*(CursHandle) gCursor[curFrame]);
		}
	}
}


/************************************************************
  TearDownCursors - This routine undoes what SetUpCursors
  did.  It disposes of any loaded handles and sets the
  cursor to the cursor having the ID stored in
  gFinalCursorID, which is where SetUpCursors stores the
  final cursor's resource ID.

*************************************************************/

void TearDownCursors()
{
	short		numCursors, cu;
	Handle		newCursor;

// If we have cursors loaded, go through and dispose of them all, and
// their accompanying 'acur' resource.  Note that these aren't really
// resources, since we called DetachResource on them earlier.

	if (gACur != nil)
	{
		numCursors = (*gACur)->numCursors;

		for (cu = 0; cu < numCursors; cu++)
		{
			if (gCursor[cu]) DisposHandle((Handle) gCursor[cu]);
			gCursor[cu] = nil;
		}
		
		DisposHandle((Handle) gACur);
		gACur = nil;
	}

// If we didn't get switched into the background, set the cursor to
// whatever is in gFinalCursor.  If it's kArrowCursor, we use the
// global arrow cursor (which doesn't have an ID).  We look for a
// 'crsr' first, and if that's not found, try to load a 'CURS'.

	if (gFinalCursorID == kArrowCursor && !gInBackground)
		SetCursor(&qd.arrow);
	else
	{
		newCursor = (Handle) GetCCursor(gFinalCursorID);
	
		if (newCursor != nil && !gInBackground)
			SetCCursor((CCrsrHandle) newCursor);
		else
		{
			newCursor = (Handle) GetCursor(gFinalCursorID);
	
			if (newCursor != nil && !gInBackground)
				SetCursor(*(CursHandle) newCursor);
		}
	}
}


/************************************************************
  SpoolInstall - This routine installs a new getPicProc
  bottleneck routine in the current port which will spool
  an image from disk.
  
  The WindowInfoHdl passed in is our private structure
  which tells us which PICT file to spool in.
  
  The cursorToUseWhenDone value is a resource ID of the
  cursor to use when spooling is completed.  Normally, this
  will be either watchCursor or kArrowCursor.  This value
  is passed to SetUpCursors, so that TearDownCursors can
  find it.

*************************************************************/

OSErr SpoolInstall(WindowInfoHdl windInfo, short cursorToUseWhenDone)
{
	GrafPtr		curPort;
	OSErr		err = noErr;

// Set our global error value to noErr, save the final cursor ID,
// lock down our private window info and open the file to spool.

	gError = noErr;
	HLock((Handle) windInfo);
	
	if (gHaveNewSFCalls)
		err = FSpOpenDF(&(*windInfo)->oFSSpec, fsCurPerm, &gSpoolRefNum);
	else
		err = FSOpen(&(*windInfo)->reply.fName, (*windInfo)->reply.vRefNum,
					 &gSpoolRefNum);

	HUnlock((Handle) windInfo);
	if (err) return err;


// Position 512 bytes plus the 72 dpi bounds and obsolete size parameter
// into the file.  This is because we already have the size and bounds
// read into out PicHandle from LoadAPICT.  Create a spool-data cache
// that's 1/2 the maximum block size available (this is a good amount and
// leaves some memory available for others).  Also, set up our animated
// spooling cursors and get the current port.

	err = SetFPos(gSpoolRefNum, fsFromStart, 512 + sizeof(Picture));

	gSpoolFirstTimeThru = true;					// This must be set before 1st call to
												// SpoolPicProc if we're using a cache.

	gSpoolCacheSize = MaxBlock() /2;				// Allocate a handle this size that we
	gSpoolCache = NewHandle(gSpoolCacheSize);		// can use as a cache for SpoolPicProc.

	SetUpCursors(r_spoolSpinCursor, cursorToUseWhenDone);
	GetPort(&curPort);


// Install the bottleneck routines.  If we are using a CGrafPort, we
// must use CQDProcs, otherwise we install QDProcs.  We save the
// old bottlenecks, if any, so that we can restore them when we're
// done.

	if (curPort->portBits.rowBytes < 0)	// CGrafPort
	{
		gSavedBottlenecks = (Ptr) ((CGrafPtr) curPort)->grafProcs;
		
		if (gSavedBottlenecks != nil)
			BlockMove(gSavedBottlenecks, &gCQDBottlenecks, sizeof(CQDProcs));
		else
			SetStdCProcs(&gCQDBottlenecks);

		gCQDBottlenecks.getPicProc = (Ptr) SpoolPicProc;
		((CGrafPtr) curPort)->grafProcs = &gCQDBottlenecks;
	}
	else								// GrafPort
	{
		gSavedBottlenecks = (Ptr) curPort->grafProcs;
		
		if (gSavedBottlenecks != nil)
			BlockMove(gSavedBottlenecks, &gQDBottlenecks, sizeof(QDProcs));
		else
			SetStdProcs(&gQDBottlenecks);

		gQDBottlenecks.getPicProc = (Ptr) SpoolPicProc;
		curPort->grafProcs = &gQDBottlenecks;
	}

// If there was an error, undo what we just did.  Return any errors.

	if (err) SpoolRemove();

	return err;
}


/************************************************************
  SpoolRemove - This routine removes what SpoolInstall has
  installed.

*************************************************************/

OSErr SpoolRemove()
{
	GrafPtr		curPort;

// Get the current port, restore the original QuickDraw bottlenecks, close
// the spool file, dispose of the spool-data cache if we created one, and
// tear down the cursors.  Finally, return the global error, if any, from
// spooling.

	GetPort(&curPort);

	if (curPort->portBits.rowBytes < 0)	// CGrafPort
		((CGrafPtr) curPort)->grafProcs = (CQDProcsPtr) gSavedBottlenecks;
	else								// GrafPort
		curPort->grafProcs = (QDProcsPtr) gSavedBottlenecks;

	if (gSpoolRefNum) FSClose(gSpoolRefNum);

	if (gSpoolCache)
	{
		DisposHandle(gSpoolCache);
		gSpoolCache = nil;
	}

	TearDownCursors();

	return gError;
}


/************************************************************
  SpoolPicProc - This routine is our bottleneck replacement
  for getPicProc.
  
  This routine must be guaranteed to be around when it gets
  called through the bottlenecks.  Therefore, I've put it
  in the Main segment, which is always around.

  NOTE: In the descriptions and code comments below, "cache"
  refers to the application created cache for this routine.
  It's just a handle that we pre-fetch data into.  Don't be
  confused into thinking that this cache has anything to do
  with other hardware or software caches on the Macintosh.

  If gSpoolCache is not nil, we use that handle as a cache.
  We read in a full buffer of data, and then feed that to
  the caller whenever they ask for more data.  When we reach
  the end of the cache, we read in another full buffer of
  data and continue.
  
  If gSpoolCache is nil, we just read the amount of data
  that's requested directly into the passed buffer.
  
  You'll normally want a cache buffer, since the difference
  in speed is significant.  Typically, this routine will be
  called with many requests to read in a byte or two, or a
  few K worth of data.  With a cache you'll do far fewer
  disk reads, doing memory accesses instead.  If you're
  curious about the speed difference, set gSpoolCache to
  nil before installing this bottleneck replacement.  You
  can do this in SpoolInstall where the cache handle is
  being created.

*************************************************************/

#pragma segment Main

pascal void SpoolPicProc(Ptr buffer, short amtToRead)
{
	Ptr		killBuffer, srcPtr, destPtr;
	long	count, amtToMove, amtMoved;
	OSErr	err;

// If we already have a previous error, drop down to our error handling code.
// Otherwise, see if we're using a cache.

	if (!(err = gError))
	{
		if (gSpoolCache != nil)					// Using data cache.
		{

// Using a cache.  Lock down the handle so we can reference it by pointer.  Set
// our initial source pointer to the cache plus the current offset into it.  Set
// the initial destination pointer to the passed buffer, and indicate that we
// haven't moved any bytes yet.

			HLock(gSpoolCache);
	
			srcPtr = *gSpoolCache +gSpoolCacheIdx;
			destPtr = buffer;
			amtMoved = 0;


// While we haven't moved all the data we need to, bump our spinning cursor, and
// check if this is our first time in the routine.  If so, we skip over the data
// moving code, because we need to fill the cache first.  If this isn't the first
// time through (or we've filled the cache), move as much data as we can to the
// passed buffer.  This amount will be either the amount of data left to move or
// the amount of data remaining in our cache.  After the move, update our index
// into the cache, the amount of data we've moved, and the destination pointer.

			while (!err && (amtMoved != amtToRead))
			{
				BumpCursor();

				if (!gSpoolFirstTimeThru)
				{
					amtToMove = MIN(amtToRead -amtMoved, gSpoolCacheSize -gSpoolCacheIdx);
					BlockMove(srcPtr, destPtr, amtToMove);
					gSpoolCacheIdx += amtToMove;
					amtMoved += amtToMove;
					destPtr += amtToMove;
				}
				else
					gSpoolFirstTimeThru = false;


// If we haven't read all we need to for this call, or if the cache index now
// points to one byte beyond the end of the cache, we need to refill the cache
// from disk.  Read another buffer-full and reset the cache index to 0.  It's
// possible we may end up with an "end of file" error because we're trying to
// fill the cache without worrying about how large the actual file is.  Therefore,
// if we generate an "end of file" error, we clear it.  After reading, go back
// up and move more data if we need to.  When done with all the moving, unlock
// the handle to the cache and set gError to our current error value.

				if ((amtMoved != amtToRead) || (gSpoolCacheIdx == gSpoolCacheSize))
				{
					gSpoolCacheIdx = 0;
					srcPtr = *gSpoolCache;
					count = gSpoolCacheSize;
					err = FSRead(gSpoolRefNum, &count, srcPtr);
					if (err == eofErr) err = noErr;
				}
			}
	
			HUnlock(gSpoolCache);
			gError = err;
		}
		else		// Not using cache.
		{

// If we're not using a cache, things are much simpler (but slower, since we'll
// typically be asked to read very little on every pass through this bottleneck
// routine).  Just bump the spinning cursor and read in the requested amount of
// data.

			BumpCursor();
			count = amtToRead;
			gError = FSRead(gSpoolRefNum, &count, buffer);
		}
	}

// If an error occurred, fill the caller's buffer with FFs.  The reason we do
// this is so that the next time the caller looks for a pict opcode, he'll
// find FF, which is the end-of-picture opcode.  Therefore, we may be able to
// fake the caller into stopping early, rather than asking us to continue
// reading, which we can't do anyway (because we have an error).

	if (gError)
		for (killBuffer = buffer; killBuffer != ((long) buffer + count); killBuffer++)
			*killBuffer = 0xFF;
}

#pragma segment File


/************************************************************
  LoadAPICT - This routine loads all or part of a picture
  from a PICT file.  If the picture will use up more than
  kMaxPICTHeapPrcnt of the maximum contiguous block in the
  heap, the picture will be spooled in (in DoUpdate), rather
  than loaded all at once.  In that case, the only thing
  that will be in the returned PicHandle will be the
  picture's 72 dpi bounds and its (obsolete) size field.
  
  Parameters:
  
  WindowInfoHdl windInfo
  
  A handle (with space allocated for it) that points to
  the window information record for this picture.  The
  file information fields will be initialized, and we
  use those to locate the file to load.  The picture
  will be loaded into the windowPict field of this
  record, and all other fields will be initialized
  after successful execution of this routine.
  
  Rect *pictRect

  pictRect is passed in with the maximum area the picture
  can occupy.  It corresponds to the size of the largest
  window which will fit on the screen.  Upon return, it
  will be set to the size the window should be created at.
  This rectangle will be a proportionally scaled version of
  the image's picFrame, so that the aspect ratio is
  preserved.  

*************************************************************/

OSErr LoadAPICT(WindowInfoHdl windInfo, Rect *pictRect)
{
	OSErr				err;
	Ptr					buff;
	long				count;
	Rect				drawRect;
	Fixed				scale;
	short				pictRefNum;
	FSSpecPtr			opFSSpec;
	SFReply				*reply;

// Lock down our private window information handle, initialize
// a few fields in it and store pointers to others.

	MoveHHi((Handle) windInfo);
	HLock((Handle) windInfo);

	(*windInfo)->loaded = true;
	(*windInfo)->windowPict = nil;

	opFSSpec = &(*windInfo)->oFSSpec;
	reply = &(*windInfo)->reply;

	
// Open the requested file.

	if (gHaveNewSFCalls)
		err = FSpOpenDF(opFSSpec,fsCurPerm, &pictRefNum);
	else
		err = FSOpen(reply->fName, reply->vRefNum, &pictRefNum);

	if (err)
	{
		HUnlock((Handle) windInfo);
		return err;
	}


// Move past the 512K byte header and create a PicHandle big enough
// to hold the picture's bounds and size parameters.

	err = SetFPos(pictRefNum, fsFromStart, 512);

	(*windInfo)->windowPict = (PicHandle) NewHandle(sizeof(Picture));
	err = MemError();
	
	if (err)
	{
		HUnlock((Handle) windInfo);
		FSClose(pictRefNum);
		(*windInfo)->windowPict = nil;
		return err;
	}

// Lock down the handle, read in the picture's bounds and size,
// and close the file.

	HLock((Handle) (*windInfo)->windowPict);
	count = sizeof(Picture);
	buff = (Ptr) *(*windInfo)->windowPict;
	err = FSRead(pictRefNum, &count, buff);
	HUnlock((Handle) (*windInfo)->windowPict);
	

// Now locate the picture's version number.

	if (!err)
	{
		unsigned char	tempChar = 0;
		
		count = sizeof(unsigned char);
		buff = (Ptr) &tempChar;


// Search for version picture opCode.

		while (!err && (tempChar != 0x11))
			err = FSRead(pictRefNum, &count, buff);


// Found it.  Now read in version.

		if (!err)
		{
			count = sizeof(short);
			buff = (Ptr) &(*windInfo)->partialPict.pictVersion;
			err = FSRead(pictRefNum, &count, buff);
			(*windInfo)->partialPict.pictVersion >>= 8;
		}
	

// If this is a version 2 picture, now search for the extended version 2
// picture header.

		if (!err && ((*windInfo)->partialPict.pictVersion == 2))
		{
			short	tempShort = 0;
		
			count = sizeof(short);
			buff = (Ptr) &tempShort;


// Search for picture header opCode.

			while (!err && (tempShort != 0x0C00))
				err = FSRead(pictRefNum, &count, buff);


// Found it.  Now read in header.

			if (!err)
			{
				count = sizeof(PictureHeader);
				buff = (Ptr) &(*windInfo)->partialPict.pictHeader;
				err = FSRead(pictRefNum, &count, buff);
			}
		}
	}

// If we had any errors, unlock the window info handle, kill our picture,
// close the PICT file and return the error.

	if (err)
	{
		HUnlock((Handle) windInfo);
		KillPicture((*windInfo)->windowPict);
		(*windInfo)->windowPict = nil;
		FSClose(pictRefNum);
		return err;
	}

// Align the passed rectangle and the picture's picFrame, and see
// if we need to scale the picture to fit it within the available
// bounds.  If so, do it and set pictRect to the actual bounds
// we'll use.

	drawRect = (*(*windInfo)->windowPict)->picFrame;

	OffsetRect(&drawRect, -drawRect.left, -drawRect.top);
	OffsetRect(pictRect, -pictRect->left, -pictRect->top);

	if ((drawRect.right > pictRect->right) || (drawRect.bottom > pictRect->bottom))
	{
		if ((drawRect.right - pictRect->right) > (drawRect.bottom - pictRect->bottom))
			scale = FixRatio(pictRect->right, drawRect.right);
		else
			scale = FixRatio(pictRect->bottom, drawRect.bottom);
	
		drawRect.right = FixMul(scale, drawRect.right <<16) >>16;
		drawRect.bottom = FixMul(scale, drawRect.bottom <<16) >>16;
	}

	*pictRect = drawRect;


// Find the end of the file and subtract 512 bytes in order to calculate
// how large a PicHandle we'd need to read in the whole thing.  If this
// amount is more than kMaxPICTHeapPrcnt * the largest free block available,
// we'll spool the image.  This spool/no-spool decision making isn't fool-
// proof, but it works well for the most part.

	err = GetEOF(pictRefNum, &count);
	count -= 512;

	if (!err)
	{
		if (count > (kMaxPICTHeapPrcnt * MaxBlock()))
			(*windInfo)->loaded = false;	// we spool.
		else								// we don't spool.
		{

// If we're not spooling, resize the picture's handle and read in
// the entire picture.  The SetHandleSize below should never failed
// because we just checked for the available memory.  We're only
// reading in the picture data after the size and picFrame, because
// that's already in our PicHandle.

			SetHandleSize((Handle) (*windInfo)->windowPict, count);
			HLock((Handle) (*windInfo)->windowPict);
			buff = (Ptr) *(*windInfo)->windowPict + (Ptr) sizeof(Picture);
			count -= sizeof(Picture);

			err = SetFPos(pictRefNum, fsFromStart, 512 + sizeof(Picture));
			if (!err) err = FSRead(pictRefNum, &count, buff);
			HUnlock((Handle) (*windInfo)->windowPict);
		}
	}
	
	if (err)
	{
		KillPicture((*windInfo)->windowPict);
		(*windInfo)->windowPict = nil;
	}


// Unlock our window info handle, close the file and return any errors.

	HUnlock((Handle) windInfo);
	FSClose(pictRefNum);
	return err;
}


/************************************************************
  SpoolPictToWindow - This routine spools a picture into
  the window passed.  We use the information in the window
  information handle that we've hung off the window's
  refCon field to locate the file to spool.

*************************************************************/

OSErr SpoolPictToWindow(WindowPtr wp)
{
	WindowInfoHdl	windInfo;
	Rect			drawRect;
	PicHandle		thePict;
	OSErr			err;

// Get the information stored with this window, then install
// spooling, draw the picture, and remove spooling.

	windInfo = (WindowInfoHdl) GetWRefCon(wp);
	thePict = (*windInfo)->windowPict;
	drawRect = wp->portRect;
	SetPort(wp);

	err = SpoolInstall(windInfo, kArrowCursor);

	if (!err)
	{
		DrawPicture(thePict, &drawRect);
		err = SpoolRemove();
	}

	return err;
}


/************************************************************
  GetOpenName - This routine performs a StandardGetFile
  or SFGetFile, depending on the system software being
  used.  it allows the user to select a PICT file to
  display.

*************************************************************/

void GetOpenName(StandardFileReply *toReply7, SFReply *toReply)
{
	SFTypeList	typeList;
	Point		pt;
	
	SetPt(&pt, kDILeft, kDITop);
	typeList[0] = 'PICT';
	
	if (gHaveNewSFCalls)
		StandardGetFile(nil, 1, &typeList, toReply7);
	else
		SFGetFile(pt, "\pOpen which PICT:", nil, 1, &typeList, nil, toReply);
}


/************************************************************
  LoadOrPrintFiles - This routine opens or prints any
  documents we're launched with for pre-7.0 versions
  of system software.  (Later system software versions
  use AppleEvents and the HandleODOC or HandlePDOC
  routines.)

*************************************************************/

void LoadOrPrintFiles()
{
	short		message, numFiles, fileNo;
	long		chCount;
	AppFile		theFile;
	SFReply		reply;
	OSErr		err;
	WindowPtr	wp;

// See how many files we were launched with and whether
// the message is appOpen or appPrint.

	CountAppFiles(&message, &numFiles);


// If we were launched with at least one file, loop through
// them all, retrieving their info.

	for (fileNo = 1; fileNo <= numFiles; fileNo++)
	{
		GetAppFiles(fileNo, &theFile);


// Make sure the file type is 'PICT', and then set up a
// SFReply structure based on the name and vRefNum for
// each (remember, we only use this routine pre-7.0.)

		if (theFile.fType == 'PICT')
		{
			reply.vRefNum = theFile.vRefNum;
			chCount = theFile.fName[0] +1;
			BlockMove(&theFile.fName, &reply.fName, chCount);


// Based on the message that was used, either open or print
// this document.  If the message is appOpen, load the
// document and display it in a window.  If the message is
// appPrint, do all that, then print the document and close
// it.

			switch (message)
			{
				case appOpen:	// Load selected docs.
					err = CreatePictWindow((FSSpecPtr) nil, &reply);
					AlertUser(err);
					break;
					
				case appPrint:	// Load and print selected docs, then quit.
					err = CreatePictWindow((FSSpecPtr) nil, &reply);
					AlertUser(err);

					if (!err)
					{
						wp = FrontWindow();
						err = PrintOne(wp);
						AlertUser(err);
						DoCloseWindow(wp);
					}
					break;
			}
		}


// Clear the file we just handled from the document list, then
// go handle the next.  When we're all done, set gQuitting to
// true if we were printing.  This way, the application will
// quit after printing all files, which is the proper behavior.

		ClrAppFiles(fileNo);
	}
	
	if (message == appPrint) gQuitting = true;
}
