/*********************************************************************

	Color Printing.c
	
	This file contains printing functions for the "Adventures in
	Color Printing" color/grayscale sample, "Color Adventures."
	
	Dave Hersey, Apple Developer Technical Support.
	
	 Edit Trail 
	
	creation/reorg/rebirth:						 6/28/92  - dmh
	sync'd with other samples:					 11/7/92  - dmh
	cleaned it up:								11/14/92  - dmh
	final clean up and sync with other samples:	11/29/92  - dmh

	Hint: If you start looking at "PrintOne," you'll be able
	      to follow the chain of command easier.
	
	* Related functions are grouped together in the Mark menu. *
	
*********************************************************************/

#pragma segment Printing

#include "Color Headers.h"

static Boolean		gRectNotSetYet;		// Flag used by ColorBitsProc.
static Fixed		gPMHRes;			// Horizontal resolution,
static Fixed		gPMVRes;			// vertical resolution,
static Rect			gSrcRect;			// and bounds of PixMap currently being
										// examined in ColorBitsProc.

/************************************************************
  PrintOne - This routine starts the printing process for
  the indicated window.  We determine the bounds to use when
  printing the image, set the printer port's resolution,
  display the job dialog and so forth.  If everything
  succeeds, we pass control on to the DoPrintLoop routine.

*************************************************************/

OSErr PrintOne(WindowPtr whichWindow)
{
	QDErr				err;
	GWorldPtr			savedGW;
	GDHandle			savedGDH;
	short				pxDPI, pyDPI, ixDPI, iyDPI, itemMark;
	Boolean				useDrawPicture;
	DialogPtr			printStatDlog = nil;
	Rect				srcRect, destRect;
	Str255				tempStr, orgTitle;
	Fixed				scale;
	PicHandle			wPICT;
	WindowInfoHdl		windInfo;

// Save the current GWorld/GDevice so that we can restore it later.
// See if the user selected "Use DrawPicture" from the Print Options
// menu, and retrieve the picture to print from our private
// "WindowInfoHdl" structure that's hanging off the window's refCon.

	GetGWorld(&savedGW, &savedGDH);
	GetItemMark(GetMHandle(mOptions), iDrawPicture, &itemMark);
	useDrawPicture = (itemMark != noMark);
	windInfo = (WindowInfoHdl) GetWRefCon(whichWindow);
	wPICT = (*windInfo)->windowPict;


// Get the picture's resolution.  We'll want to set the printer port's
// resolution as close to that as possible.  In case a perfect match 
// isn't available, we indicate that 300 x 300 is the highest resolution
// that the GetBestDPI routine should substitute.

	GetImageRes(windInfo, &ixDPI, &iyDPI, &srcRect);

	pxDPI = ixDPI;
	pyDPI = iyDPI;
	GetBestDPI(&pxDPI, &pyDPI, 300, 300, true);


// Handle the print job dialog and, if there are no errors, save the
// print handle.  We set the printer resolution here (rather than above)
// because the DoJobDlog call will reset it to 72 dpi.  For more info on
// why it does this, see DoJobDlog.  We could have set the resolution
// before the dialogs (and changed the dialog code slightly), but I've
// done it this way to maintain consistency with the final sample.

	err = DoJobDlog(ghPrint);

	if (!err)
	{
		SavePHdl(ghPrint, gAppResRefNum);
		SetPrinterPortDPI(ghPrint, &pxDPI, &pyDPI);


// destRect will be the destination rectangle for this image.  First, set
// it to the native resolution bounds we found from GetImageRes.  Then,
// align the top-left corner of the rectangle to (0,0) and scale the area
// to the printer port's resolution, if necessary.

		destRect = srcRect;
		OffsetRect(&destRect, -destRect.left, -destRect.top);
	
		if ((pxDPI != ixDPI) || (pyDPI != iyDPI))
		{
			scale = FixRatio(pxDPI, ixDPI);
			destRect.right = FixMul(scale, (long) destRect.right << 16) >> 16;
			scale = FixRatio(pyDPI, iyDPI);
			destRect.bottom = FixMul(scale, (long) destRect.bottom << 16) >> 16;
		}

// Display a status dialog, jump to the print loop routine, and hide the status
// dialog.  Finally, restore the current GWorld to what it was when we entered,
// and return.

		if (useDrawPicture)
			sprintf(tempStr, "%s %i %s %i %s", " at", pxDPI, "dpi horizontally by",
					pyDPI, "dpi vertically.  (Using DrawPicture.)");
		else
			sprintf(tempStr, "%s %i %s %i %s", " at", pxDPI, "dpi horizontally by",
					pyDPI, "dpi vertically.  (Using banding.)");

		GetWTitle(whichWindow, orgTitle);
		ShowStatusMsg(&printStatDlog, "Printing ", orgTitle, tempStr, true);
		err = DoPrintLoop(ghPrint, whichWindow, &destRect, useDrawPicture);
		HideStatusMsg(&printStatDlog);
	}

	SetGWorld(savedGW, savedGDH);
	return err;
}


/************************************************************
  DoPrintLoop - This routine does the actual printing.  The
  picture to be printed is attached to the window passed
  (through our WindowInfoHdl structure).
  
  If useDrawPicture is true, we simply print using
  DrawPicture, otherwise we print by banding the picture
  through a 32-bit GWorld.

*************************************************************/

OSErr DoPrintLoop(THPrint hPrint, WindowPtr wp, Rect *destRect, Boolean useDrawPicture)
{
	OSErr			err;
	TPPrPort		printPort;
	TPrStatus		theStatus;
	Rect			rPage, bandRect, pictFrame, prtRect;
	GWorldPtr		prtGWorld, savedGWorld, bandGWorld = nil;
	GDHandle		prtGDH, savedGDH;
	PixMapHandle	srcPix;
	Boolean			moreToDo, spooling;
	WindowInfoHdl	windInfo;
	PicHandle		pict;
	short			copy, iFstPage, iLstPage, rowWidth;
	long			bandHt, bandWidthInBytes, maxBlockSize;

// Get the window's picture and find out if we're spooling.

	windInfo = (WindowInfoHdl) GetWRefCon(wp);
	pict = (*windInfo)->windowPict;
	spooling = !(*windInfo)->loaded;


// Get the cursor spinning (we'll keep it spinning in our idle proc).  We
// specify r_stdCurs4 as the final cursor because we may be switching
// between the spinning spool cursors and the spinning standard cursors.
// Using the last standard cursor as the final cursor makes the transition
// look smoother.
//
// Save the current GWorld/GDevice so that we can restore them when we leave.
// Open the printer driver and validate the print record so that the Printing
// Manager pulls our document title from the title of the frontmost window.
// The reason we save and restore the iFstPage and iLstPage fields is that the
// LaserWriter driver (through at least v7.1.2) has a bug in PrValidate which
// causes these values to be reset.  While we aren't using iFstPage or iLstPage
// below, you never know where this code might end up in this crazy cut & paste
// world.

	SetUpCursors(r_stdSpinCursor, r_stdCurs4);
		
	GetGWorld(&savedGWorld, &savedGDH);

	PrOpen();

	if (!(err = PrError()))
	{
		iFstPage = (*hPrint)->prJob.iFstPage;
		iLstPage = (*hPrint)->prJob.iLstPage;
		
		PrValidate(hPrint);
		
		(*hPrint)->prJob.iFstPage = iFstPage;
		(*hPrint)->prJob.iLstPage = iLstPage;


// Add an idle proc to spin the cursor.  Get the printer driver's page
// rectangle, and offset and center the picture frame in it. It's ok if
// pictFrame is larger than rPage, because we handle that below.

		(*hPrint)->prJob.pIdleProc = (PrIdleProcPtr) IdleProc;
		
		rPage = (*hPrint)->prInfo.rPage;
		pictFrame = *destRect;
		OffsetRect(&pictFrame, -pictFrame.left, -pictFrame.top);
		CenterRect(&pictFrame, &rPage, &pictFrame);


// If we're not using DrawPicture (if we're banding), get 75% of
// the maximum block size available and use that figure to determine
// the size of the bands we create.  The calculation is more straight-
// forward than it may appear.  We determine the maximum buffer size we
// have to work with (maxBlockSize), then determine the width of each row
// (rowWidth).  Note that if the picture's rows are wider than the page
// rectangle, we clip to the page rectangle.  From the rowWidth, we
// determine how many bytes are required for each row (at 32 bits per
// pixel).  We then make sure this number is longword aligned (as 32-
// bit QuickDraw will make it when we allocate the GWorld).  Finally,
// if we have enough room for the entire image, we use the height of the
// image as the bandHt.  Otherwise, we use (maxBlockSize/BandWidthInBytes)
// as bandHt. Actually, bandHt is really the height of a band -1.

		if (!useDrawPicture)
		{
			maxBlockSize = MaxBlock() * .75;
			rowWidth = MIN(pictFrame.right - pictFrame.left, rPage.right - rPage.left);
	
			bandWidthInBytes = 4 * rowWidth;			// (32 bits/pixel *
														// rowWidth pixels)  8 bits/byte

			bandWidthInBytes += bandWidthInBytes % 4;	// Make sure the value is
														// longword aligned.

			bandHt = MIN(maxBlockSize / bandWidthInBytes,
						 pictFrame.bottom - pictFrame.top);


// Set up bandRect, which is the bounds of our band buffer, and prtRect
// which is the same area, just aligned with our page rectangle.  Create
// a new GWorld the size of bandRect, to use as our band buffer.

			SetRect(&bandRect, pictFrame.left, pictFrame.top,
					pictFrame.left + rowWidth, pictFrame.top +bandHt);

			prtRect = bandRect;
			OffsetRect(&prtRect, -rPage.left, -rPage.top);
			if (!err) err = NewGWorld(&bandGWorld, 32, &bandRect, nil, nil, 0);
		}

// Loop for the number of copies we need to make, opening a document
// and a page for each.

		for (copy = 1; !err && (copy <= (*hPrint)->prJob.iCopies); copy++)
		{
			printPort = PrOpenDoc(hPrint, nil, nil);

			if (!(err = PrError()))
			{
				PrOpenPage(printPort, nil);
				err = PrError();
				
				if (!err)
				{

// If we're using DrawPicture, clip to rPage and call DrawPicture.
// If we're also spooling, first tear down the spinning cursors
// (because spooling installs its own) and install spooling.
// After we've drawn, remove spooling and set the spinning cursors
// back up again.

					if (useDrawPicture)
					{
						if (spooling)
						{
							TearDownCursors();
							err = SpoolInstall(windInfo, r_stdCurs4);
						}
						
						if (!err)
						{
							ClipRect(&rPage);
							DrawPicture(pict, &pictFrame);

							if (spooling)
							{
								err = SpoolRemove();
								SetUpCursors(r_stdSpinCursor, r_stdCurs4);
							}
						}
					}
					else
					{

// If we're using banding instead of using DrawPicture, save off the current
// GWorld/GDevice settings (which can get us back to the printer port later).

						moreToDo = true;
						GetGWorld(&prtGWorld, &prtGDH);


// For each pass through the loop (each band), set the current GWorld to
// our bandGWorld, erase it, clip to it, and call DrawPicture.  pictFrame
// is offset below, so that successive passes will write successive bands
// into our bandGWorld.  If this picture is being spooled in, we need to
// install and remove spooling (and remove and install our spinning
// cursors) around the DrawPicture call.  (The spooling routines use their
// own spinning cursors, so we need to take down the current ones first.)

						while (moreToDo && !err)
						{
							SetGWorld(bandGWorld, nil);
		
							if (spooling)
							{
								TearDownCursors();
								err = SpoolInstall(windInfo, r_stdCurs4);
							}

							EraseRect(&bandRect);
							ClipRect(&bandRect);
							DrawPicture(pict, &pictFrame);

							if (spooling)
							{
								err = SpoolRemove();
								SetUpCursors(r_stdSpinCursor, r_stdCurs4);
							}

// Set the current GWorld (or port) back to the printer port and check for
// command-periods.  If none are found, copy the band from our bandGWorld
// to the printer port.  Afterwards, if the whole picture hasn't been
// drawn (pictFrame.bottom > bandRect.top) and we haven't drawn all that
// will fit on a page (prtRect.bottom < rPage.bottom), we have more to draw.

							SetGWorld(prtGWorld, prtGDH);
							if (!err) err = CommandPeriod();
			
							if (!err)
							{
								srcPix = GetGWPixMap(bandGWorld);
								LockPixels(srcPix);
								ClipRect(&prtRect);
								CopyBits((BitMap *) *srcPix,
										 &(printPort->gPort.portBits), &bandRect,
										 &prtRect, srcCopy, nil);

								UnlockPixels(srcPix);
								
								moreToDo = (pictFrame.bottom > bandRect.top) &&
										   (prtRect.bottom < rPage.bottom);
							}

// If we have more to do, and there were no errors, shift pictFrame up by bandHt.
// This means that next time through, one band lower will drawn.  Also offset
// prtRect so that we copy the next band one band lower in the printer port.  Next,
// calculate the new bandHt, which will be either the current bandHt, the height
// of the remaining area on the page that we can draw, or the height of the
// remaining part of the image to draw, whichever is smallest.  Set the new band
// size by adjusting prtRect and bandRect appropriately.

							if (moreToDo && !err)
							{
								pictFrame.top -= bandHt;
								pictFrame.bottom -= bandHt;
								prtRect.top += bandHt;
								bandHt = MIN(bandHt, (rPage.bottom - prtRect.bottom));
								bandHt = MIN(bandHt, (pictFrame.bottom - bandRect.top));
								prtRect.bottom = prtRect.top + bandHt;
								bandRect.bottom = bandRect.top + bandHt;
							}
						}

// All done banding.  Make sure that the current GWorld (port) is the printer port,
// so that the Printing Manager stays happy when we make the remaining print calls.
// If an error occurred, set a print error.  Close the page since all drawing
// for this page is completed.  When all pages are completed, close the document.


						SetGWorld(prtGWorld, prtGDH);
					}
				}
				
				if (err) PrSetError(err);
				PrClosePage(printPort);
			}
		
			PrCloseDoc(printPort);
		}

// When finished printing, dispose of our bandGWorld if we banded, call
// PrPicFile (if necessary), stop the cursor's spinning, close the printer
// driver and restore the GWorld/GDevice which were active upon entry.

		if (bandGWorld)
			DisposeGWorld(bandGWorld);

		if (!err) err = PrError();

		if (!err && ((*hPrint)->prJob.bJDocLoop == bSpoolLoop))
			PrPicFile(hPrint, nil, nil, nil, &theStatus);
	}

	(*hPrint)->prJob.pIdleProc = nil;
	TearDownCursors();
	SetCursor(&qd.arrow);				// We'd still be a beachball without this.

	PrClose();

	SetGWorld(savedGWorld, savedGDH);
	return err;
}


/************************************************************
  GetBestDPI - This routine obtains the best horizontal &
  vertical resolutions to use for printing with the
  selected driver.
  
  pxDPI, pyDPI --  The caller places an "ideal" resolution
  				   pair here (what they really want to use).
				   Also, this is where the routine returns
				   the resolutions it decides on.

  xDPI_ceiling,
  yDPI_ceiling --  The caller places the maximum resolution
  				   they want in either direction.  For
				   example, if you didn't want values larger
				   than 300 dpi returned, you would put 300
				   in both of these parameters.

  wantSquareDPI -- If true, only square resolutions (equal
  				   horizontal and vertical components) will
				   be considered.

  The printer driver is expected to be closed upon entry to
  this routine, and is therefore opened and closed around
  the PrGeneral code.  If PrGeneral is not supported by this
  driver, or if an error occurs, the routine returns
  72x72 dpi, which is the default for Macintosh printer
  drivers.
  
  If the "ideal" resolution the caller passes in is
  available, we choose that, ignoring wantSquareDPI,
  xDPI_ceiling and yDPI_ceiling.  We figure that the calling
  routine knows more about the ideal resolution it requests
  than we do.

*************************************************************/

void GetBestDPI(short *pxDPI, short *pyDPI, short xDPI_ceiling,
				short yDPI_ceiling, Boolean wantSquareDPI)
{
	TGetRslBlk		getResRec;
	Boolean			exactMatch = false;
	short			bestResX, bestResY, xDPI, yDPI,
					desiredResX, desiredResY, rec;

// Store the desired resolution that the caller passed to us through
// the pxDPI and pyDPI parameters.  If 72x72 dpi is requested, return
// those values, since that resolution must always be supported by a
// Mac printer driver (even though it may not appear in the PrGeneral
// getRsl list) for some (confused) printer drivers.

	desiredResX = *pxDPI;
	desiredResY = *pyDPI;

	if ((desiredResX == 72) && (desiredResY == 72)) return;


// Open the driver for our PrGeneral call.  Assume that we'll return
// 72x72 DPI until we find otherwise, and ask PrGeneral for the
// resolution records for this driver.

	PrOpen();
	
	bestResX = bestResY = 72;

	if (!PrError())
	{
		getResRec.iOpCode = getRslDataOp;
		PrGeneral((Ptr) &getResRec);

		if ((!ResError()) && (!getResRec.iError))
		{

// First check for the exact resolution pair that the caller
// requested.  We'll start by checking the range of resolutions
// supported to see if the pair is within that.

			if ((getResRec.xRslRg.iMin <= desiredResX) &&
				(getResRec.xRslRg.iMax >= desiredResX) &&
				(getResRec.yRslRg.iMin <= desiredResY) &&
				(getResRec.yRslRg.iMax >= desiredResY))
				exactMatch = true;


// If we didn't find an exact match, check the driver's
// discrete resolutions to see if we have one there.

			for (rec = 0; (!exactMatch) && (rec < getResRec.iRslRecCnt); rec++)
				if ((getResRec.rgRslRec[rec].iXRsl == desiredResX) &&
					(getResRec.rgRslRec[rec].iYRsl == desiredResY))
					exactMatch = true;


// If we found an exact match use it.  Otherwise, loop through
// each resolution record and find the one which best matches
// our criteria.

			if (exactMatch)	
			{
				bestResX = desiredResX;
				bestResY = desiredResY;
			}
			else
				for (rec = 0; (rec < getResRec.iRslRecCnt); rec++)
				{
					xDPI = getResRec.rgRslRec[rec].iXRsl;
					yDPI = getResRec.rgRslRec[rec].iYRsl;
	
					if ((xDPI <= xDPI_ceiling) && (yDPI <= yDPI_ceiling) &&
						(!wantSquareDPI || (xDPI == yDPI)) &&
						((xDPI * yDPI) > (bestResX * bestResY)))
					{
						bestResX = xDPI;
						bestResY = yDPI;
					}
				}
		}
	}
		
// Return the best resolution pair we found and close the driver.
	
	*pxDPI = bestResX;
	*pyDPI = bestResY;
	PrClose();
}


/************************************************************
  SetPrinterPortDPI - This routine sets the resolution of
  the printer port to the passed horizontal and vertical
  resolutions.  hPrint is the current print handle.

  The printer driver is expected to be closed upon entry to
  this routine, and is therefore opened and closed around
  the PrGeneral code.  If the current driver doesn't support
  PrGeneral, or if an error occurs, the passed horizontal
  and vertical resolutions are changed to 72 dpi, which is
  what printer drivers on the Macintosh default to.

*************************************************************/

void SetPrinterPortDPI(THPrint hPrint, short *pxDPI, short *pyDPI)
{
	TSetRslBlk		setResRec;
	OSErr			err1, err2;

	PrOpen();

	if (!PrError())
	{
		setResRec.iOpCode = setRslOp;
		setResRec.hPrint = hPrint;
		setResRec.iXRsl = *pxDPI;
		setResRec.iYRsl = *pxDPI;
		PrGeneral((Ptr) &setResRec);
	
		err1 = ResError();
		err2 = setResRec.iError;
		
		if (err1 || err2)
			*pxDPI = *pyDPI = 72;
	}

	PrClose();
}


/************************************************************
  DoStyleDlog - This routine handles the style (Page Setup)
  dialog.

  The printer driver is expected to be closed upon entry,
  and is therefore opened and closed in this routine.

  This routine assures that when the handle is returned,
  it is at 72 dpi, regardless of what it was before.  For
  reasons mentioned below, this is important.

*************************************************************/

OSErr DoStyleDlog(THPrint hPrint)
{
	OSErr		err;
	Boolean		doIt;
	short		resetThedpi;

// Make sure we're working with a default resolution print record, and open
// the printer driver.  Since (in the final sample) we don't know what resolution
// we'll be setting the printer port to until after the dialogs, it's best to
// use a default dpi handle now so that options (which may be ignored when we
// print at high resolutions) are at least available in case they *can* be used.
//
// What makes things tricky is that different printer drivers support
// different ways of resetting the printer port resolution to 72 dpi.  You 
// always use PrGeneral, but the ImageWriter driver, for example, goes to "true"
// default resolution mode only when you try to set an invalid resolution such
// as 0x0 dpi, and the LaserWriter makes you pass 72x72 to get to 72 dpi.
// Neither respects the other's way of doing things.  (For an example, try setting
// the ImageWriter's print handle to 72x72 dpi with PrGeneral, and then doing a
// style dialog with that.)
// 
// One solution which I've found to work well is to first try to set the port to
// 72x72 dpi, and then try to set it to 0x0 dpi.  Drivers which reset when passed
// 72x72 ignore the 0x0 setting, and drivers that need to be passed 0x0, are happy
// as well.  Remember that the order is important here.  If you tell the IW driver
// that you want 72x72 dpi, it will disable options in the print dialog that
// shouldn't be.  Only when it tries to set the bogus 0x0 dpi will it clean house.
//
// Unfortunately, this is one of the design short-comings of the current
// printing architecture, or more specifically, the current printer drivers.
// If drivers don't disable options at different resolutions, we have no
// problem at all.

	resetThedpi = 72;				// Pass 72x72 dpi to force the defaults.
	SetPrinterPortDPI(hPrint, &resetThedpi, &resetThedpi);

	resetThedpi = 0;				// Pass a bogus pair to force the defaults.
	SetPrinterPortDPI(hPrint, &resetThedpi, &resetThedpi);
									// One of those should work.

// Handle the style dialog, close the driver and clean up any windows that
// need updating.

	PrOpen();

	err = PrError();
	
	if (err == fnfErr)	// Here, "file not found" really means "no printer chosen."
		err = noPrinterChosenErr;
	
	if (!err)
	{
		doIt = PrStlDialog(hPrint);
		err = PrError();
		if (!err && !doIt) err = iPrAbort;
	}

	PrClose();
	DoAnyUpdates();
	return err;
}


/************************************************************
  DoJobDlog - This routine handles the print job dialog.

  The printer driver is expected to be closed upon entry,
  and is therefore opened and closed in this routine.

  This routine assures that when the handle is returned,
  it is at 72 dpi, regardless of what it was before.  For
  reasons mentioned below, this is important.

*************************************************************/

OSErr DoJobDlog(THPrint hPrint)
{
	OSErr		err;
	Boolean		doIt;
	short		resetThedpi;

// Make sure we're working with a default resolution print record, and open
// the printer driver.  Since (in the final sample) we don't know what resolution
// we'll be setting the printer port to until after the dialogs, it's best to
// use a default dpi handle now so that options (which may be ignored when we
// print at high resolutions) are at least available in case they *can* be used.
//
// What makes things tricky is that different printer drivers support
// different ways of resetting the printer port resolution to 72 dpi.  You 
// always use PrGeneral, but the ImageWriter driver, for example, goes to "true"
// default resolution mode only when you try to set an invalid resolution such
// as 0x0 dpi, and the LaserWriter makes you pass 72x72 to get to 72 dpi.
// Neither respects the other's way of doing things.  (For an example, try setting
// the ImageWriter's print handle to 72x72 dpi with PrGeneral, and then doing a
// style dialog with that.)
// 
// One solution which I've found to work well is to first try to set the port to
// 72x72 dpi, and then try to set it to 0x0 dpi.  Drivers which reset when passed
// 72x72 ignore the 0x0 setting, and drivers that need to be passed 0x0, are happy
// as well.  Remember that the order is important here.  If you tell the IW driver
// that you want 72x72 dpi, it will disable options in the print dialog that
// shouldn't be.  Only when it tries to set the bogus 0x0 dpi will it clean house.
//
// Unfortunately, this is one of the design short-comings of the current
// printing architecture, or more specifically, the current printer drivers.
// If drivers don't disable options at different resolutions, we have no
// problem at all.

	resetThedpi = 72;				// Pass 72x72 dpi to force the defaults.
	SetPrinterPortDPI(hPrint, &resetThedpi, &resetThedpi);

	resetThedpi = 0;				// Pass a bogus pair to force the defaults.
	SetPrinterPortDPI(hPrint, &resetThedpi, &resetThedpi);
									// One of those should work.

// Handle the job dialog, close the driver and clean up any windows that
// need updating.

	PrOpen();

	err = PrError();
	
	if (err == fnfErr)	// Here, "file not found" really means "no printer chosen."
		err = noPrinterChosenErr;
	
	if (!err)
	{
		doIt = PrJobDialog(hPrint);
		err = PrError();
		if (!err && !doIt) err = iPrAbort;
	}

	PrClose();
	DoAnyUpdates();
	return err;
}


/************************************************************
  SetupPrintHdl - This routine returns a default print
  handle for the current printer driver.

  The printer driver is expected to be closed upon entry,
  and is therefore opened and closed in this routine.

*************************************************************/

OSErr SetupPrintHdl(THPrint *hPrint)
{
	OSErr err;
	
// Create a handle, open the printer driver, set the
// default values for the handle, and close the driver.
// Be on the lookout for errors.

	*hPrint = (THPrint) NewHandle(sizeof(TPrint));

	if (*hPrint != nil)
	{
		PrOpen();

		err = PrError();
		
		if (err == fnfErr)	// Here, "file not found" really means "no printer chosen."
			err = noPrinterChosenErr;
		
		if (err) return err;

		PrintDefault(*hPrint);
		err = PrError();

		PrClose();
	}
	else
		err = MemError();
	
	return err;
}


/************************************************************
  GetResIDString - This routine creates a unique resource
  ID by adding (the driver version * 10000) and the high
  byte of the wDev.  All values are derived from the print
  handle passed and the currently opened printer driver.
  Once we have the ID, it's converted to a string to use
  for the resource's name.  (Our ID might be 8 digits long,
  so we can't use it as a normal resource ID which is only 5
  digits long.)
  
  Because we call PrDrvrVers, this function must be called
  between PrOpen & PrClose.

*************************************************************/

void GetResIDString(THPrint hPrint, Str255 resIDStr)
{
	short		wDev;
	long		drvrVers, codeID;

	wDev = (*hPrint)->prStl.wDev >> 8;
	drvrVers = PrDrvrVers();
	codeID = (drvrVers * 10000) + wDev;
	NumToString(codeID, resIDStr);
}


/************************************************************
  SavePHdl - This routine saves the passed print handle in
  the indicated resource fork, using a name which identifies
  it uniquely for the printer driver and version being used.
  If there is already a print record resource with this name,
  it's deleted and replaced.

  The printer driver is expected to be closed upon entry,
  and is therefore opened and closed in this routine.

*************************************************************/

OSErr SavePHdl(THPrint hPrint, short resRefNum)
{
	short		savedRefNum, uniqID;
	OSErr		err;
	Handle		oldHdl;
	Str255		resIDStr;

// Save the current resource file refNum, open the printer
// driver, get our unique name for the passed print handle
// and close the driver.

	savedRefNum = CurResFile();
	PrOpen();
	err = PrError();
	
	if (err == fnfErr)	// Here, "file not found" really means "no printer chosen."
		err = noPrinterChosenErr;
	
	if (err) return err;

	if (!err) GetResIDString(hPrint, resIDStr);
	PrClose();
	

// If there's already a resource with this name in the passed
// resource file, delete it.

	UseResFile(resRefNum);
	
	if (!err)
	{
		oldHdl = GetNamedResource(r_printHdlType, resIDStr);

		if (oldHdl)
		{
			RmveResource(oldHdl);
			DisposHandle(oldHdl);
		}

// Save the new resource with a unique ID and our coded name.
	
		uniqID = UniqueID(r_printHdlType);
		AddResource((Handle) hPrint, r_printHdlType, uniqID, resIDStr);
		WriteResource((Handle) hPrint);
		err = ResError();


// Detach the resource, update the resource file and switch back
// to the original resource file.
	

		if (!err)
			DetachResource((Handle) hPrint);
	}
	
	UpdateResFile(resRefNum);
	UseResFile(savedRefNum);
	return err;
}


/************************************************************
  LoadPHdl - This routine loads a print handle resource
  which matches the name we uniquely encode.  If there is
  no print record resource in the indicated resource file
  which matches our uniquely encoded name, the function
  returns resNotFound.
  
  Upon entry, hPrint is expected to be a valid print handle
  or nil.

  The printer driver is expected to be closed upon entry,
  and is therefore opened and closed in this routine.

*************************************************************/

OSErr LoadPHdl(THPrint *hPrint, short resRefNum)
{
	short		savedRefNum;
	OSErr		err;
	Handle		newHdl;
	Str255		resIDStr;

// Save the current resource file refNum, open the printer driver,
// get the name that we would have encoded this resource with and
// close the driver.

	savedRefNum = CurResFile();
	PrOpen();

	err = PrError();
	
	if (err == fnfErr)	// Here, "file not found" really means "no printer chosen."
		err = noPrinterChosenErr;
	
	if (err) return err;

	GetResIDString(*hPrint, resIDStr);
	PrClose();


// Get our resource, if there is one, out of the indicated resource
// file.  If there isn't one, return resNotFound.  Finally, switch
// back to the original resource file.

	UseResFile(resRefNum);
	
	newHdl = GetNamedResource(r_printHdlType, resIDStr);

	if (newHdl)
	{

// Got one.  Detach it from the resource map.  If we previously had
// a print handle, dispose of that before passing this new one back.

		DetachResource(newHdl);
		if (*hPrint) DisposHandle(*(Handle *) hPrint);
		*hPrint = (THPrint) newHdl;
		err = noErr;
	}
	else
		err = resNotFound;
	
	UseResFile(savedRefNum);
	return err;
}


/************************************************************
  ColorBitsProc - This routine is our QuickDraw bottleneck
  replacement for stdBits.  We use this routine when we're
  trying to determine the resolution of a picture based
  solely on the resolution of its contents.

*************************************************************/

pascal void ColorBitsProc(BitMap *srcBits, Rect *srcRect, Rect *dstRect,
						  short mode, RgnHandle maskRgn)
{
    PixMapPtr  pm;

#pragma unused(srcRect, dstRect, mode, maskRgn);

// If this is a PixMap (rowBytes < 0), store the resolution
// and bounds of the PixMap in our global variables.  Since
// BitMaps don't have resolution fields, ignore those.  We
// keep expanding the bounds everytime we find a pixmap in
// case the image was banded (we want gSrcRect to encompass
// all bands).

	if (srcBits->rowBytes < 0)
	{
		pm = (PixMapPtr) srcBits;
		gPMHRes = MAX(gPMHRes, pm->hRes);
		gPMVRes = MAX(gPMVRes, pm->vRes);

		if (gRectNotSetYet)
		{
			gSrcRect = pm->bounds;
			gRectNotSetYet = false;
		}
		else
		{
			gSrcRect.top = MIN(gSrcRect.top, pm->bounds.top);
			gSrcRect.left = MIN(gSrcRect.left, pm->bounds.left);
			gSrcRect.bottom = MAX(gSrcRect.bottom, pm->bounds.bottom);
			gSrcRect.right = MAX(gSrcRect.right, pm->bounds.right);
		}
	}
}


/************************************************************
  GetPixMapRes - This routine returns the resolution of the
  contents of a picture (specifically the resolution of its
  PixMap).  We assume that there will be at most one PixMap
  in our image, and given the type of images we're dealing 
  with, that's a valid assumption.  It's not clear how we
  would determine the correct resolution in a situation
  where multiple PixMaps were involved anyway.
  
  This routine is used if the original picture was not
  created with OpenCPicture.  If it was, we can get the
  resolution by just checking the picture header, as in
  GetImageRes below.

*************************************************************/

OSErr GetPixMapRes(WindowInfoHdl windInfo, Fixed *hRes, Fixed *vRes, Rect *srcRect)
{
	OSErr		err = noErr;
	PicHandle	aPict;
	CQDProcs	bottlenecks;
	GrafPtr		oldPort;
	CGrafPort	cPort;
	Boolean		spoolThisPict;

// Assume the image is 72x72 dpi and that the picture's bounds
// are the same as its picFrame.

    spoolThisPict = !(*windInfo)->loaded;
	aPict = (*windInfo)->windowPict;
	gPMHRes = gPMVRes = Long2Fix(72);
	gSrcRect = (*aPict)->picFrame;
	gRectNotSetYet = true;


// Save off the current port, and open up a new color
// port to work in.  We do this because we *must* be drawing
// into a CGrafPort for us to have PixMaps passed to our
// bottleneck routine without being converted into BitMaps.

	GetPort(&oldPort);
	OpenCPort(&cPort);


// Get the standard CGrafPort bottleneck routines, and add
// our replacement for stdBits.  This replacement routine
// will snoop for a PixMap, and store the resolution and
// bounds of any it finds in our global variables.  If
// none are found, we'll still have the defaults set up
// above.

	if (spoolThisPict)
		err = SpoolInstall(windInfo, kArrowCursor);
	else
	{
		SetStdCProcs(&bottlenecks);
		cPort.grafProcs = &bottlenecks;
	}

	if (!err)
	{
		cPort.grafProcs->bitsProc = (Ptr) ColorBitsProc;


// Draw the picture so our bottleneck routine is called.
// Afterwards, set back to the old port we were using and
// Close the CGrafPort we created.  Finally, return the
// resolution and bounds rectangle for this image.

		DrawPicture(aPict, &(*aPict)->picFrame);

		if (spoolThisPict)
			err = SpoolRemove();
	}

	SetPort(oldPort);
	CloseCPort(&cPort);

	*hRes = gPMHRes;
	*vRes = gPMVRes;
	*srcRect = gSrcRect;

	return err;
}


/************************************************************
  GetImageRes - This routine returns the native resolution
  of a picture, and the appropriate drawing rectangle for
  the image at that resolution.

*************************************************************/

OSErr GetImageRes(WindowInfoHdl windInfo, short *ixDPI, short *iyDPI, Rect *srcRect)
{
	PicHandle			wPICT;
	Fixed				hRes, vRes;
	PartialPicture		partialPict;
	Boolean				extndV2PICT = false;
	OSErr				err = noErr;

	wPICT = (*windInfo)->windowPict;
	HLock((Handle) windInfo);
	BlockMove(&(*windInfo)->partialPict, &partialPict, sizeof(PartialPicture));
	HUnlock((Handle) windInfo);

// If we have a version 2 picture, find out if it's an extended
// version 2 picture.  If so, the version field in the picture's
// header will be -2.  In that case, pull the resolution and bounding
// rect out of the picture header.

	if (partialPict.pictVersion == 2)
		if (partialPict.pictHeader.version == -2)
		{
			extndV2PICT = true;

			hRes = partialPict.pictHeader.hRes;
			vRes = partialPict.pictHeader.vRes;
			*srcRect = partialPict.pictHeader.srcRect;
		}

// If we have a version 1 picture, or a non-extended version 2
// picture, set the resolution and bounding rect for the
// picture based on the PixMap it contains.

	if (!extndV2PICT)
		err = GetPixMapRes(windInfo, &hRes, &vRes, srcRect);

	*ixDPI = FixRound(hRes);
	*iyDPI = FixRound(vRes);

	return err;
}


/************************************************************
  CenterRect - This routine centers srcRect within destRect
  and returns it in ctrRect.  It's mainly used for centering
  output on the printed page and for centering windows on
  the screen.  The source rectangle should not be larger
  than the destination rectangle in either direction.  If
  it is, the rectangle will not be adjusted in that
  direction.

*************************************************************/

void CenterRect(Rect *srcRect, Rect *destRect, Rect *ctrRect)
{
	Rect	srcR, dstR;
	short	srcWidth, srcHeight, destWidth, destHeight, dx, dy;

	srcR = *srcRect;
	dstR = *destRect;

	OffsetRect(&srcR, dstR.left - srcR.left, dstR.top - srcR.top);

	srcWidth = srcR.right - srcR.left +1;
	srcHeight = srcR.bottom - srcR.top +1;
	destWidth = dstR.right - dstR.left +1;
	destHeight = dstR.bottom - dstR.top +1;
	
	dx = MAX(0, (destWidth-srcWidth)/2);
	dy = MAX(0, (destHeight-srcHeight)/2);

	*ctrRect = srcR;
	OffsetRect(ctrRect, dx, dy);
}


/************************************************************
  IdleProc - This routine is a PrIdleProcPtr that spins our
  cursor and looks for command-period presses during
  printing.

*************************************************************/

pascal void IdleProc()
{
	OSErr		err;
	
	BumpCursor();
	if (err = CommandPeriod()) PrSetError(err);
}


/************************************************************
  CommandPeriod - This routine checks to see if the user has
  pressed command-period.  The Printing Manager does this
  for us, but a routine like this is helpful to allow users
  to cancel in the middle of a rendering routine, rather
  than only Printing Manager operations.
  
  If there's a command-period in the event queue, the
  routine returns iPrAbort just like the Printing Manager.
  Otherwise, it returns noErr.

*************************************************************/

OSErr CommandPeriod()
{
	Boolean		cancel = false;
	EventRecord	evtRec;
	
	while (!cancel && (WaitNextEvent(keyDownMask | autoKeyMask, &evtRec, 0, nil)))
		cancel = (evtRec.modifiers & cmdKey) &&
				 ((evtRec.message & charCodeMask) == '.');

	return ((cancel)? iPrAbort: noErr);
}
