/*********************************************************************

	HT Printing.c
	
	This file contains printing functions for the "Adventures in
	Color Printing" halftone sample, "Halftone Adventures."
	
	Dave Hersey, Apple Developer Technical Support.
	
	 Edit Trail 
	
	creation/reorg/rebirth:   					  6/26/92  - dmh
	Added ditherCopy example: 					  6/29/92  - dmh
	Fixed brain-dead switch statement:			  8/14/92  - dmh
	sync'd with other samples:					 11/15/92  - dmh
	major commenting spree:						 11/16/92  - dmh
	Fixed scaling problem with SimulateHalftone: 11/21/92  - dmh
	Added spinning cursors during halftoning:	 11/21/92  - dmh
	final clean up and sync with other samples:	 11/29/92  - dmh

	Hint: If you start looking at "PrintOne," you'll be able
	      to follow the chain of command easier.
	
	* Related functions are grouped together in the Mark menu. *
	
*********************************************************************/

#pragma segment Printing

#include "HT Headers.h"

static Boolean		gRectNotSetYet;		// Flag used by ColorBitsProc.
static Fixed		gPMHRes;			// Horizontal resolution,
static Fixed		gPMVRes;			// vertical resolution,
static Rect			gSrcRect;			// and bounds of PixMap currently being
										// examined in ColorBitsProc.

/************************************************************
  PrintOne - This routine starts the printing process for
  the indicated window.  We set the printer port's
  resolution, display the job dialog and so forth, then pass
  control on to DoCopyBitsDither, DoHT0 or DoHT45 for
  rendering.  Those routines in turn call the DoPrintLoop
  routine to print the image.

*************************************************************/

OSErr PrintOne(WindowPtr whichWindow)
{
	QDErr				err;
	GWorldPtr			savedGW;
	GDHandle			savedGDH;
	short				itemMark, pxDPI, pyDPI, ixDPI, iyDPI;
	WindowInfoHdl		windInfo;
	Rect				pictFrame;
	Fixed				scale;
	enum {CopyBitsDither,
		  HT0deg,
		  HT45deg}		htType;

// Save the current GWorld/GDevice so that we can restore it later.
// See which rendering option the user selected, retrieve our private
// "WindowInfoHdl" structure that's hanging off the window's refCon
// field and calculate the image's native resolution and dimensions.

	GetGWorld(&savedGW, &savedGDH);
	GetItemMark(GetMHandle(mOptions), iCopyBitsDither, &itemMark);

	if (itemMark != noMark)
		htType = CopyBitsDither;
	else
	{
		GetItemMark(GetMHandle(mOptions), i0degHalftone, &itemMark);
		htType = (itemMark == noMark)? HT45deg: HT0deg;
	}

	windInfo = (WindowInfoHdl) GetWRefCon(whichWindow);
	GetImageRes(windInfo, &ixDPI, &iyDPI, &pictFrame);


// We want to print at the printer's physical resolution for minimal
// distortion, so find that.

	pxDPI = pyDPI = GetPrinterRes();				// rather than guessing the 
													// physical resolution of the
													// printer, find it.

//	GetBestDPI(&pxDPI, &pyDPI, 300, 300, true);		// Not really necessary unless
													// we're just guessing.  Left here
													// for reference in case you go
													// the guessing route.


// Handle the print job dialog and, if there are no errors, save the
// print handle.  We set the printer resolution here (rather than above)
// because the DoJobDlog call will reset it to 72 dpi.  For more info on
// why it does this, see DoJobDlog.  We could have set the resolution
// before the dialogs (and changed the dialog code slightly), but I've
// done it this way to maintain consistency with the final sample.
//
// Set the cursor to the watch, align the top-left corner of our pictFrame
// to (0,0), and then scale the image (if necessary) to the printer's port's
// resolution.

	err = DoJobDlog(ghPrint);

	if (!err)
	{
		SetCursor(*GetCursor(watchCursor));
		SavePHdl(ghPrint, gAppResRefNum);
		SetPrinterPortDPI(ghPrint, &pxDPI, &pyDPI);

		OffsetRect(&pictFrame, -pictFrame.left, -pictFrame.top);

		if (htType != HT45deg)	// The 45 halftone method handles scaling internally.
		{
			if ((pxDPI != ixDPI) || (pyDPI != iyDPI))
			{
				scale = FixRatio(pxDPI, ixDPI);
				pictFrame.right = FixMul(scale, (long) pictFrame.right << 16) >> 16;
				scale = FixRatio(pyDPI, iyDPI);
				pictFrame.bottom = FixMul(scale, (long) pictFrame.bottom << 16) >> 16;
			}
		}

// Spin off to the appropriate rendering and printing routine.  Note that these
// routines expect pxDPI = pyDPI so we only pass pxDPI.  When we return, restore
// the GWorld/GDevice we saved earlier, set the cursor to the arrow, and bop on home.

		switch (htType)
		{
			case CopyBitsDither:
				err = DoCopyBitsDither(whichWindow, pyDPI, &pictFrame);
				break;
			
			case HT0deg:
				err = DoHT0(whichWindow, pxDPI, &pictFrame);
				break;
			
			case HT45deg:
				err = DoHT45(whichWindow, pxDPI, ixDPI, iyDPI, &pictFrame);
				break;
		}
	}

	SetGWorld(savedGW, savedGDH);
	SetCursor(&qd.arrow);

	return err;
}


/************************************************************
  DoCopyBitsDither - This routine creates a 1-bit GWorld
  containing a 1-bit dithered representation of the picture
  associated with the passed window.  If all goes well, the
  image is passed to the DoPrintLoop routine and printed.

  pDPI is the printer's resolution (horizontal and vertical
  resolutions will be equal.

  destRect is the destination rectangle for the image at
  the printer's resolution.  Its top-left corner is aligned
  to (0,0).

*************************************************************/

OSErr DoCopyBitsDither(WindowPtr wp, short pDPI, Rect *destRect)
{
	QDErr			err;
	Str255			orgTitle, tempStr;
	GWorldPtr		savedGW, ditherGWorld;
	GDHandle		savedGDH;
	DialogPtr		printStatDlog = nil;
	Rect			printRect, rPage;
	PicHandle		wPICT;
	WindowInfoHdl	windInfo;

// Save the current GWorld/GDevice for restoration later.  Also, retrieve the
// picture we're supposed to print.

	GetGWorld(&savedGW, &savedGDH);
	windInfo = (WindowInfoHdl) GetWRefCon(wp);
	wPICT = (*windInfo)->windowPict;


// Get the current window title, display a status dialog saying that we're
// about to dither, then do so.

	GetWTitle(wp, orgTitle);
	sprintf(tempStr, "%s %i %s %i %s", " at", pDPI, "dpi horizontally by", pDPI,
			"dpi vertically.");

	ShowStatusMsg(&printStatDlog, "Creating 1-bit dither of ", orgTitle, tempStr, true);

	printRect = *destRect;
	rPage = (*ghPrint)->prInfo.rPage;
	SectRect(&printRect, &rPage, &printRect);

	err = NewGWorld(&ditherGWorld, 1, &printRect, nil, nil, 0);

	if (!err)
		err = BandPicture(wp, destRect, ditherGWorld, true);


// If there were no errors, change our status message and print the image.  When
// done, dispose of our dithered image, hide the status dialog and restore the
// current GWorld/GDevice to what we saved earlier.

	if (!err)
	{
		ShowStatusMsg(&printStatDlog, "Printing ", orgTitle, ".", true);
		err = DoPrintLoop(ghPrint, &ditherGWorld, &printRect);
	}

	if (ditherGWorld) DisposeGWorld(ditherGWorld);
	HideStatusMsg(&printStatDlog);

	SetGWorld(savedGW, savedGDH);
	return err;
}


/************************************************************
  DoHT0 - This routine creates a 1-bit GWorld containing a
  simulated halftone (created with 4x4 pixel patterns) of
  the picture associated with the passed window.  If all
  goes well, the image is passed to the DoPrintLoop routine
  and printed.

  pDPI is the printer's resolution (horizontal and vertical
  resolutions will be equal.
  
  destRect is the destination rectangle for the image at
  the printer's resolution.  Its top-left corner is aligned
  to (0,0).

*************************************************************/

OSErr DoHT0(WindowPtr wp, short pDPI, Rect *destRect)
{
	QDErr			err;
	Str255			orgTitle, tempStr;
	Rect			srcRect, printRect, rPage;
	GWorldPtr		fourBitGWorld = nil, halftoneGWorld = nil, savedGW;
	GDHandle		savedGDH;
	DialogPtr		printStatDlog = nil;
	PicHandle		wPICT;
	WindowInfoHdl	windInfo;

// Save the current GWorld/GDevice for restoration later.  Also, retrieve the
// picture we're supposed to print and the window's title, then display a
// status dialog.

	GetGWorld(&savedGW, &savedGDH);
	windInfo = (WindowInfoHdl) GetWRefCon(wp);
	wPICT = (*windInfo)->windowPict;

	GetWTitle(wp, orgTitle);
	ShowStatusMsg(&printStatDlog, "Converting ", orgTitle, " to 16 grays.", true);


// Determine how much of the image will appear on the page, and store that in
// printRect.  We're going to create a GWorld that's 1/4 the size of printRect,
// so we make sure that that rectangle can be evenly divided by 4 and adjust it
// if not.  printRect will be the actual area of the halftone image that we copy
// to the printer port.  We set srcRect to 1/4 this size and use it to create a
// 4-bit GWorld to dither to.  Finally, we pass the picture and GWorld to
// BandPicture which will dither the PICT for us.

	printRect = *destRect;			// NOTE: destRect is top-left aligned to (0,0).

	rPage = (*ghPrint)->prInfo.rPage;				// Clip the output rectangle to
	SectRect(&printRect, &rPage, &printRect);		// the printable page area.

	printRect.right  &= 0xFFFC;						// Make the print rectangle bounds
	printRect.bottom &= 0xFFFC;						// divisible by 4.

	srcRect = printRect;							// srcRect = printRect /4.
	srcRect.right     >>= 2;
	srcRect.bottom    >>= 2;

	destRect->right   >>= 2;						// destRect = original destRect /4.
	destRect->bottom  >>= 2;						// This insures that picture
													// rect passed to BandPicture is
													// proportionally 1/4 as big as the
													// original, which is how we must
													// render it in fourBitGWorld.

	err = NewGWorld(&fourBitGWorld, 4, &srcRect, GetCTable(32 +4), nil, 0);
	if (!err) err = BandPicture(wp, destRect, fourBitGWorld, true);


// If there were no errors, change our status message and create our simulated
// halftone.  When finished, dispose of our 4-bit GWorld.

	if (!err)
	{
		sprintf(tempStr, "%s %i %s %i %s", "using patterns at", pDPI,
				"dpi horizontally by", pDPI, "dpi vertically.");

		ShowStatusMsg(&printStatDlog, "Simulating  0 halftone of ", orgTitle,
					  tempStr, true);

		err = SimulateHalftone(&printRect, fourBitGWorld, &halftoneGWorld);

		DisposeGWorld(fourBitGWorld);
		fourBitGWorld = nil;


// Tell the user we're going to print, then do so.  When finished, get rid of
// the status dialog, make sure that our GWorlds have been disposed, and restore
// the current GWorld and GDevice to our stored values.

		if (!err)
		{
			ShowStatusMsg(&printStatDlog, "Printing ", orgTitle, ".", true);
			err = DoPrintLoop(ghPrint, &halftoneGWorld, &printRect);
		}
	}

	HideStatusMsg(&printStatDlog);

	if (fourBitGWorld)  DisposeGWorld(fourBitGWorld);
	if (halftoneGWorld) DisposeGWorld(halftoneGWorld);

	SetGWorld(savedGW, savedGDH);
	return err;
}


/************************************************************
  DoHT45 - This routine creates a true halftone of the
  picture associated with the passed window.  If all goes
  well, the halftone is passed to the DoPrintLoop routine
  and printed.
  
  pDPI is the printer's resolution (horizontal and vertical
  resolutions will be equal.
  
  ixDPI and iyDPi are the horizontal and vertical
  resolutions of the source image.
  
  destRect is the destination rectangle for the image at
  the printer's resolution.  Its top-left corner is aligned
  to (0,0).

*************************************************************/

OSErr DoHT45(WindowPtr wp, short pDPI, short ixDPI, short iyDPI, Rect *destRect)
{
	QDErr				err;
	Str255				orgTitle, tempStr;
	GWorldPtr			srcGWorld = nil, halftoneGWorld = nil, savedGW;
	GDHandle			savedGDH;
	DialogPtr			printStatDlog = nil;
	Rect				rPage, printRect;
	PicHandle			wPICT;
	WindowInfoHdl		windInfo;
	Fixed				scale;
	PixMapHandle		srcPix;

// Save the current GWorld/GDevice for restoration later.  Also, retrieve the
// picture we're supposed to print.

	GetGWorld(&savedGW, &savedGDH);
	GetWTitle(wp, orgTitle);
	windInfo = (WindowInfoHdl) GetWRefCon(wp);
	wPICT = (*windInfo)->windowPict;


// Set printRect to the amount of the image that will actually fit on the page
// at the destination resolution.  This way, we don't work with any more of the
// image than necessary.  Finally, create a new 32-bit deep GWorld to store the
// image in, and use BandPicture to copy to it.

	printRect = *destRect;

	if (pDPI != ixDPI || pDPI != iyDPI)			// Scale printRect to the printer's res.
	{
		scale = FixRatio(pDPI, ixDPI);
		printRect.right = FixMul(scale, (long) printRect.right << 16) >> 16;
		scale = FixRatio(pDPI, iyDPI);
		printRect.bottom = FixMul(scale, (long) printRect.bottom << 16) >> 16;
	}

	rPage = (*ghPrint)->prInfo.rPage;			// Clip the output rectangle to the
	SectRect(&printRect, &rPage, &printRect);	// printable page area.

	if (pDPI != ixDPI || pDPI != iyDPI)			// Scale the clipped printRect back to
	{											// the source image's resolution.
		scale = FixRatio(ixDPI, pDPI);
		printRect.right = FixMul(scale, (long) printRect.right << 16) >> 16;
		scale = FixRatio(iyDPI, pDPI);
		printRect.bottom = FixMul(scale, (long) printRect.bottom << 16) >> 16;
	}

	err = NewGWorld(&srcGWorld, 32, &printRect, nil, nil, 0);

	if (!err)
	{
		ShowStatusMsg(&printStatDlog, "Preparing ", orgTitle, "for halftoning", true);
		err = BandPicture(wp, destRect, srcGWorld, false);


// Tell the user we're going to create a halftone of the image, and then go do it.
// When we come back, dispose of the source GWorld and print.

		sprintf(tempStr, "%s %i %s %i %s", " 45 at", pDPI, "dpi horizontally by",
				pDPI, "dpi vertically.");

		ShowStatusMsg(&printStatDlog, "Halftoning ", orgTitle, tempStr, true);

		if (!err) err = HalftonePixMap(srcGWorld, &halftoneGWorld, ixDPI, pDPI);

		DisposeGWorld(srcGWorld);
		srcGWorld = nil;

		if (!err)
		{
			srcPix = GetGWPixMap(halftoneGWorld);
			printRect = (*srcPix)->bounds;

			ShowStatusMsg(&printStatDlog, "Printing ", orgTitle, ".", true);
			err = DoPrintLoop(ghPrint, &halftoneGWorld, &printRect);
		}

// Tear down the status dialog, make sure our GWorlds are disposed of and restore
// the current GWorld/GDevice as we exit.

		HideStatusMsg(&printStatDlog);
	}
	
	if (srcGWorld) DisposeGWorld(srcGWorld);
	if (halftoneGWorld) DisposeGWorld(halftoneGWorld);

	SetGWorld(savedGW, savedGDH);
	return err;
}


/************************************************************
  DoPrintLoop - This routine does the actual printing.  The
  picture to be printed is attached to the window passed
  (through our WindowInfoHdl structure).

*************************************************************/

OSErr DoPrintLoop(THPrint hPrint, GWorldPtr *srcWorld, Rect *srcRect)
{
	OSErr			err;
	TPPrPort		printPort;
	TPrStatus		theStatus;
	GWorldPtr		savedGWorld;
	GDHandle		savedGDH;
	PixMapHandle	srcPix;
	Rect			rPage, outRect;
	short			copy, iFstPage, iLstPage;

// Set the cursor spinning (we'll keep it spinning in our idle proc).  Open the
// printer driver and validate the print record so that the Printing Manager
// pulls our document title from the title of the frontmost window.  The reason
// we save and restore the iFstPage and iLstPage fields is that the LaserWriter
// driver (through at least v7.1.2) has a bug in PrValidate which causes these
// values to be reset.  While we aren't using iFstPage or iLstPage below, you
// never know where this code might end up in this crazy cut & paste world.

	GetGWorld(&savedGWorld, &savedGDH);
	SetUpCursors(r_stdSpinCursor, kArrowCursor);

	PrOpen();

	if (!(err = PrError()))
	{
		iFstPage = (*hPrint)->prJob.iFstPage;
		iLstPage = (*hPrint)->prJob.iLstPage;
		
		PrValidate(hPrint);
		
		(*hPrint)->prJob.iFstPage = iFstPage;
		(*hPrint)->prJob.iLstPage = iLstPage;


// Install our idle proc, then center the image's rectangle within the page
// rectangle, then loop for the number of copies we need to make, opening a
// document and a page for each.

		(*hPrint)->prJob.pIdleProc = (PrIdleProcPtr) IdleProc;

		rPage = (*hPrint)->prInfo.rPage;
		outRect = *srcRect;
		CenterRect(&outRect, &rPage, &outRect);

		for (copy = 1; !err && (copy <= (*hPrint)->prJob.iCopies); copy++)
		{
			printPort = PrOpenDoc(hPrint, nil, nil);
		
			if (!(err = PrError()))
			{
				PrOpenPage(printPort, nil);


// Lock down our source PixMap, then copy it to the printer port.  When done,
// unlock its pixels, close the page, and the document.

				if (!(err = PrError()))
				{
					srcPix = GetGWPixMap(*srcWorld);
					LockPixels(srcPix);
					CopyBits((BitMap *) *srcPix, &printPort->gPort.portBits, srcRect, &outRect, srcCopy, nil);
					UnlockPixels(srcPix);
				}
				
				PrClosePage(printPort);
			}
		
			PrCloseDoc(printPort);
		}

// Dispose of the source image, call PrPicFile if necessary, remove our idle
// proc, stop our cursors from spinning, close down the Printing Manager and
// just LEAVE.

		DisposeGWorld(*srcWorld);
		*srcWorld = nil;

		if (!err) err = PrError();
		
		if (!err && ((*hPrint)->prJob.bJDocLoop == bSpoolLoop))
			PrPicFile(hPrint, nil, nil, nil, &theStatus);
	}

	(*hPrint)->prJob.pIdleProc = nil;
	TearDownCursors();
	PrClose();

	return err;
}


/************************************************************
  GetPrinterRes - This routine returns the highest physical
  square resolution of the selected printer.

*************************************************************/

short GetPrinterRes()
{
	TGetRslBlk		getResRec;
	short			bestRes, rec;
	
// Open the driver for our PrGeneral call.  Assume that
// we'll return 72x72 DPI until we find otherwise.

	PrOpen();
	bestRes = 72;
	
	if (!PrError())
	{

// Ask PrGeneral for the resolution records for this driver and find the
// find the largest square discrete resolution.

		getResRec.iOpCode = getRslDataOp;
		PrGeneral((Ptr) &getResRec);

		if ((!ResError()) && (!getResRec.iError))
			for (rec = 0; rec < getResRec.iRslRecCnt; rec++)
				if ((getResRec.rgRslRec[rec].iXRsl == getResRec.rgRslRec[rec].iYRsl) &&
					(getResRec.rgRslRec[rec].iXRsl > bestRes))
					bestRes = getResRec.rgRslRec[rec].iXRsl;
	}
		
// Close the driver and return the highest square discrete resolution.
	
	PrClose();
	return bestRes;
}


/************************************************************
  GetBestDPI - This routine obtains the best horizontal &
  vertical resolutions to use for printing with the
  selected driver.
  
  pxDPI, pyDPI --  The caller places an "ideal" resolution
  				   pair here (what they really want to use).
				   Also, this is where the routine returns
				   the resolutions it decides on.

  xDPI_ceiling,
  yDPI_ceiling --  The caller places the maximum resolution
  				   they want in either direction.  For
				   example, if you didn't want values larger
				   than 300 dpi returned, you would put 300
				   in both of these parameters.

  wantSquareDPI -- If true, only square resolutions (equal
  				   horizontal and vertical components) will
				   be considered.

  The printer driver is expected to be closed upon entry to
  this routine, and is therefore opened and closed around
  the PrGeneral code.  If PrGeneral is not supported by this
  driver, or if an error occurs, the routine returns
  72x72 dpi, which is the default for Macintosh printer
  drivers.
  
  If the "ideal" resolution the caller passes in is
  available, we choose that, ignoring wantSquareDPI,
  xDPI_ceiling and yDPI_ceiling.  We figure that the calling
  routine knows more about the ideal resolution it requests
  than we do.

*************************************************************/

void GetBestDPI(short *pxDPI, short *pyDPI, short xDPI_ceiling,
				short yDPI_ceiling, Boolean wantSquareDPI)
{
	TGetRslBlk		getResRec;
	Boolean			exactMatch = false;
	short			bestResX, bestResY, xDPI, yDPI,
					desiredResX, desiredResY, rec;

// Store the desired resolution that the caller passed to us through
// the pxDPI and pyDPI parameters.  If 72x72 dpi is requested, return
// those values, since that resolution must always be supported by a
// Mac printer driver (even though it may not appear in the PrGeneral
// getRsl list) for some (confused) printer drivers.

	desiredResX = *pxDPI;
	desiredResY = *pyDPI;

	if ((desiredResX == 72) && (desiredResY == 72)) return;


// Open the driver for our PrGeneral call.  Assume that we'll return
// 72x72 DPI until we find otherwise, and ask PrGeneral for the
// resolution records for this driver.

	PrOpen();
	
	bestResX = bestResY = 72;

	if (!PrError())
	{
		getResRec.iOpCode = getRslDataOp;
		PrGeneral((Ptr) &getResRec);

		if ((!ResError()) && (!getResRec.iError))
		{

// First check for the exact resolution pair that the caller
// requested.  We'll start by checking the range of resolutions
// supported to see if the pair is within that.

			if ((getResRec.xRslRg.iMin <= desiredResX) &&
				(getResRec.xRslRg.iMax >= desiredResX) &&
				(getResRec.yRslRg.iMin <= desiredResY) &&
				(getResRec.yRslRg.iMax >= desiredResY))
				exactMatch = true;


// If we didn't find an exact match, check the driver's
// discrete resolutions to see if we have one there.

			for (rec = 0; (!exactMatch) && (rec < getResRec.iRslRecCnt); rec++)
				if ((getResRec.rgRslRec[rec].iXRsl == desiredResX) &&
					(getResRec.rgRslRec[rec].iYRsl == desiredResY))
					exactMatch = true;


// If we found an exact match use it.  Otherwise, loop through
// each resolution record and find the one which best matches
// our criteria.

			if (exactMatch)	
			{
				bestResX = desiredResX;
				bestResY = desiredResY;
			}
			else
				for (rec = 0; (rec < getResRec.iRslRecCnt); rec++)
				{
					xDPI = getResRec.rgRslRec[rec].iXRsl;
					yDPI = getResRec.rgRslRec[rec].iYRsl;
	
					if ((xDPI <= xDPI_ceiling) && (yDPI <= yDPI_ceiling) &&
						(!wantSquareDPI || (xDPI == yDPI)) &&
						((xDPI * yDPI) > (bestResX * bestResY)))
					{
						bestResX = xDPI;
						bestResY = yDPI;
					}
				}
		}
	}
		
// Return the best resolution pair we found and close the driver.
	
	*pxDPI = bestResX;
	*pyDPI = bestResY;
	PrClose();
}


/************************************************************
  SetPrinterPortDPI - This routine sets the resolution of
  the printer port to the passed horizontal and vertical
  resolutions.  hPrint is the current print handle.

  The printer driver is expected to be closed upon entry to
  this routine, and is therefore opened and closed around
  the PrGeneral code.  If the current driver doesn't support
  PrGeneral, or if an error occurs, the passed horizontal
  and vertical resolutions are changed to 72 dpi, which is
  what printer drivers on the Macintosh default to.

*************************************************************/

void SetPrinterPortDPI(THPrint hPrint, short *pxDPI, short *pyDPI)
{
	TSetRslBlk		setResRec;
	OSErr			err1, err2;

	PrOpen();

	if (!PrError())
	{
		setResRec.iOpCode = setRslOp;
		setResRec.hPrint = hPrint;
		setResRec.iXRsl = *pxDPI;
		setResRec.iYRsl = *pxDPI;
		PrGeneral((Ptr) &setResRec);
	
		err1 = ResError();
		err2 = setResRec.iError;
		
		if (err1 || err2)
			*pxDPI = *pyDPI = 72;
	}

	PrClose();
}


/************************************************************
  DoStyleDlog - This routine handles the style (Page Setup)
  dialog.

  The printer driver is expected to be closed upon entry,
  and is therefore opened and closed in this routine.

  This routine assures that when the handle is returned,
  it is at 72 dpi, regardless of what it was before.  For
  reasons mentioned below, this is important.

*************************************************************/

OSErr DoStyleDlog(THPrint hPrint)
{
	OSErr		err;
	Boolean		doIt;
	short		resetThedpi;

// Make sure we're working with a default resolution print record, and open
// the printer driver.  Since (in the final sample) we don't know what resolution
// we'll be setting the printer port to until after the dialogs, it's best to
// use a default dpi handle now so that options (which may be ignored when we
// print at high resolutions) are at least available in case they *can* be used.
//
// What makes things tricky is that different printer drivers support
// different ways of resetting the printer port resolution to 72 dpi.  You 
// always use PrGeneral, but the ImageWriter driver, for example, goes to "true"
// default resolution mode only when you try to set an invalid resolution such
// as 0x0 dpi, and the LaserWriter makes you pass 72x72 to get to 72 dpi.
// Neither respects the other's way of doing things.  (For an example, try setting
// the ImageWriter's print handle to 72x72 dpi with PrGeneral, and then doing a
// style dialog with that.)
// 
// One solution which I've found to work well is to first try to set the port to
// 72x72 dpi, and then try to set it to 0x0 dpi.  Drivers which reset when passed
// 72x72 ignore the 0x0 setting, and drivers that need to be passed 0x0, are happy
// as well.  Remember that the order is important here.  If you tell the IW driver
// that you want 72x72 dpi, it will disable options in the print dialog that
// shouldn't be.  Only when it tries to set the bogus 0x0 dpi will it clean house.
//
// Unfortunately, this is one of the design short-comings of the current
// printing architecture, or more specifically, the current printer drivers.
// If drivers don't disable options at different resolutions, we have no
// problem at all.

	resetThedpi = 72;				// Pass 72x72 dpi to force the defaults.
	SetPrinterPortDPI(hPrint, &resetThedpi, &resetThedpi);

	resetThedpi = 0;				// Pass a bogus pair to force the defaults.
	SetPrinterPortDPI(hPrint, &resetThedpi, &resetThedpi);
									// One of those should work.

// Handle the style dialog, close the driver and clean up any windows that
// need updating.

	PrOpen();

	err = PrError();
	
	if (err == fnfErr)	// Here, "file not found" really means "no printer chosen."
		err = noPrinterChosenErr;
	
	if (!err)
	{
		doIt = PrStlDialog(hPrint);
		err = PrError();
		if (!err && !doIt) err = iPrAbort;
	}

	PrClose();
	DoAnyUpdates();
	return err;
}


/************************************************************
  DoJobDlog - This routine handles the print job dialog.

  The printer driver is expected to be closed upon entry,
  and is therefore opened and closed in this routine.

  This routine assures that when the handle is returned,
  it is at 72 dpi, regardless of what it was before.  For
  reasons mentioned below, this is important.

*************************************************************/

OSErr DoJobDlog(THPrint hPrint)
{
	OSErr		err;
	Boolean		doIt;
	short		resetThedpi;

// Make sure we're working with a default resolution print record, and open
// the printer driver.  Since (in the final sample) we don't know what resolution
// we'll be setting the printer port to until after the dialogs, it's best to
// use a default dpi handle now so that options (which may be ignored when we
// print at high resolutions) are at least available in case they *can* be used.
//
// What makes things tricky is that different printer drivers support
// different ways of resetting the printer port resolution to 72 dpi.  You 
// always use PrGeneral, but the ImageWriter driver, for example, goes to "true"
// default resolution mode only when you try to set an invalid resolution such
// as 0x0 dpi, and the LaserWriter makes you pass 72x72 to get to 72 dpi.
// Neither respects the other's way of doing things.  (For an example, try setting
// the ImageWriter's print handle to 72x72 dpi with PrGeneral, and then doing a
// style dialog with that.)
// 
// One solution which I've found to work well is to first try to set the port to
// 72x72 dpi, and then try to set it to 0x0 dpi.  Drivers which reset when passed
// 72x72 ignore the 0x0 setting, and drivers that need to be passed 0x0, are happy
// as well.  Remember that the order is important here.  If you tell the IW driver
// that you want 72x72 dpi, it will disable options in the print dialog that
// shouldn't be.  Only when it tries to set the bogus 0x0 dpi will it clean house.
//
// Unfortunately, this is one of the design short-comings of the current
// printing architecture, or more specifically, the current printer drivers.
// If drivers don't disable options at different resolutions, we have no
// problem at all.

	resetThedpi = 72;				// Pass 72x72 dpi to force the defaults.
	SetPrinterPortDPI(hPrint, &resetThedpi, &resetThedpi);

	resetThedpi = 0;				// Pass a bogus pair to force the defaults.
	SetPrinterPortDPI(hPrint, &resetThedpi, &resetThedpi);
									// One of those should work.

// Handle the job dialog, close the driver and clean up any windows that
// need updating.

	PrOpen();

	err = PrError();
	
	if (err == fnfErr)	// Here, "file not found" really means "no printer chosen."
		err = noPrinterChosenErr;
	
	if (!err)
	{
		doIt = PrJobDialog(hPrint);
		err = PrError();
		if (!err && !doIt) err = iPrAbort;
	}

	PrClose();
	DoAnyUpdates();
	return err;
}


/************************************************************
  SetupPrintHdl - This routine returns a default print
  handle for the current printer driver.

  The printer driver is expected to be closed upon entry,
  and is therefore opened and closed in this routine.

*************************************************************/

OSErr SetupPrintHdl(THPrint *hPrint)
{
	OSErr err;
	
// Create a handle, open the printer driver, set the
// default values for the handle, and close the driver.
// Be on the lookout for errors.

	*hPrint = (THPrint) NewHandle(sizeof(TPrint));

	if (*hPrint != nil)
	{
		PrOpen();

		err = PrError();
		
		if (err == fnfErr)	// Here, "file not found" really means "no printer chosen."
			err = noPrinterChosenErr;
		
		if (err) return err;

		PrintDefault(*hPrint);
		err = PrError();

		PrClose();
	}
	else
		err = MemError();
	
	return err;
}


/************************************************************
  GetResIDString - This routine creates a unique resource
  ID by adding (the driver version * 10000) and the high
  byte of the wDev.  All values are derived from the print
  handle passed and the currently opened printer driver.
  Once we have the ID, it's converted to a string to use
  for the resource's name.  (Our ID might be 8 digits long,
  so we can't use it as a normal resource ID which is only 5
  digits long.)
  
  Because we call PrDrvrVers, this function must be called
  between PrOpen & PrClose.

*************************************************************/

void GetResIDString(THPrint hPrint, Str255 resIDStr)
{
	short		wDev;
	long		drvrVers, codeID;

	wDev = (*hPrint)->prStl.wDev >> 8;
	drvrVers = PrDrvrVers();
	codeID = (drvrVers * 10000) + wDev;
	NumToString(codeID, resIDStr);
}


/************************************************************
  SavePHdl - This routine saves the passed print handle in
  the indicated resource fork, using a name which identifies
  it uniquely for the printer driver and version being used.
  If there is already a print record resource with this name,
  it's deleted and replaced.

  The printer driver is expected to be closed upon entry,
  and is therefore opened and closed in this routine.

*************************************************************/

OSErr SavePHdl(THPrint hPrint, short resRefNum)
{
	short		savedRefNum, uniqID;
	OSErr		err;
	Handle		oldHdl;
	Str255		resIDStr;

// Save the current resource file refNum, open the printer
// driver, get our unique name for the passed print handle
// and close the driver.

	savedRefNum = CurResFile();
	PrOpen();
	err = PrError();
	
	if (err == fnfErr)	// Here, "file not found" really means "no printer chosen."
		err = noPrinterChosenErr;
	
	if (err) return err;

	if (!err) GetResIDString(hPrint, resIDStr);
	PrClose();
	

// If there's already a resource with this name in the passed
// resource file, delete it.

	UseResFile(resRefNum);
	
	if (!err)
	{
		oldHdl = GetNamedResource(r_printHdlType, resIDStr);

		if (oldHdl)
		{
			RmveResource(oldHdl);
			DisposHandle(oldHdl);
		}

// Save the new resource with a unique ID and our coded name.
	
		uniqID = UniqueID(r_printHdlType);
		AddResource((Handle) hPrint, r_printHdlType, uniqID, resIDStr);
		WriteResource((Handle) hPrint);
		err = ResError();


// Detach the resource, update the resource file and switch back
// to the original resource file.
	

		if (!err)
			DetachResource((Handle) hPrint);
	}
	
	UpdateResFile(resRefNum);
	UseResFile(savedRefNum);
	return err;
}


/************************************************************
  LoadPHdl - This routine loads a print handle resource
  which matches the name we uniquely encode.  If there is
  no print record resource in the indicated resource file
  which matches our uniquely encoded name, the function
  returns resNotFound.
  
  Upon entry, hPrint is expected to be a valid print handle
  or nil.

  The printer driver is expected to be closed upon entry,
  and is therefore opened and closed in this routine.

*************************************************************/

OSErr LoadPHdl(THPrint *hPrint, short resRefNum)
{
	short		savedRefNum;
	OSErr		err;
	Handle		newHdl;
	Str255		resIDStr;

// Save the current resource file refNum, open the printer driver,
// get the name that we would have encoded this resource with and
// close the driver.

	savedRefNum = CurResFile();
	PrOpen();

	err = PrError();
	
	if (err == fnfErr)	// Here, "file not found" really means "no printer chosen."
		err = noPrinterChosenErr;
	
	if (err) return err;

	GetResIDString(*hPrint, resIDStr);
	PrClose();


// Get our resource, if there is one, out of the indicated resource
// file.  If there isn't one, return resNotFound.  Finally, switch
// back to the original resource file.

	UseResFile(resRefNum);
	
	newHdl = GetNamedResource(r_printHdlType, resIDStr);

	if (newHdl)
	{

// Got one.  Detach it from the resource map.  If we previously had
// a print handle, dispose of that before passing this new one back.

		DetachResource(newHdl);
		if (*hPrint) DisposHandle(*(Handle *) hPrint);
		*hPrint = (THPrint) newHdl;
		err = noErr;
	}
	else
		err = resNotFound;
	
	UseResFile(savedRefNum);
	return err;
}


/************************************************************
  ColorBitsProc - This routine is our QuickDraw bottleneck
  replacement for stdBits.  We use this routine when we're
  trying to determine the resolution of a picture based
  solely on the resolution of its contents.

*************************************************************/

pascal void ColorBitsProc(BitMap *srcBits, Rect *srcRect, Rect *dstRect,
						  short mode, RgnHandle maskRgn)
{
    PixMapPtr  pm;

#pragma unused(srcRect, dstRect, mode, maskRgn);

// If this is a PixMap (rowBytes < 0), store the resolution
// and bounds of the PixMap in our global variables.  Since
// BitMaps don't have resolution fields, ignore those.  We
// keep expanding the bounds everytime we find a pixmap in
// case the image was banded (we want gSrcRect to encompass
// all bands).

	if (srcBits->rowBytes < 0)
	{
		pm = (PixMapPtr) srcBits;
		gPMHRes = MAX(gPMHRes, pm->hRes);
		gPMVRes = MAX(gPMVRes, pm->vRes);
		
		if (gRectNotSetYet)
		{
			gSrcRect = pm->bounds;
			gRectNotSetYet = false;
		}
		else
		{
			gSrcRect.top = MIN(gSrcRect.top, pm->bounds.top);
			gSrcRect.left = MIN(gSrcRect.left, pm->bounds.left);
			gSrcRect.bottom = MAX(gSrcRect.bottom, pm->bounds.bottom);
			gSrcRect.right = MAX(gSrcRect.right, pm->bounds.right);
		}
	}
}


/************************************************************
  GetPixMapRes - This routine returns the resolution of the
  contents of a picture (specifically the resolution of its
  PixMap).  We assume that there will be at most one PixMap
  in our image, and given the type of images we're dealing 
  with, that's a valid assumption.  It's not clear how we
  would determine the correct resolution in a situation
  where multiple PixMaps were involved anyway.
  
  This routine is used if the original picture was not
  created with OpenCPicture.  If it was, we can get the
  resolution by just checking the picture header, as in
  GetImageRes below.

*************************************************************/

OSErr GetPixMapRes(WindowInfoHdl windInfo, Fixed *hRes, Fixed *vRes, Rect *srcRect)
{
	OSErr		err = noErr;
	PicHandle	aPict;
	CQDProcs	bottlenecks;
	GrafPtr		oldPort;
	CGrafPort	cPort;
	Boolean		spoolThisPict;

// Assume the image is 72x72 dpi and that the picture's bounds
// are the same as its picFrame.

    spoolThisPict = !(*windInfo)->loaded;
	aPict = (*windInfo)->windowPict;
	gPMHRes = gPMVRes = Long2Fix(72);
	gSrcRect = (*aPict)->picFrame;
	gRectNotSetYet = true;


// Save off the current port, and open up a new color
// port to work in.  We do this because we *must* be drawing
// into a CGrafPort for us to have PixMaps passed to our
// bottleneck routine without being converted into BitMaps.

	GetPort(&oldPort);
	OpenCPort(&cPort);


// Get the standard CGrafPort bottleneck routines, and add
// our replacement for stdBits.  This replacement routine
// will snoop for a PixMap, and store the resolution and
// bounds of any it finds in our global variables.  If
// none are found, we'll still have the defaults set up
// above.

	if (spoolThisPict)
		err = SpoolInstall(windInfo, kArrowCursor);
	else
	{
		SetStdCProcs(&bottlenecks);
		cPort.grafProcs = &bottlenecks;
	}

	if (!err)
	{
		cPort.grafProcs->bitsProc = (Ptr) ColorBitsProc;


// Draw the picture so our bottleneck routine is called.
// Afterwards, set back to the old port we were using and
// Close the CGrafPort we created.  Finally, return the
// resolution and bounds rectangle for this image.

		DrawPicture(aPict, &(*aPict)->picFrame);

		if (spoolThisPict)
			err = SpoolRemove();
	}

	SetPort(oldPort);
	CloseCPort(&cPort);

	*hRes = gPMHRes;
	*vRes = gPMVRes;
	*srcRect = gSrcRect;

	return err;
}


/************************************************************
  GetImageRes - This routine returns the native resolution
  of a picture, and the appropriate drawing rectangle for
  the image at that resolution.

*************************************************************/

OSErr GetImageRes(WindowInfoHdl windInfo, short *ixDPI, short *iyDPI, Rect *srcRect)
{
	PicHandle			wPICT;
	Fixed				hRes, vRes;
	PartialPicture		partialPict;
	Boolean				extndV2PICT = false;
	OSErr				err = noErr;

	wPICT = (*windInfo)->windowPict;
	HLock((Handle) windInfo);
	BlockMove(&(*windInfo)->partialPict, &partialPict, sizeof(PartialPicture));
	HUnlock((Handle) windInfo);

// If we have a version 2 picture, find out if it's an extended
// version 2 picture.  If so, the version field in the picture's
// header will be -2.  In that case, pull the resolution and bounding
// rect out of the picture header.

	if (partialPict.pictVersion == 2)
		if (partialPict.pictHeader.version == -2)
		{
			extndV2PICT = true;

			hRes = partialPict.pictHeader.hRes;
			vRes = partialPict.pictHeader.vRes;
			*srcRect = partialPict.pictHeader.srcRect;
		}

// If we have a version 1 picture, or a non-extended version 2
// picture, set the resolution and bounding rect for the
// picture based on the PixMap it contains.

	if (!extndV2PICT)
		err = GetPixMapRes(windInfo, &hRes, &vRes, srcRect);

	*ixDPI = FixRound(hRes);
	*iyDPI = FixRound(vRes);
	
	return err;
}


/************************************************************
  SimulateHalftone - This routine creates a simulated
  halftone image by converting the passed GWorld into a
  1-bit representation using 4x4 pixel values to represent
  each pixel's color.
  
  The source pixmap is expected to by 4 bits deep, using the
  standard gray color table for that depth as obtained from
  GetCTable(32 +4).
  
  htGWorld is created and contains the image that's created
  upon exit.

*************************************************************/

OSErr SimulateHalftone(Rect *srcRect, GWorldPtr srcGWorld, GWorldPtr *htGWorld)
{
	register Ptr			srcByte, destByte, prevSrc, prevDest;
	register long			rowOffset[5];
	register short			col, row, pixRow;
	QDErr					err;
	PixMapHandle			srcPix, destPix;
	Rect					hToneRect;
	short					aPixel, rowBytes1bit, rowBytes4bit;
	short					lastRow, lastCol;
	signed char				mode;
	GWorldPtr				curGWorld;
	GDHandle				curGDH;

// The grid below corresponds to figure 6 in the "Adventures in Color Printing"
// article in develop #13.
//
// The patterns are stored in an optimized pattern as such:
//
// Every pixel will have a value from 0 to 15.  This corresponds to a
// column of the pixPattern matrix.
//
// To figure out which pixels in the 4 x 4 pattern to turn on, go to the row
// representing the pixel's value.  Since every pixel will be stored in 4 bits,
// our code must handle high or low nibble setting.  We do this by storing the
// value for an entire nibble in an entry of our pixPattern matrix.
// 
// For example, the pattern for a nibble with an pixel value of 5 (the 6th
// column from the left of the table) will be:
//
//              0    -------- (in binary) ------->   0 0 0 0   (A)
//              6    -------- (in binary) ------->   0 1 1 0   (B)
//              6    -------- (in binary) ------->   0 1 1 0   (C)
//              4    -------- (in binary) ------->   0 1 0 0   (D)
//
// which agrees with the entry for a pixle value of 5 shown in figure 6 of
// the article.  We store nibble (A) in row 1 of the current pattern,
// nibble (B) in row 2, (c) in row 3 and (D) in row 4.  This way we set up
// an entire 4x4 pattern by only storing 4 values.  This method may seem
// obscure, but it lets us quickly calculate and set the different pixel
// values of every nibble in the image.

	char		pixPattern[4][16] =
				{
					0, 0, 0, 0, 0, 0, 0, 2,  2,  2,  2,  3,  3,  3,  3, 15,
					0, 0, 0, 2, 6, 6, 6, 6, 14, 14, 15, 15, 15, 15, 15, 15,
					0, 4, 6, 6, 6, 6, 7, 7,  7,  7,  7,  7, 15, 15, 15, 15,
					0, 0, 0, 0, 0, 4, 4, 4,  4,  6,  6,  6,  6, 14, 15, 15
				};

// Save the current GWorld/GDevice so that we can restore them when we
// leave.  Create a 1-bit GWorld for our source image and get and lock
// our source and destination PixMaps, since we'll be walking them pixel
// by pixel.  Also, start our cursor spinning so the user knows we're
// alive.

	GetGWorld(&curGWorld, &curGDH);

	err = NewGWorld(htGWorld, 1, srcRect, nil, nil, 0);
	if (err) return err;
	
	srcPix = GetGWPixMap(srcGWorld);
	destPix = GetGWPixMap(*htGWorld);
	LockPixels(srcPix);
	LockPixels(destPix);
	
	SetUpCursors(r_stdSpinCursor, watchCursor);


// Erase our halftone GWorld so that we can skip over pixel values of zero,
// without having to draw an all white pattern.  This can improve our speed.

	SetGWorld(*htGWorld, nil);
	EraseRect(srcRect);
	SetGWorld(curGWorld, curGDH);
	

// By dividing the source rectangle by 4 in each direction, we can determine
// how many rows and columns of patterns we need to lay down.  Note that
// before we entered this routine (in DoHT0) we made sure that the bounds
// were evenly divisible by 4.  Also, we made sure that srcRect had its
// top-left corner at (0,0).

	hToneRect = *srcRect;
	hToneRect.right >>= 2;
	hToneRect.bottom >>= 2;
	lastRow = hToneRect.bottom;
	lastCol = hToneRect.right;
	

// Store the rowByte values of our source and destination GWorlds, masking
// off the high bit to get the true rowBytes4bit values.  While we're at it
// calculate the values we need to add to a byte position in the source
// GWorld's pixmap to get to the 4 different rows of the destination
// pixmap.

	rowBytes4bit = (*srcPix)->rowBytes  & 0x7FFF;
	rowBytes1bit = (*destPix)->rowBytes & 0x7FFF;

	for (pixRow = 0; pixRow <= 4; pixRow++)
		rowOffset[pixRow] = (long) pixRow * (long) rowBytes1bit;


// Store the baseAddresses of our source and destination PixMaps and begin
// looping through the pixels.  On each pass we check to see if the user
// has become impatient and pressed command=period.  Also, because we may
// be accessing bytes in a 32-bit pointer (for example if the 824 GC card
// is being used, we MUST insure that we're in 32-bit addressing mode before
// using the baseAddr pointers.  Since it would be a bad thing to make
// system calls (such as those in CommandPeriod) if we've changed the MMU
// mode, we set and restore it around the code that accesses the baseAddr
// pointers.  This way, the system is always set up correctly when we go
// to CommandPeriod.

	srcByte  = GetPixBaseAddr(srcPix);
	destByte = GetPixBaseAddr(destPix);

	for (row = 0; ((row < lastRow) && !(err = CommandPeriod())); row++)
	{
		BumpCursor();						// Spin the cursor.

		mode = true32b;						// Make sure we're in 32-bit addressing
		SwapMMUMode(&mode);					// mode before touching those pixels!


// Store the previous (current) pointers so that we can easily update the
// new pointers when we get to the end of this row.  For each byte in our
// source PixMap, we handle the high and low nibbles separately to keep
// things simple and efficient.
//
// If (col & 1) == 0, we're working on an even pixel, whose value will be
// stored in the high nibble of a byte.  We grab the value of the pixel and
// store it in aPixel.  Note that we can just index off of the pixel value
// instead of tracing back to the color table since the color table we used
// is a linear ramp of grays.  Index into the pixPattern matrix and set the
// current destination nibble and the 3 underneath that to the values found
// in the table.  This sets 4x4 pixels in the destination pixmap with only 4
// storage operations, and completes the storage of the halftone pattern for
// the pixel.
// 
// If we're working on an odd pixel ((col & 1) == 1) then we do everything
// the same, except we work on the low order nibble of the source and we
// make sure that we don't destroy the high order nibbles of our destination
// bytes.  After every odd pixel (low order nibble) we increment our byte
// pointers.

		prevSrc  = srcByte;
		prevDest = destByte;
		
		for (col = 0; col < lastCol; col++)
			if ((col & 1) == 0)					// even pixel.
			{
				aPixel = (*srcByte & 0xF0) >> 4;

				if (aPixel)
					for (pixRow = 0; pixRow < 4; pixRow++)
						*(Ptr) ((long) destByte +
								rowOffset[pixRow]) = pixPattern[pixRow][aPixel] << 4;
			}
			else								// odd pixel.
			{
				aPixel = *srcByte & 0x0F;

				if (aPixel)
					for (pixRow = 0; pixRow < 4; pixRow++)
						*(Ptr) ((long) destByte +
								rowOffset[pixRow]) |= pixPattern[pixRow][aPixel];

				++srcByte;
				++destByte;
			}

// Update our source pointer to point to the next row, and our destination pointer
// to point 4 rows lower.  Remember that every pixel in the destination PixMap is
// actually 4 pixels high and 4 pixels wide, and this is sort of like moving a 4
// pixel thick pen to the next row to draw in.  We restore the old MMU mode so that
// CommandPeriod doesn't die a horrid death when the system was booted in 24-bit mode.

		srcByte  = rowBytes4bit + prevSrc;
		destByte = rowOffset[4] + prevDest;
		
		SwapMMUMode(&mode);						// Go back to native addressing mode.
	}

// Unlock all our PixMaps, restore the current GWorld/GDevice, tear down our spinning
// cursors, and take a hike.

	UnlockPixels(srcPix);
	UnlockPixels(destPix);

	SetGWorld(curGWorld, curGDH);
	
	TearDownCursors();

	return err;
}


/************************************************************
  BandPicture - This routine bands the passed picture into
  the destination GWorld, using ditherCopy if desired.
  
  wp is the window containing the picture.  It has one of
  our WindowInfoHdl handles hanging off the window refCon
  field.  From that structure, we can either locate the
  picture's handle in memory or spool the picture in from
  disk.

*************************************************************/

OSErr BandPicture(WindowPtr	wp, Rect *srcRect, GWorldPtr destGWorld,
				  Boolean useDitherCopy)
{
	OSErr			err = noErr;
	Rect			pictFrame, bandFrame, ditherFrame, ditherRect1, ditherRect2;
	GWorldPtr		curGWorld, bandGWorld = nil, ditherGWorld = nil;
	PixMapHandle	srcPix, destPix, ditherPix;
	PicHandle		srcPict;
	Boolean			moreToDo, spoolThisPict, useSecondGWorld;
	GDHandle		curGDH;
	WindowInfoHdl	windInfo;
	long			bandHeight, bandWidthInBytes, maxBlockSize;
	short			bitsPerPixel, rowWidth, padAmt;

	windInfo = (WindowInfoHdl) GetWRefCon(wp);
	srcPict = (*windInfo)->windowPict;
	spoolThisPict = !(*windInfo)->loaded;
	useSecondGWorld = useDitherCopy;


// Save the current GWorld for restoration later, then lock down the destination
// GWorld's bits and erase them so that no garbage is left in the GWorld.  Lastly,
// get the picture's bounds, putting the top-left corner at (0, 0).

	GetGWorld(&curGWorld, &curGDH);

	destPix = GetGWPixMap(destGWorld);
	LockPixels(destPix);
	
	SetGWorld(destGWorld, nil);
	EraseRect(&(*destPix)->bounds);
	SetGWorld(curGWorld, curGDH);

	pictFrame = *srcRect;
	OffsetRect(&pictFrame, -pictFrame.left, -pictFrame.top);

// Determine the size bands to use.  We first get the maximum contiguous block 
// possible in our heap.  Taking 85% of that value, we figure out how many
// horizontal bands could fit in that.
//
// If we're dithering, we'll need an extra GWorld for dither-distortion correction.
// The process we'll use will require a "pad" of a few extra rows of pixels in
// each GWorld (padAmt).  Therefore, our calculated band size will be:
//
// for srcCopy:    (#of bytes available / number of bytes per row) -padAmt
//
// Remember that the number of bytes per row will be rounded up to the nearest
// long word, since 32-bit QuickDraw makes GWorlds which are 32-bit aligned.
//
// For ditherCopy we do the same calculations as above, but we adjust bitsPerPixel to
// reflect the sum of the pixels in the 32 bit GWorld and those in the additional dither
// GWorld.
//
// If our calculations come up with a band size that's bigger than the original image,
// then just use the height of the original image.

	bandFrame = (*destPix)->bounds;

	padAmt = (useDitherCopy)? kPadAmount: 0;
	maxBlockSize = MaxBlock() * .85;
	bitsPerPixel = (useDitherCopy)? 32 + (*destPix)->pixelSize: 32;
	rowWidth = (bandFrame.right - bandFrame.left +1);
	bandWidthInBytes = (bitsPerPixel * rowWidth)/8;
	bandWidthInBytes += bandWidthInBytes % 4;		// Make sure we're on a LW boundary.

	bandHeight = (maxBlockSize / bandWidthInBytes) -padAmt;
	
	if ((bandHeight +padAmt) >= (bandFrame.bottom - bandFrame.top))
	{
		useSecondGWorld = false;
		padAmt = 0;
		bandHeight = (bandFrame.bottom - bandFrame.top);
	}

// Create the rectangle for the banding GWorld.  The band buffer will be as wide as
// the original image and as tall as either the available band size or the original
// image, whichever is smaller.

	SetRect(&bandFrame, 0, -padAmt, bandFrame.right, bandHeight +padAmt);


// If we're dithering, and using the second GWorld, create the dither GWorld which
// we use to filter our some of the effects of ditherer banding.  This GWorld will
// mirror the destination, except for the bounds we use.

	if (useSecondGWorld)
	{
		ditherFrame = bandFrame;
		err = NewGWorld(&ditherGWorld, (*destPix)->pixelSize, &ditherFrame, (*destPix)->pmTable,
						GetGWorldDevice(destGWorld), 0);
	}


// While there's more to do, band our image.  On every pass we create a 32-bit GWorld
// to band the image through, which we dispose of at the end of the loop.  (UpdateGWorld
// is not used because its behaviour is sometimes questionable when we're only changing
// the bounds of a GWorld.)  After locking down the pixels, we set this as the
// current GWorld and draw the picture into it, clipping to the GWorld's bounds.  This
// creates a band which we then CopyBits to the destination GWorld, or pass through
// ditherGWorld if we're dithering and banding.
//
// Of particular importance are the rectangles we're using:
//
// When we cut the bands out of the picture, we call DrawPicture using pictFrame, and
// clipping to bandFrame.  If we used bandFrame in DrawPicture, every band would be a
// squished up representation of the original image, which obviously isn't what we want.
// bandFrame is updated on each pass so that it drops one band lower, and the DrawPicture
// call copies the next logical band to the 32-bit GWorld.
//
// When dithering and banding, the padAmt is non-zero, and useSecondGWorld is true.  In
// this case, we draw the picture into our 32-bit GWorld and then dither from there to
// our ditherGWorld.  Finally, we srcCopy that image to the destination GWorld, using
// a rectangle that leaves out the top padAmt rows of the dithered image.  By doing this,
// we negate some of the effects that dithering has on the fringes of our bands.  We
// still end up copying all the rows, but we don't copy the pixels that are dithered
// on the top fringe of ditherGWorld between consecutive bands.
//
// We also check for command-period presses while we do this, in case the user gets
// bored and wants to cancel.

	moreToDo = true;
	
	if (!err)
	do
	{
		err = NewGWorld(&bandGWorld, 32, &bandFrame, nil, nil, 0);
		if (!err)
		{
			srcPix = GetGWPixMap(bandGWorld);
			LockPixels(srcPix);

			SetGWorld(bandGWorld, nil);
			ClipRect(&bandFrame);


// If we need to spool this image, install the spooling bottleneck procedure.
 
 			if (spoolThisPict)
				err = SpoolInstall(windInfo, r_stdCurs4);

			if (!err)
			{
				DrawPicture(srcPict, &pictFrame);


// If we spooled this image, remove the spooling bottleneck procedure.
 
				if (spoolThisPict)
					err = SpoolRemove();

				SetGWorld(destGWorld, nil);
				
				if (useSecondGWorld)
				{
					ditherPix = GetGWPixMap(ditherGWorld);
					LockPixels(ditherPix);

					CopyBits((BitMap *) *srcPix, (BitMap *) *ditherPix, &bandFrame,
							 &ditherFrame, ditherCopy, nil);

					ditherRect1 = ditherFrame;	// Offset the rectangles so we don't
					ditherRect1.top += padAmt;	// copy the topmost dithered pixels.
					ditherRect2 = bandFrame;
					ditherRect2.top += padAmt;
	
					CopyBits((BitMap *) *ditherPix, (BitMap *) *destPix, &ditherRect1,
							 &ditherRect2, srcCopy, nil);

					UnlockPixels(ditherPix);
				}
				else
					CopyBits((BitMap *) *srcPix, (BitMap *) *destPix, &bandFrame,
							 &bandFrame, ((useDitherCopy)? ditherCopy: srcCopy), nil);
				
				UnlockPixels(srcPix);
			
				if (useSecondGWorld)
					moreToDo =  (ditherRect2.bottom < pictFrame.bottom);
				else
					moreToDo =  (bandFrame.bottom < pictFrame.bottom);
				
				OffsetRect(&bandFrame, 0, padAmt + MIN(bandHeight, pictFrame.bottom - bandFrame.bottom));
				DisposeGWorld(bandGWorld);
				bandGWorld = nil;
	
				SetGWorld(curGWorld, curGDH);
				if (!err) err = CommandPeriod();
			}
		}
	}
	while (moreToDo && (!err));


// Unlock the destination bits, set the current GWorld to what we started with, and
// dispose of the GWorlds.  Return any errors.

	UnlockPixels(destPix);

	SetGWorld(curGWorld, curGDH);
	if (bandGWorld) DisposeGWorld(bandGWorld);
	if (ditherGWorld) DisposeGWorld(ditherGWorld);

	return err;
}


/************************************************************
  CenterRect - This routine centers srcRect within destRect
  and returns it in ctrRect.  It's mainly used for centering
  output on the printed page and for centering windows on
  the screen.  The source rectangle should not be larger
  than the destination rectangle in either direction.  If
  it is, the rectangle will not be adjusted in that
  direction.

*************************************************************/

void CenterRect(Rect *srcRect, Rect *destRect, Rect *ctrRect)
{
	Rect	srcR, dstR;
	short	srcWidth, srcHeight, destWidth, destHeight, dx, dy;

	srcR = *srcRect;
	dstR = *destRect;

	OffsetRect(&srcR, dstR.left - srcR.left, dstR.top - srcR.top);

	srcWidth = srcR.right - srcR.left +1;
	srcHeight = srcR.bottom - srcR.top +1;
	destWidth = dstR.right - dstR.left +1;
	destHeight = dstR.bottom - dstR.top +1;
	
	dx = MAX(0, (destWidth-srcWidth)/2);
	dy = MAX(0, (destHeight-srcHeight)/2);

	*ctrRect = srcR;
	OffsetRect(ctrRect, dx, dy);
}


/************************************************************
  IdleProc - This routine is a PrIdleProcPtr that spins our
  cursor and looks for command-period presses during
  printing.

*************************************************************/

pascal void IdleProc()
{
	OSErr		err;
	
	BumpCursor();
	if (err = CommandPeriod()) PrSetError(err);
}


/************************************************************
  CommandPeriod - This routine checks to see if the user has
  pressed command-period.  The Printing Manager does this
  for us, but a routine like this is helpful to allow users
  to cancel in the middle of a rendering routine, rather
  than only Printing Manager operations.
  
  If there's a command-period in the event queue, the
  routine returns iPrAbort just like the Printing Manager.
  Otherwise, it returns noErr.

*************************************************************/

OSErr CommandPeriod()
{
	Boolean		cancel = false;
	EventRecord	evtRec;
	
	while (!cancel && (WaitNextEvent(keyDownMask | autoKeyMask, &evtRec, 0, nil)))
		cancel = (evtRec.modifiers & cmdKey) &&
				 ((evtRec.message & charCodeMask) == '.');

	return ((cancel)? iPrAbort: noErr);
}
