/*********************************************************************

	HT develop 9 Halftone.c
	
	This file contains the halftone sample code written by Dan Lipton
	for the article "Making the Most of Color on 1-bit Devices," by
	Dan Lipton and Konstantin Othmer.  The article appears in Apple's
	develop magazine, issue #9, and the sample code can be found on
	the Developer CD series.


			>>  Subtle changes from develop #9 code  <<
	
	HalftonePixMap now caches luminance values for indexed PixMaps,
	therefore the routine LUMVAL is only called for direct PixMaps.
	However, this sample always passes 32-bit data, so this change
	turns out to be inconsequential.
	
	Added command-period exit and GetGWPixMap calls to
	HalftonePixMap so that we work well under system 6 or 7.
	
	HalftonePixMap returns the halftoned image in a GWorld now,
	not a PicHandle.  This helps keep the memory requirements down.

	Reformatted to keep consistency between routines in other
	source files.
	
	Dave Hersey, Apple Developer Technical Support.
	
	 Edit Trail 

	Incorporated code and did the above.		- dmh  11/15/92

	Added SwapMMUModes to make sure that we
	only access the PixMap data directly in
	32-bit addressing mode.						- dmh  11/21/92

	Added spinning cursor stuff.				- dmh  11/21/92
	finalized and sync'd with other samples.	- dmh  11/28/92

*********************************************************************/

#pragma segment Printing

#include "HT Headers.h"


/************************************************************
  LUMVAL - This routine returns the luminance for a pixel of
  depth 8 or 32 bits.
	
  pPixel points to the pixel in the PixMap.
  pMap is the Pixmap's PixMapPtr.

  For more information, see the article "Making the Most of
  Color on 1-bit Devices," by Dan Lipton and Konstantin
  Othmer which appears in develop issue #9.

*************************************************************/

long LUMVAL(Ptr pPixel, PixMapPtr pMap)
{
	long	red, green, blue;

	if (pMap->pixelSize == 32)
	{
		red = (long)(unsigned char)*(++pPixel);				// Skip alpha, get red
		green = (long)(unsigned char)*(++pPixel);			// get green,
		blue = (long)(unsigned char)*(++pPixel);			// get blue.
		
		return ((30 * red + 59 * green + 11 * blue) /100);
	
	}
	else
		if (pMap->pixelSize == 8)
		{
			RGBColor	*theColor;

			theColor = &((*(pMap->pmTable))->ctTable[(unsigned char)*pPixel].rgb);
			
			return ((30 * (theColor->red >> 8) +
					 59 * (theColor->green >> 8) +
					 11 * (theColor->blue >> 8)) /100);
		}
}


/************************************************************
  HalftonePixMap - This routine takes a pixmap and halftones
  it at the passed destination resolution into a 1-bit
  GWorld, which is returned to the caller.

  Note, if destination resolution is less than 72, source
  scaling code does not work.  It assumes target resolution
  is always >= 72 so pointer arithmetic in source image
  doesn't need to do a multiply every time.  This could be
  fixed at the expense of performance, but in this sample
  the destination (printer) resolution will always be >= 72.
	
  For more information, see the article "Making the Most of
  Color on 1-bit Devices," by Dan Lipton and Konstantin
  Othmer which appears in develop issue #9.

*************************************************************/

OSErr HalftonePixMap(GWorldPtr srcGWorld, GWorldPtr *destGWorld,
					 short srcDPI, short destDPI)
{
	QDErr							err;
	register unsigned long			mask, destLong;
	short							widthAtDestDPI, heightAtDestDPI, lastX, lastY,
									destX, destY, lumins[256], idx;

	Rect							boundsAtDestDPI;
	Fixed							scale, sourceX, sourceY;
	Ptr								pRaster, pPixel, pDestRaster;
	long							*nextDestLong, luminance,
									*pMatrix, *pMatrixRow, M = 8, N = 8, I, J;

	long			   matrix[8][8] = { 7,  111, 183, 239, 231, 191, 71, 15, 
										87, 47, 135, 175, 199, 159, 55, 127, 
										215, 143, 39, 79, 119, 63, 151, 207, 
										247, 223, 103, 31, 23, 95, 167, 255 ,
										231, 191, 71, 15, 7, 111, 183, 239 ,
										199, 159, 55, 127, 87, 47, 135, 175, 
										119, 63, 151, 207, 215, 143, 39, 79, 
										23, 95, 167, 255, 247, 223, 103, 31
									   }; 

	Boolean							directPMap;
	RGBColor						*anRGB;
	PixMap							*pSource;
	PixMapHandle					offScreen, hSource;
	signed char						mode;
	
// If our destination resolution is < 72 dpi, or our source PixMap is not 8 or
// 32-bits deep, return an error.  Otherwise, get the source PixMap, lock its
// pixels and PixMapHandle down and get a pointer to the PixMap.

	if (destDPI < 72)
		return paramErr;

	hSource = GetGWPixMap(srcGWorld);
	
	if ((*hSource)->pixelSize != 8 && (*hSource)->pixelSize != 32)
		return paramErr;
	
	LockPixels(hSource);
	HLock((Handle) hSource);
	pSource = (PixMap *) StripAddress(*hSource);


// If we're not using a direct PixMap, cache the luminance values of all entries
// in the PixMap's color table.  This can speed up luminance calculations for
// indexed PixMaps.

	directPMap = (pSource->pixelType == RGBDirect);
	
	if (!directPMap)
		for (idx = 0; idx <= (*(pSource->pmTable))->ctSize; idx++)
		{
			anRGB = &(*(pSource->pmTable))->ctTable[ (unsigned char) idx].rgb;
			lumins[idx] = (30 * (anRGB->red   >> 8) + 59 *
								(anRGB->green >> 8) + 11 *
								(anRGB->blue  >> 8)) /100;
		}

// Compute our halftoned image's width and bounds (which will be the source image's
// bounds at the destination resolution.

	scale = FixRatio(destDPI, srcDPI);

	widthAtDestDPI  = FixMul(scale, (long) (pSource->bounds.right -
							 pSource->bounds.left) << 16) >>16;
	heightAtDestDPI = FixMul(scale, (long) (pSource->bounds.bottom -
							 pSource->bounds.top) << 16) >>16;

	boundsAtDestDPI.top    = pSource->bounds.top;
	boundsAtDestDPI.left   = pSource->bounds.left;
	boundsAtDestDPI.bottom = boundsAtDestDPI.top + heightAtDestDPI;
	boundsAtDestDPI.right  = boundsAtDestDPI.left + widthAtDestDPI;


// Allocate a 1-bit GWorld to create the halftoned image in.  If we succeed,
// get the GWorld's PixMap, lock it down, and set up our spinning cursors.

	err = NewGWorld(destGWorld, 1, &boundsAtDestDPI, nil, nil, 0);

	if (err)
	{
		UnlockPixels(hSource);
		HUnlock((Handle) hSource);
		return err;
	}
	
	offScreen = GetGWPixMap(*destGWorld);
	LockPixels(offScreen);

	SetUpCursors(r_stdSpinCursor, watchCursor);


// Compute the scale for walking source pixels.  sourceX, sourceY will walk source
// image in fixed-point numbers.  Integer part will be saved to see when we've crossed
// into a new pixel boundary for updating pointers and luminance value.

	scale = FixRatio(srcDPI, destDPI);
	sourceY = pSource->bounds.top << 16;
	lastY = sourceY >> 16;
	

// Initialize raster pointers.

	pRaster =  GetPixBaseAddr(hSource);
	pDestRaster = GetPixBaseAddr(offScreen);
	
	I = 0;								// Y index into dither matrix.
	pMatrixRow = &(matrix[0][0]);		// Point to next row of matrix.


// Begin halftoning.  We walk the source image's rows for every row in the
// destination PixMap.

	for (destY = 0; !err && (destY < heightAtDestDPI); ++destY)
	{
		BumpCursor();					// Spin our cursor.

		mode = true32b;					// Make sure we're in 32-bit addressing
		SwapMMUMode(&mode);				// mode before touching those pixels!

		pPixel = pRaster;
		luminance = (directPMap)? LUMVAL(pPixel, pSource):
								  lumins[(unsigned char) *pPixel];
		mask = 0x80000000;
		destLong = 0;
		nextDestLong = (long*) pDestRaster;

		sourceX = pSource->bounds.left << 16;
		lastX = sourceX >> 16;

		J = 0;							// X index into dither matrix.
		pMatrix = pMatrixRow;			// Point into current row of dither matrix.


// Walk the source image's columns for every column in the destination
// PixMap.

		for (destX = 0; destX < widthAtDestDPI; ++destX)
		{

// If the luminance for this pixel is less than matrix value, turn on the pixel. 

			if (luminance < *pMatrix)
				destLong |= mask;
				
// Adjust masks and longs for next device pixel.  Pixels are set in a long-word
// register. When all 32 bits in the register have been set, the register is
// laid into the PixMap.
			
			mask >>= 1;

			if (mask == 0)				// If we've done 32 pixels, lay the long word
			{							// into the offscreen.
			
				mask = 0x80000000;		// Reset the mask and increment the destination
										// pointer.

				*(nextDestLong++) = destLong;
				destLong = 0;
			}


// Update the pointer into the dither matrix: Instead of using matrix[X % M, Y % N],
// we use a pointer and counters. A Mod function combined with an array index
// calculation would be more time-consuming.
			
			++pMatrix;					// Next element in row.

			if (++J >= N)
			{							// Get J+1 mod N, faster than
										// doing real mod every pixel.
				J = 0;
				pMatrix = pMatrixRow;	// Start at beginning of matrix row again.
			}

			sourceX += scale;			// Add fixed scale increment to source X value.


// Check to see if we crossed into the next source pixel.  Note: Getting integer
// portion can be done faster than a shift; this is left as an exercise for the
// reader.

			if ((sourceX >> 16) > lastX)
			{
				lastX = sourceX >> 16;	// Get new integer portion of X coordinate,
										// point to the next pixel, and get the
										// pixel's luminance value.

				if (pSource->pixelSize == 8)
					++pPixel;			// 8-bit source image.
				else
					pPixel += 4;		// 32-bit source image.
				
				luminance = (directPMap)? LUMVAL(pPixel, pSource):
										  lumins[(unsigned char) *pPixel];
			}
		}
	
// Make sure the last long-word was laid into the target image.  If the 1 bit in the
// mask is not all the way left, it means there are pixels in the next long, but not
// enough to force the write within the raster loop, so lay it into the block.

		if (mask != 0x80000000)
			*nextDestLong = destLong;


// Move to the next destination raster, add the fixed scale increment to source Y
// value, and check to see if we crossed into next source raster.

		pDestRaster += ((*offScreen)->rowBytes & 0x7FFF);
		sourceY += scale;

		if ((sourceY >> 16) > lastY)
		{
			lastY = sourceY >> 16;		// Get the integer portion of Y coordinate.

			pRaster += (pSource->rowBytes & 0x7FFF);	// Get the next source raster.
		
		}

// Update our pointer into the dither matrix.  At the end of a row of halftoning,
// check to see if the user has pressed command-period.
		
		pMatrixRow += M;				// Point to the next row of the matrix.
		if (++I >= M)
		{								// Get I+1 mod M, faster than doing a real
										// mod every raster.
			I = 0;
			pMatrixRow = &(matrix[0][0]);	// Point to next row of matrix.
		}
		
		SwapMMUMode(&mode);				// Put us back to the native addressing mode
										// so we can call CommandPeriod.
		err = CommandPeriod();
	 }

// Unlock our source and destination PixMaps, stop our spinning cursors and return
// any error we turned up.

	HUnlock((Handle) hSource);
	UnlockPixels(hSource);
	UnlockPixels(offScreen);

	TearDownCursors();
	return err;	
}

