/*********************************************************************

	Final Color Printing.c
	
	This file contains all of the color/grayscale-specific printing
	functions for the all-in-one "Adventures in Color Printing"
	sample, "The Final Adventure."
	
	Dave Hersey, Apple Developer Technical Support.
	
	 Edit Trail 
	
	creation:									11/24/92  - dmh
	sync'd with other samples:					11/30/92  - dmh
	
*********************************************************************/

#pragma segment Printing

#include "Final Headers.h"


/************************************************************
  DoColorPrint - This routine starts the printing process for
  the indicated window.  We determine the bounds to use when
  printing the image, set the printer port's resolution,
  display the job dialog and so forth.  If everything
  succeeds, we pass control on to the DoColorPrintLoop
  routine.
  
  This function is a minimally-modified version of the
  "Color Adventures" sample's PrintOne routine.

*************************************************************/

OSErr DoColorPrint(WindowPtr whichWindow, Boolean useBanding)
{
	QDErr				err;
	GWorldPtr			savedGW;
	GDHandle			savedGDH;
	short				pxDPI, pyDPI, ixDPI, iyDPI;
	Boolean				useDrawPicture;
	DialogPtr			printStatDlog = nil;
	Rect				srcRect, destRect;
	Str255				tempStr, orgTitle;
	Fixed				scale;
	PicHandle			wPICT;
	WindowInfoHdl		windInfo;

// Save the current GWorld/GDevice so that we can restore it later.
// See if we're supposed to use DrawPicture, and retrieve the
// picture to print from our private "WindowInfoHdl" structure
// that's hanging off the window's refCon.

	GetGWorld(&savedGW, &savedGDH);

	useDrawPicture = !useBanding;

	windInfo = (WindowInfoHdl) GetWRefCon(whichWindow);
	wPICT = (*windInfo)->windowPict;


// Get the picture's resolution and set the printer port's resolution
// as close to that as possible.  In case a perfect match isn't
// available, we indicate that 300 x 300 is the highest resolution
// that the GetBestDPI routine should substitute.  Set the printer
// port to the resolution we selected.

	GetImageRes(windInfo, &ixDPI, &iyDPI, &srcRect);
	pxDPI = ixDPI;
	pyDPI = iyDPI;
	GetBestDPI(&pxDPI, &pyDPI, 300, 300, true);
	SetPrinterPortDPI(ghPrint, &pxDPI, &pyDPI);


// destRect will be the destination rectangle for this image.  First, set
// it to the native resolution bounds we found from GetImageRes.  Then,
// align the top-left corner of the rectangle to (0,0) and scale the area
// to the printer port's resolution, if necessary.

	destRect = srcRect;
	OffsetRect(&destRect, -destRect.left, -destRect.top);

	if ((pxDPI != ixDPI) || (pyDPI != iyDPI))
	{
		scale = FixRatio(pxDPI, ixDPI);
		destRect.right = FixMul(scale, (long) destRect.right << 16) >> 16;
		scale = FixRatio(pyDPI, iyDPI);
		destRect.bottom = FixMul(scale, (long) destRect.bottom << 16) >> 16;
	}

// Display a status dialog, jump to the print loop routine, and hide the
// status dialog.  Finally, restore the current GWorld to what it was
// when we entered, and return.

	if (useDrawPicture)
		sprintf(tempStr, "%s %i %s %i %s", " at", pxDPI, "dpi horizontally by",
				pyDPI, "dpi vertically.  (Using DrawPicture.)");
	else
		sprintf(tempStr, "%s %i %s %i %s", " at", pxDPI, "dpi horizontally by",
				pyDPI, "dpi vertically.  (Using banding.)");

	GetWTitle(whichWindow, orgTitle);
	ShowStatusMsg(&printStatDlog, "Printing ", orgTitle, tempStr, true);
	err = DoColorPrintLoop(ghPrint, whichWindow, &destRect, useDrawPicture);
	HideStatusMsg(&printStatDlog);

	SetGWorld(savedGW, savedGDH);
	return err;
}


/************************************************************
  DoColorPrintLoop - This routine does the actual printing.
  The picture to be printed is attached to the window passed
  (through our WindowInfoHdl structure).
  
  If useDrawPicture is true, we simply print using
  DrawPicture, otherwise we print by banding the picture
  through a 32-bit GWorld.
  
  This function is a minimally-modified version of the
  "Color Adventures" sample's DoPrintLoop routine.

*************************************************************/

OSErr DoColorPrintLoop(THPrint hPrint, WindowPtr wp, Rect *destRect, Boolean useDrawPicture)
{
	OSErr			err;
	TPPrPort		printPort;
	TPrStatus		theStatus;
	Rect			rPage, bandRect, pictFrame, prtRect;
	GWorldPtr		prtGWorld, savedGWorld, bandGWorld = nil;
	GDHandle		prtGDH, savedGDH;
	PixMapHandle	srcPix;
	Boolean			moreToDo, spooling;
	WindowInfoHdl	windInfo;
	PicHandle		pict;
	short			copy, iFstPage, iLstPage, rowWidth;
	long			bandHt, bandWidthInBytes, maxBlockSize;

// Get the window's picture and find out if we're spooling.

	windInfo = (WindowInfoHdl) GetWRefCon(wp);
	pict = (*windInfo)->windowPict;
	spooling = !(*windInfo)->loaded;


// Get the cursor spinning (we'll keep it spinning in our idle proc).  We
// specify r_stdCurs4 as the final cursor because we may be switching
// between the spinning spool cursors and the spinning standard cursors.
// Using the last standard cursor as the final cursor makes the transition
// look smoother.
//
// Save the current GWorld/GDevice so that we can restore them when we leave.
// Open the printer driver and validate the print record so that the Printing
// Manager pulls our document title from the title of the frontmost window.
// The reason we save and restore the iFstPage and iLstPage fields is that the
// LaserWriter driver (through at least v7.1.2) has a bug in PrValidate which
// causes these values to be reset.  While we aren't using iFstPage or iLstPage
// below, you never know where this code might end up in this crazy cut & paste
// world.

	SetUpCursors(r_stdSpinCursor, r_stdCurs4);
		
	GetGWorld(&savedGWorld, &savedGDH);

	PrOpen();

	if (!(err = PrError()))
	{
		iFstPage = (*hPrint)->prJob.iFstPage;
		iLstPage = (*hPrint)->prJob.iLstPage;
		
		PrValidate(hPrint);
		
		(*hPrint)->prJob.iFstPage = iFstPage;
		(*hPrint)->prJob.iLstPage = iLstPage;


// Add an idle proc to spin the cursor.  Get the printer driver's page
// rectangle, and offset and center the picture frame in it. It's ok if
// pictFrame is larger than rPage, because we handle that below.

		(*hPrint)->prJob.pIdleProc = (PrIdleProcPtr) IdleProc;
		
		rPage = (*hPrint)->prInfo.rPage;
		pictFrame = *destRect;
		OffsetRect(&pictFrame, -pictFrame.left, -pictFrame.top);
		CenterRect(&pictFrame, &rPage, &pictFrame);


// If we're not using DrawPicture (if we're banding), get 75% of
// the maximum block size available and use that figure to determine
// the size of the bands we create.  The calculation is more straight-
// forward than it may appear.  We determine the maximum buffer size we
// have to work with (maxBlockSize), then determine the width of each row
// (rowWidth).  Note that if the picture's rows are wider than the page
// rectangle, we clip to the page rectangle.  From the rowWidth, we
// determine how many bytes are required for each row (at 32 bits per
// pixel).  We then make sure this number is longword aligned (as 32-
// bit QuickDraw will make it when we allocate the GWorld).  Finally,
// if we have enough room for the entire image, we use the height of the
// image as the bandHt.  Otherwise, we use (maxBlockSize/BandWidthInBytes)
// as bandHt. Actually, bandHt is really the height of a band -1.

		if (!useDrawPicture)
		{
			maxBlockSize = MaxBlock() * .75;
			rowWidth = MIN(pictFrame.right - pictFrame.left, rPage.right - rPage.left);
	
			bandWidthInBytes = 4 * rowWidth;			// (32 bits/pixel *
														// rowWidth pixels)  8 bits/byte

			bandWidthInBytes += bandWidthInBytes % 4;	// Make sure the value is
														// longword aligned.

			bandHt = MIN(maxBlockSize / bandWidthInBytes,
						 pictFrame.bottom - pictFrame.top);


// Set up bandRect, which is the bounds of our band buffer, and prtRect
// which is the same area, just aligned with our page rectangle.  Create
// a new GWorld the size of bandRect, to use as our band buffer.

			SetRect(&bandRect, pictFrame.left, pictFrame.top,
					pictFrame.left + rowWidth, pictFrame.top +bandHt);

			prtRect = bandRect;
			OffsetRect(&prtRect, -rPage.left, -rPage.top);
			if (!err) err = NewGWorld(&bandGWorld, 32, &bandRect, nil, nil, 0);
		}

// Loop for the number of copies we need to make, opening a document
// and a page for each.

		for (copy = 1; !err && (copy <= (*hPrint)->prJob.iCopies); copy++)
		{
			printPort = PrOpenDoc(hPrint, nil, nil);

			if (!(err = PrError()))
			{
				PrOpenPage(printPort, nil);
				err = PrError();
				
				if (!err)
				{

// If we're using DrawPicture, clip to rPage and call DrawPicture.
// If we're also spooling, first tear down the spinning cursors
// (because spooling installs its own) and install spooling.
// After we've drawn, remove spooling and set the spinning cursors
// back up again.

					if (useDrawPicture)
					{
						if (spooling)
						{
							TearDownCursors();
							err = SpoolInstall(windInfo, r_stdCurs4);
						}
						
						if (!err)
						{
							ClipRect(&rPage);
							DrawPicture(pict, &pictFrame);

							if (spooling)
							{
								err = SpoolRemove();
								SetUpCursors(r_stdSpinCursor, r_stdCurs4);
							}
						}
					}
					else
					{

// If we're using banding instead of using DrawPicture, save off the current
// GWorld/GDevice settings (which can get us back to the printer port later).

						moreToDo = true;
						GetGWorld(&prtGWorld, &prtGDH);


// For each pass through the loop (each band), set the current GWorld to
// our bandGWorld, erase it, clip to it, and call DrawPicture.  pictFrame
// is offset below, so that successive passes will write successive bands
// into our bandGWorld.  If this picture is being spooled in, we need to
// install and remove spooling (and remove and install our spinning
// cursors) around the DrawPicture call.  (The spooling routines use their
// own spinning cursors, so we need to take down the current ones first.)

						while (moreToDo && !err)
						{
							SetGWorld(bandGWorld, nil);
		
							if (spooling)
							{
								TearDownCursors();
								err = SpoolInstall(windInfo, r_stdCurs4);
							}

							EraseRect(&bandRect);
							ClipRect(&bandRect);
							DrawPicture(pict, &pictFrame);

							if (spooling)
							{
								err = SpoolRemove();
								SetUpCursors(r_stdSpinCursor, r_stdCurs4);
							}

// Set the current GWorld (or port) back to the printer port and check for
// command-periods.  If none are found, copy the band from our bandGWorld
// to the printer port.  Afterwards, if the whole picture hasn't been
// drawn (pictFrame.bottom > bandRect.top) and we haven't drawn all that
// will fit on a page (prtRect.bottom < rPage.bottom), we have more to draw.

							SetGWorld(prtGWorld, prtGDH);
							if (!err) err = CommandPeriod();
			
							if (!err)
							{
								srcPix = GetGWPixMap(bandGWorld);
								LockPixels(srcPix);
								ClipRect(&prtRect);
								CopyBits((BitMap *) *srcPix,
										 &(printPort->gPort.portBits), &bandRect,
										 &prtRect, srcCopy, nil);

								UnlockPixels(srcPix);
								
								moreToDo = (pictFrame.bottom > bandRect.top) &&
										   (prtRect.bottom < rPage.bottom);
							}

// If we have more to do, and there were no errors, shift pictFrame up by bandHt.
// This means that next time through, one band lower will drawn.  Also offset
// prtRect so that we copy the next band one band lower in the printer port.  Next,
// calculate the new bandHt, which will be either the current bandHt, the height
// of the remaining area on the page that we can draw, or the height of the
// remaining part of the image to draw, whichever is smallest.  Set the new band
// size by adjusting prtRect and bandRect appropriately.

							if (moreToDo && !err)
							{
								pictFrame.top -= bandHt;
								pictFrame.bottom -= bandHt;
								prtRect.top += bandHt;
								bandHt = MIN(bandHt, (rPage.bottom - prtRect.bottom));
								bandHt = MIN(bandHt, (pictFrame.bottom - bandRect.top));
								prtRect.bottom = prtRect.top + bandHt;
								bandRect.bottom = bandRect.top + bandHt;
							}
						}

// All done banding.  Make sure that the current GWorld (port) is the printer port,
// so that the Printing Manager stays happy when we make the remaining print calls.
// If an error occurred, set a print error.  Close the page since all drawing
// for this page is completed.  When all pages are completed, close the document.


						SetGWorld(prtGWorld, prtGDH);
					}
				}
				
				if (err) PrSetError(err);
				PrClosePage(printPort);
			}
		
			PrCloseDoc(printPort);
		}

// When finished printing, dispose of our bandGWorld if we banded, call
// PrPicFile (if necessary), stop the cursor's spinning, close the printer
// driver and restore the GWorld/GDevice which were active upon entry.

		if (bandGWorld)
			DisposeGWorld(bandGWorld);

		if (!err) err = PrError();

		if (!err && ((*hPrint)->prJob.bJDocLoop == bSpoolLoop))
			PrPicFile(hPrint, nil, nil, nil, &theStatus);
	}

	(*hPrint)->prJob.pIdleProc = nil;
	TearDownCursors();
	SetCursor(&qd.arrow);				// We'd still be a beachball without this.

	PrClose();

	SetGWorld(savedGWorld, savedGDH);
	return err;
}
