/*********************************************************************

	Final Halftone.c
	
	This file contains all of the halftone-specific printing
	functions for the all-in-one "Adventures in Color Printing"
	color sample, "The Final Adventure."
	
	Dave Hersey, Apple Developer Technical Support.
	
	 Edit Trail 

	created.									- dmh  11/24/92

*********************************************************************/

#pragma segment Printing

#include "Final Headers.h"


/************************************************************
  DoHTPrint - This routine starts the printing process for
  the indicated window.  We set the printer port's
  resolution, display the job dialog and so forth, then pass
  control on to DoHT0 or DoHT45 for rendering.  Those
  routines in turn call the DoHTPrintLoop routine to print
  the image.
  
  This function is a minimally-modified version of the
  "Halftone Adventures" sample's PrintOne routine.

*************************************************************/

OSErr DoHTPrint(WindowPtr whichWindow, Boolean useTrueHalftone)
{
	QDErr					err;
	GWorldPtr				savedGW;
	GDHandle				savedGDH;
	short					pxDPI, pyDPI, ixDPI, iyDPI;
	enum{HT0deg, HT45deg}	htType;
	WindowInfoHdl			windInfo;
	Rect					pictFrame;
	Fixed					scale;

// Save the current GWorld/GDevice so that we can restore it later.
// See which rendering option we're using, retrieve our private
// "WindowInfoHdl" structure that's hanging off the window's refCon
// field and calculate the image's native resolution and dimensions.

	GetGWorld(&savedGW, &savedGDH);
	htType = (useTrueHalftone)? HT45deg: HT0deg;

	windInfo = (WindowInfoHdl) GetWRefCon(whichWindow);
	GetImageRes(windInfo, &ixDPI, &iyDPI, &pictFrame);


// We want to print at the printer's physical resolution for minimal
// distortion.  We find out the physical resolution of the printer
// and set the printer port to that.

	pxDPI = pyDPI = GetPrinterRes();				// rather than guessing the 
													// physical resolution of the
													// printer, find it.

//	GetBestDPI(&pxDPI, &pyDPI, 300, 300, true);		// Not really necessary unless
													// we're just guessing.  Left here
													// for reference in case you go
													// the guessing route.

	SetPrinterPortDPI(ghPrint, &pxDPI, &pyDPI);


// Set the cursor to the watch, align the top-left corner of our pictFrame to
// (0,0), and then scale the image (if necessary) to the printer's port's
// resolution.

	SetCursor(*GetCursor(watchCursor));

	OffsetRect(&pictFrame, -pictFrame.left, -pictFrame.top);

	if (htType != HT45deg)	// The 45 halftone method handles scaling internally.
	{
		if ((pxDPI != ixDPI) || (pyDPI != iyDPI))
		{
			scale = FixRatio(pxDPI, ixDPI);
			pictFrame.right = FixMul(scale, (long) pictFrame.right << 16) >> 16;
			scale = FixRatio(pyDPI, iyDPI);
			pictFrame.bottom = FixMul(scale, (long) pictFrame.bottom << 16) >> 16;
		}
	}

// Spin off to the appropriate rendering and printing routine.  Note that these
// routines expect pxDPI = pyDPI so we only pass pxDPI.  When we return, restore
// the GWorld/GDevice we saved earlier, set the cursor to the arrow, and bop on home.

	switch (htType)
	{
		case HT0deg:
			err = DoHT0(whichWindow, pxDPI, &pictFrame);
			break;
		
		case HT45deg:
			err = DoHT45(whichWindow, pxDPI, ixDPI, iyDPI, &pictFrame);
			break;
	}

	SetGWorld(savedGW, savedGDH);
	SetCursor(&qd.arrow);

	return err;
}


/************************************************************
  DoHT0 - This routine creates a 1-bit GWorld containing a
  simulated halftone (created with 4x4 pixel patterns) of
  the picture associated with the passed window.  If all
  goes well, the image is passed to the DoHTPrintLoop routine
  and printed.

  pDPI is the printer's resolution (horizontal and vertical
  resolutions will be equal.
  
  destRect is the destination rectangle for the image at
  the printer's resolution.  Its top-left corner is aligned
  to (0,0).

*************************************************************/

OSErr DoHT0(WindowPtr wp, short pDPI, Rect *destRect)
{
	QDErr			err;
	Str255			orgTitle, tempStr;
	Rect			srcRect, printRect, rPage;
	GWorldPtr		fourBitGWorld = nil, halftoneGWorld = nil, savedGW;
	GDHandle		savedGDH;
	DialogPtr		printStatDlog = nil;
	PicHandle		wPICT;
	WindowInfoHdl	windInfo;

// Save the current GWorld/GDevice for restoration later.  Also, retrieve the
// picture we're supposed to print and the window's title, then display a
// status dialog.

	GetGWorld(&savedGW, &savedGDH);
	windInfo = (WindowInfoHdl) GetWRefCon(wp);
	wPICT = (*windInfo)->windowPict;

	GetWTitle(wp, orgTitle);
	ShowStatusMsg(&printStatDlog, "Converting ", orgTitle, " to 16 grays.", true);


// Determine how much of the image will appear on the page, and store that in
// printRect.  We're going to create a GWorld that's 1/4 the size of printRect,
// so we make sure that that rectangle can be evenly divided by 4 and adjust it
// if not.  printRect will be the actual area of the halftone image that we copy
// to the printer port.  We set srcRect to 1/4 this size and use it to create a
// 4-bit GWorld to dither to.  Finally, we pass the picture and GWorld to
// BandPicture which will dither the PICT for us.

	printRect = *destRect;			// NOTE: destRect is top-left aligned to (0,0).

	rPage = (*ghPrint)->prInfo.rPage;				// Clip the output rectangle to
	SectRect(&printRect, &rPage, &printRect);		// the printable page area.

	printRect.right  &= 0xFFFC;						// Make the print rectangle bounds
	printRect.bottom &= 0xFFFC;						// divisible by 4.

	srcRect = printRect;							// srcRect = printRect /4.
	srcRect.right	  >>= 2;
	srcRect.bottom    >>= 2;

	destRect->right   >>= 2;						// destRect = original destRect /4.
	destRect->bottom  >>= 2;						// This insures that picture
													// rect passed to BandPicture is
													// proportionally 1/4 as big as the
													// original, which is how we must
													// render it in fourBitGWorld.

	err = NewGWorld(&fourBitGWorld, 4, &srcRect, GetCTable(32 +4), nil, 0);
	if (!err) err = BandPicture(wp, destRect, fourBitGWorld, true);


// If there were no errors, change our status message and create our simulated
// halftone.  When finished, dispose of our 4-bit GWorld.

	if (!err)
	{
		sprintf(tempStr, "%s %i %s %i %s", "using patterns at", pDPI,
				"dpi horizontally by", pDPI, "dpi vertically.");

		ShowStatusMsg(&printStatDlog, "Simulating  0 halftone of ", orgTitle,
					  tempStr, true);

		err = SimulateHalftone(&printRect, fourBitGWorld, &halftoneGWorld);

		DisposeGWorld(fourBitGWorld);
		fourBitGWorld = nil;


// Tell the user we're going to print, then do so.  When finished, get rid of
// the status dialog, make sure that our GWorlds have been disposed, and restore
// the current GWorld and GDevice to our stored values.

		if (!err)
		{
			ShowStatusMsg(&printStatDlog, "Printing ", orgTitle, ".", true);
			err = DoHTPrintLoop(ghPrint, &halftoneGWorld, &printRect);
		}
	}

	HideStatusMsg(&printStatDlog);

	if (fourBitGWorld)  DisposeGWorld(fourBitGWorld);
	if (halftoneGWorld) DisposeGWorld(halftoneGWorld);

	SetGWorld(savedGW, savedGDH);
	return err;
}


/************************************************************
  DoHT45 - This routine creates a true halftone of the
  picture associated with the passed window.  If all goes
  well, the halftone is passed to the DoHTPrintLoop routine
  and printed.
  
  pDPI is the printer's resolution (horizontal and vertical
  resolutions will be equal.
  
  ixDPI and iyDPi are the horizontal and vertical
  resolutions of the source image.
  
  destRect is the destination rectangle for the image at
  the printer's resolution.  Its top-left corner is aligned
  to (0,0).

*************************************************************/

OSErr DoHT45(WindowPtr wp, short pDPI, short ixDPI, short iyDPI, Rect *destRect)
{
	QDErr				err;
	Str255				orgTitle, tempStr;
	GWorldPtr			srcGWorld = nil, halftoneGWorld = nil, savedGW;
	GDHandle			savedGDH;
	DialogPtr			printStatDlog = nil;
	Rect				rPage, printRect;
	PicHandle			wPICT;
	WindowInfoHdl		windInfo;
	Fixed				scale;
	PixMapHandle		srcPix;

// Save the current GWorld/GDevice for restoration later.  Also, retrieve the
// picture we're supposed to print.

	GetGWorld(&savedGW, &savedGDH);

	GetWTitle(wp, orgTitle);
	windInfo = (WindowInfoHdl) GetWRefCon(wp);
	wPICT = (*windInfo)->windowPict;


// Set printRect to the amount of the image that will actually fit on the page
// at the destination resolution.  This way, we don't work with any more of the
// image than necessary.  Create a new 32-bit deep GWorld to store the image in,
// then use BandPicture to copy the picture into it.

	printRect = *destRect;

	if (pDPI != ixDPI || pDPI != iyDPI)			// Scale printRect to the printer's res.
	{
		scale = FixRatio(pDPI, ixDPI);
		printRect.right = FixMul(scale, (long) printRect.right << 16) >> 16;
		scale = FixRatio(pDPI, iyDPI);
		printRect.bottom = FixMul(scale, (long) printRect.bottom << 16) >> 16;
	}

	rPage = (*ghPrint)->prInfo.rPage;			// Clip the output rectangle to the
	SectRect(&printRect, &rPage, &printRect);	// printable page area.

	if (pDPI != ixDPI || pDPI != iyDPI)			// Scale the clipped printRect back to
	{											// the source image's resolution.
		scale = FixRatio(ixDPI, pDPI);
		printRect.right = FixMul(scale, (long) printRect.right << 16) >> 16;
		scale = FixRatio(iyDPI, pDPI);
		printRect.bottom = FixMul(scale, (long) printRect.bottom << 16) >> 16;
	}

	err = NewGWorld(&srcGWorld, 32, &printRect, nil, nil, 0);

	if (!err)
	{
		ShowStatusMsg(&printStatDlog, "Preparing ", orgTitle, "for halftoning", true);
		err = BandPicture(wp, destRect, srcGWorld, false);


// Tell the user we're going to create a halftone of the image, then go do
// it.  When we come back, dispose of the source GWorld and print.

		sprintf(tempStr, "%s %i %s %i %s", " 45 at", pDPI, "dpi horizontally by",
				pDPI, "dpi vertically.");

		ShowStatusMsg(&printStatDlog, "Halftoning ", orgTitle, tempStr, true);
		if (!err) err = HalftonePixMap(srcGWorld, &halftoneGWorld, ixDPI, pDPI);

		DisposeGWorld(srcGWorld);
		srcGWorld = nil;

		if (!err)
		{
			srcPix = GetGWPixMap(halftoneGWorld);
			printRect = (*srcPix)->bounds;

			ShowStatusMsg(&printStatDlog, "Printing ", orgTitle, ".", true);
			err = DoHTPrintLoop(ghPrint, &halftoneGWorld, &printRect);
		}

// Tear down the status dialog, make sure our GWorlds are disposed of and restore
// the current GWorld/GDevice as we exit.

		HideStatusMsg(&printStatDlog);
	}
	
	if (srcGWorld) DisposeGWorld(srcGWorld);
	if (halftoneGWorld) DisposeGWorld(halftoneGWorld);

	SetGWorld(savedGW, savedGDH);
	return err;
}


/************************************************************
  DoHTPrintLoop - This routine does the actual printing.  The
  picture to be printed is attached to the window passed
  (through our WindowInfoHdl structure).
  
  This function is a minimally-modified version of the
  "Halftone Adventures" sample's DoPrintLoop routine.

*************************************************************/

OSErr DoHTPrintLoop(THPrint hPrint, GWorldPtr *srcWorld, Rect *srcRect)
{
	OSErr			err;
	TPPrPort		printPort;
	TPrStatus		theStatus;
	GWorldPtr		savedGWorld;
	GDHandle		savedGDH;
	PixMapHandle	srcPix;
	Rect			rPage, outRect;
	short			copy, iFstPage, iLstPage;

// Set the cursor spinning (we'll keep it spinning in our idle proc).  Open the
// printer driver and validate the print record so that the Printing Manager
// pulls our document title from the title of the frontmost window.  The reason
// we save and restore the iFstPage and iLstPage fields is that the LaserWriter
// driver (through at least v7.1.2) has a bug in PrValidate which causes these
// values to be reset.  While we aren't using iFstPage or iLstPage below, you
// never know where this code might end up in this crazy cut & paste world.

	GetGWorld(&savedGWorld, &savedGDH);
	SetUpCursors(r_stdSpinCursor, kArrowCursor);

	PrOpen();

	if (!(err = PrError()))
	{
		iFstPage = (*hPrint)->prJob.iFstPage;
		iLstPage = (*hPrint)->prJob.iLstPage;
		
		PrValidate(hPrint);
		
		(*hPrint)->prJob.iFstPage = iFstPage;
		(*hPrint)->prJob.iLstPage = iLstPage;


// Install our idle proc, then center the image's rectangle within the page
// rectangle, then loop for the number of copies we need to make, opening a
// document and a page for each.

		(*hPrint)->prJob.pIdleProc = (PrIdleProcPtr) IdleProc;

		rPage = (*hPrint)->prInfo.rPage;
		outRect = *srcRect;
		CenterRect(&outRect, &rPage, &outRect);

		for (copy = 1; !err && (copy <= (*hPrint)->prJob.iCopies); copy++)
		{
			printPort = PrOpenDoc(hPrint, nil, nil);
		
			if (!(err = PrError()))
			{
				PrOpenPage(printPort, nil);


// Lock down our source PixMap, then copy it to the printer port.  When done,
// unlock its pixels, close the page, and the document.

				if (!(err = PrError()))
				{
					srcPix = GetGWPixMap(*srcWorld);
					LockPixels(srcPix);
					CopyBits((BitMap *) *srcPix, &printPort->gPort.portBits, srcRect, &outRect, srcCopy, nil);
					UnlockPixels(srcPix);
				}
				
				PrClosePage(printPort);
			}
		
			PrCloseDoc(printPort);
		}

// Dispose of the source image, call PrPicFile if necessary, remove our idle
// proc, stop our cursors from spinning, close down the Printing Manager and
// just LEAVE.

		DisposeGWorld(*srcWorld);
		*srcWorld = nil;

		if (!err) err = PrError();
		
		if (!err && ((*hPrint)->prJob.bJDocLoop == bSpoolLoop))
			PrPicFile(hPrint, nil, nil, nil, &theStatus);
	}

	(*hPrint)->prJob.pIdleProc = nil;
	TearDownCursors();
	PrClose();

	return err;
}


/************************************************************
  SimulateHalftone - This routine creates a simulated
  halftone image by converting the passed GWorld into a
  1-bit representation using 4x4 pixel values to represent
  each pixel's color.
  
  The source pixmap is expected to by 4 bits deep, using the
  standard gray color table for that depth as obtained from
  GetCTable(32 +4).
  
  htGWorld is created and contains the image that's created
  upon exit.

*************************************************************/

OSErr SimulateHalftone(Rect *srcRect, GWorldPtr srcGWorld, GWorldPtr *htGWorld)
{
	register Ptr			srcByte, destByte, prevSrc, prevDest;
	register long			rowOffset[5];
	register short			col, row, pixRow;
	QDErr					err;
	PixMapHandle			srcPix, destPix;
	Rect					hToneRect;
	short					aPixel, rowBytes1bit, rowBytes4bit;
	short					lastRow, lastCol;
	signed char				mode;
	GWorldPtr				curGWorld;
	GDHandle				curGDH;

// The grid below corresponds to figure 6 in the "Adventures in Color Printing"
// article in develop #13.
//
// The patterns are stored in an optimized pattern as such:
//
// Every pixel will have a value from 0 to 15.  This corresponds to a
// column of the pixPattern matrix.
//
// To figure out which pixels in the 4 x 4 pattern to turn on, go to the row
// representing the pixel's value.  Since every pixel will be stored in 4 bits,
// our code must handle high or low nibble setting.  We do this by storing the
// value for an entire nibble in an entry of our pixPattern matrix.
// 
// For example, the pattern for a nibble with an pixel value of 5 (the 6th
// column from the left of the table) will be:
//
//              0    -------- (in binary) ------->   0 0 0 0   (A)
//              6    -------- (in binary) ------->   0 1 1 0   (B)
//              6    -------- (in binary) ------->   0 1 1 0   (C)
//              4    -------- (in binary) ------->   0 1 0 0   (D)
//
// which agrees with the entry for a pixle value of 5 shown in figure 6 of
// the article.  We store nibble (A) in row 1 of the current pattern,
// nibble (B) in row 2, (c) in row 3 and (D) in row 4.  This way we set up
// an entire 4x4 pattern by only storing 4 values.  This method may seem
// obscure, but it lets us quickly calculate and set the different pixel
// values of every nibble in the image.

	char		pixPattern[4][16] =
				{
					0, 0, 0, 0, 0, 0, 0, 2,  2,  2,  2,  3,  3,  3,  3, 15,
					0, 0, 0, 2, 6, 6, 6, 6, 14, 14, 15, 15, 15, 15, 15, 15,
					0, 4, 6, 6, 6, 6, 7, 7,  7,  7,  7,  7, 15, 15, 15, 15,
					0, 0, 0, 0, 0, 4, 4, 4,  4,  6,  6,  6,  6, 14, 15, 15
				};

// Save the current GWorld/GDevice so that we can restore them when we
// leave.  Create a 1-bit GWorld for our source image and get and lock
// our source and destination PixMaps, since we'll be walking them pixel
// by pixel.  Also, start our cursor spinning so the user knows we're
// alive.

	GetGWorld(&curGWorld, &curGDH);

	err = NewGWorld(htGWorld, 1, srcRect, nil, nil, 0);
	if (err) return err;
	
	srcPix = GetGWPixMap(srcGWorld);
	destPix = GetGWPixMap(*htGWorld);
	LockPixels(srcPix);
	LockPixels(destPix);
	
	SetUpCursors(r_stdSpinCursor, watchCursor);


// Erase our halftone GWorld so that we can skip over pixel values of zero,
// without having to draw an all white pattern.  This can improve our speed.

	SetGWorld(*htGWorld, nil);
	EraseRect(srcRect);
	SetGWorld(curGWorld, curGDH);
	

// By dividing the source rectangle by 4 in each direction, we can determine
// how many rows and columns of patterns we need to lay down.  Note that
// before we entered this routine (in DoHT0) we made sure that the bounds
// were evenly divisible by 4.  Also, we made sure that srcRect had its
// top-left corner at (0,0).

	hToneRect = *srcRect;
	hToneRect.right >>= 2;
	hToneRect.bottom >>= 2;
	lastRow = hToneRect.bottom;
	lastCol = hToneRect.right;
	

// Store the rowByte values of our source and destination GWorlds, masking
// off the high bit to get the true rowBytes4bit values.  While we're at it
// calculate the values we need to add to a byte position in the source
// GWorld's pixmap to get to the 4 different rows of the destination
// pixmap.

	rowBytes4bit = (*srcPix)->rowBytes  & 0x7FFF;
	rowBytes1bit = (*destPix)->rowBytes & 0x7FFF;

	for (pixRow = 0; pixRow <= 4; pixRow++)
		rowOffset[pixRow] = (long) pixRow * (long) rowBytes1bit;


// Store the baseAddresses of our source and destination PixMaps and begin
// looping through the pixels.  On each pass we check to see if the user
// has become impatient and pressed command=period.  Also, because we may
// be accessing bytes in a 32-bit pointer (for example if the 824 GC card
// is being used, we MUST insure that we're in 32-bit addressing mode before
// using the baseAddr pointers.  Since it would be a bad thing to make
// system calls (such as those in CommandPeriod) if we've changed the MMU
// mode, we set and restore it around the code that accesses the baseAddr
// pointers.  This way, the system is always set up correctly when we go
// to CommandPeriod.

	srcByte  = GetPixBaseAddr(srcPix);
	destByte = GetPixBaseAddr(destPix);

	for (row = 0; ((row < lastRow) && !(err = CommandPeriod())); row++)
	{
		BumpCursor();						// Spin the cursor.

		mode = true32b;						// Make sure we're in 32-bit addressing
		SwapMMUMode(&mode);					// mode before touching those pixels!


// Store the previous (current) pointers so that we can easily update the
// new pointers when we get to the end of this row.  For each byte in our
// source PixMap, we handle the high and low nibbles separately to keep
// things simple and efficient.
//
// If (col & 1) == 0, we're working on an even pixel, whose value will be
// stored in the high nibble of a byte.  We grab the value of the pixel and
// store it in aPixel.  Note that we can just index off of the pixel value
// instead of tracing back to the color table since the color table we used
// is a linear ramp of grays.  Index into the pixPattern matrix and set the
// current destination nibble and the 3 underneath that to the values found
// in the table.  This sets 4x4 pixels in the destination pixmap with only 4
// storage operations, and completes the storage of the halftone pattern for
// the pixel.
// 
// If we're working on an odd pixel ((col & 1) == 1) then we do everything
// the same, except we work on the low order nibble of the source and we
// make sure that we don't destroy the high order nibbles of our destination
// bytes.  After every odd pixel (low order nibble) we increment our byte
// pointers.

		prevSrc  = srcByte;
		prevDest = destByte;
		
		for (col = 0; col < lastCol; col++)
			if ((col & 1) == 0)					// even pixel.
			{
				aPixel = (*srcByte & 0xF0) >> 4;

				if (aPixel)
					for (pixRow = 0; pixRow < 4; pixRow++)
						*(Ptr) ((long) destByte +
								rowOffset[pixRow]) = pixPattern[pixRow][aPixel] << 4;
			}
			else								// odd pixel.
			{
				aPixel = *srcByte & 0x0F;

				if (aPixel)
					for (pixRow = 0; pixRow < 4; pixRow++)
						*(Ptr) ((long) destByte +
								rowOffset[pixRow]) |= pixPattern[pixRow][aPixel];

				++srcByte;
				++destByte;
			}

// Update our source pointer to point to the next row, and our destination pointer
// to point 4 rows lower.  Remember that every pixel in the destination PixMap is
// actually 4 pixels high and 4 pixels wide, and this is sort of like moving a 4
// pixel thick pen to the next row to draw in.  We restore the old MMU mode so that
// CommandPeriod doesn't die a horrid death when the system was booted in 24-bit mode.

		srcByte  = rowBytes4bit + prevSrc;
		destByte = rowOffset[4] + prevDest;
		
		SwapMMUMode(&mode);						// Go back to native addressing mode.
	}

// Unlock all our PixMaps, restore the current GWorld/GDevice, tear down our spinning
// cursors, and take a hike.

	UnlockPixels(srcPix);
	UnlockPixels(destPix);

	SetGWorld(curGWorld, curGDH);
	TearDownCursors();

	return err;
}


/************************************************************
  LUMVAL - This routine returns the luminance for a pixel of
  depth 8 or 32 bits.
	
  pPixel points to the pixel in the PixMap.
  pMap is the Pixmap's PixMapPtr.

  For more information, see the article "Making the Most of
  Color on 1-bit Devices," by Dan Lipton and Konstantin
  Othmer which appears in develop issue #9.  This code comes
  from the article's sample code.

*************************************************************/

long LUMVAL(Ptr pPixel, PixMapPtr pMap)
{
	long	red, green, blue;

	if (pMap->pixelSize == 32)
	{
		red = (long)(unsigned char)*(++pPixel);				// Skip alpha, get red
		green = (long)(unsigned char)*(++pPixel);			// get green,
		blue = (long)(unsigned char)*(++pPixel);			// get blue.
		
		return ((30 * red + 59 * green + 11 * blue) /100);
	
	}
	else
		if (pMap->pixelSize == 8)
		{
			RGBColor	*theColor;

			theColor = &((*(pMap->pmTable))->ctTable[(unsigned char)*pPixel].rgb);
			
			return ((30 * (theColor->red >> 8) +
					 59 * (theColor->green >> 8) +
					 11 * (theColor->blue >> 8)) /100);
		}
}


/************************************************************
  HalftonePixMap - This routine takes a pixmap and halftones
  it at the passed destination resolution into a 1-bit
  GWorld, which is returned to the caller.

  Note, if destination resolution is less than 72, source
  scaling code does not work.  It assumes target resolution
  is always >= 72 so pointer arithmetic in source image
  doesn't need to do a multiply every time.  This could be
  fixed at the expense of performance, but in this sample
  the destination (printer) resolution will always be >= 72.
	
  For more information, see the article "Making the Most of
  Color on 1-bit Devices," by Dan Lipton and Konstantin
  Othmer which appears in develop issue #9.  This function
  is adapted from the article's sample code.

*************************************************************/

OSErr HalftonePixMap(GWorldPtr srcGWorld, GWorldPtr *destGWorld,
					 short srcDPI, short destDPI)
{
	QDErr							err;
	register unsigned long			mask, destLong;
	short							widthAtDestDPI, heightAtDestDPI, lastX, lastY,
									destX, destY, lumins[256], idx;

	Rect							boundsAtDestDPI;
	Fixed							scale, sourceX, sourceY;
	Ptr								pRaster, pPixel, pDestRaster;
	long							*nextDestLong, luminance,
									*pMatrix, *pMatrixRow, M = 8, N = 8, I, J;

	long			   matrix[8][8] = { 7,  111, 183, 239, 231, 191, 71, 15, 
										87, 47, 135, 175, 199, 159, 55, 127, 
										215, 143, 39, 79, 119, 63, 151, 207, 
										247, 223, 103, 31, 23, 95, 167, 255 ,
										231, 191, 71, 15, 7, 111, 183, 239 ,
										199, 159, 55, 127, 87, 47, 135, 175, 
										119, 63, 151, 207, 215, 143, 39, 79, 
										23, 95, 167, 255, 247, 223, 103, 31
									   }; 

	Boolean							directPMap;
	RGBColor						*anRGB;
	PixMap							*pSource;
	PixMapHandle					offScreen, hSource;
	signed char						mode;
	
// If our destination resolution is < 72 dpi, or our source PixMap is not 8 or
// 32-bits deep, return an error.  Otherwise, get the source PixMap, lock its
// pixels and PixMapHandle down and get a pointer to the PixMap.

	if (destDPI < 72)
		return paramErr;

	hSource = GetGWPixMap(srcGWorld);
	
	if ((*hSource)->pixelSize != 8 && (*hSource)->pixelSize != 32)
		return paramErr;
	
	LockPixels(hSource);
	HLock((Handle) hSource);
	pSource = (PixMap *) StripAddress(*hSource);


// If we're not using a direct PixMap, cache the luminance values of all entries
// in the PixMap's color table.  This can speed up luminance calculations for
// indexed PixMaps.

	directPMap = (pSource->pixelType == RGBDirect);
	
	if (!directPMap)
		for (idx = 0; idx <= (*(pSource->pmTable))->ctSize; idx++)
		{
			anRGB = &(*(pSource->pmTable))->ctTable[ (unsigned char) idx].rgb;
			lumins[idx] = (30 * (anRGB->red   >> 8) + 59 *
								(anRGB->green >> 8) + 11 *
								(anRGB->blue  >> 8)) /100;
		}

// Compute our halftoned image's width and bounds (which will be the source image's
// bounds at the destination resolution.

	scale = FixRatio(destDPI, srcDPI);

	widthAtDestDPI  = FixMul(scale, (long) (pSource->bounds.right -
							 pSource->bounds.left) << 16) >>16;
	heightAtDestDPI = FixMul(scale, (long) (pSource->bounds.bottom -
							 pSource->bounds.top) << 16) >>16;

	boundsAtDestDPI.top    = pSource->bounds.top;
	boundsAtDestDPI.left   = pSource->bounds.left;
	boundsAtDestDPI.bottom = boundsAtDestDPI.top + heightAtDestDPI;
	boundsAtDestDPI.right  = boundsAtDestDPI.left + widthAtDestDPI;


// Allocate a 1-bit GWorld to create the halftoned image in.  If we succeed,
// get the GWorld's PixMap, lock it down, and set up our spinning cursors.

	err = NewGWorld(destGWorld, 1, &boundsAtDestDPI, nil, nil, 0);

	if (err)
	{
		UnlockPixels(hSource);
		HUnlock((Handle) hSource);
		return err;
	}
	
	offScreen = GetGWPixMap(*destGWorld);
	LockPixels(offScreen);

	SetUpCursors(r_stdSpinCursor, watchCursor);


// Compute the scale for walking source pixels.  sourceX, sourceY will walk source
// image in fixed-point numbers.  Integer part will be saved to see when we've crossed
// into a new pixel boundary for updating pointers and luminance value.

	scale = FixRatio(srcDPI, destDPI);
	sourceY = pSource->bounds.top << 16;
	lastY = sourceY >> 16;
	

// Initialize raster pointers.

	pRaster =  GetPixBaseAddr(hSource);
	pDestRaster = GetPixBaseAddr(offScreen);
	
	I = 0;								// Y index into dither matrix.
	pMatrixRow = &(matrix[0][0]);		// Point to next row of matrix.


// Begin halftoning.  We walk the source image's rows for every row in the
// destination PixMap.

	for (destY = 0; !err && (destY < heightAtDestDPI); ++destY)
	{
		BumpCursor();					// Spin our cursor.

		mode = true32b;					// Make sure we're in 32-bit addressing
		SwapMMUMode(&mode);				// mode before touching those pixels!

		pPixel = pRaster;
		luminance = (directPMap)? LUMVAL(pPixel, pSource):
								  lumins[(unsigned char) *pPixel];
		mask = 0x80000000;
		destLong = 0;
		nextDestLong = (long*) pDestRaster;

		sourceX = pSource->bounds.left << 16;
		lastX = sourceX >> 16;

		J = 0;							// X index into dither matrix.
		pMatrix = pMatrixRow;			// Point into current row of dither matrix.


// Walk the source image's columns for every column in the destination
// PixMap.

		for (destX = 0; destX < widthAtDestDPI; ++destX)
		{

// If the luminance for this pixel is less than matrix value, turn on the pixel. 

			if (luminance < *pMatrix)
				destLong |= mask;
				
// Adjust masks and longs for next device pixel.  Pixels are set in a long-word
// register. When all 32 bits in the register have been set, the register is
// laid into the PixMap.
			
			mask >>= 1;

			if (mask == 0)				// If we've done 32 pixels, lay the long word
			{							// into the offscreen.
			
				mask = 0x80000000;		// Reset the mask and increment the destination
										// pointer.

				*(nextDestLong++) = destLong;
				destLong = 0;
			}


// Update the pointer into the dither matrix: Instead of using matrix[X % M, Y % N],
// we use a pointer and counters. A Mod function combined with an array index
// calculation would be more time-consuming.
			
			++pMatrix;					// Next element in row.

			if (++J >= N)
			{							// Get J+1 mod N, faster than
										// doing real mod every pixel.
				J = 0;
				pMatrix = pMatrixRow;	// Start at beginning of matrix row again.
			}

			sourceX += scale;			// Add fixed scale increment to source X value.


// Check to see if we crossed into the next source pixel.  Note: Getting integer
// portion can be done faster than a shift; this is left as an exercise for the
// reader.

			if ((sourceX >> 16) > lastX)
			{
				lastX = sourceX >> 16;	// Get new integer portion of X coordinate,
										// point to the next pixel, and get the
										// pixel's luminance value.

				if (pSource->pixelSize == 8)
					++pPixel;			// 8-bit source image.
				else
					pPixel += 4;		// 32-bit source image.
				
				luminance = (directPMap)? LUMVAL(pPixel, pSource):
										  lumins[(unsigned char) *pPixel];
			}
		}
	
// Make sure the last long-word was laid into the target image.  If the 1 bit in the
// mask is not all the way left, it means there are pixels in the next long, but not
// enough to force the write within the raster loop, so lay it into the block.

		if (mask != 0x80000000)
			*nextDestLong = destLong;


// Move to the next destination raster, add the fixed scale increment to source Y
// value, and check to see if we crossed into next source raster.

		pDestRaster += ((*offScreen)->rowBytes & 0x7FFF);
		sourceY += scale;

		if ((sourceY >> 16) > lastY)
		{
			lastY = sourceY >> 16;		// Get the integer portion of Y coordinate.

			pRaster += (pSource->rowBytes & 0x7FFF);	// Get the next source raster.
		
		}

// Update our pointer into the dither matrix.  At the end of a row of halftoning,
// check to see if the user has pressed command-period.
		
		pMatrixRow += M;				// Point to the next row of the matrix.
		if (++I >= M)
		{								// Get I+1 mod M, faster than doing a real
										// mod every raster.
			I = 0;
			pMatrixRow = &(matrix[0][0]);	// Point to next row of matrix.
		}
		
		SwapMMUMode(&mode);				// Put us back to the native addressing mode
										// so we can call CommandPeriod.
		err = CommandPeriod();
	 }

// Unlock our source and destination PixMaps, stop our spinning cursors and return
// any error we turned up.

	HUnlock((Handle) hSource);
	UnlockPixels(hSource);
	UnlockPixels(offScreen);

	TearDownCursors();
	return err;	
}

