/*********************************************************************

	Final Print Dialogs.c
	
	This file contains print dialog modification and handling
	functions for the all-in-one "Adventures in Color Printing"
	sample, "The Final Adventure."
	
	Dave Hersey, Apple Developer Technical Support.
	
	 Edit Trail 
	
	creation:   								11/25/92  - dmh
	added dialog-appending stuff:   			11/27/92  - dmh
	
*********************************************************************/

#pragma segment Printing

#include "Final Headers.h"

Boolean				gPrintInColor;		// If true, we set the "color" button that we
										// append to the ImageWriter print dialog.

Boolean				gPrintMaxDetail;	// If true, we set the "print with max. detail"
										// checkbox that we append to the print dialog.

static TPPrDlg		gPrtJobDialog;		// A pointer to our print job dialog.
static Ptr			gPrPItemProc;		// A pointer to the driver's item handling proc.
static short		gCurAppendDITLResID;// The resource ID of the DITL we've appended.
static long			gFirstItem;			// The adjusted item number of our first
										// appended DITL item.
 
/************************************************************
  DoStyleDlog - This routine handles the style (Page Setup)
  dialog.

  The printer driver is expected to be closed upon entry,
  and is therefore opened and closed in this routine.

  This routine assures that when the handle is returned,
  it is at 72 dpi, regardless of what it was before.  For
  reasons mentioned below, this is important.

*************************************************************/

OSErr DoStyleDlog(THPrint hPrint)
{
	OSErr		err;
	Boolean		doIt;
	short		resetThedpi;

// Make sure we're working with a default resolution print record, and open
// the printer driver.  Since (in this sample) we don't know what resolution
// we'll be setting the printer port to until after the dialogs, it's best to
// use a default dpi handle now so that options (which may be ignored when we
// print at high resolutions) are at least available in case they *can* be used.
//
// What makes things tricky is that different printer drivers support
// different ways of resetting the printer port resolution to 72 dpi.  You 
// always use PrGeneral, but the ImageWriter driver, for example, goes to "true"
// default resolution mode only when you try to set an invalid resolution such
// as 0x0 dpi, and the LaserWriter makes you pass 72x72 to get to 72 dpi.
// Neither respects the other's way of doing things.  (For an example, try setting
// the ImageWriter's print handle to 72x72 dpi with PrGeneral, and then doing a
// style dialog with that.)
// 
// One solution which I've found to work well is to first try to set the port to
// 72x72 dpi, and then try to set it to 0x0 dpi.  Drivers which reset when passed
// 72x72 ignore the 0x0 setting, and drivers that need to be passed 0x0, are happy
// as well.  Remember that the order is important here.  If you tell the IW driver
// that you want 72x72 dpi, it will disable options in the print dialog that
// shouldn't be.  Only when it tries to set the bogus 0x0 dpi will it clean house.
//
// Unfortunately, this is one of the design short-comings of the current
// printing architecture, or more specifically, the current printer drivers.
// If drivers don't disable options at different resolutions, we have no
// problem at all.

	resetThedpi = 72;				// Pass 72x72 dpi to force the defaults.
	SetPrinterPortDPI(hPrint, &resetThedpi, &resetThedpi);

	resetThedpi = 0;				// Pass a bogus pair to force the defaults.
	SetPrinterPortDPI(hPrint, &resetThedpi, &resetThedpi);
									// One of those should work.

// Handle the style dialog, close the driver and clean up any windows that
// need updating.

	PrOpen();

	err = PrError();
	
	if (err == fnfErr)	// Here, "file not found" really means "no printer chosen."
		err = noPrinterChosenErr;
	
	if (!err)
	{
		doIt = PrStlDialog(hPrint);
		err = PrError();
		if (!err && !doIt) err = iPrAbort;
	}

	PrClose();
	DoAnyUpdates();
	return err;
}


/************************************************************
  DoJobDlog - This routine handles the print job dialog,
  appending a DITL to it if an ID (other than 0) is passed.

  The printer driver is expected to be closed upon entry,
  and is therefore opened and closed in this routine.

  This routine assures that when the handle is returned,
  it is at 72 dpi, regardless of what it was before.  For
  reasons mentioned below, this is important.

*************************************************************/

OSErr DoJobDlog(THPrint hPrint, short aDITLResID)
{
	OSErr		err;
	Boolean		doIt;
	short		resetThedpi;

// Make sure we're working with a default resolution print record, and open
// the printer driver.  Since (in this sample) we don't know what resolution
// we'll be setting the printer port to until after the dialogs, it's best to
// use a default dpi handle now so that options (which may be ignored when we
// print at high resolutions) are at least available in case they *can* be used.
//
// What makes things tricky is that different printer drivers support
// different ways of resetting the printer port resolution to 72 dpi.  You 
// always use PrGeneral, but the ImageWriter driver, for example, goes to "true"
// default resolution mode only when you try to set an invalid resolution such
// as 0x0 dpi, and the LaserWriter makes you pass 72x72 to get to 72 dpi.
// Neither respects the other's way of doing things.  (For an example, try setting
// the ImageWriter's print handle to 72x72 dpi with PrGeneral, and then doing a
// style dialog with that.)
// 
// One solution which I've found to work well is to first try to set the port to
// 72x72 dpi, and then try to set it to 0x0 dpi.  Drivers which reset when passed
// 72x72 ignore the 0x0 setting, and drivers that need to be passed 0x0, are happy
// as well.  Remember that the order is important here.  If you tell the IW driver
// that you want 72x72 dpi, it will disable options in the print dialog that
// shouldn't be.  Only when it tries to set the bogus 0x0 dpi will it clean house.
//
// Unfortunately, this is one of the design short-comings of the current
// printing architecture, or more specifically, the current printer drivers.
// If drivers don't disable options at different resolutions, we have no
// problem at all.

	resetThedpi = 72;				// Pass 72x72 dpi to force the defaults.
	SetPrinterPortDPI(hPrint, &resetThedpi, &resetThedpi);

	resetThedpi = 0;				// Pass a bogus pair to force the defaults.
	SetPrinterPortDPI(hPrint, &resetThedpi, &resetThedpi);
									// One of those should work.

// Open the printer driver and check for errors.

	PrOpen();

	err = PrError();
	
	if (err == fnfErr)	// Here, "file not found" really means "no printer chosen."
		err = noPrinterChosenErr;
	

// If no errors, validate the print record and call PrJobDialog if we're not
// appending a DITL to the dialog.  If we are appending a DITL, call PrJobInit
// to get a pointer to the job dialog, then call PrDlgMain, passing our print
// handle and dialog init proc.

	if (!err)
	{
	 	PrValidate(hPrint);

		if (aDITLResID == nil)
			doIt = PrJobDialog(hPrint);
		else
		{
			gCurAppendDITLResID = aDITLResID;
			gPrtJobDialog = PrJobInit(hPrint);
			err = PrError();
	
			if (!err)
				doIt = PrDlgMain(hPrint, (PDlgInitProcPtr) DoJobDlgInit);
		}
	
// If no errors, but the user cancelled, set the return code to iPrAbort.
// Close up the driver and head back.

		err = PrError();

		if (!err && !doIt)
			err = iPrAbort;
	}

	PrClose();
	DoAnyUpdates();
	return err;
}


/************************************************************
  Append_DITL - This routine appends one DITL to another.
  this code is a C version of the code that appears in
  the Tech Note "How to Add Items to the Print Dialogs."  For
  more information, see that Tech Note.  This code also
  corrects a bug in the Tech Note's sample code, which added
  an extra short word to the concatenated DITL.
  
  This routine returns the adjusted item number of the first
  item of the appended DITL for the new (concatenated) DITL.

*************************************************************/

short Append_DITL(TPPrDlg dialog, short aDITLResID)
{
    Point         	offset;
    Rect            max_rect;
    ItemListHandle 	append_item_list;
    DialogItemPtr 	item;
    ItemListHandle	dlg_item_list;
    short          	first_item;
    short          	new_items, data_size, i;
    ByteAccess    	usb;
    OSErr          	err;
 
// Using the original dialog:
//
// 1. Remember the original window Size.
// 2. Set the offset Point to be the bottom of the original window.
// 3. Subtract 5 pixels from bottom and right, to be added back later after we
//	  have possibly expanded window.
// 4. Get working Handle to original item list.
// 5. Calculate our first item number to be returned to caller.
// 6. Get locked Handle to DITL to be appended.
// 7. Calculate count of new items.
 
    max_rect = ((DialogPeek) dialog)->window.port.portRect;
    offset.v = max_rect.bottom;
    offset.h = 0;
    max_rect.bottom -= 5;
    max_rect.right -= 5;
 
    dlg_item_list = (ItemListHandle) ((DialogPeek) dialog)->items;
    first_item = (*dlg_item_list)->max_index +2;
 
    append_item_list = (ItemListHandle) GetResource('DITL', aDITLResID);
    if (append_item_list == nil)
        return first_item;
 
    HLock((Handle) append_item_list);
    new_items = (*append_item_list)->max_index +1;
 
// For each item,
//	1. Offset the rectangle to follow the original window.
//	2. Make the original window larger if necessary.
//	3. fill in item handle according to type.
 
    item = (*append_item_list)->items;
    for (i = 0; i < new_items; i++)
    {
        OffsetRect(&item->bounds, offset.h, offset.v);
        UnionRect(&item->bounds, &max_rect, &max_rect);
        usb.integer = 0;
        usb.bytes[1] = item->data[0];
 
        switch ( item->type & 0x7F )
        {
            case ctrlItem + btnCtrl :
            case ctrlItem + chkCtrl :
            case ctrlItem + radCtrl :
                item->handle = (Handle) NewControl((DialogPtr) dialog,
												   &item->bounds,
												   (StringPtr)item->data,
												   true, 0, 0, 1, item->type & 0x03, 0);
            	break;
 
            case ctrlItem + resCtrl :
                item->handle = (Handle) GetNewControl(*(short*) (item->data +1),
													  (DialogPtr) dialog);
                (*(ControlHandle) item->handle)->contrlRect = item->bounds;
            	break;

            case statText :
            case editText :
                err = PtrToHand(item->data +1, &item->handle, usb.integer);
    			break;

            case iconItem :
                item->handle = GetIcon(*(short*) (item->data +1));
				break;
 
            case picItem :
                item->handle = (Handle) GetPicture(*(short*) (item->data +1));
            	break;
 
            default :
                item->handle = nil;
        }
 
        data_size = (usb.integer +1) & 0xFFFE;
        item = (DialogItemPtr) ((char*) item + data_size + sizeof(DialogItem));
    }
 
 // This is where the bug used to be.  The size of the data we're adding is
 // actually the size of the appended DITL, minus it's item count.
 
    err = PtrAndHand((*append_item_list)->items, (Handle) dlg_item_list,
                     GetHandleSize((Handle) append_item_list) -sizeof(short));

    (*dlg_item_list)->max_index += new_items;
    HUnlock((Handle) append_item_list);
    ReleaseResource((Handle) append_item_list);
 
    max_rect.bottom += 5;
    max_rect.right  += 5;
    SizeWindow((WindowPtr) dialog, max_rect.right, max_rect.bottom, true);
 
    return first_item;
}


/************************************************************
  DoJobDlgInit - This routine handles initialization and
  appending of our additional print dialog items.  It also
  gives us a chance to replace the driver's item handler
  with our own.

*************************************************************/

pascal TPPrDlg DoJobDlgInit(THPrint hPrint)
{
    short      itemType, oldResRefNum;
    Handle     itemH;
    Rect       itemBox;
 
#pragma unused(hPrint);

// Append our DITL to the print job dialog, then get the control handle of
// the "print maximum detail" checkbox and set it according to the user's
// last preference. We temporarily switch to our resource file to make
// sure we find *our* DITL, not somebody else's with the same ID.

	oldResRefNum = CurResFile();
	UseResFile(gAppResRefNum);

	gFirstItem = Append_DITL(gPrtJobDialog, gCurAppendDITLResID);

	UseResFile(oldResRefNum);

	GetDItem((DialogPtr) gPrtJobDialog, d_maxDetailItem + gFirstItem -1, &itemType,
			 &itemH, &itemBox);

	SetCtlValue((ControlHandle) itemH, (short) gPrintMaxDetail);


// If we're appending the ImageWriter DITL, we also have a couple radio buttons
// to intialize.  Set them both based on the user's last preferences.

	if (gCurAppendDITLResID == r_IW_AppendDITL)
	{
		GetDItem((DialogPtr) gPrtJobDialog, d_halftoneItem + gFirstItem -1, &itemType,
				 &itemH, &itemBox);

		SetCtlValue((ControlHandle) itemH, (short) !gPrintInColor);

		GetDItem((DialogPtr) gPrtJobDialog, d_colorItem + gFirstItem -1, &itemType,
				 &itemH, &itemBox);

		SetCtlValue((ControlHandle) itemH, (short) gPrintInColor);
	}

// Finally, store the driver's current item handler and set it to ours.  (Our
// item handler will in turn call the saved one, if an item other than ours
// is selected.  Return a pointer to the print job dialog, cause that's what's
// expected.

    gPrPItemProc = (Ptr) gPrtJobDialog->pItemProc;
    gPrtJobDialog->pItemProc = (PItemProcPtr) HandleJobItems;
    return gPrtJobDialog;
}


/************************************************************
  HandleJobItems - This routine handles selection of items
  in the print dialog.  If one of our items is selected, we
  handle it, otherwise we pass it to the driver's original
  item handler.

*************************************************************/

pascal void HandleJobItems(TPPrDlg theDialog, short itemNo)
{
	short   appendItemNo, itemType;
	Handle  itemH;
	Rect    itemBox;

// Translate the item number passed into an itme number with our DITL.  If
// the resulting item number is positive, the item is ours, else it's not.
// Handle the item if it's ours, otherwise pass it to the driver's routine.

	appendItemNo = itemNo - gFirstItem +1;
	
	if (appendItemNo > 0)
	{
		switch (appendItemNo)
		{
			case d_maxDetailItem:	// "Print maximum detail" checkbox.  Set it and
									// gPrintMaxDetail appropriately.

				GetDItem((DialogPtr) gPrtJobDialog, d_maxDetailItem + gFirstItem -1,
						 &itemType, &itemH, &itemBox);

				gPrintMaxDetail = !(Boolean) GetCtlValue((ControlHandle) itemH);
				SetCtlValue((ControlHandle) itemH, (short) gPrintMaxDetail);
				break;


// The next two items are only used by ImageWriter DITL appendage.

			case d_halftoneItem:	// "Halftone" radio button.
			case d_colorItem:		// "Color" radio button.
									// Set the controls and gPrintInColor appropriately.

				gPrintInColor = (appendItemNo == d_colorItem);

				GetDItem((DialogPtr) gPrtJobDialog, d_halftoneItem + gFirstItem -1,
						 &itemType, &itemH, &itemBox);

				SetCtlValue((ControlHandle) itemH, (short) !gPrintInColor);

				GetDItem((DialogPtr) gPrtJobDialog, d_colorItem + gFirstItem -1,
						 &itemType, &itemH, &itemBox);

				SetCtlValue((ControlHandle) itemH, (short) gPrintInColor);
				break;
		}
	}
	else	// T'aint ours - let the driver deal with it.
	{
		(pascal void (*) (TPPrDlg, short)) gPrPItemProc (theDialog, itemNo);
	}
}
