/*********************************************************************

	Final Printing.c
	
	This file contains all of the generic printing functions (except
	the print dialog routines) for the all-in-one "Adventures in
	Color Printing" sample, "The Final Adventure."
	
	Dave Hersey, Apple Developer Technical Support.
	
	 Edit Trail 
	
	creation:   								11/24/92  - dmh
	sync'd with other samples:					11/30/92  - dmh

	Hint: If you start looking at "PrintOne," you'll be able
	      to follow the chain of command easier.
	
	* Related functions are grouped together in the Mark menu. *
	
*********************************************************************/

#pragma segment Printing

#include "Final Headers.h"

static Boolean		gRectNotSetYet;		// Flag used by ColorBitsProc.
static Fixed		gPMHRes;			// Horizontal resolution,
static Fixed		gPMVRes;			// vertical resolution,
static Rect			gSrcRect;			// and bounds of PixMap currently being
										// examined in ColorBitsProc.

/************************************************************
  PrintOne - This routine is called when we print.  The
  routine determines which of the three printing methods to
  use (color ImageWriter printing, color printing, or
  halftone printing) and passes control on to that. 

*************************************************************/

OSErr PrintOne(WindowPtr wp)
{
	OSErr			err;
	unsigned char	devID;
	Boolean			isImageWriter, isColorPort;

// Get the print handle's wDev ID, and examine the high byte, and see
// if we're using the ImageWriter or ImageWriter LQ drivers.  If so,
// call DoJobDialog passing our ImageWriter dialog appendage ID. If not,
// pass the non-ImageWriter one.

	devID = (*ghPrint)->prStl.wDev >>8;
	isImageWriter = ((devID == kIW_wdevID) || (devID == kIWLQ_wdevID));	

	if (isImageWriter)
		err = DoJobDlog(ghPrint, r_IW_AppendDITL);
	else
		err = DoJobDlog(ghPrint, r_nonIW_AppendDITL);


// If no errors, save the print handle and see if we have a color port.
// If so, call the printing routine for CGrafPorts.  If not, call the
// color ImageWriter routine if we're using an ImageWriter and the user
// wants to print in color, otherwise call the halftone routine.

	if (!err)
	{
		err = SavePHdl(ghPrint, gAppResRefNum);

		if (!err)
		{
			isColorPort = HaveColorPrPort(ghPrint, &err);
			SetGWorld((GWorldPtr) FrontWindow(), nil);

			if (!err)
			{
				if (isColorPort)
					return DoColorPrint(wp, gPrintMaxDetail);
				else
					if (isImageWriter && gPrintInColor)
						return DoColorIWPrint(wp, gPrintMaxDetail);
			}
		}
	}
	
	if (!err)
		return DoHTPrint(wp, gPrintMaxDetail);
	else
		return err;
}


/************************************************************
  HaveColorPrPort - This routine tells us if the selected
  driver will generate a GrafPort or CGrafPort when used
  with the passed handle.  The routine returns true if
  we'll be using a CGrafPort, false otherwise.

  The printer driver is expected to be closed upon entry,
  and is therefore opened and closed in this routine.

*************************************************************/

Boolean HaveColorPrPort(THPrint hPrint, OSErr *anErr)
{
	Boolean		haveCGrafPort = false;
	TPPrPort	dummyPort;
	TPrStatus	statusRec;
	
	if (hPrint)
	{

// Open a document and check for errors.
		
		PrOpen();
		*anErr = PrError();
		
		if (*anErr == noErr)
		{
			dummyPort = PrOpenDoc(hPrint, nil, nil);
			*anErr = PrError();
		}


// If no errors, check the port's rowBytes value.
		
		if (*anErr == noErr)
		{
			haveCGrafPort = (dummyPort->gPort.portBits.rowBytes < 0);
		
			
// We don't want to print yet, so kill the job by setting an error.
// Clean up by closing the document and calling PrPicFile to delete
// any spool file we may have created.  Finally, clear the error we
// set.

			PrSetError(iPrAbort);
			PrCloseDoc(dummyPort);
			
			if ((*hPrint)->prJob.bJDocLoop == bSpoolLoop)
				PrPicFile(hPrint, dummyPort, nil, nil, &statusRec);

			PrSetError(noErr);
		}

		PrClose();
	}
	else
		*anErr = nilHandleErr;

	return haveCGrafPort;
}


/************************************************************
  GetPrinterRes - This routine returns the highest physical
  square resolution of the selected printer.

  The printer driver is expected to be closed upon entry,
  and is therefore opened and closed in this routine.

*************************************************************/

short GetPrinterRes()
{
	TGetRslBlk		getResRec;
	short			bestRes, rec;
	
// Open the driver for our PrGeneral call.  Assume that we'll
// return 72x72 DPI until we find otherwise.

	PrOpen();
	bestRes = 72;
	
	if (!PrError())
	{

// Ask PrGeneral for the resolution records for this driver and find the
// find the largest square discrete resolution.

		getResRec.iOpCode = getRslDataOp;
		PrGeneral((Ptr) &getResRec);

		if ((!ResError()) && (!getResRec.iError))
			for (rec = 0; rec < getResRec.iRslRecCnt; rec++)
				if ((getResRec.rgRslRec[rec].iXRsl == getResRec.rgRslRec[rec].iYRsl) &&
					(getResRec.rgRslRec[rec].iXRsl > bestRes))
					bestRes = getResRec.rgRslRec[rec].iXRsl;
	}
		
// Close the driver and return the highest square discrete resolution.
	
	PrClose();
	return bestRes;
}


/************************************************************
  GetBestDPI - This routine obtains the best horizontal &
  vertical resolutions to use for printing with the
  selected driver.
  
  pxDPI, pyDPI --  The caller places an "ideal" resolution
  				   pair here (what they really want to use).
				   Also, this is where the routine returns
				   the resolutions it decides on.

  xDPI_ceiling,
  yDPI_ceiling --  The caller places the maximum resolution
  				   they want in either direction.  For
				   example, if you didn't want values larger
				   than 300 dpi returned, you would put 300
				   in both of these parameters.

  wantSquareDPI -- If true, only square resolutions (equal
  				   horizontal and vertical components) will
				   be considered.

  The printer driver is expected to be closed upon entry to
  this routine, and is therefore opened and closed around
  the PrGeneral code.  If PrGeneral is not supported by this
  driver, or if an error occurs, the routine returns
  72x72 dpi, which is the default for Macintosh printer
  drivers.
  
  If the "ideal" resolution the caller passes in is
  available, we choose that, ignoring wantSquareDPI,
  xDPI_ceiling and yDPI_ceiling.  We figure that the calling
  routine knows more about the ideal resolution it requests
  than we do.

*************************************************************/

void GetBestDPI(short *pxDPI, short *pyDPI, short xDPI_ceiling,
				short yDPI_ceiling, Boolean wantSquareDPI)
{
	TGetRslBlk		getResRec;
	Boolean			exactMatch = false;
	short			bestResX, bestResY, xDPI, yDPI,
					desiredResX, desiredResY, rec;

// Store the desired resolution that the caller passed to us through
// the pxDPI and pyDPI parameters.  If 72x72 dpi is requested, return
// those values, since that resolution must always be supported by a
// Mac printer driver (even though it may not appear in the PrGeneral
// getRsl list) for some (confused) printer drivers.

	desiredResX = *pxDPI;
	desiredResY = *pyDPI;

	if ((desiredResX == 72) && (desiredResY == 72)) return;


// Open the driver for our PrGeneral call.  Assume that we'll return
// 72x72 DPI until we find otherwise, and ask PrGeneral for the
// resolution records for this driver.

	PrOpen();
	
	bestResX = bestResY = 72;

	if (!PrError())
	{
		getResRec.iOpCode = getRslDataOp;
		PrGeneral((Ptr) &getResRec);

		if ((!ResError()) && (!getResRec.iError))
		{

// First check for the exact resolution pair that the caller
// requested.  We'll start by checking the range of resolutions
// supported to see if the pair is within that.

			if ((getResRec.xRslRg.iMin <= desiredResX) &&
				(getResRec.xRslRg.iMax >= desiredResX) &&
				(getResRec.yRslRg.iMin <= desiredResY) &&
				(getResRec.yRslRg.iMax >= desiredResY))
				exactMatch = true;


// If we didn't find an exact match, check the driver's
// discrete resolutions to see if we have one there.

			for (rec = 0; (!exactMatch) && (rec < getResRec.iRslRecCnt); rec++)
				if ((getResRec.rgRslRec[rec].iXRsl == desiredResX) &&
					(getResRec.rgRslRec[rec].iYRsl == desiredResY))
					exactMatch = true;


// If we found an exact match use it.  Otherwise, loop through
// each resolution record and find the one which best matches
// our criteria.

			if (exactMatch)	
			{
				bestResX = desiredResX;
				bestResY = desiredResY;
			}
			else
				for (rec = 0; (rec < getResRec.iRslRecCnt); rec++)
				{
					xDPI = getResRec.rgRslRec[rec].iXRsl;
					yDPI = getResRec.rgRslRec[rec].iYRsl;
	
					if ((xDPI <= xDPI_ceiling) && (yDPI <= yDPI_ceiling) &&
						(!wantSquareDPI || (xDPI == yDPI)) &&
						((xDPI * yDPI) > (bestResX * bestResY)))
					{
						bestResX = xDPI;
						bestResY = yDPI;
					}
				}
		}
	}
		
// Return the best resolution pair we found and close the driver.
	
	*pxDPI = bestResX;
	*pyDPI = bestResY;
	PrClose();
}


/************************************************************
  SetPrinterPortDPI - This routine sets the resolution of
  the printer port to the passed horizontal and vertical
  resolutions.  hPrint is the current print handle.

  The printer driver is expected to be closed upon entry to
  this routine, and is therefore opened and closed around
  the PrGeneral code.  If the current driver doesn't support
  PrGeneral, or if an error occurs, the passed horizontal
  and vertical resolutions are changed to 72 dpi, which is
  what printer drivers on the Macintosh default to.

*************************************************************/

void SetPrinterPortDPI(THPrint hPrint, short *pxDPI, short *pyDPI)
{
	TSetRslBlk		setResRec;
	OSErr			err1, err2;

	PrOpen();

	if (!PrError())
	{
		setResRec.iOpCode = setRslOp;
		setResRec.hPrint = hPrint;
		setResRec.iXRsl = *pxDPI;
		setResRec.iYRsl = *pxDPI;
		PrGeneral((Ptr) &setResRec);
	
		err1 = ResError();
		err2 = setResRec.iError;
		
		if (err1 || err2)
			*pxDPI = *pyDPI = 72;
	}

	PrClose();
}


/************************************************************
  SetupPrintHdl - This routine returns a default print
  handle for the current printer driver.

  The printer driver is expected to be closed upon entry,
  and is therefore opened and closed in this routine.
  
  NOTE: This function has been slightly modified from the
  other samples in order to initialize our appended options
  for the print dialog.  This is necessary in the event that
  a previously saved print record can't be loaded with the
  user settings.

*************************************************************/

OSErr SetupPrintHdl(THPrint *hPrint)
{
	OSErr err;
	
// Initialize our dialog settings, create a handle, open the
// printer driver, set the default values for the handle, and
// close the driver.  Be on the lookout for errors.

	gPrintInColor = gPrintMaxDetail = false;
	*hPrint = (THPrint) NewHandle(sizeof(TPrint));

	if (*hPrint != nil)
	{
		PrOpen();

		err = PrError();
		
		if (err == fnfErr)	// Here, "file not found" really means "no printer chosen."
			err = noPrinterChosenErr;
		
		if (err) return err;

		PrintDefault(*hPrint);
		err = PrError();

		PrClose();
	}
	else
		err = MemError();
	
	return err;
}


/************************************************************
  GetResIDString - This routine creates a unique resource
  ID by adding (the driver version * 10000) and the high
  byte of the wDev.  All values are derived from the print
  handle passed and the currently opened printer driver.
  Once we have the ID, it's converted to a string to use
  for the resource's name.  (Our ID might be 8 digits long,
  so we can't use it as a normal resource ID which is only 5
  digits long.)
  
  Because we call PrDrvrVers, this function must be called
  between PrOpen & PrClose.

*************************************************************/

void GetResIDString(THPrint hPrint, Str255 resIDStr)
{
	short		wDev;
	long		drvrVers, codeID;

	wDev = (*hPrint)->prStl.wDev >> 8;
	drvrVers = PrDrvrVers();
	codeID = (drvrVers * 10000) + wDev;
	NumToString(codeID, resIDStr);
}


/************************************************************
  SavePHdl - This routine saves the passed print handle in
  the indicated resource fork, using a name which identifies
  it uniquely for the printer driver and version being used.
  If there is already a print record resource with this name,
  it's deleted and replaced.

  The printer driver is expected to be closed upon entry,
  and is therefore opened and closed in this routine.
  
  NOTE: This function has been slightly modified from the
  other samples in order to store our appended options from
  the print dialog.  To do this, we've introduced a new
  data type: PrintSettings, which contains a print record
  and our additional settings.  See the data type in the
  header file for more info.

*************************************************************/

OSErr SavePHdl(THPrint hPrint, short resRefNum)
{
	short				savedRefNum, uniqID;
	OSErr				err;
	Handle				oldHdl;
	Str255				resIDStr;
	PrintSettingsHdl	settingsHdl;

// Save the current resource file refNum, open the printer
// driver, get our unique name for the passed print handle
// and close the driver.

	savedRefNum = CurResFile();
	PrOpen();
	err = PrError();
	
	if (err == fnfErr)	// Here, "file not found" really means "no printer chosen."
		err = noPrinterChosenErr;
	
	if (err) return err;

	if (!err) GetResIDString(hPrint, resIDStr);
	PrClose();
	

// If there's already a resource with this name in the passed
// resource file, delete it.

	UseResFile(resRefNum);
	
	if (!err)
	{
		oldHdl = GetNamedResource(r_newPrintHdlType, resIDStr);

		if (oldHdl)
		{
			RmveResource(oldHdl);
			DisposHandle(oldHdl);
		}

// Save a new "print handle & settings" resource with a unique ID and
// our coded name.
	
		uniqID = UniqueID(r_newPrintHdlType);

		settingsHdl = (PrintSettingsHdl) NewHandle(sizeof(PrintSettings));

		if (!(err = MemError()))
		{
			BlockMove(*hPrint, &(*settingsHdl)->printRec, sizeof(TPrint));
			(*settingsHdl)->printInColor = gPrintInColor;
			(*settingsHdl)->printMaxDetail = gPrintMaxDetail;
			
			AddResource((Handle) settingsHdl, r_newPrintHdlType, uniqID, resIDStr);
			err = ResError();
			
			if (!err)
			{
				WriteResource((Handle) settingsHdl);
				err = ResError();
			}


// Get rid of the data handle, update the resource file and switch back
// to the original resource file.

			if (!err)
				ReleaseResource((Handle) settingsHdl);
			else
				DisposeHandle((Handle) settingsHdl);
		}
	}
	
	UpdateResFile(resRefNum);
	UseResFile(savedRefNum);
	return err;
}


/************************************************************
  LoadPHdl - This routine loads a print handle resource
  which matches the name we uniquely encode.  If there is
  no print record resource in the indicated resource file
  which matches our uniquely encoded name, the function
  returns resNotFound.
  
  Upon entry, hPrint is expected to be a valid print handle
  or nil.

  The printer driver is expected to be closed upon entry,
  and is therefore opened and closed in this routine.
  
  NOTE: This function has been slightly modified from the
  other samples in order to load our appended options for
  the print dialog.  To do this, we've introduced a new
  data type: PrintSettings, which contains a print record
  and our additional settings.  See the data type in the
  header file for more info.

*************************************************************/

OSErr LoadPHdl(THPrint *hPrint, short resRefNum)
{
	short					savedRefNum;
	OSErr					err;
	PrintSettingsHdl		newHdl;
	Str255					resIDStr;

// Save the current resource file refNum, open the printer driver,
// get the name that we would have encoded this resource with and
// close the driver.

	savedRefNum = CurResFile();
	PrOpen();

	err = PrError();
	
	if (err == fnfErr)	// Here, "file not found" really means "no printer chosen."
		err = noPrinterChosenErr;
	
	if (err) return err;

	GetResIDString(*hPrint, resIDStr);
	PrClose();


// Get our resource, if there is one, out of the indicated resource
// file.  If there isn't one, return resNotFound.  Finally, switch
// back to the original resource file.

	UseResFile(resRefNum);
	
	newHdl = (PrintSettingsHdl) GetNamedResource(r_newPrintHdlType, resIDStr);

	if (newHdl)
	{

// Got one.  Store the settings for our appended dialog items in their
// global variables, then detach the resource and convert its handle

		gPrintInColor = (*newHdl)->printInColor;
		gPrintMaxDetail = (*newHdl)->printMaxDetail;
		DetachResource((Handle) newHdl);
		SetHandleSize((Handle) newHdl, sizeof(TPrint));
		if (*hPrint) DisposHandle(*(Handle *) hPrint);
		*hPrint = (THPrint) newHdl;
		err = noErr;
	}
	else
		err = resNotFound;
	
	UseResFile(savedRefNum);
	return err;
}


/************************************************************
  ColorBitsProc - This routine is our QuickDraw bottleneck
  replacement for stdBits.  We use this routine when we're
  trying to determine the resolution of a picture based
  solely on the resolution of its contents.

*************************************************************/

pascal void ColorBitsProc(BitMap *srcBits, Rect *srcRect, Rect *dstRect,
						  short mode, RgnHandle maskRgn)
{
    PixMapPtr  pm;

#pragma unused(srcRect, dstRect, mode, maskRgn);

// If this is a PixMap (rowBytes < 0), store the resolution
// and bounds of the PixMap in our global variables.  Since
// BitMaps don't have resolution fields, ignore those.  We
// keep expanding the bounds everytime we find a pixmap in
// case the image was banded (we want gSrcRect to encompass
// all bands).

	if (srcBits->rowBytes < 0)
	{
		pm = (PixMapPtr) srcBits;
		gPMHRes = MAX(gPMHRes, pm->hRes);
		gPMVRes = MAX(gPMVRes, pm->vRes);
		
		if (gRectNotSetYet)
		{
			gSrcRect = pm->bounds;
			gRectNotSetYet = false;
		}
		else
		{
			gSrcRect.top = MIN(gSrcRect.top, pm->bounds.top);
			gSrcRect.left = MIN(gSrcRect.left, pm->bounds.left);
			gSrcRect.bottom = MAX(gSrcRect.bottom, pm->bounds.bottom);
			gSrcRect.right = MAX(gSrcRect.right, pm->bounds.right);
		}
	}
}


/************************************************************
  GetPixMapRes - This routine returns the resolution of the
  contents of a picture (specifically the resolution of its
  PixMap).  We assume that there will be at most one PixMap
  in our image, and given the type of images we're dealing 
  with, that's a valid assumption.  It's not clear how we
  would determine the correct resolution in a situation
  where multiple PixMaps were involved anyway.
  
  This routine is used if the original picture was not
  created with OpenCPicture.  If it was, we can get the
  resolution by just checking the picture header, as in
  GetImageRes below.

*************************************************************/

OSErr GetPixMapRes(WindowInfoHdl windInfo, Fixed *hRes, Fixed *vRes, Rect *srcRect)
{
	OSErr		err = noErr;
	PicHandle	aPict;
	CQDProcs	bottlenecks;
	GrafPtr		oldPort;
	CGrafPort	cPort;
	Boolean		spoolThisPict;

// Assume the image is 72x72 dpi and that the picture's bounds
// are the same as its picFrame.

    spoolThisPict = !(*windInfo)->loaded;
	aPict = (*windInfo)->windowPict;
	gPMHRes = gPMVRes = Long2Fix(72);
	gSrcRect = (*aPict)->picFrame;
	gRectNotSetYet = true;


// Save off the current port, and open up a new color
// port to work in.  We do this because we *must* be drawing
// into a CGrafPort for us to have PixMaps passed to our
// bottleneck routine without being converted into BitMaps.

	GetPort(&oldPort);
	OpenCPort(&cPort);


// Get the standard CGrafPort bottleneck routines, and add
// our replacement for stdBits.  This replacement routine
// will snoop for a PixMap, and store the resolution and
// bounds of any it finds in our global variables.  If
// none are found, we'll still have the defaults set up
// above.

	if (spoolThisPict)
		err = SpoolInstall(windInfo, kArrowCursor);
	else
	{
		SetStdCProcs(&bottlenecks);
		cPort.grafProcs = &bottlenecks;
	}

	if (!err)
	{
		cPort.grafProcs->bitsProc = (Ptr) ColorBitsProc;


// Draw the picture so our bottleneck routine is called.
// Afterwards, set back to the old port we were using and
// Close the CGrafPort we created.  Finally, return the
// resolution and bounds rectangle for this image.

		DrawPicture(aPict, &(*aPict)->picFrame);

		if (spoolThisPict)
			err = SpoolRemove();
	}

	SetPort(oldPort);
	CloseCPort(&cPort);

	*hRes = gPMHRes;
	*vRes = gPMVRes;
	*srcRect = gSrcRect;

	return err;
}


/************************************************************
  GetImageRes - This routine returns the native resolution
  of a picture, and the appropriate drawing rectangle for
  the image at that resolution.

*************************************************************/

OSErr GetImageRes(WindowInfoHdl windInfo, short *ixDPI, short *iyDPI, Rect *srcRect)
{
	PicHandle			wPICT;
	Fixed				hRes, vRes;
	PartialPicture		partialPict;
	Boolean				extndV2PICT = false;
	OSErr				err = noErr;

	wPICT = (*windInfo)->windowPict;
	HLock((Handle) windInfo);
	BlockMove(&(*windInfo)->partialPict, &partialPict, sizeof(PartialPicture));
	HUnlock((Handle) windInfo);

// If we have a version 2 picture, find out if it's an extended
// version 2 picture.  If so, the version field in the picture's
// header will be -2.  In that case, pull the resolution and bounding
// rect out of the picture header.

	if (partialPict.pictVersion == 2)
		if (partialPict.pictHeader.version == -2)
		{
			extndV2PICT = true;

			hRes = partialPict.pictHeader.hRes;
			vRes = partialPict.pictHeader.vRes;
			*srcRect = partialPict.pictHeader.srcRect;
		}

// If we have a version 1 picture, or a non-extended version 2
// picture, set the resolution and bounding rect for the
// picture based on the PixMap it contains.

	if (!extndV2PICT)
		err = GetPixMapRes(windInfo, &hRes, &vRes, srcRect);

	*ixDPI = FixRound(hRes);
	*iyDPI = FixRound(vRes);
	
	return err;
}


/************************************************************
  BandPicture - This routine bands the passed picture into
  the destination GWorld, using ditherCopy if desired.
  
  wp is the window containing the picture.  It has one of
  our WindowInfoHdl handles hanging off the window refCon
  field.  From that structure, we can either locate the
  picture's handle in memory or spool the picture in from
  disk.

*************************************************************/

OSErr BandPicture(WindowPtr	wp, Rect *srcRect, GWorldPtr destGWorld,
				  Boolean useDitherCopy)
{
	OSErr			err = noErr;
	Rect			pictFrame, bandFrame, ditherFrame, ditherRect1, ditherRect2;
	GWorldPtr		curGWorld, bandGWorld = nil, ditherGWorld = nil;
	PixMapHandle	srcPix, destPix, ditherPix;
	PicHandle		srcPict;
	Boolean			moreToDo, spoolThisPict, useSecondGWorld;
	GDHandle		curGDH;
	WindowInfoHdl	windInfo;
	long			bandHeight, bandWidthInBytes, maxBlockSize;
	short			bitsPerPixel, rowWidth, padAmt;

	windInfo = (WindowInfoHdl) GetWRefCon(wp);
	srcPict = (*windInfo)->windowPict;
	spoolThisPict = !(*windInfo)->loaded;
	useSecondGWorld = useDitherCopy;


// Save the current GWorld for restoration later, then lock down the destination
// GWorld's bits and erase them so that no garbage is left in the GWorld.  Lastly,
// get the picture's bounds, putting the top-left corner at (0, 0).

	GetGWorld(&curGWorld, &curGDH);

	destPix = GetGWPixMap(destGWorld);
	LockPixels(destPix);
	
	SetGWorld(destGWorld, nil);
	EraseRect(&(*destPix)->bounds);
	SetGWorld(curGWorld, curGDH);

	pictFrame = *srcRect;
	OffsetRect(&pictFrame, -pictFrame.left, -pictFrame.top);

// Determine the size bands to use.  We first get the maximum contiguous block 
// possible in our heap.  Taking 85% of that value, we figure out how many
// horizontal bands could fit in that.
//
// If we're dithering, we'll need an extra GWorld for dither-distortion correction.
// The process we'll use will require a "pad" of a few extra rows of pixels in
// each GWorld (padAmt).  Therefore, our calculated band size will be:
//
// for srcCopy:    (#of bytes available / number of bytes per row) -padAmt
//
// Remember that the number of bytes per row will be rounded up to the nearest
// long word, since 32-bit QuickDraw makes GWorlds which are 32-bit aligned.
//
// For ditherCopy we do the same calculations as above, but we adjust bitsPerPixel to
// reflect the sum of the pixels in the 32 bit GWorld and those in the additional dither
// GWorld.
//
// If our calculations come up with a band size that's bigger than the original image,
// then just use the height of the original image.

	bandFrame = (*destPix)->bounds;

	padAmt = (useDitherCopy)? kPadAmount: 0;
	maxBlockSize = MaxBlock() * .85;
	bitsPerPixel = (useDitherCopy)? 32 + (*destPix)->pixelSize: 32;
	rowWidth = (bandFrame.right - bandFrame.left +1);
	bandWidthInBytes = (bitsPerPixel * rowWidth)/8;
	bandWidthInBytes += bandWidthInBytes % 4;		// Make sure we're on a LW boundary.

	bandHeight = (maxBlockSize / bandWidthInBytes) -padAmt;
	
	if ((bandHeight +padAmt) >= (bandFrame.bottom - bandFrame.top))
	{
		useSecondGWorld = false;
		padAmt = 0;
		bandHeight = (bandFrame.bottom - bandFrame.top);
	}

// Create the rectangle for the banding GWorld.  The band buffer will be as wide as
// the original image and as tall as either the available band size or the original
// image, whichever is smaller.

	SetRect(&bandFrame, 0, -padAmt, bandFrame.right, bandHeight +padAmt);


// If we're dithering, and using the second GWorld, create the dither GWorld which
// we use to filter our some of the effects of ditherer banding.  This GWorld will
// mirror the destination, except for the bounds we use.

	if (useSecondGWorld)
	{
		ditherFrame = bandFrame;
		err = NewGWorld(&ditherGWorld, (*destPix)->pixelSize, &ditherFrame, (*destPix)->pmTable,
						GetGWorldDevice(destGWorld), 0);
	}


// While there's more to do, band our image.  On every pass we create a 32-bit GWorld
// to band the image through, which we dispose of at the end of the loop.  (UpdateGWorld
// is not used because its behaviour is sometimes questionable when we're only changing
// the bounds of a GWorld.)  After locking down the pixels, we set this as the
// current GWorld and draw the picture into it, clipping to the GWorld's bounds.  This
// creates a band which we then CopyBits to the destination GWorld, or pass through
// ditherGWorld if we're dithering and banding.
//
// Of particular importance are the rectangles we're using:
//
// When we cut the bands out of the picture, we call DrawPicture using pictFrame, and
// clipping to bandFrame.  If we used bandFrame in DrawPicture, every band would be a
// squished up representation of the original image, which obviously isn't what we want.
// bandFrame is updated on each pass so that it drops one band lower, and the DrawPicture
// call copies the next logical band to the 32-bit GWorld.
//
// When dithering and banding, the padAmt is non-zero, and useSecondGWorld is true.  In
// this case, we draw the picture into our 32-bit GWorld and then dither from there to
// our ditherGWorld.  Finally, we srcCopy that image to the destination GWorld, using
// a rectangle that leaves out the top padAmt rows of the dithered image.  By doing this,
// we negate some of the effects that dithering has on the fringes of our bands.  We
// still end up copying all the rows, but we don't copy the pixels that are dithered
// on the top fringe of ditherGWorld between consecutive bands.
//
// We also check for command-period presses while we do this, in case the user gets
// bored and wants to cancel.

	moreToDo = true;
	
	if (!err)
	do
	{
		err = NewGWorld(&bandGWorld, 32, &bandFrame, nil, nil, 0);
		if (!err)
		{
			srcPix = GetGWPixMap(bandGWorld);
			LockPixels(srcPix);

			SetGWorld(bandGWorld, nil);
			ClipRect(&bandFrame);


// If we need to spool this image, install the spooling bottleneck procedure.
 
 			if (spoolThisPict)
				err = SpoolInstall(windInfo, r_stdCurs4);

			if (!err)
			{
				DrawPicture(srcPict, &pictFrame);


// If we spooled this image, remove the spooling bottleneck procedure.
 
				if (spoolThisPict)
					err = SpoolRemove();

				SetGWorld(destGWorld, nil);
				
				if (useSecondGWorld)
				{
					ditherPix = GetGWPixMap(ditherGWorld);
					LockPixels(ditherPix);

					CopyBits((BitMap *) *srcPix, (BitMap *) *ditherPix, &bandFrame,
							 &ditherFrame, ditherCopy, nil);

					ditherRect1 = ditherFrame;	// Offset the rectangles so we don't
					ditherRect1.top += padAmt;	// copy the topmost dithered pixels.
					ditherRect2 = bandFrame;
					ditherRect2.top += padAmt;
	
					CopyBits((BitMap *) *ditherPix, (BitMap *) *destPix, &ditherRect1,
							 &ditherRect2, srcCopy, nil);

					UnlockPixels(ditherPix);
				}
				else
					CopyBits((BitMap *) *srcPix, (BitMap *) *destPix, &bandFrame,
							 &bandFrame, ((useDitherCopy)? ditherCopy: srcCopy), nil);
				
				UnlockPixels(srcPix);
			
				if (useSecondGWorld)
					moreToDo =  (ditherRect2.bottom < pictFrame.bottom);
				else
					moreToDo =  (bandFrame.bottom < pictFrame.bottom);
				
				OffsetRect(&bandFrame, 0, padAmt + MIN(bandHeight, pictFrame.bottom - bandFrame.bottom));
				DisposeGWorld(bandGWorld);
				bandGWorld = nil;
	
				SetGWorld(curGWorld, curGDH);
				if (!err) err = CommandPeriod();
			}
		}
	}
	while (moreToDo && (!err));


// Unlock the destination bits, set the current GWorld to what we started with, and
// dispose of the GWorlds.  Return any errors.

	UnlockPixels(destPix);

	SetGWorld(curGWorld, curGDH);
	if (bandGWorld) DisposeGWorld(bandGWorld);
	if (ditherGWorld) DisposeGWorld(ditherGWorld);

	return err;
}


/************************************************************
  CenterRect - This routine centers srcRect within destRect
  and returns it in ctrRect.  It's mainly used for centering
  output on the printed page and for centering windows on
  the screen.  The source rectangle should not be larger
  than the destination rectangle in either direction.  If
  it is, the rectangle will not be adjusted in that
  direction.

*************************************************************/

void CenterRect(Rect *srcRect, Rect *destRect, Rect *ctrRect)
{
	Rect	srcR, dstR;
	short	srcWidth, srcHeight, destWidth, destHeight, dx, dy;

	srcR = *srcRect;
	dstR = *destRect;

	OffsetRect(&srcR, dstR.left - srcR.left, dstR.top - srcR.top);

	srcWidth = srcR.right - srcR.left +1;
	srcHeight = srcR.bottom - srcR.top +1;
	destWidth = dstR.right - dstR.left +1;
	destHeight = dstR.bottom - dstR.top +1;
	
	dx = MAX(0, (destWidth-srcWidth)/2);
	dy = MAX(0, (destHeight-srcHeight)/2);

	*ctrRect = srcR;
	OffsetRect(ctrRect, dx, dy);
}


/************************************************************
  IdleProc - This routine is a PrIdleProcPtr that spins our
  cursor and looks for command-period presses during
  printing.

*************************************************************/

pascal void IdleProc()
{
	OSErr		err;
	
	BumpCursor();
	if (err = CommandPeriod()) PrSetError(err);
}


/************************************************************
  CommandPeriod - This routine checks to see if the user has
  pressed command-period.  The Printing Manager does this
  for us, but a routine like this is helpful to allow users
  to cancel in the middle of a rendering routine, rather
  than only Printing Manager operations.
  
  If there's a command-period in the event queue, the
  routine returns iPrAbort just like the Printing Manager.
  Otherwise, it returns noErr.

*************************************************************/

OSErr CommandPeriod()
{
	Boolean		cancel = false;
	EventRecord	evtRec;
	
	while (!cancel && (WaitNextEvent(keyDownMask | autoKeyMask, &evtRec, 0, nil)))
		cancel = (evtRec.modifiers & cmdKey) &&
				 ((evtRec.message & charCodeMask) == '.');

	return ((cancel)? iPrAbort: noErr);
}
