/*
	File:		Moe.c

	Contains:	Moe component routines.

				Refer to develop Issue 15, "Managing Component Registration",
				for details on this code.

	Written by:	Gary Woodcock

	Copyright:	 1993 by Apple Computer, Inc.

	Change History (most recent first):

*/

//-----------------------------------------------------------------------
// Includes

#include "MoePrivate.h"

#include <Errors.h>
#include <Memory.h>
#include <Packages.h>
#include <SysEqu.h>
#include <OSUtils.h>

//-----------------------------------------------------------------------

#ifdef BUILD_LINKED

// Use this declaration when we're running linked (for debugging)
pascal ComponentResult
MoeDispatcher (ComponentParameters *params, Handle storage)
											 
#else

// Use this declaration when we're a standalone component
pascal ComponentResult
main (ComponentParameters *params, Handle storage)

#endif BUILD_LINKED

{
	// This routine is the main dispatcher for the component
	
	ComponentResult		result = noErr;
	ComponentFunction	moeFunction = nil;
	
	// Did we get a Component Manager request code (< 0)?
	if (params->what < 0)
	{
		switch (params->what)
		{
			case kComponentOpenSelect:			// Open request
			{
				moeFunction = _MoeOpen;
				break;
			}
			case kComponentCloseSelect:			// Close request
			{
				moeFunction = _MoeClose;
				break;
			}
			case kComponentCanDoSelect:			// Can Do request
			{
				result = CallComponentFunction (params, 
					(ComponentFunction) _MoeCanDo);
				break;
			}
			case kComponentVersionSelect:		// Version request
			{
				result = CallComponentFunction (params,
					(ComponentFunction) _MoeVersion);
				break;
			}
			case kComponentRegisterSelect:		// Register request
			{
				result = CallComponentFunction (params,
					(ComponentFunction) _MoeRegister);
				break;
			}
			case kComponentTargetSelect:		// Target request not supported
			case kComponentUnregisterSelect:	// Unregister request not supported
			default:							// Unknown request
			{
				result = badComponentSelector;
				break;
			}
		}
	}
	else	// Unknown request
	{
		result = badComponentSelector;
	}
	if (moeFunction != nil)
	{
		result = CallComponentFunctionWithStorage (storage, params, moeFunction);
	}
	return (result);
}
											 
//-----------------------------------------------------------------------

pascal ComponentResult
_MoeOpen (Handle storage, ComponentInstance self)
{
	#pragma	unused (storage)
	
	ComponentResult	result = noErr;
	
	#ifdef THINK_C
	SetComponentInstanceA5 (self, *(long *) CurrentA5);
	#endif THINK_C
	
	// Can we open another instance?
	if (CountComponentInstances ((Component) self) <= kMaxMoeInstances)
	{
		// Set up our instance storage		
		MoePrivateGlobalsHdl	globals = (MoePrivateGlobalsHdl) NewHandleClear (sizeof (MoePrivateGlobals));
		
		// Did we get our storage?
		if (globals != nil)
		{
			// Keep a reference to self
			(**globals).self = (Component) self;
						
			// Set storage ref
			SetComponentInstanceStorage (self, (Handle) globals);
		}
		else	// NewHandleClear failed
		{
			result = MemError();
		}
	}
	else	// No more instances can be opened
	{
		result = -1L;	// Return anonymous error
	}
	return (result);
}

//-----------------------------------------------------------------------

pascal ComponentResult
_MoeClose (Handle storage, ComponentInstance self)
{
	MoePrivateGlobalsHdl	globals = (MoePrivateGlobalsHdl) storage;
	ComponentResult			result = noErr;
	
	// Do we have any clean up to do?
	if (globals != nil)
	{
		// Dispose globals
		DisposHandle ((Handle) globals);
	}
	return (result);
}

//-----------------------------------------------------------------------

pascal ComponentResult
_MoeCanDo (short selector)
{
	// Case on the request code
	switch (selector)
	{
		// Component Manager request codes
		case kComponentOpenSelect:
		case kComponentCloseSelect:
		case kComponentCanDoSelect:
		case kComponentVersionSelect:
		case kComponentRegisterSelect:
		{
			return (true);
		}
		
		// Unsupported or unknown request codes
		case kComponentTargetSelect:		// Target request not supported
		case kComponentUnregisterSelect:	// Unregister request not supported
		default:							// Not a request code we recognize
		{
			return (false); 
		}
	}
}

//-----------------------------------------------------------------------

pascal ComponentResult
_MoeVersion (void)
{
	// Return the version info
	return (moeInterfaceRev);
}

//-----------------------------------------------------------------------

pascal ComponentResult
_MoeRegister (void)
{
	long	dummy;
	
	// Beep once to let us know Moe's registered
	SysBeep(5);
	
	// Delay for a sec (just to keep all the components' beeps from
	// running together)
	Delay (60, &dummy);
	
	return (noErr);
}

//-----------------------------------------------------------------------

#ifdef THINK_C
#ifdef BUILD_LINKED

Component
RegisterMoe (void)
{
	ComponentDescription	theDesc;
	Handle					name;
	Component				theComp;

	// Set up component description
  	theDesc.componentType = 'moe ';
  	theDesc.componentSubType = kAnyComponentSubType;
	theDesc.componentManufacturer = 'appl';
  	theDesc.componentFlags = cmpWantsRegisterMessage;
  	theDesc.componentFlagsMask = kAnyComponentFlagsMask;

	// Name the component
	PtrToHand ("\pMoe (linked)", &name, 12);
	
	// Register with the component main entry point
	theComp = RegisterComponent (&theDesc, (void *)MoeDispatcher, 0, name, 0, 0);
	
	// Clean up
	if (name != nil)
	{
		DisposHandle (name);
	}
	return (theComp);
}

#endif BUILD_LINKED
#endif THINK_C

//-----------------------------------------------------------------------



