/* ------------------------------------------------------------------------------

FILENAME
  PrintingManager.h

DESCRIPTION
    This file defines all the public level data types
    and interface calls for the Printing Manager.
  
COPYRIGHT
    Apple Computer, Inc. 1990, 1991, 1992, 1993
   All rights reserved. 

-------------------------------------------------------------------------------- */

#ifndef __PRINTINGMANAGER__
#define __PRINTINGMANAGER__


#ifndef __PRINTING__
  #include <PrintTraps.h>
#endif

#ifndef __FILES__
  #include <Files.h>
#endif

#ifndef __COLLECTIONMANAGER__
  #include <Collections.h>
#endif

#ifndef __PRINTINGERRORS__
  #include <PrintingErrors.h>
#endif

#ifndef gxGraphicsTypesIncludes
  #include <graphics types.h>
#endif

#ifndef gxFontTypesIncludes
  #include <font types.h>
#endif

#ifndef gxScalerTypeIncludes
  #include <scaler types.h>
#endif

/* ------------------------------------------------------------------------------

              Printing Manager API Contants and Types

-------------------------------------------------------------------------------- */



/* Printing Manager Gestalt Selector... */


enum {
  gestaltPrintingMgrVersion = 'pmgr'
};


/*--------------------------*/
/* abstract data types...   */
/*--------------------------*/



typedef struct gxPrivatePrinterRecord *gxPrinter;
typedef struct gxPrivateJobRecord *gxJob;
typedef struct gxPrivateFormatRecord *gxFormat;
typedef struct gxPrivatePaperTypeRecord *gxPaperType;
typedef struct gxPrivatePrintFileRecord *gxPrintFile;



/*----------------------------------------*/
/* new function constants and types...    */
/*----------------------------------------*/



typedef Boolean gxLoopStatus;

enum {            // possible values for gxLoopStatus
  gxStopLooping = false,
  gxKeepLooping = true
};

typedef pascal gxLoopStatus (*gxViewDeviceProc) (gxViewDevice, void *refCon);
typedef pascal gxLoopStatus (*gxFormatProc) (gxFormat, void *refCon);
typedef pascal gxLoopStatus (*gxPaperTypeProc) (gxPaperType, void *refCon);

typedef pascal OSErr (*gxPrintingFlattenProc) (long size, void *data, void *refCon);



/*----------------------------------------------*/
/* printing collection constants and types... */
/*----------------------------------------------*/


// The following constants are used to set collection item flags in printing collections.
// The Printing Manager purges certain items whenever a driver switch occurs.
// If the formatting driver changes, all items marked as gxVolatileFormattingDriverCategory
// will be purged.  If the output driver changes, all items marked as
// gxVolatileOutputDriverCategory will be purged.
// Note that to prevent items from being flattened when GXFlattenJob is called, you should
// unset the collectionPersistenceBit (defined in Collections.h), which is on by default.


typedef short gxCollectionCategory;   // stored in collection items' user attribute bits

enum {
  gxNoCollectionCategory      = (gxCollectionCategory) 0x0000,
  gxOutputDriverCategory      = (gxCollectionCategory) 0x0001,
  gxFormattingDriverCategory  = (gxCollectionCategory) 0x0002,
  gxDriverVolatileCategory    = (gxCollectionCategory) 0x0004,
  
  gxVolatileOutputDriverCategory = gxOutputDriverCategory + gxDriverVolatileCategory,
  gxVolatileFormattingDriverCategory = gxFormattingDriverCategory + gxDriverVolatileCategory
};



/*-----------------------------*/
/* printing collection tags... */
/*-----------------------------*/



enum { gxPrintingTagID = -28672 };      // all pre-defined tags have this ID



/*--------------------------*/
/* job collection tags...   */
/*--------------------------*/



enum { gxJobTag = 'job ' };

  enum {                    // priorities for jobs
    gxPrintJobUrgent  = 0x00000001,
    gxPrintJobAtTime  = 0x00000002, 
    gxPrintJobASAP  = 0x00000003
  };
  
  enum { gxPrintJobHoldingBit = 0x00001000 }; // reserved bit in the priority field means the job is on hold
  
  enum {
    gxPrintJobHolding     = (gxPrintJobHoldingBit + gxPrintJobASAP),
    gxPrintJobHoldingAtTime = (gxPrintJobHoldingBit + gxPrintJobAtTime),
    gxPrintJobHoldingUrgent = (gxPrintJobHoldingBit + gxPrintJobUrgent)
  };
  
  enum {                    // ===== gxJob Alert User constants (jobAlert field) =====
    gxNoPrintTimeAlert  = 0,          // don't alert user when we print
    gxAlertBefore     = 1,          // alert user before we print
    gxAlertAfter      = 2,          //  alert user after we print
    gxAlertBothTimes    = 3         //  alert before and after we print
  };
    
  enum {                    // ===== gxJob Timeout constants (jobTimeout field) =====
    gxThirtySeconds   = 1800,       // 30 seconds in ticks
    gxTwoMinutes      = 7200        // 2 minutes in ticks
  };  

  typedef struct {
    long  numPages;
    long  priority;
    long  timeToPrint;
    long  jobTimeout;       // in ticks
    long  firstPageToPrint;   // start printing from this page
    short jobAlert;
    Str31 appName;
    Str31 documentName;
    Str31 userName;
  } gxJobInfo;

enum { gxCollationTag = 'sort' };

  typedef struct {
    Boolean collation;          // true if copies are to be collated
  } gxCollationInfo;

enum { gxCopiesTag = 'copy' };

  typedef struct {
    long copies;              // number of copies of document to print
  } gxCopiesInfo;

enum { gxPageRangeTag = 'rang' };

  enum {                    // for optionChosen field in gxSimplePageRangeInfo
    gxDefaultPageRange    = (char) 0,
    gxReplacePageRange    = (char) 1,
    gxCustomizePageRange  = (char) 2
  };

  typedef struct {
    char      optionChosen;         // from options listed above
    Boolean   printAll;           // true if user wants to print all pages
    long      fromPage;           // for gxDefaultPageRange, current value
    long      toPage;             // for gxDefaultPageRange, current value
  } gxSimplePageRangeInfo;

  typedef struct {
    gxSimplePageRangeInfo simpleRange;    // info which will be returned for GXGetJobPageRange    
    Str31           fromString;     // for gxCustomizePageRange, current value
    Str31           toString;     // for gxCustomizePageRange, current value
    long            minFromPage;    // for gxDefaultPageRange, we parse with this, ignored if nil
    long            maxToPage;      // for gxDefaultPageRange, we parse with this, ignored if nil
    char            replaceString[1]; // for gxReplacePageRange, string to display
  } gxPageRangeInfo;

enum { gxQualityTag = 'qual' };

  typedef struct {
    Boolean disableQuality;     // true to disable standard quality controls
    short   defaultQuality;
    short   currentQuality;
    short   qualityCount;       // number of quality menu items in popup menu
    char    qualityNames[1];      // array of packed pascal strings for popup menu titles
  } gxQualityInfo;

enum { gxFileDestinationTag = 'dest' };

  typedef struct {
    Boolean toFile;           // true if destination is a file
  } gxFileDestinationInfo;

enum { gxFileLocationTag = 'floc' };

  typedef struct {
    FSSpec fileSpec;            // location to put file, if destination is file
  } gxFileLocationInfo;

enum { gxFileFormatTag = 'ffmt' };

  typedef struct {
    Str31 fileFormatName;       // name of file format if destination is file
  } gxFileFormatInfo;

enum { gxFileFontsTag = 'incf' };

  typedef struct {
    Boolean includeFonts;       // true if including fonts, if destination is file
  } gxFileFontsInfo;

enum { gxPaperFeedTag = 'feed' };

  typedef struct {
    Boolean gxAutoFeed;         // true if automatic feed, false if manual
  } gxPaperFeedInfo;

typedef long gxTrayIndex;           // type of tray index reference

enum { gxTrayFeedTag = 'tray' };      // public Printing Manager tray setting

  typedef struct {
    gxTrayIndex feedTrayIndex;      // tray to feed paper from
  } gxTrayFeedInfo;

enum { gxManualFeedTag = 'manf' };

  typedef struct {
    long  numPaperTypeNames;      // number of paperTypes to manually feed
    Str31 paperTypeNames[1];      // array of names of paperTypes to manually feed
  } gxManualFeedInfo;

enum { gxNormalMappingTag = 'nmap' };

  typedef struct {
    Boolean normalPaperMapping;   // true if not overriding normal paper gxMapping
  } gxNormalMappingInfo;

enum { gxSpecialMappingTag = 'smap' };

  enum {                    // for paperMapping field in gxSpecialMappingInfo
    gxRedirectPages     = (char) 1,
    gxScalePages        = (char) 2,
    gxTilePages       = (char) 3
  };

  typedef struct {
    char specialMapping;          // enumerated redirect, scale or tile setting
  } gxSpecialMappingInfo;

enum { gxTrayMappingTag = 'tmap' };

  typedef struct {
    gxTrayIndex mapPaperToTray;     // tray to map all paper to
  } gxTrayMappingInfo;

enum { gxPaperMappingTag = 'pmap' };

  // This collection item contains a flattened paper type resource

enum { gxPrintPanelTag = 'ppan' };

  typedef struct {
    Str31 startPanelName;       // name of starting panel in job print dialog
  } gxPrintPanelInfo;

enum { gxFormatPanelTag = 'fpan' };

  typedef struct {
    Str31 startPanelName;       // name of starting panel in format dialog
  } gxFormatPanelInfo;



/*--------------------------*/
/* format collection tags...*/
/*--------------------------*/



enum { gxPaperTypeLockTag = 'ptlk' };   // paperType lock gxTag

  typedef struct {
    Boolean paperTypeLocked;      // true if format's paperType is locked
  } gxPaperTypeLockInfo;

enum { gxOrientationTag = 'layo' };

  enum {                    // for orientation field in gxOrientationTag
    gxPortraitLayout      = (char) 0,
    gxLandscapeLayout       = (char) 1,
    gxRotatedPortraitLayout   = (char) 2,
    gxRotatedLandscapeLayout  = (char) 3
  };
  
  typedef struct {
    char orientation;           // an enumerated orientation value
  } gxOrientationInfo;

enum { gxScalingTag = 'scal' };

  typedef struct {
    Fixed horizontalScaleFactor;    // current horizontal scaling factor
    Fixed verticalScaleFactor;      // current vertical scaling factor
    short minScaling;           // minimum current scaling
    short maxScaling;           // maximum current scaling
  } gxScalingInfo;

enum { gxDirectModeTag = 'dirm' };

  typedef struct {
    Boolean directModeOn;       // true if direct mode is enabled
  } gxDirectModeInfo;

enum { gxFormatHalftoneTag = 'half' };

  typedef struct {
    long numHalftones;          // how many halftones
    gxHalftone  halftones[1];         // any number of halftones
  } gxFormatHalftoneInfo;

enum { gxInvertPageTag = 'invp' };

  typedef struct  {
    Boolean   invert;         // true -> invert the page
  } gxInvertPageInfo;             // missing | false -> don't invert the page

enum { gxFlipPageHorizontalTag = 'flph' };

  typedef struct  {
    Boolean   flipHorizontal;   // true -> flip x coordinates on the page
  } gxFlipPageHorizontalInfo;       // missing | false -> don't flip horizontally
  
enum { gxFlipPageVerticalTag = 'flpv' };

  typedef struct  {
    Boolean   flipVertical;     // true -> flip y coordinates on the page
  } gxFlipPageVerticalInfo;         // missing | false -> don't flip vertically
  
enum { gxPreciseBitmapsTag = 'pbmp' };
  
  typedef struct {
    Boolean   preciseBitmaps;   // true -> scale the page by 96%
  } gxPreciseBitmapInfo;            // missing | false -> don't scale the page by 96%



/*-----------------------------*/
/* paperType collection tags...*/
/*-----------------------------*/



enum { gxBaseTag = 'base' };

  enum {                    // for baseType field in gxBaseTag
    gxUnknownBase   = 0,            // base paper type from which this paper type is
    gxUsLetterBase  = 1,            // derived.  This is not a complete set.
    gxUsLegalBase   = 2,
    gxA4LetterBase  = 3,
    gxB5LetterBase  = 4,
    gxTabloidBase   = 5
  };
  
  typedef struct {
    long baseType;              // paperType's base type
  } gxBaseInfo;

enum { gxCreatorTag = 'crea' };

  enum {                    // for creator field in gxCreatorTag
    gxSysPaperType      = 'sypt',   // system paper type creator
    gxUserPaperType     = 'uspt'    // user paper type creator
                            // if driver paperTypes then = driver's creator type
  };
  
  typedef struct {
    OSType creator;           // paperType's creator
  } gxCreatorInfo;

enum { gxUnitsTag = 'unit' };

  enum {                    // for units field in gxUnitsTag
    gxPicas           = (char) 0, // pica measurement
    gxMms           = (char) 1, // millimeter measurement
    gxInches          = (char) 2    // gxInches measurement
  };
  
  typedef struct {
    char units;               // paperType's units
  } gxUnitsInfo;

enum { gxFlagsTag = 'flag' };

  enum {                    // paper type flags (bit positions)
    gxOldPaperTypeFlag        = 0x00800000, // indicates a paper type with 7.0 settings
    gxNewPaperTypeFlag        = 0x00400000, // indicates a paper type with post 7.0 settings
    gxOldAndNewPaperTypeFlag    = 0x00C00000, // indicates a paper type that's both old and new
    gxFixedPrintArea          = 0x00200000, //  indicates the margins cannot change when format for a different device
    gxDefaultPaperTypeFlag      = 0x00100000, //  indicates the default paper type in the group 
    gxDefaultBasePT         = 0x00080000  //  indicates the default paper type for a specific base type (e.g. US Letter)
  };
  
  typedef struct {
    long flags;               // paperType's flags
  } gxFlagsInfo;

enum { gxCommentTag = 'cmnt' };

  typedef struct {
    Str255 comment;           // paperType's comment
  } gxCommentInfo;



/*---------------------------*/
/* printer gxViewDevice tags...*/
/*---------------------------*/



enum { gxPenTableTag = 'pent' };

  enum {          // possible values for the flags field in the gxPenTableEntry structure
    gxDeviceUnits = 0,
    gxMmUnits = 1,
    gxInchesUnits = 2
  };
  
  enum {    
      gxPenNotLoaded   = -1
  };
  
  typedef struct { 
    Str31      penName;     // name of the pen
    gxColor      penColor;      // gxColor that is part of the gxColor set
    fixed      penThickness;  // size of the pen
    short      penUnits;      // specifies units in which pen thickness is defined
    short      penPosition;   // pen position in the carousel, -1 (kPenNotLoaded) if not loaded
  } gxPenTableEntry;
     
     
  typedef struct {
    long        numPens;    // number of pen entries in the following array
    gxPenTableEntry pens[1];    // array of pen entries
  } PenTable, *PenTablePtr, **PenTableHdl;



/*----------------------------------------*/
/* dialog-related constants and types...  */
/*----------------------------------------*/



typedef long gxDialogResult;

enum {
  gxCancelSelected    = (gxDialogResult) 0,
  gxOkSelected      = (gxDialogResult) 1,
  gxRevertSelected    = (gxDialogResult) 2
};


typedef struct {
  short editMenuID;
  short cutItem;
  short copyItem;
  short pasteItem;
  short clearItem;
  short undoItem;
} gxEditMenuRecord;


/*-----------------------------------------*/
/* job format mode constants and types...  */
/*-----------------------------------------*/



typedef long gxQueryType;

enum {
  gxGetJobFormatLineConstraintQuery   = (gxQueryType)  0,
  gxGetJobFormatFontsQuery          = (gxQueryType)  1,
  gxGetJobFormatFontCommonStylesQuery = (gxQueryType)  2,
  gxGetJobFormatFontConstraintQuery   = (gxQueryType)  3,
  gxSetStyleJobFormatCommonStyleQuery = (gxQueryType)  4
};


typedef OSType gxJobFormatMode;

enum {
  gxGraphicsJobFormatMode = (gxJobFormatMode) 'grph',
  gxTextJobFormatMode     = (gxJobFormatMode) 'text',
  gxPostscriptJobFormatMode = (gxJobFormatMode) 'post'
};


typedef struct {
  long        numModes;
  gxJobFormatMode modes[1];   // any number of modes
} JobFormatModeTable, *JobFormatModeTablePtr, **JobFormatModeTableHdl;


enum { gxConstraintRange = -1 };  // in numSizes field of PositionConstraintTable to indicate a range in sizes array


typedef struct {            // for text mode  
  gxPoint   phase;
  gxPoint   offset;
  long    numSizes;
  Fixed   sizes[1];       // any number of sizes
} PositionConstraintTable, *PositionConstraintTablePtr, **PositionConstraintTableHdl;

typedef struct {            // for text mode
  long      numStyleNames;      // number of gxStyle names
  Str255    styleNames[1];      // any number of gxStyle names
} StyleNameTable, *StyleNameTablePtr, **StyleNameTableHdl;

typedef struct {            // for text mode
  long      numFonts;       // number of gxFont references
  gxFont      fonts[1];       // any number of gxFont references
} FontTable, *FontTablePtr, **FontTableHdl;



/* ------------------------------------------------------------------------------

                Printing Manager API Functions

-------------------------------------------------------------------------------- */


#define GXPUBLICAPIGLUE(selector) {0x203C, 0x0000, selector, 0xABFE}


#ifdef __cplusplus
extern "C" {
#endif


// Global Routines (Basic)


pascal OSErr GXInitPrinting (void)
    = GXPUBLICAPIGLUE(0);

pascal OSErr GXExitPrinting (void)
    = GXPUBLICAPIGLUE(1);


// gxJob Routines (Basic)


pascal OSErr GXNewJob (gxJob *)
    = GXPUBLICAPIGLUE(2);
  
pascal OSErr GXDisposeJob (gxJob)
    = GXPUBLICAPIGLUE(3);
   
pascal void GXFlattenJob (gxJob, gxPrintingFlattenProc, void *)
    = GXPUBLICAPIGLUE(4);
  
pascal gxJob GXUnflattenJob (gxJob, gxPrintingFlattenProc, void *)
    = GXPUBLICAPIGLUE(5);
   
pascal Handle GXFlattenJobToHdl (gxJob, Handle)
    = GXPUBLICAPIGLUE(6);
  
pascal gxJob GXUnflattenJobFromHdl (gxJob, Handle)
    = GXPUBLICAPIGLUE(7);

pascal void GXInstallApplicationOverride (gxJob, short messageID, void *override)
  = GXPUBLICAPIGLUE(8);


// gxFormat Routines (Basic)


pascal gxFormat GXNewFormat (gxJob)
    = GXPUBLICAPIGLUE(9);
  
pascal void GXDisposeFormat (gxFormat)
    = GXPUBLICAPIGLUE(10);
  
                
// gxPaperType Routines (Basic)


pascal gxPaperType GXNewPaperType (gxJob, Str31 name, gxRectangle *pageSize, gxRectangle *paperSize)
    = GXPUBLICAPIGLUE(11);
  
pascal void GXDisposePaperType (gxPaperType)
    = GXPUBLICAPIGLUE(12);

pascal gxPaperType GXGetNewPaperType (gxJob, short resID)
    = GXPUBLICAPIGLUE(13);


// Error-Handling Routines (Basic)
          
                
pascal OSErr GXGetJobError (gxJob)
  = GXPUBLICAPIGLUE(14);

pascal void GXSetJobError (gxJob, OSErr)
  = GXPUBLICAPIGLUE(15);


// Dialog Routines (Basic)
          
                
pascal gxDialogResult GXJobDefaultFormatDialog (gxJob, gxEditMenuRecord *)
   = GXPUBLICAPIGLUE(16);

pascal gxDialogResult GXJobPrintDialog (gxJob, gxEditMenuRecord *)
   = GXPUBLICAPIGLUE(17);

pascal gxDialogResult GXFormatDialog (gxFormat, gxEditMenuRecord *, StringPtr title)
  = GXPUBLICAPIGLUE(18);

                
// gxFormat Routines (Basic)
          
                
pascal gxFormat GXGetJobFormat (gxJob, long whichFormat)
    = GXPUBLICAPIGLUE(19);
  
pascal gxJob GXGetFormatJob (gxFormat)
  = GXPUBLICAPIGLUE(20);

pascal gxPaperType GXGetFormatPaperType (gxFormat)
    = GXPUBLICAPIGLUE(21);
  
pascal void GXGetFormatDimensions (gxFormat, gxRectangle *pageSize, gxRectangle *paperSize)
    = GXPUBLICAPIGLUE(22);
  

// Spooling Routines (Basic)
          
                
pascal void GXGetJobPageRange (gxJob theJob, long *firstPage, long *lastPage)
  = GXPUBLICAPIGLUE(23);
  
pascal void GXStartJob (gxJob, StringPtr docName, long pageCount)
    = GXPUBLICAPIGLUE(24);
  
pascal void GXPrintPage (gxJob, long pageNumber, gxFormat, gxShape thePage)
   = GXPUBLICAPIGLUE(25);
    
pascal Boolean GXStartPage (gxJob, long pageNumber, gxFormat, long numViewPorts, gxViewPort *viewPortList)
  = GXPUBLICAPIGLUE(26);

pascal void GXFinishPage (gxJob)
  = GXPUBLICAPIGLUE(27);

pascal void GXFinishJob (gxJob)
    = GXPUBLICAPIGLUE(28);
  

// gxJob Routines (Advanced)
                      

pascal Collection GXGetJobCollection (gxJob)
  = GXPUBLICAPIGLUE(29);

pascal void *GXGetJobRefCon (gxJob)
  = GXPUBLICAPIGLUE(30);

pascal void GXSetJobRefCon (gxJob, void *refCon)
  = GXPUBLICAPIGLUE(31);
  
pascal gxJob GXCopyJob (gxJob srcJob, gxJob dstJob)
  = GXPUBLICAPIGLUE(32);

pascal void GXSelectJobFormattingPrinter (gxJob, Str31 printerName)
  = GXPUBLICAPIGLUE(33);
  
pascal void GXSelectJobOutputPrinter (gxJob, Str31 printerName)
  = GXPUBLICAPIGLUE(34);

pascal void GXForEachJobFormatDo (gxJob, gxFormatProc, void *refCon)
  = GXPUBLICAPIGLUE(35);

pascal long GXCountJobFormats (gxJob)
  = GXPUBLICAPIGLUE(36);

pascal Boolean GXUpdateJob (gxJob)
  = GXPUBLICAPIGLUE(37);
  
pascal void GXConvertPrintRecord (gxJob, THPrint)
  = GXPUBLICAPIGLUE(38);


// gxPrinter Routines (Advanced)


pascal gxPrinter GXGetJobFormattingPrinter (gxJob)
  = GXPUBLICAPIGLUE(39);
  
pascal gxPrinter GXGetJobOutputPrinter (gxJob)
  = GXPUBLICAPIGLUE(40);
  
pascal gxPrinter GXGetJobPrinter (gxJob)
  = GXPUBLICAPIGLUE(41);
  
pascal gxJob GXGetPrinterJob (gxPrinter)
  = GXPUBLICAPIGLUE(42);

pascal void GXForEachPrinterViewDeviceDo (gxPrinter, gxViewDeviceProc, void *refCon)
  = GXPUBLICAPIGLUE(43);

pascal long GXCountPrinterViewDevices (gxPrinter)
  = GXPUBLICAPIGLUE(44);

pascal gxViewDevice GXGetPrinterViewDevice (gxPrinter, long whichViewDevice)
  = GXPUBLICAPIGLUE(45);
  
pascal void GXSelectPrinterViewDevice (gxPrinter, long  whichViewDevice)
  = GXPUBLICAPIGLUE(46);
  
pascal void GXGetPrinterName (gxPrinter, Str31)
  = GXPUBLICAPIGLUE(47);

pascal OSType GXGetPrinterType (gxPrinter)
  = GXPUBLICAPIGLUE(48);

pascal void GXGetPrinterDriverName (gxPrinter, Str31)
  = GXPUBLICAPIGLUE(49);

pascal OSType GXGetPrinterDriverType (gxPrinter)
  = GXPUBLICAPIGLUE(50);


// gxFormat Routines (Advanced)


pascal Collection GXGetFormatCollection (gxFormat)
  = GXPUBLICAPIGLUE(51);

pascal void GXChangedFormat (gxFormat)
  = GXPUBLICAPIGLUE(52);

pascal gxFormat GXCopyFormat (gxFormat srcFormat, gxFormat dstFormat)
  = GXPUBLICAPIGLUE(53);
  
pascal gxFormat GXCloneFormat (gxFormat)
  = GXPUBLICAPIGLUE(54);
    
pascal long GXCountFormatOwners (gxFormat)
  = GXPUBLICAPIGLUE(55);
    
pascal void GXGetFormatMapping (gxFormat, gxMapping *)
  = GXPUBLICAPIGLUE(56);

pascal gxShape GXGetFormatForm (gxFormat, gxShape *mask)
  = GXPUBLICAPIGLUE(57);
  
pascal void GXSetFormatForm (gxFormat, gxShape form, gxShape mask)
  = GXPUBLICAPIGLUE(58);


// gxJob gxFormat Modes Routines (Advanced)


pascal void GXSetAvailableJobFormatModes (gxJob, JobFormatModeTableHdl)
  = GXPUBLICAPIGLUE(59);

pascal gxJobFormatMode GXGetPreferredJobFormatMode (gxJob, Boolean *directOnly)
  = GXPUBLICAPIGLUE(60);

pascal gxJobFormatMode GXGetJobFormatMode (gxJob)
  = GXPUBLICAPIGLUE(61);

pascal void GXSetJobFormatMode (gxJob, gxJobFormatMode)
  = GXPUBLICAPIGLUE(62);

pascal void GXJobFormatModeQuery (gxJob, gxQueryType, void *srcData, void *dstData)
  = GXPUBLICAPIGLUE(63);


// Dialog Routines  (Advanced)


pascal void GXEnableJobScalingPanel (gxJob, Boolean enabled)
  = GXPUBLICAPIGLUE(64);

pascal void GXGetJobPanelDimensions (gxJob, Rect *)
  = GXPUBLICAPIGLUE(65);


// gxPaperType Routines (Advanced)


pascal long GXCountJobPaperTypes (gxJob, Boolean forFormatDevice)
  = GXPUBLICAPIGLUE(66);

pascal gxPaperType GXGetJobPaperType (gxJob, long whichPaperType, Boolean forFormatDevice, gxPaperType)
  = GXPUBLICAPIGLUE(67);

pascal void GXForEachJobPaperTypeDo (gxJob, gxPaperTypeProc, void *refCon, Boolean forFormattingPrinter)
  = GXPUBLICAPIGLUE(68);

pascal gxPaperType GXCopyPaperType (gxPaperType srcPaperType, gxPaperType dstPaperType)
  = GXPUBLICAPIGLUE(69);

pascal void GXGetPaperTypeName (gxPaperType, Str31)
  = GXPUBLICAPIGLUE(70);

pascal void GXGetPaperTypeDimensions (gxPaperType, gxRectangle *pageSize, gxRectangle *paperSize)
  = GXPUBLICAPIGLUE(71);

pascal gxJob GXGetPaperTypeJob (gxPaperType)
  = GXPUBLICAPIGLUE(72);

pascal Collection GXGetPaperTypeCollection (gxPaperType)
  = GXPUBLICAPIGLUE(73);


// PrintViewer Routines (Advanced)


pascal gxPrintFile GXOpenPrintFile (gxJob, FSSpecPtr, char permission)
  = GXPUBLICAPIGLUE(74);

pascal void GXClosePrintFile (gxPrintFile)
  = GXPUBLICAPIGLUE(75);

pascal gxJob GXGetPrintFileJob (gxPrintFile)
  = GXPUBLICAPIGLUE(76);

pascal long GXCountPrintFilePages (gxPrintFile)
  = GXPUBLICAPIGLUE(77);

pascal void GXReadPrintFilePage (gxPrintFile, long pageNumber, long numViewPorts, gxViewPort *viewPortList, gxFormat *, gxShape *)
  = GXPUBLICAPIGLUE(78);

pascal void GXReplacePrintFilePage (gxPrintFile, long pageNumber, gxFormat, gxShape)
  = GXPUBLICAPIGLUE(79);

pascal void GXInsertPrintFilePage (gxPrintFile, long atPageNumber, gxFormat, gxShape)
  = GXPUBLICAPIGLUE(80);

pascal void GXDeletePrintFilePageRange (gxPrintFile, long fromPageNumber, long toPageNumber)
  = GXPUBLICAPIGLUE(81);

pascal void GXSavePrintFile (gxPrintFile, FSSpec *)
  = GXPUBLICAPIGLUE(82);


#ifdef __cplusplus
}
#endif


#endif __PRINTINGMANAGER__
