/* layout types.h -- Line Layout types

  by Dave Opstad and Eric Mader,
  with special thanks to the following folks:
     Mark Davis, for starting it all off;
     Russ Sonnenschein, for many good ideas;
     Cary Clark, Dave Good, Rob Johnson, Keith McGreggor, and Chris Yerga for the graphics system;
     Richard Becker, Lynn Bekkala and Mike Reed for fonts and the gxFont system;
     Robert Dierkes and the SQE team for testing and good ideas;
     Parry Kejriwal and Richard Klein, for gxFont tools;
     Diane Patterson, for great patience in writing documentation;
     Hugo Ayala, Tom Dowdy, Dan Lipton, Sean Parent and the rest of the PrintShop;
     Mitch Allen for managing to manage us; and
     many other good folk at Apple!
*/

/* Copyright 1988-1992 Apple Computer, Inc.  All rights reserved. */

#ifndef gxLayoutTypesIncludes
#define gxLayoutTypesIncludes

#ifndef gxMathTypesIncludes
#include "math types.h"
#endif gxMathTypesIncludes

#ifndef gxGraphicsTypesIncludes
#include "graphics types.h"
#endif gxGraphicsTypesIncludes

typedef unsigned short gxGlyphcode;   /* single glyph in a gxFont */
typedef long gxByteOffset;            /* byte offset within backing store */

/* The actual constants for feature types and selectors have been moved to a library. */

typedef unsigned short gxRunFeatureType;
typedef unsigned short gxRunFeatureSelector;

/* If tracking is not desired, specify the following value in the track field in the
    gxRunControls record (note that a track of 0 does *not* mean to turn tracking off;
    rather, it means to use normal tracking). */

#define gxNoTracking gxNegativeInfinity

/* The special "gxNoStake" value is returned by the GXGetLayoutBreakOffset call to indicate
    the absence of a character offset that is stable with respect to metamorphosis and
    contextual kerning. */

#define gxNoStake -1

/* A glyph's behavior with respect to other glyphs on its gxLine is defined in part by its
    gxBaselineType. These types allow correct alignment of the baselines of all glyphs on
    the gxLine. */

enum {
  gxRomanBaseline = 0,
  gxIdeographicCenterBaseline,
  gxIdeographicLowBaseline,
  gxHangingBaseline,
  gxMathBaseline,
  gxLastBaseline = 31,
  gxNumberOfBaselineTypes = gxLastBaseline + 1,
  gxNoOverrideBaseline = 255
  };
typedef unsigned long gxBaselineType;

typedef fixed gxBaselineDeltas[gxNumberOfBaselineTypes];

/* gxJustificationPriority defines the gxPoint during the justification process at which a
    glyph will begin to receive deltas before and after itself. */

enum {
  gxKashidaPriority           = 0,
  gxWhiteSpacePriority        = 1,
  gxInterCharPriority         = 2,
  gxNullJustificationPriority = 3,
  gxNumberOfJustificationPriorities
  };
typedef unsigned char gxJustificationPriority;

/* gxJustificationFlags are used to control which fields of a gxWidthDeltaRecord are to be
    overridden and which are not if a gxPriorityJustificationOverride or
    gxGlyphJustificationOverride (qq.v.) is specified. */

enum {
  gxOverridePriority          = 0x8000, /* use priority value from override */
  gxOverrideLimits            = 0x4000, /* use limits values from override */
  gxOverrideUnlimited         = 0x2000, /* use unlimited flag from override */
  gxUnlimitedGapAbsorption    = 0x1000, /* glyph can take unlimited gap */
  gxJustificationPriorityMask = 0x000F, /* justification priority */

  gxAllJustificationFlags     = gxOverridePriority|gxOverrideLimits|gxOverrideUnlimited|
                              gxUnlimitedGapAbsorption|gxJustificationPriorityMask
  };
typedef unsigned short gxJustificationFlags;

/* The directional behavior of a glyph can be overridden using a gxDirectionOverride. */

enum {
  gxNoDirectionOverride = 0,
  gxImposeLeftToRight   = 1,
  gxImposeRightToLeft   = 2,
  gxImposeArabic        = 3
  };
typedef unsigned short gxDirectionOverride;

/* gxRunControlFlags describe the nonparametric layout options contained in a gxStyle. */

#define gxNoLigatureSplits        0x80000000
#define gxNoCaretAngle            0x40000000
#define gxImposeWidth             0x20000000
#define gxNoCrossKerning          0x10000000
#define gxNoOpticalAlignment      0x08000000
#define gxForceHanging            0x04000000
#define gxNoSpecialJustification  0x02000000
#define gxDirectionOverrideMask   0x00000003

#define gxAllRunControlFlags (gxNoLigatureSplits|gxNoCaretAngle|gxImposeWidth|gxNoCrossKerning|gxNoOpticalAlignment|gxForceHanging|gxNoSpecialJustification|gxDirectionOverrideMask)

typedef unsigned long gxRunControlFlags;

/* gxHighlightType is used to distinguish various highlighting methods, both in terms of
    character offset based vs. visual based, and in terms of angled sides vs. non-angled
    sides. */

enum {
  gxHighlightStraight     = 0,  /* straight-edged simple highlighting */
  gxHighlightAverageAngle = 1   /* takes average of two edge angles */
  };
typedef unsigned long gxHighlightType;

/* gxCaretType is used to control whether the caret that is returned from GXGetLayoutCaret
    is a split caret or a (keyboard-syncronized) single caret. */

enum {
  gxSplitCaretType         = 0,  /* returns Mac-gxStyle split caret (default) */
  gxLeftRightKeyboardCaret = 1,  /* single caret in left-right position */
  gxRightLeftKeyboardCaret = 2   /* single caret in right-left position */
  };
typedef unsigned long gxCaretType;

/* gxLayoutOffsetState describes the characteristics of a given gxByteOffset in some
    layout. It is returned by the GXGetOffsetGlyphs call. Note that the gxOffsetInsideLigature
    value is returned in addition to the 8/16 (or invalid) indication. */

enum {
  gxOffset8_8            = 0,
  gxOffset8_16           = 1,
  gxOffset16_8           = 2,
  gxOffset16_16          = 3,
  gxOffsetInvalid        = 4,
  gxOffsetInsideLigature = 0x8000
  };
typedef unsigned short gxLayoutOffsetState;

/* gxLayoutOptionsFlags are single-bit flags contained in a gxLayoutOptions record. We also
    define here some utility constants that are useful in setting various fields in the
    gxLayoutOptions record. */

#define gxNoLayoutOptions 0

#define gxLineIsDisplayOnly   0x00000001

#define gxAllLayoutOptionsFlags (gxLineIsDisplayOnly)

#define gxMaxRunLevel 15
#define gxFlushLeft 0
#define gxFlushCenter (frac1/2)
#define gxFlushRight frac1

#define gxNoJustification 0
#define gxFullJustification frac1

typedef unsigned long gxLayoutOptionsFlags;

/* A gxRunFeature describes a feature and a level for that feature. */

typedef struct {
  gxRunFeatureType      featureType;
  gxRunFeatureSelector  featureSelector;
  } gxRunFeature;

/* A gxWidthDeltaRecord contains all of the information needed to describe the behavior of one
    class of glyphs during the justification process. */

typedef struct {
  fixed               beforeGrowLimit;    /* ems AW can grow by at most on LT */
  fixed               beforeShrinkLimit;  /* ems AW can shrink by at most on LT */
  fixed               afterGrowLimit;     /* ems AW can grow by at most on RB */
  fixed               afterShrinkLimit;   /* ems AW can shrink by at most on RB */
  gxJustificationFlags  growFlags;          /* flags controlling grow case */
  gxJustificationFlags  shrinkFlags;        /* flags controlling shrink case */
  } gxWidthDeltaRecord;

/* gxPriorityJustificationOverride contains an array of WidthDeltaRecords, one for each
    gxJustificationPriority. */

typedef struct {
  gxWidthDeltaRecord deltas[gxNumberOfJustificationPriorities]; /* overrides for each of the priorities */
  } gxPriorityJustificationOverride;

/* gxGlyphJustificationOverride contains a gxWidthDeltaRecord that is to be used for a
    specific glyph in a specific run (this limitation is because glyphcodes vary by gxFont). */

typedef struct {
  gxGlyphcode        glyph;
  gxWidthDeltaRecord override;
  } gxGlyphJustificationOverride;

/* gxRunControls contains flags, shifts, imposed widths and overrides for a run. */
/* NOTE: a value of "gxNoTracking" in track disables tracking */

typedef struct {
  gxRunControlFlags flags;
  fixed           beforeWithStreamShift;
  fixed           afterWithStreamShift;
  fixed           crossStreamShift;
  fixed           imposedWidth;
  fixed           track;
  fract           hangingInhibitFactor;
  fract           kerningInhibitFactor;
  fixed           decompositionAdjustmentFactor;
  gxBaselineType    baselineType;
  } gxRunControls;

/* A gxGlyphSubstitution describes one client-provided substitution that occurs after all
    other automatic glyph changes have happened. */

typedef struct {
  gxGlyphcode originalGlyph;    /* Whenever you see this glyph... */
  gxGlyphcode substituteGlyph;  /* ...change it to this one. */
  } gxGlyphSubstitution;

/* gxKerningAdjustmentFactors specify an adjustment to automatic kerning.
    The adjustment is ax + b where x is the automatic kerning value, a
    is scaleFactor, and b is adjustmentPointSizeFactor times the run's
    gxPoint size. */
    
typedef struct {
  fract scaleFactor;
  fixed adjustmentPointSizeFactor;
} gxKerningAdjustmentFactors;

/* A gxKerningAdjustment identifies with- and cross-stream kerning adjustments
    for specific glyph pairs. */

typedef struct {
  gxGlyphcode                 firstGlyph;
  gxGlyphcode                 secondGlyph;
  gxKerningAdjustmentFactors  withStreamFactors;
  gxKerningAdjustmentFactors  crossStreamFactors;
  } gxKerningAdjustment;

/* a value of gxResetCrossStreamFactor in crossStreamFactors.adjustmentPointSizeFactor
    will reset the cross-stream kerning to the baseline. */
    
#define gxResetCrossStreamFactor gxNegativeInfinity

/* gxLayoutHitInfo contains the output from the GXHitTestLayout call. */

typedef struct {
  fixed      firstPartialDist;
  fixed      lastPartialDist;
  gxByteOffset hitSideOffset;
  gxByteOffset nonHitSideOffset;
  gxBoolean    leadingEdge;
  gxBoolean    inLoose;
  } gxLayoutHitInfo;

/* A gxLineBaselineRecord contains the deltas from 0 to all the different baselines for
    the layout. It can be filled via a call to GetBaselineDeltas (q.v.). */

typedef struct {
  gxBaselineDeltas  deltas;
  } gxLineBaselineRecord;

/* The gxLayoutOptions type contains information about the layout characteristics of the
    whole gxLine. */

typedef struct {
  fixed               width;
  fract               flush;
  fract               just;
  gxLayoutOptionsFlags  flags;
  gxLineBaselineRecord  *baselineRec;
  } gxLayoutOptions;

#endif gxLayoutTypesIncludes
