/*  scaler types.h
**
**    These are the data types relevent to the API of a scaler.  Included are the
**  types for managing memory blocks, and for communicating with scalers in the input
**  and output directions.
**
**  Copyright:   1992 by Apple Computer, Inc., all rights reserved.
**
**  Change History:
**
**  <1>   03/15/92 ABB  Created
*/

#ifndef gxScalerTypeIncludes
#define gxScalerTypeIncludes

#ifndef gxMathTypesIncludes
  #include "math types.h"
#endif

#ifndef gxSfntTypesIncludes
  #include "sfnt types.h"
#endif

#define gxTruetypeFontFormat    0x74727565  /* 'true' */
#define gxType1FontFormat   0x74797031  /* 'typ1' */
#define gxNfntFontFormat      0x6e666e74  /* 'nfnt' */

typedef enum {
  scaler_no_problem = 0,          /* Everything went OK */
  
  scaler_first_error,           /* Generally fatal errors */
  scaler_null_context = scaler_first_error, /* Client passed a null context pointer */
  scaler_null_input,            /* Client passed a null input pointer */
  scaler_invalid_context,         /* There was a problem with the context */
  scaler_invalid_input,         /* There was a problem with an input */
  scaler_invalid_font_data,         /* A portion of the gxFont was corrupt */
  scaler_new_block_failed,          /* A call to NewBlock() failed */
  scaler_bitmap_alloc_failed,       /* Call to alloc gxBitmap perm. block failed */
  scaler_outline_alloc_failed,        /* Call to alloc outline perm. block failed */
  scaler_required_table_missing,      /* A needed gxFont table was not found */
  scaler_unsupported_outline_format,      /* Couldn't create an outline of the desired format */
  scaler_unsupported_stream_format,     /* ScalerStreamFont() call can't supply any requested format */
  scaler_unsupported_font_format,     /* No scaler supports the gxFont format */
  scaler_hinting_error,         /* An error occurred during hinting */
  scaler_scan_error,            /* An error in scan conversion */
  scaler_internal_error,          /* Scaler has a bug */
  scaler_invalid_matrix,          /* The gxTransform matrix was unusable */
  scaler_fixed_overflow,          /* An overflow ocurred during matrix operations */
  scaler_API_version_mismatch,        /* Scaler requires a newer/older version of the scaler API */
  scaler_no_kerning_applied,        /* ScalerKernGlyphs() found no kerning data, or it had no effect */
  scaler_last_error = scaler_no_kerning_applied,
  
  scaler_first_warning,         /* Informative warnings */
  scaler_no_output = scaler_first_warning,    /* Couldn't fulfill any glyph request. */
  scaler_fake_metrics,          /* Returned metrics aren't based on info in the gxFont */
  scaler_fake_linespacing,          /* Linespacing metrics not based on info in the gxFont */
  scaler_glyph_substitution,        /* Requested glyph out of range, a substitute was used */
  scaler_last_warning = scaler_glyph_substitution
} gxScalerErrors;

typedef long gxScalerError;

/*** ScalerOpen output type ***/

typedef struct {
  gxFontFormatTag format;     /* Font format supported by this scaler */
  fixed     scalerVersion;    /* Version number of the scaler */
  fixed     APIVersion;   /* Version of API implemented (compare with version in gxScalerContext) */
} gxScalerInfo;

/*** ScalerNewFont output type ***/

typedef enum {
  gxRequiresLayoutFont = 1,
  gxHasNormalLayoutFont = 2,
  gxCanReorderFont = 4,
  gxCanRearrangeFont = 8,
  gxHasOutlinesFont = 16
} gxScalerFontFlags;

typedef unsigned long gxScalerFontFlag;

typedef struct {
  unsigned long   unitsPerEm;
  gxScalerFontFlag    flags;
  unsigned long   numGlyphs;
} gxScalerFontInfo;

/*** ScalerNewTransform input types ***/

typedef enum {
  gxApplyHintsTransform = 1,      /* Execute hinting instructions (grid fit) */
  gxExactBitmapTransform = 2,   /* Use embedded gxBitmap iff exact size */
  gxUseThresholdTransform = 4,    /* Use scaled gxBitmap (if any) if below outline threshold */
  gxVerticalTransform = 8     /* Glyphs will be in vertical orientation */
} gxScalerTransformFlags;

typedef unsigned long gxScalerTransformFlag;

typedef struct {
  gxScalerTransformFlag   flags;      /* Hint, embedded gxBitmap control, etc. */
  fixed         pointSize;      /* The desired pointsize */
  gxMapping       *fontMatrix;    /* The 3x3 matrix to apply to glyphs */
  gxPoint         resolution;   /* 2D device resolution */
  gxPoint         spotSize;     /* 2D pixel size */
} gxScalerTransform;

/*** ScalerNewTransform output type ***/

typedef struct {
  gxPoint   before;     /* Spacing of the gxLine before */
  gxPoint   after;      /* Spacing of the gxLine after */
  gxPoint   caretAngle;   /* Rise (y) and run (x) of the insertion caret */
  gxPoint   caretOffset;    /* adjustment to caret for variants like italic */
} gxScalerTransformInfo;

/*** ScalerNewGlyph input types ***/

typedef enum {
  gxNoImageGlyph = 1        /* Don't return the bit image for this glyph */
} gxScalerGlyphFlags;

typedef unsigned long gxScalerGlyphFlag;

#define gxPathOutline   0x70617468  /* 'gxPath' QuickDraw GX outline */

typedef long gxOutlineFormat;

typedef struct {
  long        glyphIndex; /* Index of the glyph to be considered */
  long        bandingTop; /* Banding controls (scanline numbers) top==bottom==0 means no banding */
  long        bandingBottom;
  gxOutlineFormat   format;   /* gxFormat of outline to return, ignored if no outline desired */
  gxScalerGlyphFlag   flags;
} gxScalerGlyph;

/*** ScalerNewGlyph output types ***/

typedef struct {
  gxPoint   advance, sideBearing, otherSideBearing;
} gxScalerMetrics;

typedef struct {
  long  xMin;
  long  yMin;
  long  xMax;
  long  yMax;
} gxScalerRectangle;

typedef struct {
  char        *image;   /* Pointer to pixels */
  gxPoint       topLeft;    /* Bitmap positioning relative to client's origin */
  gxScalerRectangle   bounds;   /* Bounding box of gxBitmap */
  long        rowBytes; /* Width in bytes */
#ifdef cpuPrinter
  long        actualBitmapSize;
  long        scanConvertScratchSize;
  long        dropoutScratchSize;
#endif
} gxScalerBitmap;

/*** ScalerKernGlyphs input/output types ***/

enum {
  gxLineStartKern   = 1,      /* Array of glyphs starts a gxLine */
  gxLineEndKern   = 2,      /* Array of glyphs ends a gxLine */
  gxNoCrossKern   = 4,      /* Prohibit cross kerning */
  gxDeviceValuesKern  = 8     /* Return device (vs. fractional) values */
};
typedef long gxScalerKerningFlags;

enum {
  gxNoStakeNote = 1,
  gxCrossStreamResetNote = 2
};
typedef unsigned short gxScalerKerningNotes;

typedef struct {
  long        numGlyphs;    /* Number of glyphs in the glyphs array */
  fract       scaleFactor;    /* Amount of kerning to apply (0 == none, frac1 == all) */
  gxScalerKerningFlags  flags;      /* Various control flags */
  unsigned short    *glyphs;      /* Pointer to the array of glyphs to be kerned */
}  gxScalerKerning;

/*** ScalerStream input/output types ***/

typedef enum {
  gxTruetypeStream  = 0x0001,
  gxType1Stream   = 0x0002,
  gxType3Stream   = 0x0004,
  gxType42Stream    = 0x0008,
  gxType42GXStream  = 0x0010,
  gxPortableStream    = 0x0020,
  gxFlattenedStream = 0x0040
} gxStreamTypes;            /* Possible streamed gxFont formats */

typedef unsigned long gxStreamType;

typedef enum {
  gxDownloadFlag = 0,
  gxAsciiDownloadFlag,
  gxFontSizeQueryFlag,
  gxEncodingOnlyFlag
} gxScalerStreamFlags;

typedef long gxScalerStreamFlag;

enum {
  gxSelectAllVariations = -1
};

typedef struct gxScalerStream {
  void        *streamRefCon;  /* <- private reference for client */
  gxStreamType    type;     /* <->  data stream format */
  gxScalerStreamFlag  flags;      /* <- what action to take */
  unsigned long   memorySize;   /* -> worst case vm in printer */
  long        variationCount;   /* <- the number of variations, or gxSelectAllVariations */
  gxFontVariation   *variations;    /* <-     a pointer to an array the variations */
  short     *encoding;    /* <- Intention is * short[256] */
  long        *glyphBits;   /* <->  sizeof(glyphcount)/bitsperlong */
  char        *gxFontName;    /* <->  the printer gxFont name to use/used */
} gxScalerStream;

typedef struct {
  long    hexFlag;
  long    byteCount;
  void      *data;
} gxScalerStreamData;

typedef enum {
  gxScalerScratchBlock = -1,    /* Scaler alloced/freed temporary memory */

  gxScalerOpenBlock,      /* Five permanent input/state block types */
  gxScalerFontBlock,
  gxScalerVariationBlock,
  gxScalerTransformBlock,
  gxScalerGlyphBlock,
  gxScalerBlockCount,     /* Number of permanent block types */
  
  gxScalerOutlineBlock = gxScalerBlockCount,      /* Two output block types */
  gxScalerBitmapBlock
  
} gxScalerBlockTypes;

typedef long gxScalerBlockType;

#define gxSfntDirectoryTag      0x64697220  /* 'dir ' */

typedef struct gxScalerContext {
  fixed     version;
  void      *theFont;
  gxFontFormatTag format;
  
  long    (*GXGetFontTable)(struct gxScalerContext *context, gxFontTableTag tableTag, long offset, long length, void *data);
  void    (*ReleaseFontTable)(struct gxScalerContext *context, void* fontData);
  void* (*NewBlock)(struct gxScalerContext *context, long size, gxScalerBlockType, void* oldBlock);
  void    (*DisposeBlock)(struct gxScalerContext *context, void* scratchData, gxScalerBlockType theType);
  void    (*StreamFunction)(struct gxScalerContext *context,  gxScalerStream* streamInfo, gxScalerStreamData *dataInfo);
  void    (*ScanLineFunction)(struct gxScalerContext *context,  gxScalerBitmap* scanLine);
  void    (*PostErrorFunction)(struct gxScalerContext *context, gxScalerError theProblem);
#ifdef cpuPrinter
  long    (*FindExtrema)( void *, void * );
  long    (*ScanChar2)( void *, void *, void *, short, short, long, struct gxScalerContext * );
#endif
  void* scalerBlocks[gxScalerBlockCount];
  void    (*ScalerFunction)(struct gxScalerContext *context, void* data);
} gxScalerContext;

#endif
