/*
	File:		FontHIFontUtils.c

	Contains:	internal font-related utility routines supporting FontHI library

	Written by:	Parry Kejriwal, Barton House, Dennis Buchheim

	Copyright:	 1992, 1993 by Apple Computer, Inc., All rights reserved.

	Change History (most recent first):


*/

#include <Memory.h>
#include <OSUtils.h>

#include "font routines.h"
#include "font library.h"
#include "font library additions.h"

#include "math routines.h"

#include "FontHIUtils.h"
#include "FontHIFontUtils.h"


#define	numberOfSetsToOrder	4

#define TrakTag				'trak'


typedef struct {
	Fixed			trackValue;
	short			nameID;
	short			offset;
} TrackTableEntry;

typedef struct {
	short			numTracks;
	short			numSizes;
	long			sizeTableOffset;
	TrackTableEntry	tracks[];
	Fixed			sizes[];
} TrackDataRec, *TrackDataPtr, **TrackDataHan;

typedef struct {
	Fixed			version;
	short			format;
	short			horizOffset;
	short			vertOffset;
	short			reserved;
	TrackDataRec	horizData[];
	TrackDataRec	vertData[];
} TrackTable, *TrackTablePtr, **TrackTableHan;


/* this should use the feature registry header file */
static short	featureSortOrder[numberOfSetsToOrder] = { 3, 6, 7, 10 };


/*------------------------------------------------------- FHIVariationToIndex */
short FHIVariationToIndex (font theFontID, short numCmpVariations, fontVariation *cmpVariations)
{
	short			numInstances, numVariations;
	fontVariation	*variations;
	short			i, j;
	short			equalValues;
	Boolean			instanceMatchFound = false;
	
	if (numCmpVariations == 0)
		return (1);

	numInstances = CountFontInstances (theFontID);
	numVariations = CountFontVariations (theFontID);
	
	if (numVariations != numCmpVariations)
		return (0);

	variations = (fontVariation *) NewPtr (numVariations * sizeof (fontVariation));
	if (MemError () != noErr) return (0);
	
	for (i = 1; i <= numInstances; i++)
	{
		equalValues = 0;

		GetFontInstance (theFontID, i, variations);

		for (j = 0; j < numVariations; j++)
			if (cmpVariations[j].name == variations[j].name && 
				cmpVariations[j].value == variations[j].value)
				equalValues++;
		
		if (equalValues == numVariations)
		{
			instanceMatchFound = true;
			break;
		}
	}
	
	if (variations)
		DisposPtr ((Ptr) variations);
	
	if (instanceMatchFound)
		return (i);
	else
		return (0);
}

/* ------------------------------------------------------- FHIGetInstances */
FHI_InstanceHan FHIGetInstances ( font theFontID )
{
	long				numInstances;
	fontName			theNameID;
	short				i = 0;
	long				nameLength;
	Str255				name;
	FHI_InstanceHan		theInstanceH = NULL;
	
	numInstances = CountFontInstances ( theFontID );
	
	theInstanceH = (FHI_InstanceHan) NewHandle (sizeof (FHI_InstanceRec) + 
												sizeof (FHI_NameRec) * numInstances);
	if ( !theInstanceH ) return ( NULL );
	
	(**theInstanceH).numberOfInstances = numInstances;
		
	if ( numInstances ) {

		for ( i = 1; i <= numInstances; i++ ) {
		
			theNameID = GetFontInstance (theFontID, i, nil);
			
			FindFontPName ( theFontID, theNameID, name );
			(**theInstanceH).names[i-1].theNameID = theNameID;
			FHI_Pstrcpy ( (**theInstanceH).names[i-1].name, name );
		}
	}
	
	return ( theInstanceH );
}

/*------------------------------------------------------- FeatureIsExclusive */
static Boolean FeatureIsExclusive(FHI_FeatMenuInfoHan featH, fontRunFeature * feature)
{
	FHI_FeatItemInfoRec		* fp;
	short					i;
	short					numFeatures;
	
	fp = (**featH).features;
	numFeatures = (**featH).numOfFeatures;
	
	for(i=0; i<numFeatures; i++, fp++)
		if(fp->featureType == feature->featureType)
			return((fp->featureFlag & mutuallyExclusiveFeature) == mutuallyExclusiveFeature);
	
	return(false);
}

/*------------------------------------------------------- RemoveFeatureConflict */
static void RemoveFeatureConflict(FHI_FeatMenuInfoHan featH, short * numFeaturesPtr,
								fontRunFeature ** features, fontRunFeature * feature)
{
	short				numFeatures;
	fontRunFeature		*fp;
	short				i;
	long				size;
	
	numFeatures = *numFeaturesPtr;
	
	if(FeatureIsExclusive(featH, feature)) {
	
		fp = *features;
		for(i=0; i<numFeatures; i++, fp++)
			if(fp->featureType == feature->featureType) {
				size = sizeof(fontRunFeature) * (numFeatures - i - 1);
				BlockMove((Ptr) (fp+1), (Ptr) fp, size);
				numFeatures--;
			}
			
	} else {
	
		fp = *features;
		for(i=0; i<numFeatures; i++, fp++)
			if(fp->featureType == feature->featureType &&
			   (fp->setting & ~1) == (feature->setting & ~1)) {
				size = sizeof(fontRunFeature) * (numFeatures - i - 1);
				BlockMove((Ptr) (fp+1), (Ptr) fp, size);
				numFeatures--;
			}
	}
	
	SetHandleSize((Handle) features, sizeof(fontRunFeature) * numFeatures);
	*numFeaturesPtr = numFeatures;
}

/*------------------------------------------------------- FHIApplyFeaturesToDefault */
void FHIApplyFeaturesToDefault( FHI_FeatMenuInfoHan featH, 
							short srcNumFeatures, fontRunFeature * srcFeatures, 
							short * dstNumFeatures, fontRunFeature *** dstFeatures)
{
	short				numFeatures;
	fontRunFeature		** features;
	fontRunFeature		* fp;
	short				i;
	
	numFeatures = (**featH).numDefaultFeatures;
	features = (**featH).defaultFeatures;
	
	HandToHand((Handle *) &features);
	fp = srcFeatures;
	
	for(i=0; i<srcNumFeatures; i++, fp++) {
	
		RemoveFeatureConflict(featH, &numFeatures, features, fp);
		
		PtrAndHand((Ptr) fp, (Handle) features, sizeof(fontRunFeature));
		numFeatures++;
	}
	
	*dstNumFeatures = numFeatures;
	*dstFeatures = features;
}

/*------------------------------------------------------- SpecialFeatureType */
static Boolean SpecialFeatureType ( fontFeature	theFeature )
{
	short	i;
	
	for ( i=0; i < numberOfSetsToOrder; i++ )
		if ( featureSortOrder[i] == theFeature )
			return ( true );
	return ( false );

}

/*------------------------------------------------------- SortRunFeatures */
static FHI_FeatMenuInfoHan SortFontFeatures ( FHI_FeatMenuInfoHan featH )
{
	FHI_FeatMenuInfoHan	sortedFeatH;
	FHI_FeatMenuInfoPtr	sortedFeatP, featP;
	short		i, j;
	short		numOfFeatures;
	fontFeature	theFeature;
	short		sortIndex = 0;
	
	
	sortedFeatH = ( FHI_FeatMenuInfoHan ) NewHandle ( GetHandleSize ( (Handle) featH ) );
	if ( !sortedFeatH ) return ( NULL );

	sortedFeatP = *sortedFeatH;
	featP = *featH;
	
	numOfFeatures = sortedFeatP->numOfFeatures = featP->numOfFeatures;

	sortIndex = 0;
	
	/* first sort by the order defined above */
	for ( i=0; i < numberOfSetsToOrder; i++ ) {
	
		theFeature = featureSortOrder[i];
		for ( j =0; j < numOfFeatures; j++ ) {
			if ( featP->features[j].featureType == theFeature ) {
				sortedFeatP->features[sortIndex] = featP->features[j];
				sortIndex++;
			}
		}
	}
	/* now stuff the rest into the struct in the order they were read */
	for ( j =0; j < numOfFeatures; j++ ) {
		if ( !SpecialFeatureType ( featP->features[j].featureType ) ) {
			sortedFeatP->features[sortIndex] = featP->features[j];
			sortIndex++;
		}
	}
	return ( sortedFeatH );
}

/*------------------------------------------------------- AddFontFeature */
/*	
 *	the routine adds all the settings for the feature and add it in the 
 *	bottom of the feature structure - does not check for duplicates, etc.
 */
static short AddFontFeature (FHI_FeatMenuInfoHan	featH,
							 fontFeatureFlag		flags,
							 long					count,
							 fontFeatureSetting		setting[],
							 fontFeature			feature)
{

	long				numOfFeatures;
	FHI_FeatMenuInfoPtr	featP;
	short				err, i;
	
	err = noErr;
	numOfFeatures = (**featH).numOfFeatures;
	
	SetHandleSize ( (Handle) featH, GetHandleSize ( (Handle) featH ) + count * sizeof(FHI_FeatItemInfoRec));
	if ( err = MemError() )
		return ( err );
	
	featP = *featH;
	
	for ( i=0; i< count; i++ ) {
		
		featP->features[ numOfFeatures + i ].featureType = feature;
		featP->features[ numOfFeatures + i ].featureFlag = flags;
		featP->features[ numOfFeatures + i ].featureSetting = setting[i];

		featP->features[ numOfFeatures + i ].menuItem = 0;	/* set this later */
		featP->features[ numOfFeatures + i ].selected = false;
	}
	featP->numOfFeatures += count;
	
	return ( err );
}

/*------------------------------------------------------- FHIGetFeatures */
FHI_FeatMenuInfoHan	FHIGetFeatures ( font theFontID )
{
	long				numFontFeatures;
	long				i;
	fontFeatureFlag		flags;
	long				settingCount;
	fontFeature			feature;
	fontFeatureSetting	*settingPtr;
	fontName			nameID;
	FHI_FeatMenuInfoHan	featH, sortedFeatH;
	
	
	featH = (FHI_FeatMenuInfoHan) NewHandle ( sizeof ( FHI_FeatMenuInfoRec ) );
	
	if ( !featH )
		return ( NULL );

	(**featH).numFeatGroups = 1;
	(**featH).numOfFeatures = 0;	
	
	numFontFeatures = CountFontFeatures ( theFontID );

	for ( i= 1; i <= numFontFeatures; i++ ) {

		nameID = GetFontFeature(theFontID, i, &flags, &settingCount, nil, &feature);
		
		if ( settingCount ) {
			settingPtr = (fontFeatureSetting *) NewPtr (settingCount * sizeof (fontFeatureSetting)); 
			GetFontFeature (theFontID, i, &flags, &settingCount, settingPtr, &feature);
			AddFontFeature (featH, flags, settingCount, settingPtr, feature);
			DisposePtr ((Ptr) settingPtr);
		}
	}
	sortedFeatH = SortFontFeatures ( featH );
	if ( sortedFeatH )
		(**sortedFeatH).numFeatGroups = (**featH).numFeatGroups;

	DisposeHandle ((Handle) featH);
	
	HLock ((Handle) sortedFeatH);
	GetFontDefaultFeatures (theFontID, &(**sortedFeatH).numDefaultFeatures, &(**sortedFeatH).defaultFeatures);
	HUnlock ((Handle) sortedFeatH);
	
	return ( sortedFeatH );
}


/*------------------------------------------------------- FHIGetTrackTableInfo */
FHI_TrackingHan FHIGetTrackTableInfo (font theFontID)
{
	long						tableIndex, length;
	Handle						theTrTableH = NULL;
	register TrackDataPtr		theTrDataP;
	FHI_TrackingHan				theTrackingH = NULL;
	register FHI_TrackingPtr	theTrackingP;	
	short						i;
	Str255						trackName;
	
	length = FindFontTable (theFontID, TrakTag, nil, &tableIndex);
	
	/*	get values from 'trak' table
	 */
	if (length)
	{
		theTrTableH =  NewHandle (length);
		if (MemError () != noErr) goto errorReturn;
		
		if (theTrTableH)
			length = FindFontTable (theFontID, TrakTag, *theTrTableH, &tableIndex);
	
		HLock (theTrTableH);
		theTrDataP = (TrackDataPtr) (*theTrTableH + (**((TrackTableHan) theTrTableH)).horizOffset);
	
		theTrackingH =
			(FHI_TrackingHan) NewHandle (sizeof (FHI_TrackingRec) +
								(theTrDataP->numTracks * sizeof (FHI_TrackInfoRec)));
		if (MemError () != noErr) goto errorReturn;
		
		HLock ((Handle) theTrackingH);
		theTrackingP = (FHI_TrackingPtr) (*theTrackingH);

		theTrackingP->numTracks = theTrDataP->numTracks;
		for (i = 0; i < theTrackingP->numTracks; i++)
		{
			theTrackingP->tracks[i].trackValue = theTrDataP->tracks[i].trackValue;
			theTrackingP->tracks[i].nameID = theTrDataP->tracks[i].nameID;
			FindFontPName (theFontID, theTrDataP->tracks[i].nameID + 1, trackName);
			FHI_Pstrcpy (theTrackingP->tracks[i].name, trackName);
		}
		
		HUnlock ((Handle) theTrackingH);
		HUnlock ((Handle) theTrTableH);		

		DisposeHandle ((Handle) theTrTableH );
	}
	
	return (theTrackingH);

errorReturn:
	if (theTrTableH)
		DisposeHandle ((Handle) theTrTableH);
	return (NULL);
}
