/*
 *	File:		FontHIMenuData.c
 *
 *	Contains:	internal routines for FontHI Font and Style menus
 *
 *	Written by:	Dennis Buchheim
 *				some code from QuickDraw GX "font menu library"
 *
 *	Copyright:	 1992, 1993 by Apple Computer, Inc., all rights reserved.
 *
 *	Change History (most recent first):
 *
 */

#include <Resources.h>
#include <Memory.h>
#include <ToolUtils.h>
#include <Errors.h>

#include "font routines.h"
#include "font library.h"

#include "FontHIResources.h"
#include "FontHIUtils.h"
#include "FontHIMenuData.h"


#ifdef MacintoshIncludes
	#define GetFontName gGetFontName
#endif


#define BoldStr			((char *) "\pBold")
#define ItalicStr		((char *) "\pItalic")
#define BoldItalicStr	((char *) "\pBold Italic")

#define MaxStrLen		256

#define LargeVal		32767

#define DashMark		'\055'


typedef struct {
	short			nameLen;
	long			nameOffset;
} NameHeaderRec, *NameHeaderPtr, **NameHeaderHan;

typedef struct {
	short			numNames;
	NameHeaderHan	nameHdrH;
	Handle			nameDataH;
} NameListRec, *NameListPtr, **NameListHan;

typedef struct {
	Str255			name;
	Style			textfaceToApply;
} TextfaceInfoRec, *TextfaceInfoPtr;

typedef struct {
	short			numStyles;
	Str255			styleNames[];
} StyleListRec, *StyleListPtr, **StyleListHan;


/*
 *	sorts the text, command char, and item mark
 *	this way any submenus will sort with the  item
 *
 *	THIS SHOULD CALL IUMagString
 */
static void SortMenu(MenuHandle menu)
{
	short i, j, count = CountMItems(menu);
	
	for (i = 2; i <= count; i++)
		for (j = count; j >= i; --j) {
			short lowercmdChar, uppercmdChar;
			short lowermarkChar, uppermarkChar;
			Str255 lower, upper;
			
			GetItem(menu, j, lower);
			GetItem(menu, j-1, upper);
			GetItemCmd ( menu, j, &lowercmdChar );
			GetItemCmd ( menu, j-1, &uppercmdChar );
			GetItemMark ( menu, j, &lowermarkChar );
			GetItemMark ( menu, j-1, &uppermarkChar );
			
			if (FHIStr255Cmp(lower, upper) < 0) {
				SetItem(menu, j, upper);
				SetItem(menu, j-1, lower);
				SetItemCmd ( menu, j, uppercmdChar );
				SetItemCmd ( menu, j-1, lowercmdChar );
				SetItemMark ( menu, j, uppermarkChar );
				SetItemMark ( menu, j-1, lowermarkChar );
			}
		}
}


/*------------------------------------------------------ AddFamilyToMenuData */
static short AddFamilyToMenuData (FHI_MenuDataHan fontMenuDataH, Str255 familyName,
								  short numStyles, short *familyIndex)
{
	short						err = noErr;
	FHI_StyleHan				newStyleH = NULL;
	register FHI_MenuDataPtr	fontMenuDataP;
	short						i, numFamilies = (**fontMenuDataH).numFamilies;
	Ptr							blockSrcPtr, blockDestPtr;
	Size						blockSize;
	
	*familyIndex = 0;
	
	newStyleH = (FHI_StyleHan) NewHandle (numStyles * sizeof (FHI_StyleRec));
	if ((err = MemError ()) != noErr) return (err);
	
	/*	find index for new family (keep families alphabetical)
	 */
	for (i = 0; i < numFamilies; i++)
		if (FHIStr255Cmp ((**fontMenuDataH).families[i].name, familyName) < 0)
			(*familyIndex)++;
		else
			break;

	/*	shift data to account for new entry
	 */
	blockSrcPtr = (Ptr) ((**fontMenuDataH).families + *familyIndex);
	blockDestPtr = (Ptr) ((**fontMenuDataH).families + *familyIndex + 1);
	blockSize = (numFamilies - *familyIndex) * sizeof (FHI_FamilyRec);
	if (blockSize)
		BlockMove (blockSrcPtr, blockDestPtr, blockSize); 

	FHI_Pstrcpy ((**fontMenuDataH).families[*familyIndex].name, familyName);
	
	fontMenuDataP = (FHI_MenuDataPtr) (*fontMenuDataH);
	fontMenuDataP->families[*familyIndex].nameIndex = 0;
	fontMenuDataP->families[*familyIndex].familyMenuItem = 0;
	fontMenuDataP->families[*familyIndex].numStyles = 0;
	fontMenuDataP->families[*familyIndex].defaultStyleIndex = 0;
	(**fontMenuDataH).families[*familyIndex].styleH = newStyleH;
	
	(**fontMenuDataH).numFamilies++;
	
	return (noErr);
}


/*------------------------------------------------------ GetEquivStyleName */
static void GetEquivStyleName (Str255 styleName)
{
	Boolean			equivFound = false;
	Handle			theStrH = NULL;
	short			numEquivStrs;
	StyleListHan	equivStylesH = NULL;
	short			i, j;
	
	for (i = RegEquivStylesStrID; !equivFound; i++)
	{
		theStrH = GetResource ('STR#', i);
		numEquivStrs = theStrH ? *(short *)(*theStrH) : 0;
		
		if (theStrH)
			ReleaseResource (theStrH);

		if (numEquivStrs > 0)
		{
			equivStylesH = (StyleListHan) NewHandle (sizeof (StyleListRec) + (numEquivStrs * sizeof (Str255)));
			if (MemError () != noErr) return;
			
			HLock ((Handle) equivStylesH);
			
			for (j = 0; j < numEquivStrs; j++)
				GetIndString ((**equivStylesH).styleNames[j], i, j+1);
			
			/*	look at each of the equivalent strings (index >= 2 in the STR# resource)
				if the passed style name matches one of these strings,
				change the name to the preferred name (index = 1 in the STR# resource)
			 */
			for (j = 1; j < numEquivStrs; j++)
				if (!FHIStr255Cmp (styleName, (**equivStylesH).styleNames[j]))
				{
					equivFound = true;
					FHI_Pstrcpy (styleName, (**equivStylesH).styleNames[0]);
				}

			HUnlock ((Handle) equivStylesH);

			if (equivStylesH)
				DisposeHandle ((Handle) equivStylesH);
		}
		else
			break;
	}
}


/*------------------------------------------------------ GetStyleIndex */
/*
 *	find index for new style;
 *	keep "textface" styles in correct positions and keep other styles alphabetical
 */
static short GetStyleIndex (FHI_MenuDataHan fontMenuDataH, short familyIndex,
							Str255 styleName, Style textface)
{
	Str255					regularStr;
	register FHI_StylePtr	styleP;
	short					styleIndex = 0, offset = 0;
	short					i, numStyles = (**fontMenuDataH).families[familyIndex].numStyles;

	FHI_Pstrcpy (regularStr, "\pRegular");
	GetEquivStyleName (regularStr);
	
	HLock ((Handle)(**fontMenuDataH).families[familyIndex].styleH);
	styleP = (FHI_StylePtr) (*(**fontMenuDataH).families[familyIndex].styleH);
	
	if (!textface)
	{
		if (!FHIPstrcmp ((char *) styleName, BoldStr) && numStyles > 0)
		{
			if (!FHIPstrcmp ((char *) styleP->name, (char *) regularStr))
				styleIndex++;
		}
		else if (!FHIPstrcmp ((char *) styleName, ItalicStr))
		{
			for (i = 0; i < italic && i < numStyles; i++)
				if (!FHIPstrcmp ((char *) (styleP + i)->name, (char *) regularStr) ||
					!FHIPstrcmp ((char *) (styleP + i)->name, BoldStr))
					styleIndex++;
		}
		else if (!FHIPstrcmp ((char *) styleName, BoldItalicStr))
		{
			for (i = 0; i < (bold | italic) && i < numStyles; i++)
				if (!FHIPstrcmp ((char *) (styleP + i)->name, (char *) regularStr) ||
					!FHIPstrcmp ((char *) (styleP + i)->name, BoldStr) ||
					!FHIPstrcmp ((char *) (styleP + i)->name, ItalicStr))
					styleIndex++;
		}
		else if (FHIPstrcmp ((char *) styleName, (char *) regularStr))
		{
			for (i = 0; i <= (bold | italic) && i < numStyles; i++)
				if (!FHIPstrcmp ((char *) (styleP + i)->name, (char *) regularStr) ||
					!FHIPstrcmp ((char *) (styleP + i)->name, BoldStr) ||
					!FHIPstrcmp ((char *) (styleP + i)->name, ItalicStr) ||
					!FHIPstrcmp ((char *) (styleP + i)->name, BoldItalicStr))
					offset++;
	
			for (i = offset, styleIndex = offset; i < numStyles; i++)
			{
				if (FHIStr255Cmp ((styleP + i)->name, styleName) < 0)
					styleIndex++;
				else
					break;
			}
		}
	}
	else
	{
		if ((**fontMenuDataH).families[familyIndex].numStyles &&
			FHIPstrcmp ((char *) styleP->name, (char *) regularStr))
			offset = -1;
		switch (textface)
		{
			case bold:
				styleIndex = bold + offset;
				break;
			case italic:
				styleIndex = italic + offset;
				break;
			case (bold | italic):
				styleIndex = (bold | italic) + offset;
				break;
		}
	}

	HUnlock ((Handle) (**fontMenuDataH).families[familyIndex].styleH);
	
	return (styleIndex);
}

/*------------------------------------------------------ AddStyleToMenuData */
static void AddStyleToMenuData (FHI_MenuDataHan fontMenuDataH, short familyIndex,
								Str255 styleName, font styleFontID, Style textface)
{
	register FHI_StylePtr	styleP;
	short					numStyles = (**fontMenuDataH).families[familyIndex].numStyles;
	short					styleIndex;
	Size					blockSize;

	GetEquivStyleName (styleName);
	
	styleIndex = GetStyleIndex (fontMenuDataH, familyIndex, styleName, textface);

	HLock ((Handle) (**fontMenuDataH).families[familyIndex].styleH);
	styleP = (FHI_StylePtr) (*(**fontMenuDataH).families[familyIndex].styleH);
	
	/*	shift data to account for new entry
	 */
	styleP += styleIndex;
	blockSize = (numStyles - styleIndex) * sizeof (FHI_StyleRec);
	if (blockSize)
		BlockMove (styleP, styleP+1, blockSize); 

	FHI_Pstrcpy (styleP->name, styleName);
	styleP->nameIndex = 0;
	styleP->styleMenuItem = 0;
	styleP->fontID = styleFontID;
	styleP->textface = textface;
	styleP->selected = false;

	HUnlock ((Handle) (**fontMenuDataH).families[familyIndex].styleH);

	(**fontMenuDataH).families[familyIndex].numStyles++;
}


/*------------------------------------------------------ AddStylesToFamily */
static void AddStylesToFamily (FHI_MenuDataHan fontMenuDataH, short familyIndex, font familyFontID)
{
	short			i;
	long			styleCount = CountFontStyles (familyFontID);
	long			length;
	unsigned char	*styleName;

	for (i = 1; i <= styleCount; i++)
	{
		font curStyle = FindFontStyle (familyFontID, i, 0, 0, 0, 0, nil);
		if ((length = FindFontName (curStyle, styleFontName, macintoshPlatform, romanScript, englishLanguage, nil, nil)) > 0)
		{
			styleName = (unsigned char *) NewPtr (length + 1);
			FindFontName (curStyle, styleFontName, macintoshPlatform, romanScript, englishLanguage, &styleName[1], nil);
			styleName[0] = (char) length;
		}
		
		AddStyleToMenuData (fontMenuDataH, familyIndex, styleName, curStyle, 0);
		
		DisposPtr ((Ptr)styleName);
	}
}


/*------------------------------------------------------ GetMissingTextfaces */
static short GetMissingTextfaces (FHI_MenuDataHan fontMenuDataH, short familyIndex,
								  short *numTextfaces, TextfaceInfoPtr *textfaces)
{
	short			err = noErr;
	Handle			theStrH = NULL;
	short			numTextfaceStrs, i, j;
	TextfaceInfoPtr	tmpTextfaces;
	Str255			textfaceName;
	short			numActualStyles;
	Boolean			actualStyleFound;

	theStrH = GetResource ('STR#', TextfaceNamesStrID);
	numTextfaceStrs = theStrH ? *(short *)(*theStrH) : 0;
	
	if (theStrH)
		ReleaseResource (theStrH);
	
	*numTextfaces = 0;
	*textfaces = NULL;
	
	if (numTextfaceStrs > 0)
	{
		tmpTextfaces = (TextfaceInfoPtr) NewPtr (numTextfaceStrs * sizeof (TextfaceInfoRec));
		if ((err = MemError ()) != noErr) return (err);
		
		for (i = 0; i < numTextfaceStrs; i++)
		{
			GetIndString (textfaceName, TextfaceNamesStrID, i+1);
			
			numActualStyles = (**fontMenuDataH).families[familyIndex].numStyles;
			actualStyleFound = false;
			for (j = 0; j < numActualStyles && !actualStyleFound; j++)
				if (!FHIStr255Cmp ((*(**fontMenuDataH).families[familyIndex].styleH)[j].name, textfaceName))
					actualStyleFound = true;
			
			if (!actualStyleFound)
			{
				FHI_Pstrcpy (tmpTextfaces[*numTextfaces].name, textfaceName);
				
				if (!FHIPstrcmp ((char *) tmpTextfaces[*numTextfaces].name, BoldStr))
					tmpTextfaces[*numTextfaces].textfaceToApply = bold;
				else if (!FHIPstrcmp ((char *) tmpTextfaces[*numTextfaces].name, ItalicStr))
					tmpTextfaces[*numTextfaces].textfaceToApply = italic;
				else if (!FHIPstrcmp ((char *) tmpTextfaces[*numTextfaces].name, BoldItalicStr))
					tmpTextfaces[*numTextfaces].textfaceToApply = bold | italic;
				else
					tmpTextfaces[*numTextfaces].textfaceToApply = 0;
				
				(*numTextfaces)++;
			}
		}
		
		*textfaces = tmpTextfaces;
	}
	
	return (noErr);
}


/*------------------------------------------------------ AddTextfacesToFamily */
static short AddTextfacesToFamily (FHI_MenuDataHan fontMenuDataH, short familyIndex)
{
	short				err = noErr;
	short				numTextfaces;
	TextfaceInfoPtr		textfaces;
	FHI_StyleHan		styleH = NULL;
	short				i;
	
	err = GetMissingTextfaces (fontMenuDataH, familyIndex, &numTextfaces, &textfaces);
	if (err != noErr) goto errorReturn;
	
	if (numTextfaces)
	{
		styleH = (**fontMenuDataH).families[familyIndex].styleH;
		SetHandleSize ((Handle) styleH, GetHandleSize ((Handle) styleH) + (numTextfaces * sizeof (FHI_StyleRec)));
		(**fontMenuDataH).families[familyIndex].styleH = styleH;
		if ((err = MemError ()) != noErr) goto errorReturn;
	
		for (i = 0; i < numTextfaces; i++)
			/* fontID for textfaces cannot be set until Style menu indices are determined
			 */
			AddStyleToMenuData (fontMenuDataH, familyIndex, textfaces[i].name, (font) -1,
								textfaces[i].textfaceToApply);
								
		DisposePtr ((Ptr) textfaces);
	}
	
	return (noErr);
	
errorReturn:
	if (textfaces)
		DisposePtr ((Ptr) textfaces);
	return (err);
}

/*------------------------------------------------------ FHIInitMenuData */
/*
 *	Initialize the application global for the Font HI Font and Style menus.
 */
FHI_MenuDataHan FHIInitMenuData ()
{
	FHI_MenuDataHan				fontMenuDataH = NULL;
	register FHI_MenuDataPtr	fontMenuDataP;
	long						i, j;
	long						familyCount = CountFontFamilies ();
	long						styleCount;
	long						length;
	short						familyIndex;
	unsigned char				*familyName;
	
	fontMenuDataH = (FHI_MenuDataHan) NewHandle (sizeof (FHI_MenuDataRec) +
											(familyCount * sizeof (FHI_FamilyRec)));
	if (MemError () != noErr) goto errorReturn;
	
	fontMenuDataP = (FHI_MenuDataPtr) (*fontMenuDataH);
	fontMenuDataP->magicNum = FHI_MenuMagic;
	fontMenuDataP->numSelectedFonts = 0;
	fontMenuDataP->selFamilyMenuItem = 0;
	fontMenuDataP->selStyleMenuItem = 0;
	fontMenuDataP->totNumActualStyles = 0;
	fontMenuDataP->numFamilies = 0;
	
	for (i = 1; i <= familyCount; i++)
	{
		font curFamily = FindFontFamily (i, 0, 0, 0, 0, nil);

		if ((length = FindFontName (curFamily, familyFontName, macintoshPlatform, romanScript, englishLanguage, nil, nil)) > 0)
		{
			familyName = (unsigned char *) NewPtr(length + 1);
			FindFontName (curFamily, familyFontName, macintoshPlatform, romanScript, englishLanguage, &familyName[1], nil);
			familyName[0] = (char) length;
			
			styleCount = CountFontStyles (curFamily);

			if (AddFamilyToMenuData (fontMenuDataH, familyName, styleCount, &familyIndex) != noErr)
				goto errorReturn;

			AddStylesToFamily (fontMenuDataH, familyIndex, curFamily);
			
			if (AddTextfacesToFamily (fontMenuDataH, familyIndex) != noErr)
				goto errorReturn;
			
			DisposPtr ((Ptr)familyName);
		}
	}
	
	return (fontMenuDataH);
	
errorReturn:
	if (fontMenuDataH)
	{
		for (i = 0; i < (**fontMenuDataH).numFamilies; i++)
			DisposeHandle ((Handle) (**fontMenuDataH).families[i].styleH);
		
		DisposeHandle ((Handle) fontMenuDataH);
	}
	return (NULL);
}


/*------------------------------------------------------ FHIDisposeMenuData */
void FHIDisposeMenuData (FHI_MenuDataHan fontMenuDataH)
{
	short			i;

	for (i = 0; i < (**fontMenuDataH).numFamilies; i++)
		DisposeHandle ((Handle) (**fontMenuDataH).families[i].styleH);
	
	DisposeHandle ((Handle) fontMenuDataH);
}


/*------------------------------------------------------ SetFamilyMenuIndices */
static void SetFamilyMenuIndices (FHI_MenuDataHan fontMenuDataH, MenuHandle familyMenu)
{
	short			i, j;
	short			numMenuItems = CountMItems (familyMenu);
	Str255			familyItemStr;
	
	for (i = 0; i < (**fontMenuDataH).numFamilies; i++)
		for (j = 1; j <= numMenuItems; j++)
		{
			GetItem (familyMenu, j, familyItemStr);
		
			if (!FHIStr255Cmp ((**fontMenuDataH).families[i].name, familyItemStr))
			{
				(**fontMenuDataH).families[i].familyMenuItem = j;
				break;
			}
		}
}


/*------------------------------------------------------ FHIFamilyMenu */
/*
 *	Build a menu with the fonts' family names
 */
short FHIFamilyMenu (FHI_MenuDataHan fontMenuDataH, MenuHandle familyMenu)
{
	short			i;
	
	if (!fontMenuDataH) return (-1);

	HLock ((Handle)fontMenuDataH);
	
	for (i = 0; i < (**fontMenuDataH).numFamilies; i++)
		AppendMenu (familyMenu, (**fontMenuDataH).families[i].name);
	
	HUnlock ((Handle) fontMenuDataH);
	
	SortMenu (familyMenu);
	
	/* determine the corresponding family menu position for each font in fontMenuDataH */
	SetFamilyMenuIndices (fontMenuDataH, familyMenu);

	return (noErr);
}


/*------------------------------------------------------ UpdateStyleMenu */
/*
 *	Uses currently selected family to determine which items in the Style menu should be
 *	enabled/disabled and marked/unmarked.
 *
 *	Should be called when a new item is chosen in the _Font_ menu.
 */
static void UpdateStyleMenu (FHI_MenuDataHan fontMenuDataH, MenuHandle styleMenu)
{
	short					i, j;
	short					numMenuItems = (**fontMenuDataH).totNumActualStyles;
	short					selectedFamilyItem = (**fontMenuDataH).selFamilyMenuItem;
	register FHI_StylePtr	styleP;
	short					numStylesSelected = 0;
	
	if (!fontMenuDataH) return;
	
	for (i = 1; i <= numMenuItems; i++)
	{
		/* de-select and deactivate all menu items */
		CheckItem (styleMenu, i, false);
		DisableItem (styleMenu, i);
	}
	
	for (i = 0; i < (**fontMenuDataH).numFamilies; i++)
		if ((**fontMenuDataH).families[i].familyMenuItem == selectedFamilyItem)
			for (j = 0; j < (**fontMenuDataH).families[i].numStyles; j++)
			{
				HLock ((Handle)(**fontMenuDataH).families[i].styleH);
				styleP = (FHI_StylePtr) (*(**fontMenuDataH).families[i].styleH + j);
				
				EnableItem (styleMenu, styleP->styleMenuItem);
				
				if (styleP->selected)
				{
					SetItemMark (styleMenu, styleP->styleMenuItem, DashMark);
					(**fontMenuDataH).selStyleMenuItem = styleP->styleMenuItem;
					numStylesSelected++;
				}

				HUnlock ((Handle) (**fontMenuDataH).families[i].styleH);
			}

	if (numStylesSelected == 1)
		CheckItem (styleMenu, (**fontMenuDataH).selStyleMenuItem, true);
}


/*------------------------------------------------------ ClearFontSelections */
static void ClearFontSelections (FHI_MenuDataHan fontMenuDataH)
{
	short					i, j;
	register FHI_StylePtr	styleP;

	/*	clear font selections; update numSelectedFonts
	 */
	for (i = 0; i < (**fontMenuDataH).numFamilies; i++)
		for (j = 0; j < (**fontMenuDataH).families[i].numStyles; j++)
		{
			styleP = (FHI_StylePtr) (*(**fontMenuDataH).families[i].styleH + j);
			styleP->selected = false;
		}
	(**fontMenuDataH).numSelectedFonts = 0;
}


/*------------------------------------------------------ InitFontMappingData */
static void InitFontMappingData (FHI_MenuDataHan fontMenuDataH,
								 short numFontMappings, FHI_FontMapPtr fontMappings)
{
	short					i, j;
	register FHI_StylePtr	styleP;
	short					mappingInd = 0;
	
	/*	set up "old" parts of fontMappings
	 */
	for (i = 0; i < (**fontMenuDataH).numFamilies; i++)
		for (j = 0; j < (**fontMenuDataH).families[i].numStyles; j++)
		{
			styleP = (FHI_StylePtr) (*(**fontMenuDataH).families[i].styleH + j);
			if (styleP->selected && mappingInd < numFontMappings)
			{
				fontMappings[mappingInd].oldFontID = styleP->fontID;
				fontMappings[mappingInd].oldTextface = styleP->textface;
				fontMappings[mappingInd].oldFamilyMenuItem = (**fontMenuDataH).families[i].familyMenuItem;
				fontMappings[mappingInd++].oldStyleMenuItem = styleP->styleMenuItem;
			}
		}
}


/*------------------------------------------------------ DoFontFamilyMapping */
static void DoFontFamilyMapping (FHI_MenuDataHan fontMenuDataH, FHI_FontMapPtr fontMappings,
								 short index, short newFamilyItem)
{
	Boolean					fontMatchFound = false;
	short					i, j, familyIndex, styleIndex;
	register FHI_StylePtr	styleP;

	/*	set up "new" parts of a given fontMapping
	 */
	for (i = 0; i < (**fontMenuDataH).numFamilies && !fontMatchFound; i++)
		if ((**fontMenuDataH).families[i].familyMenuItem == newFamilyItem)
			for (j = 0; j < (**fontMenuDataH).families[i].numStyles && !fontMatchFound; j++)
			{
				styleP = (FHI_StylePtr) (*(**fontMenuDataH).families[i].styleH + j);
				if (styleP->styleMenuItem == fontMappings[index].oldStyleMenuItem)
				{
					fontMatchFound = true;
					
					familyIndex = i;
					styleIndex = j;
				}
			}
	
	if (!fontMatchFound)
		for (i = 0; i < (**fontMenuDataH).numFamilies && !fontMatchFound; i++)
			if ((**fontMenuDataH).families[i].familyMenuItem == newFamilyItem)
			{
				familyIndex = i;
				styleIndex = (**fontMenuDataH).families[i].defaultStyleIndex;
			}
	
	styleP = (FHI_StylePtr) (*(**fontMenuDataH).families[familyIndex].styleH + styleIndex);
	fontMappings[index].newFontID = styleP->fontID;
	fontMappings[index].newTextface = styleP->textface;

	if (!styleP->selected)
		(**fontMenuDataH).numSelectedFonts++;

	styleP->selected = true;
}


/*------------------------------------------------------ FHIFamilyMenuSetItem */
/*
 *	numFontMappings = numSelectedFonts
 */
short FHIFamilyMenuSetItem (FHI_MenuDataHan fontMenuDataH,
							 MenuHandle familyMenu, MenuHandle styleMenu, short newFamilyItem,
							 short *numFontMappings, FHI_FontMapPtr *fontMappings)
{
	short					err = noErr;
	register FHI_FontMapPtr	tmpFontMappings;
	short					i;
	register FHI_StylePtr	styleP;
	short					numMenuItems;
	
	if (!fontMenuDataH) return (-1);
	
	*numFontMappings = (**fontMenuDataH).numSelectedFonts;
	*fontMappings = NULL;

	if (*numFontMappings)
	{
		tmpFontMappings = (FHI_FontMapPtr) NewPtr (*numFontMappings * sizeof (FHI_FontMapRec));
		if ((err = MemError ()) != noErr) return (err);
		
		InitFontMappingData (fontMenuDataH, *numFontMappings, tmpFontMappings);
		
		ClearFontSelections (fontMenuDataH);
		
		for (i = 0; i < *numFontMappings; i++)
			DoFontFamilyMapping (fontMenuDataH, tmpFontMappings, i, newFamilyItem);
	}
	else  /* initialize Font menu selection; first call to this routine */
	{
		for (i = 0; i < (**fontMenuDataH).numFamilies; i++)
			if ((**fontMenuDataH).families[i].familyMenuItem == newFamilyItem)
			{
				styleP = (FHI_StylePtr) (*(**fontMenuDataH).families[i].styleH + 
										 (**fontMenuDataH).families[i].defaultStyleIndex);
				styleP->selected = true;
				(**fontMenuDataH).numSelectedFonts = 1;
				break;
			}
	}

	/*	update Font and Style menus
	 */
	numMenuItems = CountMItems (familyMenu);
	for (i = 1; i <= numMenuItems; i++)
		CheckItem (familyMenu, i, false);
	
	CheckItem (familyMenu, newFamilyItem, true);
	
	(**fontMenuDataH).selFamilyMenuItem = newFamilyItem;

	UpdateStyleMenu (fontMenuDataH, styleMenu);
	
	*fontMappings = tmpFontMappings;
	
	return (noErr);
}


/*------------------------------------------------------ InitStyleList */
static StyleListHan InitStyleList (MenuHandle styleMenu)
{
	StyleListHan	theStyleListH = NULL;
	Handle			theStrH = NULL;
	short			numStdStyleStrs, i, j;
	Str255			styleStr, itemStr;
	Boolean			styleAlreadyInMenu;
	short			numMenuItems = CountMItems (styleMenu);
	
	/*	add any remaining standard styles from STR# resource to the Style menu
	 */
	theStrH = GetResource ('STR#', StdStylesStrID);
	numStdStyleStrs = theStrH ? *(short *)(*theStrH) : 0;
	
	if (theStrH)
		ReleaseResource (theStrH);

	if (numStdStyleStrs > 0)
	{
		for (i = 1; i <= numStdStyleStrs; i++)
		{
			GetIndString (styleStr, StdStylesStrID, i);
			
			styleAlreadyInMenu = false;
			for (j = 1; j <= numMenuItems; j++)
			{
				GetItem (styleMenu, j, itemStr);
				
				if (!FHIStr255Cmp (styleStr, itemStr))
				{
					styleAlreadyInMenu = true;
					break;
				}
			}
			
			if (!styleAlreadyInMenu)
			{
				AppendMenu (styleMenu, styleStr);
				numMenuItems++;
			}
		}
	}

	theStyleListH = (StyleListHan) NewHandle (sizeof (StyleListRec) + (numMenuItems * sizeof (Str255)));
	if (MemError () != noErr) return (NULL);
	
	/*	add all items in the menu to the style list
	 */
	for (i = 0; i < numMenuItems; i++)
	{
		GetItem (styleMenu, i+1, itemStr);
		FHI_Pstrcpy ((**theStyleListH).styleNames[i], itemStr);
	}
	
	(**theStyleListH).numStyles = numMenuItems;
	
	return (theStyleListH);
}


/*------------------------------------------------------ AddToStyleList */
static short AddToStyleList (StyleListHan theStyleListH, Str255 styleName)
{
	short			err = noErr;
	short			i;
	Boolean			styleInList = false;
	
	for (i = 0; i < (**theStyleListH).numStyles; i++)
	{
		/* compare styles in style list to the style to be added... */
		if (!FHIStr255Cmp ((**theStyleListH).styleNames[i], styleName))
		{
			styleInList = true;
			break;
		}
	}
	
	if (!styleInList)
	{
		SetHandleSize ((Handle) theStyleListH, GetHandleSize ((Handle) theStyleListH) + sizeof (Str255));
		if ((err = MemError ()) != noErr) return (err);
		
		FHI_Pstrcpy (((**theStyleListH).styleNames[(**theStyleListH).numStyles]), styleName);

		(**theStyleListH).numStyles++;
	}
	
	return (noErr);
}


/*------------------------------------------------------ SortStyleList */
/*
 *	Sorts the style list items alphabetically
 */
static void SortStyleList (StyleListHan theStyleListH, short startIndex)
{
	short			i, j;
	Str255			lowerStr, upperStr;
	
	for (i = startIndex + 1; i < (**theStyleListH).numStyles; i++)
		for (j = (**theStyleListH).numStyles - 1; j >= i; --j)
		{
			FHI_Pstrcpy (upperStr, (**theStyleListH).styleNames[j]);
			FHI_Pstrcpy (lowerStr, (**theStyleListH).styleNames[j-1]);
			if (FHIStr255Cmp (upperStr, lowerStr) < 0)
			{
				FHI_Pstrcpy ((**theStyleListH).styleNames[j], lowerStr);
				FHI_Pstrcpy ((**theStyleListH).styleNames[j-1], upperStr);				
			}
		}
}


/*------------------------------------------------------ SetStyleMenuIndices */
static void SetStyleMenuIndices (FHI_MenuDataHan fontMenuDataH, StyleListHan theStyleListH)
{
	short					i, j, k;
	register FHI_StylePtr	styleP;
	Str255					styleName;

	for (i = 0; i < (**fontMenuDataH).numFamilies; i++)
		for (j = 0; j < (**fontMenuDataH).families[i].numStyles; j++)
		{
			HLock ((Handle)(**fontMenuDataH).families[i].styleH);
			styleP = (FHI_StylePtr) (*(**fontMenuDataH).families[i].styleH + j);

			FHI_Pstrcpy (styleName, styleP->name);
			
			for (k = 0; k < (**theStyleListH).numStyles; k++)
				if (!FHIStr255Cmp ((**theStyleListH).styleNames[k], styleName))
				{
					styleP->styleMenuItem = k+1;
					break;
				}
			
			HUnlock ((Handle) (**fontMenuDataH).families[i].styleH);
		}
}


/*------------------------------------------------------ FHIStyleMenu */
/*
 *	Build a menu of all font styles available in one or more fonts in the System
 */
short FHIStyleMenu (FHI_MenuDataHan fontMenuDataH, MenuHandle styleMenu)
{
	short					err = noErr;
	StyleListHan			theStyleListH = NULL;
	short					numOrigMenuItems, i, j;
	register FHI_StylePtr	styleP, defaultStyleP;
	short					minStyleMenuItem;

	if (!fontMenuDataH) return (-1);

	theStyleListH = InitStyleList (styleMenu);
	if (!theStyleListH) return (memFullErr);
	numOrigMenuItems = (**theStyleListH).numStyles;

	for (i = 0; i < (**fontMenuDataH).numFamilies; i++)
		for (j = 0; j < (**fontMenuDataH).families[i].numStyles; j++)
		{
			HLock ((Handle)(**fontMenuDataH).families[i].styleH);
			styleP = (FHI_StylePtr) (*(**fontMenuDataH).families[i].styleH + j);
			
			if ((err = AddToStyleList (theStyleListH, styleP->name)) != noErr)
			{
				DisposeHandle ((Handle) theStyleListH);
				return (err);
			}
			HUnlock ((Handle) (**fontMenuDataH).families[i].styleH);
		}

	/* Sort the newly added items */
	SortStyleList (theStyleListH, numOrigMenuItems);

	/* Determine the corresponding style menu position for each font in fontMenuDataH */
	SetStyleMenuIndices (fontMenuDataH, theStyleListH);
	
	/* Determine a default style for each family; use topmost non-textface item in Style menu */
	for (i = 0; i < (**fontMenuDataH).numFamilies; i++)
	{
		minStyleMenuItem = LargeVal;
		for (j = 0; j < (**fontMenuDataH).families[i].numStyles; j++)
		{
			styleP = (FHI_StylePtr) (*(**fontMenuDataH).families[i].styleH + j);
			if (styleP->styleMenuItem < minStyleMenuItem && !styleP->textface)
			{
				minStyleMenuItem = styleP->styleMenuItem;
				(**fontMenuDataH).families[i].defaultStyleIndex = j;
			}
		}
	}
	
	/* Set textface fontIDs to fontID of default style */
	for (i = 0; i < (**fontMenuDataH).numFamilies; i++)
		for (j = 0; j < (**fontMenuDataH).families[i].numStyles; j++)
		{
			styleP = (FHI_StylePtr) (*(**fontMenuDataH).families[i].styleH + j);
			if (styleP->textface)
			{
				defaultStyleP = (FHI_StylePtr) (*(**fontMenuDataH).families[i].styleH + (**fontMenuDataH).families[i].defaultStyleIndex);
				styleP->fontID = defaultStyleP->fontID;
			}
		}

	/* Draw the added style menu items */
	HLock ((Handle)theStyleListH);
	for (i = numOrigMenuItems; i < (**theStyleListH).numStyles; i++)
		AppendMenu (styleMenu, (**theStyleListH).styleNames[i]);
	HUnlock ((Handle) theStyleListH);

	if (theStyleListH)
		DisposeHandle ((Handle) theStyleListH);

	(**fontMenuDataH).totNumActualStyles = CountMItems (styleMenu);

	return (noErr);
}


/*------------------------------------------------------ DoFontStyleMapping */
static void DoFontStyleMapping (FHI_MenuDataHan fontMenuDataH, FHI_FontMapPtr fontMappings,
								short index, short newStyleItem)
{
	Boolean					fontMatchFound = false;
	short					i, j, familyIndex, styleIndex;
	register FHI_StylePtr	styleP;

	/*	set up "new" parts of a given fontMapping
	 */
	for (i = 0; i < (**fontMenuDataH).numFamilies && !fontMatchFound; i++)
		if ((**fontMenuDataH).families[i].familyMenuItem == fontMappings[index].oldFamilyMenuItem)
			for (j = 0; j < (**fontMenuDataH).families[i].numStyles && !fontMatchFound; j++)
			{
				styleP = (FHI_StylePtr) (*(**fontMenuDataH).families[i].styleH + j);
				if (styleP->styleMenuItem == newStyleItem)
				{
					fontMatchFound = true;
					
					familyIndex = i;
					styleIndex = j;
				}
			}
	
	if (fontMatchFound)
	{
		styleP = (FHI_StylePtr) (*(**fontMenuDataH).families[familyIndex].styleH + styleIndex);
		fontMappings[index].newFontID = styleP->fontID;
		fontMappings[index].newTextface = styleP->textface;

		if (!styleP->selected)
			(**fontMenuDataH).numSelectedFonts++;

		styleP->selected = true;
	}
	else
	{
		fontMappings[index].newFontID = fontMappings[index].oldFontID;
		fontMappings[index].newTextface = fontMappings[index].oldTextface;

		for (i = 0; i < (**fontMenuDataH).numFamilies; i++)
			for (j = 0; j < (**fontMenuDataH).families[i].numStyles; j++)
			{
				styleP = (FHI_StylePtr) (*(**fontMenuDataH).families[i].styleH + j);
				if (styleP->fontID == fontMappings[index].oldFontID &&
					styleP->textface == fontMappings[index].oldTextface)
					styleP->selected = true;
			}
		(**fontMenuDataH).numSelectedFonts++;
	}
}


/*------------------------------------------------------ FHIStyleMenuSetItem */
/*
 *	numFontMappings = numSelectedFonts
 */
short FHIStyleMenuSetItem (FHI_MenuDataHan fontMenuDataH, MenuHandle styleMenu, short newStyleItem,
							short *numFontMappings, FHI_FontMapPtr *fontMappings)
{
	short					err = noErr;
	short					i, j, k, mappingInd = 0;
	Boolean					fontMatchFound;
	register FHI_StylePtr	styleP;
	short					familyIndex, styleIndex;
	register FHI_FontMapPtr	tmpFontMappings;
	
	if (!fontMenuDataH) return;
	
	*numFontMappings = (**fontMenuDataH).numSelectedFonts;
	*fontMappings = NULL;
	
	tmpFontMappings = (FHI_FontMapPtr) NewPtr (*numFontMappings * sizeof (FHI_FontMapRec));
	if ((err = MemError ()) != noErr) return (err);
	
	InitFontMappingData (fontMenuDataH, *numFontMappings, tmpFontMappings);

	ClearFontSelections (fontMenuDataH);
	
	for (i = 0; i < *numFontMappings; i++)
		DoFontStyleMapping (fontMenuDataH, tmpFontMappings, i, newStyleItem);

	/*	update Style menu
	 */
	for (i = 1; i <= (**fontMenuDataH).totNumActualStyles; i++)
		CheckItem (styleMenu, i, false);
	
	CheckItem (styleMenu, newStyleItem, true);
	
	(**fontMenuDataH).selStyleMenuItem = newStyleItem;
	
	*fontMappings = tmpFontMappings;
	
	return (noErr);
}


/*------------------------------------------------------ FHIGetMarkedFamilyItem */
short FHIGetMarkedFamilyItem (FHI_MenuDataHan fontMenuDataH)
{
	if (!fontMenuDataH) return (0);

	return ((**fontMenuDataH).selFamilyMenuItem);
}


/*------------------------------------------------------ FHIGetMarkedStyleItem */
short FHIGetMarkedStyleItem (FHI_MenuDataHan fontMenuDataH)
{
	if (!fontMenuDataH) return (0);

	return ((**fontMenuDataH).selStyleMenuItem);
}


/*------------------------------------------------------ FHICountSelectedFonts */
short FHICountSelectedFonts (FHI_MenuDataHan fontMenuDataH)
{
	if (!fontMenuDataH) return (0);

	return ((**fontMenuDataH).numSelectedFonts);
}


/*------------------------------------------------------ FHIGetSelectedFontIDs */
short FHIGetSelectedFontIDs (FHI_MenuDataHan fontMenuDataH,
							 short *numSelectedFonts, FHI_FontIDPtr *fontIDs)
{
	short					err = noErr;
	short					i, j, fontIDIndex = 0;
	FHI_FontIDPtr			tmpFontIDs;
	register FHI_StylePtr	styleP;
	
	if (!fontMenuDataH) return (0);	
	
	*numSelectedFonts = (**fontMenuDataH).numSelectedFonts;
	*fontIDs = NULL;
	
	if (*numSelectedFonts)
	{
		tmpFontIDs = (FHI_FontIDPtr) NewPtr (*numSelectedFonts * sizeof (FHI_FontIDRec));
		if ((err = MemError ()) != noErr) return (err);

		for (i = 0; i < (**fontMenuDataH).numFamilies; i++)
			for (j = 0; j < (**fontMenuDataH).families[i].numStyles; j++)
			{
				styleP = (FHI_StylePtr) &(*(**fontMenuDataH).families[i].styleH)[j];
				if (styleP->selected)
				{
					if ( fontIDIndex < *numSelectedFonts ) {
						tmpFontIDs[fontIDIndex].fontID = styleP->fontID;
						tmpFontIDs[fontIDIndex++].textface = styleP->textface;
					}
				}
			}
		*numSelectedFonts = (**fontMenuDataH).numSelectedFonts = fontIDIndex;
		*fontIDs = tmpFontIDs;
	}
	
	return (noErr);
}


/*------------------------------------------------------ FHIGetFamilyName */
void FHIGetFamilyName (FHI_MenuDataHan fontMenuDataH, font fontID, Style textface,
						Str255 familyName)
{
	short					i, j;
	register FHI_StylePtr	styleP;
	
	if (!fontMenuDataH) return;

	FHI_Pstrcpy (familyName, "\p");
	
	for (i = 0; i < (**fontMenuDataH).numFamilies; i++)
		for (j = 0; j < (**fontMenuDataH).families[i].numStyles; j++)
		{
			styleP = (FHI_StylePtr) (*(**fontMenuDataH).families[i].styleH + j);
			if (styleP->fontID == fontID && styleP->textface == textface)
			{
				FHI_Pstrcpy (familyName, (**fontMenuDataH).families[i].name);
				return;
			}
		}
}


/*------------------------------------------------------ FHIGetStyleName */
void FHIGetStyleName (FHI_MenuDataHan fontMenuDataH, font fontID, Style textface,
					   Str255 styleName)
{
	short					i, j;
	register FHI_StylePtr	styleP;
	
	if (!fontMenuDataH) return;

	FHI_Pstrcpy (styleName, "\p");
	
	for (i = 0; i < (**fontMenuDataH).numFamilies; i++)
		for (j = 0; j < (**fontMenuDataH).families[i].numStyles; j++)
		{
			styleP = (FHI_StylePtr) (*(**fontMenuDataH).families[i].styleH + j);
			if (styleP->fontID == fontID && styleP->textface == textface)
			{
				FHI_Pstrcpy (styleName, styleP->name);
				return;
			}
		}
}


/*------------------------------------------------------ FHIMarkFonts */
void FHIMarkFonts (FHI_MenuDataHan fontMenuDataH,
				   MenuHandle familyMenu, MenuHandle styleMenu,
				   short numFontsToSelect, FHI_FontIDPtr fontIDs)
{
	short						i, j, k;
	Boolean						done = false;
	short						theMark, numFamilyItemsMarked = 0, numStyleItemsMarked = 0;
	register FHI_MenuDataPtr	fontMenuDataP;
	register FHI_StylePtr		styleP;
	unsigned char				familyName[MaxStrLen];

	if (!fontMenuDataH) return;

	for (i = 1; i <= (**fontMenuDataH).numFamilies; i++)
		CheckItem (familyMenu, i, false);

	for (i = 1; i <= (**fontMenuDataH).totNumActualStyles; i++)
	{
		CheckItem (styleMenu, i, false);
		DisableItem (styleMenu, i);
	}

	HLock ((Handle)fontMenuDataH);
	fontMenuDataP = (FHI_MenuDataPtr) (*fontMenuDataH);
	fontMenuDataP->selFamilyMenuItem = 0;
	fontMenuDataP->selStyleMenuItem = 0;

	ClearFontSelections (fontMenuDataH);

	for (i = 0; i < fontMenuDataP->numFamilies && !done; i++)
		for (j = 0; j < fontMenuDataP->families[i].numStyles && !done; j++)
		{
			HLock ((Handle)fontMenuDataP->families[i].styleH);
			styleP = (FHI_StylePtr) (*(fontMenuDataP->families[i].styleH) + j);
			
			styleP->selected = false;
			
			for (k = 0; k < numFontsToSelect; k++)
			{
				if (styleP->fontID == fontIDs[k].fontID && styleP->textface == fontIDs[k].textface)
				{
					fontMenuDataP->numSelectedFonts++;
					fontMenuDataP->selFamilyMenuItem = (**fontMenuDataH).families[i].familyMenuItem;
					fontMenuDataP->selStyleMenuItem = styleP->styleMenuItem;
					styleP->selected = true;
					
					if (numFontsToSelect == 1)
					{
						CheckItem (familyMenu, fontMenuDataP->selFamilyMenuItem, true);
						CheckItem (styleMenu, fontMenuDataP->selStyleMenuItem, true);
						
						done = true;
					}
					else
					{
						/* only count the item if it hasn't already been marked */
						GetItemMark (familyMenu, fontMenuDataP->families[i].familyMenuItem, &theMark);
						if (!theMark)
							numFamilyItemsMarked++;
						GetItemMark (styleMenu, styleP->styleMenuItem, &theMark);
						if (!theMark)
							numStyleItemsMarked++;
	
						SetItemMark (familyMenu, fontMenuDataP->families[i].familyMenuItem, DashMark);
						SetItemMark (styleMenu, styleP->styleMenuItem, DashMark);
						
						if ((**fontMenuDataH).numSelectedFonts == numFontsToSelect)
							done = true;
					}
					break;
				}
			}
			
			HUnlock ((Handle) (**fontMenuDataH).families[i].styleH);
		}

	if (numFontsToSelect > 1)  /* change -'s to 's, if appropriate */
	{
		if (numFamilyItemsMarked == 1)
			CheckItem (familyMenu, fontMenuDataP->selFamilyMenuItem, true);
		if (numStyleItemsMarked == 1)
			CheckItem (styleMenu, fontMenuDataP->selStyleMenuItem, true);
	}

	/*	enable only appropriate items in Style menu (styles present in selected families)
	 */
	for (i = 0; i < numFontsToSelect; i++)
		if (FindFontPName (fontIDs[i].fontID, familyFontName, familyName))
			for (j = 0; j < fontMenuDataP->numFamilies; j++)
				if (!FHIPstrcmp ((char *) fontMenuDataP->families[j].name, (char *) familyName))
					for (k = 0; k < fontMenuDataP->families[j].numStyles; k++)
					{
						styleP = (FHI_StylePtr) (*fontMenuDataP->families[j].styleH + k);
						EnableItem (styleMenu, styleP->styleMenuItem);
					}
	
	HUnlock ((Handle) fontMenuDataH);
}
