/*
	File:		FontHIPaletteData.c

 	Contains:	internal routines for FontHI font palette data handling

	Written by:	Dennis Buchheim, Parry Kejriwal

	Copyright:	 1992, 1993 by Apple Computer, Inc., All rights reserved.

	Change History (most recent first):

*/

#include <Memory.h>
#include <Errors.h>
#include <Resources.h>

#include "math routines.h"

#include "FontHIUtils.h"
#include "FontHIFontUtils.h"
#include "FontHIPaletteWind.h"
#include "FontHIPaletteData.h"

#define TrackMultiplier			50
#define TrackBoundary			5


/*------------------------------------------------------- FHIInitPaletteData */
FHI_PalDataHan FHIInitPaletteData ()
{
	FHI_PalDataHan			fontPalDataH = NULL;
	register FHI_PalDataPtr	fontPalDataP;
	
	fontPalDataH = (FHI_PalDataHan) NewHandle (sizeof (FHI_PalDataRec));
	
	if (fontPalDataH)
	{
		fontPalDataP = (FHI_PalDataPtr) (*fontPalDataH);
		
		fontPalDataP->magicNum = FHI_PaletteMagic;
		
		fontPalDataP->paletteVisible = false;
	
		fontPalDataP->palettePort = NULL;
		fontPalDataP->paletteWictH = NULL;
		fontPalDataP->paletteWitlH = NULL;
	
		fontPalDataP->selFontPopup = NULL;
		fontPalDataP->selStylePopup = NULL;
		fontPalDataP->styleVarPopup = NULL;
		fontPalDataP->featuresPopup = NULL;
		fontPalDataP->trackingSliderH = NULL;
	
		fontPalDataP->selFontChoice = 1;
		fontPalDataP->selStyleChoice = 1;
	
		fontPalDataP->curFontIndex = 0;
		fontPalDataP->numFontsInPalette = 0;
		fontPalDataP->fontDescH = NULL;
	}
	
	return (fontPalDataH);
}

/*------------------------------------------------------- DisposePaletteFontInfo */
static void DisposePaletteFontInfo (FHI_PalDataHan fontPalDataH)
{
	short					i, numFontsInPalette = (**fontPalDataH).numFontsInPalette;
	FHI_FeatMenuInfoHan		featH;
	register FHI_PalDataPtr	fontPalDataP;

	for (i = 0; i < numFontsInPalette; i++)
	{
		if (FP_FONT_DESC_P(fontPalDataH, i)->instanceH)
			DisposeHandle ((Handle) FP_FONT_DESC_P(fontPalDataH, i)->instanceH);
			
		if (featH = FP_FONT_DESC_P(fontPalDataH, i)->featH) {
			if((**featH).defaultFeatures)
				DisposeHandle((Handle) (**featH).defaultFeatures);
			DisposeHandle ((Handle) featH);
		}
		
		if (FP_FONT_DESC_P(fontPalDataH, i)->trackingH)
			DisposeHandle ((Handle) FP_FONT_DESC_P(fontPalDataH, i)->trackingH);
	}
	DisposeHandle ((Handle) (**fontPalDataH).fontDescH);
	
	fontPalDataP = (FHI_PalDataPtr) (*fontPalDataH);
	fontPalDataP->fontDescH = NULL;
	fontPalDataP->numFontsInPalette = 0;
	fontPalDataP->curFontIndex = 0;
}

/*------------------------------------------------------- FHIDisposePaletteData */
void FHIDisposePaletteData (FHI_PalDataHan fontPalDataH)
{
	register FHI_PalDataPtr	fontPalDataP;
	
	HLock ((Handle) fontPalDataH);
	fontPalDataP = (FHI_PalDataPtr) (*fontPalDataH);
	
	if (fontPalDataP->paletteWictH)
		DisposeHandle ((Handle) fontPalDataP->paletteWictH);
	if (fontPalDataP->paletteWitlH)
		DisposeHandle ((Handle) fontPalDataP->paletteWitlH);
	
	if (fontPalDataP->selFontPopup)
		ReleaseResource ((Handle) fontPalDataP->selFontPopup);
	if (fontPalDataP->selStylePopup)
		ReleaseResource ((Handle) fontPalDataP->selStylePopup);
	if (fontPalDataP->styleVarPopup)
		ReleaseResource ((Handle) fontPalDataP->styleVarPopup);
	if (fontPalDataP->featuresPopup)
		ReleaseResource ((Handle) fontPalDataP->featuresPopup);
	if (fontPalDataP->trackingSliderH)
		ReleaseResource ((Handle) fontPalDataP->trackingSliderH);
	
	if (fontPalDataP->fontDescH)
		DisposeHandle ((Handle) fontPalDataP->fontDescH);
	
	HUnlock ((Handle) fontPalDataH);
	
	DisposeHandle ((Handle) fontPalDataH);
}

/*------------------------------------------------------- FHIPaletteVisible */
Boolean FHIPaletteVisible (FHI_PalDataHan fontPalDataH)
{
	if ((**fontPalDataH).paletteVisible)
		return (true);
	else
		return (false);
}

/*------------------------------------------------------- FHIGetPalettePort */
WindowPtr FHIGetPalettePort (FHI_PalDataHan fontPalDataH)
{
	return ((**fontPalDataH).palettePort);
}

/*------------------------------------------------------- FHIWindowIsPalette */
Boolean FHIWindowIsPalette (FHI_PalDataHan fontPalDataH, WindowPtr theWindow)
{
	if (theWindow == (**fontPalDataH).palettePort)
		return (true);
	else
		return (false);
}

/*------------------------------------------------------- FHIGetFontIDIndex */
short FHIGetFontIDIndex (FHI_PalDataHan fontPalDataH, font fontID, Style textface)
{
	short					index = -1;
	short					i, numFontsInPalette = (**fontPalDataH).numFontsInPalette;

	for (i = 0; i < numFontsInPalette; i++)
		if (FP_FONT_DESC_P(fontPalDataH, i)->fontID == fontID &&
			FP_FONT_DESC_P(fontPalDataH, i)->textface == textface)
		{
			index = i;
			break;
		}
	
	return (index);
}

/*------------------------------------------------------- FHIGetFontNameIndex */
short FHIGetFontNameIndex (FHI_PalDataHan fontPalDataH,
						   Str255 familyName, Str255 styleName)
{
	short					index = -1;
	short					i, numFontsInPalette = (**fontPalDataH).numFontsInPalette;

	for (i = 0; i < numFontsInPalette; i++)
		if (!FHIStr255Cmp (FP_FONT_DESC_P(fontPalDataH, i)->familyName, familyName) &&
			!FHIStr255Cmp (FP_FONT_DESC_P(fontPalDataH, i)->styleName, styleName))
		{
			index = i;
			break;
		}
	
	return (index);
}

/*------------------------------------------------------- FHIGetFirstMarkedInstance */
short FHIGetFirstMarkedInstance (FHI_PalDataHan fontPalDataH)
{
	register FHI_InstMenuInfoPtr	fpInstanceP;
	short							i, curFontIndex = (**fontPalDataH).curFontIndex;
	
	if (FP_FONT_DESC_P(fontPalDataH, curFontIndex)->instanceH &&
		FP_INSTANCE_P(fontPalDataH, curFontIndex)->numInstances)
		fpInstanceP = FP_INSTANCE_P(fontPalDataH, curFontIndex);
	else
		return (1);  /* menu item 1 is "None in font" if there are 0 instances */
	
	for (i = 0; i < fpInstanceP->numInstances; i++)
		if (fpInstanceP->instances[i].selected)
			return (fpInstanceP->instances[i].menuItem);
}

/*------------------------------------------------------- FHIGetPaletteFontID */
font FHIGetPaletteFontID (FHI_PalDataHan fontPalDataH)
{
	short					curFontIndex = (**fontPalDataH).curFontIndex;

	if ((**fontPalDataH).fontDescH)
		return (FP_FONT_DESC_P(fontPalDataH, curFontIndex)->fontID);
	else
		return (0);
}

/*------------------------------------------------------- FHIGetSelectedInstance */
short FHIGetSelectedInstance (FHI_PalDataHan fontPalDataH)
{
	register FHI_InstMenuInfoPtr	fpInstanceP;
	short							index = 0, i;
	short							curFontIndex = (**fontPalDataH).curFontIndex;
	
	if ((**fontPalDataH).fontDescH)
	{
		if (FP_FONT_DESC_P(fontPalDataH, curFontIndex)->instanceH)
		{
			fpInstanceP = FP_INSTANCE_P(fontPalDataH, curFontIndex);
			
			for (i = 0; i < fpInstanceP->numInstances; i++)
				if (fpInstanceP->instances[i].selected)
				{
					index = fpInstanceP->instances[i].indexInFont;
					break;
				}
		}
	}
	
	return (index);
}

/*------------------------------------------------------- FHIGetSelectedFeatures */
/*
 *	this returns the settings of all the 'ON' features and their settings
 */
short FHIGetSelectedFeatures (FHI_PalDataHan fontPalDataH, 
							  short *numFeatures, fontRunFeature **features)
{
	short				err = noErr;
	short				i;
	short				numOfFeatures;
	short				numOfSelectedFeatures;
	FHI_FeatMenuInfoHan	featH;
	FHI_FeatMenuInfoPtr	featP;
	fontRunFeature		*tmpFeatures;
	short				featureIndex;
	short				curFontIndex = (**fontPalDataH).curFontIndex;

	/* initialization stuff */
	featH = FP_FONT_DESC_P(fontPalDataH, curFontIndex)->featH;
	
	*features = NULL;
	*numFeatures = 0;
	tmpFeatures = NULL;

	numOfFeatures = (**featH).numOfFeatures;
	numOfSelectedFeatures = featureIndex = 0;

	/* this is to prepare for allocation of pointers, etc. */
	featP = *featH;
	for ( i= 0; i < numOfFeatures; i++ )
		if ( featP->features[i].selected )
			numOfSelectedFeatures++;

	/* now start stuffing the RunFeatures after allocating a pointer */
	if ( numOfSelectedFeatures ) {
		tmpFeatures = (fontRunFeature *) NewPtr (sizeof (fontRunFeature) * numOfSelectedFeatures);
		if (err = MemError()) return ( err );
		
		featP = *featH;
		for ( i= 0; i < numOfFeatures; i++ ) {
			if ( featP->features[i].selected ) {
				tmpFeatures[featureIndex].featureType = featP->features[i].featureType;
				tmpFeatures[featureIndex].setting = featP->features[i].featureSetting.setting;
				tmpFeatures++;
			}
		}
		*features = tmpFeatures;
		*numFeatures = numOfSelectedFeatures;
	}
	
	return (err);
}	

/*------------------------------------------------------- FHIGetTrackingValue */
Fixed FHIGetTrackingValue (FHI_PalDataHan fontPalDataH)
{
	float					value = 0;
	short					curFontIndex = (**fontPalDataH).curFontIndex;
	
	if ((**fontPalDataH).fontDescH)
		if (FP_FONT_DESC_P(fontPalDataH, curFontIndex)->trackingH)
		{
			value = FP_TRACKING_P(fontPalDataH, curFontIndex)->curVal;
			value /= FP_TRACKING_P(fontPalDataH, curFontIndex)->multiplier;
		}

	return (FloatToFixed (value));
}

/*------------------------------------------------------- GetFpFontInstances */
static FHI_InstMenuInfoHan GetFpFontInstances (font fontID)
{
	FHI_InstanceHan					theInstanceH = NULL;
	short							numInstances = 0;
	FHI_InstMenuInfoHan				fpInstanceH = NULL;
	register FHI_InstMenuInfoPtr	fpInstanceP;
	short							i;
	
	theInstanceH = FHIGetInstances (fontID);
	if (theInstanceH)
		numInstances = (**theInstanceH).numberOfInstances;
	
	fpInstanceH = (FHI_InstMenuInfoHan) NewHandle (sizeof (FHI_InstMenuInfoRec) +
										 (numInstances * sizeof (FHI_InstItemInfoRec)));
	if (MemError () != noErr) goto errorReturn;
	
	HLock ((Handle) fpInstanceH);
	fpInstanceP = (FHI_InstMenuInfoPtr) (*fpInstanceH);
	
	fpInstanceP->numSelInstances = 0;
	fpInstanceP->numInstances = numInstances;
	
	for (i = 0; i < numInstances; i++)
	{
		FHI_Pstrcpy (fpInstanceP->instances[i].name, (**theInstanceH).names[i].name);
		fpInstanceP->instances[i].indexInFont = i+1;
		fpInstanceP->instances[i].menuItem = i+1;
		if (i == 0)
		{
			fpInstanceP->instances[i].selected = true;
			fpInstanceP->numSelInstances++;
		}
		else
			fpInstanceP->instances[i].selected = false;			
	}
	
	HUnlock ((Handle) fpInstanceH);
	
	if (theInstanceH)
		DisposeHandle ((Handle) theInstanceH);
	
	return (fpInstanceH);

errorReturn:
	if (theInstanceH)
		DisposeHandle ((Handle) theInstanceH);
	return (NULL);
}

/*------------------------------------------------------- AssignFeatureMenuIndices */
static void AssignFeatureMenuIndices (FHI_FeatMenuInfoHan featH)
{
	short					menuIndex;
	fontFeature				featureType;
	FHI_FeatMenuInfoPtr		featP;
	short					numOfFeatures, i;

	menuIndex = 0;
	featureType = -1;
	featP = *featH;
	numOfFeatures = featP->numOfFeatures;
	
	for ( i =0; i < numOfFeatures; i++) {
	
		if ( featP->features[i].featureType != featureType ) {
		
			/* do this because we need a separator for different feature types */
			featureType = featP->features[i].featureType;
			menuIndex += 2;
		}
		else
			menuIndex++;
			
		featP->features[i].menuItem = menuIndex;
	}
}	

/*------------------------------------------------------- SelectDefaultFeatures */
static void SelectDefaultFeatures ( FHI_FeatMenuInfoHan featH )
{
	FHI_FeatMenuInfoPtr	featP;
	short				numDefaultFeatures;
	fontRunFeature		**defaultRunFeatures;
	short				i, j;

	numDefaultFeatures = (**featH).numDefaultFeatures;
	defaultRunFeatures = (**featH).defaultFeatures;

	featP = *featH;
	for ( i =0; i < numDefaultFeatures; i++ ) {
	
		for ( j = 0; j < featP->numOfFeatures; j++ ) {
		
			if ( featP->features[j].featureType == (*defaultRunFeatures)[i].featureType
					&&	featP->features[j].featureSetting.setting == (*defaultRunFeatures)[i].setting ) {
				featP->features[j].selected = true;
				break;
			}
		}
	}
}

/*------------------------------------------------------- GetFpFontFeatures */
static FHI_FeatMenuInfoHan GetFpFontFeatures (font fontID)
{
	FHI_FeatMenuInfoHan				featH;

	featH = FHIGetFeatures (fontID);
	
	if (featH) {
		SelectDefaultFeatures ( featH );
		
		AssignFeatureMenuIndices ( featH );
	}
	return (featH);
}

/*------------------------------------------------------- GetFpFontTrackingData */
static FHI_TrackSliderHan GetFpFontTrackingData (font fontID)
{
	FHI_TrackingHan				theTrackingH = NULL;
	FHI_TrackSliderHan			fpTrackingH = NULL;
	register FHI_TrackSliderPtr	fpTrackingP;
	Fixed						minValue, maxValue;
	short						i;
	float						min, max, cur;

	theTrackingH = FHIGetTrackTableInfo ( fontID );
	
	fpTrackingH = (FHI_TrackSliderHan) NewHandle (sizeof (FHI_TrackSliderRec));
	if (MemError () != noErr) goto errorReturn;
	
	if (theTrackingH)
	{
	#if 0
		minValue = maxValue = (**theTrackingH).tracks[0].trackValue;
		
		for (i = 0; i < (**theTrackingH).numTracks; i++)
		{
			if ( (**theTrackingH).tracks[i].trackValue < minValue )
				minValue = (**theTrackingH).tracks[i].trackValue;
			if ( (**theTrackingH).tracks[i].trackValue > maxValue )
				maxValue = (**theTrackingH).tracks[i].trackValue;
		}
		
		min = FixedToFloat ( minValue );
		max = FixedToFloat ( maxValue );
	#endif

		min = -TrackBoundary;
		max = TrackBoundary;

		cur = 0;
		
		fpTrackingP = (FHI_TrackSliderPtr) (*fpTrackingH);
		
		fpTrackingP->numTracks = (**theTrackingH).numTracks;
		fpTrackingP->multiplier = TrackMultiplier;
		
		if ((min == max) && (min == cur))
		{
			fpTrackingP->minVal = -fpTrackingP->multiplier;
			fpTrackingP->maxVal = fpTrackingP->multiplier;
			fpTrackingP->curVal = cur;
		}
		else
		{
			fpTrackingP->minVal = min * fpTrackingP->multiplier;
			fpTrackingP->maxVal = max * fpTrackingP->multiplier;
			fpTrackingP->curVal = cur * fpTrackingP->multiplier;
		}
		
		DisposeHandle ((Handle) theTrackingH );
	}
	else
	{
		fpTrackingP = (FHI_TrackSliderPtr) (*fpTrackingH);
		fpTrackingP->numTracks = 0;
		fpTrackingP->multiplier = 1;
		fpTrackingP->minVal = 0;
		fpTrackingP->maxVal = 0;
		fpTrackingP->curVal = 0;
	}
	
	return (fpTrackingH);

errorReturn:
	if (theTrackingH)
		DisposeHandle ((Handle) theTrackingH);
	return (NULL);
}

/*------------------------------------------------------- AddFontToFpFontDesc */
static void AddFontToFpFontDesc (FHI_PalDataHan fontPalDataH,
								 font fontID, Style textface,
								 FHI_MenuDataHan fontMenuDataH)
{
	short						fontIndex;
	Boolean						fontAlreadyInList = false;
	register FHI_FontDescPtr	fpFontDescP;
	
	fontIndex = FHIGetFontIDIndex (fontPalDataH, fontID, textface);
	if (fontIndex != -1)
		fontAlreadyInList = true;
	
	if (!fontAlreadyInList)
	{
		(**fontPalDataH).numFontsInPalette++;
		fontIndex = (**fontPalDataH).numFontsInPalette - 1;
		
		HLock ((Handle) (**fontPalDataH).fontDescH);
		fpFontDescP = FP_FONT_DESC_P(fontPalDataH, fontIndex);
		
		FontHIGetFamilyName (fontMenuDataH, fontID, textface, fpFontDescP->familyName);
		FontHIGetStyleName (fontMenuDataH, fontID, textface, fpFontDescP->styleName);
		fpFontDescP->fontID = fontID;
		fpFontDescP->textface = textface;
		
		HUnlock ((Handle) (**fontPalDataH).fontDescH);

		FP_FONT_DESC_P(fontPalDataH, fontIndex)->instanceH = GetFpFontInstances (fontID);
		FP_FONT_DESC_P(fontPalDataH, fontIndex)->featH = GetFpFontFeatures (fontID);
		FP_FONT_DESC_P(fontPalDataH, fontIndex)->trackingH = GetFpFontTrackingData (fontID);
	}
}

/*------------------------------------------------------- FHIUpdatePaletteFontInfo */
short FHIUpdatePaletteFontInfo (FHI_PalDataHan fontPalDataH, FHI_MenuDataHan fontMenuDataH)
{
	short					err = noErr;
	short					numSelectedFonts;
	FHI_FontIDPtr			fontIDs;
	short					i;
	
	if ((**fontPalDataH).fontDescH)
		DisposePaletteFontInfo (fontPalDataH);

	err = FontHIGetSelectedFontIDs (fontMenuDataH, &numSelectedFonts, &fontIDs);
	if (err != noErr) return (err);
	
	(**fontPalDataH).fontDescH =
		(FHI_FontDescHan) NewHandle (numSelectedFonts * sizeof (FHI_FontDescRec));
	if ((err = MemError ()) != noErr) return (err);
	
	for (i = 0; i < numSelectedFonts; i++)
		AddFontToFpFontDesc (fontPalDataH, fontIDs[i].fontID, fontIDs[i].textface,
							 fontMenuDataH);
	
	DisposePtr ((Ptr) fontIDs);
	
	return (noErr);
}

/*------------------------------------------------------- FHISetSelectedInstances */
/*
 *	call after a text selection is made by the user; update selected instances for
 *	each font in the selection; call FontHIPaletteFontChange BEFORE this routine
 */
void FHISetSelectedInstances (FHI_PalDataHan fontPalDataH,
							  font fontID, Style textface,
							  short numIndices, short *indices,
							  Boolean clearAll)
{
	short							fontIndex;
	register FHI_InstMenuInfoPtr	fpInstanceP;
	short							numInstances = 0, i, j;
	
	fontIndex = FHIGetFontIDIndex (fontPalDataH, fontID, textface);
	if (fontIndex == -1)
		return;

	if (FP_FONT_DESC_P(fontPalDataH, fontIndex)->instanceH)
	{
		fpInstanceP = FP_INSTANCE_P(fontPalDataH, fontIndex);
		numInstances = fpInstanceP->numInstances;
	}
	
	if (numInstances && numIndices > 0 && numIndices <= numInstances)
	{
		if (clearAll)
		{
			for (i = 0; i < numInstances; i++)
				fpInstanceP->instances[i].selected = false;
			fpInstanceP->numSelInstances = 0;
		}

		for (i = 0; i < numIndices; i++)
			for (j = 0; j < numInstances; j++)
				if (fpInstanceP->instances[j].indexInFont == indices[i])
				{
					fpInstanceP->instances[j].selected = true;
					fpInstanceP->numSelInstances++;
					break;
				}
	}

	if (FP_FONT_DESC_P(fontPalDataH, 0)->fontID == fontID &&
		FP_FONT_DESC_P(fontPalDataH, 0)->textface == textface)
		FHIBuildInstancePopup (fontPalDataH, fontID, textface);
}

/*------------------------------------------------------------------- ClearAllSelectedFeatures */
static void ClearAllSelectedFeatures ( FHI_FeatMenuInfoHan	featH )
{
	FHI_FeatMenuInfoPtr	featP;
	short				i;

	featP = *featH;
	
	for ( i =0; i < featP->numOfFeatures; i++ ) {
	
		featP->features[i].selected = false;
	}
}

/*------------------------------------------------------------------- SetSelectedFeatures */
static void SetSelectedFeatures (FHI_FeatMenuInfoHan featH,
								 short numFeatures, fontRunFeature *features)
{
	FHI_FeatMenuInfoPtr	featP;
	short				i, j,totalRunFeatures;

	featP = *featH;
	totalRunFeatures = featP->numOfFeatures;
	
	for ( i= 0; i < numFeatures; i++ ) {
	
		for ( j = 0; j < totalRunFeatures; j++ ) {
			
			if ( featP->features[j].featureType == features[i].featureType ) {
			
					if (featP->features[j].featureSetting.setting == features[i].setting) {
						featP->features[j].selected = true;
						break;
					}
			}
		}
	}
}

/*------------------------------------------------------- FHISetAllSelectedFeatures */
void FHISetAllSelectedFeatures (FHI_PalDataHan fontPalDataH,
								font fontID, Style textface,
								short numFeatureGroups, short **featureGroupSizes, 
								fontRunFeature ***featureGroups)
{
	short					j, k;
	short					fontIndex;
	FHI_FeatMenuInfoHan		featH;
	fontRunFeature			* srcFeatures;
	short					srcNumFeatures;
	fontRunFeature			** dstFeatures;
	short					dstNumFeatures;

	if ( !numFeatureGroups ) return;
	
	fontIndex = FHIGetFontIDIndex (fontPalDataH, fontID, textface);
	if (fontIndex == -1)
		return;

	featH = FP_FONT_DESC_P(fontPalDataH, fontIndex)->featH;

	(**featH).numFeatGroups = numFeatureGroups;

	ClearAllSelectedFeatures ( featH );
	
	for (j = 0; j < numFeatureGroups; j++ ) {
	
		srcNumFeatures = (*featureGroupSizes)[j];
		srcFeatures = (*featureGroups)[j];
		
		FHIApplyFeaturesToDefault( featH, srcNumFeatures, srcFeatures, &dstNumFeatures, &dstFeatures);
		
		SetSelectedFeatures (featH, dstNumFeatures, *dstFeatures);
		
		DisposeHandle((Handle) dstFeatures);
		
	}

	if (FP_FONT_DESC_P(fontPalDataH, 0)->fontID == fontID &&
		FP_FONT_DESC_P(fontPalDataH, 0)->textface == textface)
	{
		if ( numFeatureGroups > 1 )
			FHIMarkMultFeatPopup (fontPalDataH, fontID, textface);
		else
			FHIMarkFeatPopup (fontPalDataH, fontID, textface);
	}
}

/*------------------------------------------------------- FHISetTrackingValue */
/*
 *	call after a text selection is made by the user; update tracking value for
 *	each font in the selection; call FontHIPaletteFontChange BEFORE this routine
 */
void FHISetTrackingValue (FHI_PalDataHan fontPalDataH, 
						  font fontID, Style textface, float newValue)
{
	short						fontIndex;
	register FHI_TrackSliderPtr	fpTrackingP;

	fontIndex = FHIGetFontIDIndex (fontPalDataH, fontID, textface);
	if (fontIndex == -1)
		return;
		
	if (FP_FONT_DESC_P(fontPalDataH, fontIndex)->trackingH)
	{
		fpTrackingP = FP_TRACKING_P(fontPalDataH, fontIndex);
		fpTrackingP->curVal = newValue;
	}
	
	if (FP_FONT_DESC_P(fontPalDataH, 0)->fontID == fontID &&
		FP_FONT_DESC_P(fontPalDataH, 0)->textface == textface)
		UpdateTrackControl (fontPalDataH);
}
