/*
	File:		FontHIPaletteWind.c

 	Contains:	internal routines for FontHI font palette window handling

	Written by:	Dennis Buchheim, Parry Kejriwal

	Copyright:	 1992, 1993 by Apple Computer, Inc., All rights reserved.

	Change History (most recent first):

*/

#include <Resources.h>
#include <Menus.h>
#include <Memory.h>
#include <Windows.h>

#include "font library.h"

#include <FixMath.h>
#include "math routines.h"

#include "FontHIResources.h"
#include "FontHIWindowAids.h"
#include "FontHIScrolling.h"
#include "FontHIUtils.h"
#include "FontHIFontUtils.h"
#include "FontHIPaletteWind.h"


#define SelectedTextLabel		1 
#define Line2					2 
#define SelFontPopupItem		3 
#define SelStylePopupItem		4 
#define Line5					5 
#define VariationLabel			6 
#define StyleVarPopupItem		7 
#define FeaturesLabel			8
#define FeaturesPopupItem		9 
#define Rect14					14 
#define TightLabel				10 
#define NormalLabel				11 
#define LooseLabel				12 
#define TrackingSlider			13 
#define TrackingLabel			15 
#define StyleVarPict			16 
#define DesignPict				17 
#define TrackingPict			18 

#define LabelFontItem			VariationLabel

#define FeaturesPopupInset		32
#define FirstFeaturesStr		"\p "

#define DashMark				'\055'


typedef struct {
	font			fontID;
	Style			textface;
	short			numSelInstances;
	short			*selInstances;
	short			numSelFeatures;
	short			*selFeatures;
} SelectionsRec, *SelectionsPtr, **SelectionsHan;


/*------------------------------------------------------- ClearMenu */
static void ClearMenu (MenuHandle theMenu, short firstItem)
{
	short		numMenuItems;
	short		i;
	
	numMenuItems = CountMItems (theMenu);
	for (i = numMenuItems; i >= firstItem; i--)
		DelMenuItem (theMenu, i);
}

/*------------------------------------------------------- BuildFontPopup */
static void BuildFontPopup (FHI_PalDataHan fontPalDataH)
{
	register FHI_PalDataPtr	fontPalDataP;
	GrafPtr					savePort;
	short					numMenuItems, i, j;
	Rect					bounds;
	Str255					familyName, itemStr;
	Boolean					familyAlreadyInMenu;
	
	HLock ((Handle) fontPalDataH);
	fontPalDataP = (FHI_PalDataPtr) (*fontPalDataH);

	GetPort (&savePort);
	SetPort (fontPalDataP->palettePort);
	
	ClearMenu (fontPalDataP->selFontPopup, 1);
	
	for (i = 0; i < fontPalDataP->numFontsInPalette; i++)
	{
		FHI_Pstrcpy (familyName, FP_FONT_DESC_P(fontPalDataH, i)->familyName);
		
		familyAlreadyInMenu = false;
		numMenuItems = CountMItems (fontPalDataP->selFontPopup);
		for (j = 1; j <= numMenuItems; j++)
		{
			GetItem (fontPalDataP->selFontPopup, j, itemStr);
			
			if (!FHIPstrcmp ((char *) itemStr, (char *) familyName))
				familyAlreadyInMenu = true;
		}
		
		if (!familyAlreadyInMenu)
			AppendMenu (fontPalDataP->selFontPopup, familyName);
	}
	
	fontPalDataP->selFontChoice = 1;

	CheckItem (fontPalDataP->selFontPopup, fontPalDataP->selFontChoice, true);

	FHIGetWRect (fontPalDataP->paletteWitlH, SelFontPopupItem, &bounds);
	InsetRect (&bounds, 2, 2);
	InvalRect (&bounds);

	SetPort (savePort);
	
	HUnlock ((Handle) fontPalDataH);
}

/*------------------------------------------------------- BuildStylePopup */
static void BuildStylePopup (FHI_PalDataHan fontPalDataH, Str255 familyName)
{
	register FHI_PalDataPtr		fontPalDataP;
	GrafPtr						savePort;
	short						i;
	register FHI_FontDescPtr	fpFontDescP;
	Rect						bounds;

	HLock ((Handle) fontPalDataH);
	fontPalDataP = (FHI_PalDataPtr) (*fontPalDataH);

	GetPort (&savePort);
	SetPort (fontPalDataP->palettePort);
	
	ClearMenu (fontPalDataP->selStylePopup, 1);
	
	HLock ((Handle) fontPalDataP->fontDescH);
	for (i = 0; i < fontPalDataP->numFontsInPalette; i++)
	{
		fpFontDescP = FP_FONT_DESC_P(fontPalDataH, i);
		if (!FHIStr255Cmp (fpFontDescP->familyName, familyName))
			AppendMenu (fontPalDataP->selStylePopup, fpFontDescP->styleName);
	}
	HUnlock ((Handle) fontPalDataP->fontDescH);
	
	fontPalDataP->selStyleChoice = 1;
	
	CheckItem (fontPalDataP->selStylePopup, fontPalDataP->selStyleChoice, true);

	FHIGetWRect (fontPalDataP->paletteWitlH, SelStylePopupItem, &bounds);
	InsetRect (&bounds, 2, 2);
	InvalRect (&bounds);
	
	SetPort (savePort);
	
	HUnlock ((Handle) fontPalDataH);
}

/*------------------------------------------------------- FHIBuildInstancePopup */
void FHIBuildInstancePopup (FHI_PalDataHan fontPalDataH, font fontID, Style textface)
{
	register FHI_PalDataPtr			fontPalDataP;
	GrafPtr							savePort;
	short							fontIndex, i;
	register FHI_FontDescPtr		fpFontDescP;
	register FHI_InstMenuInfoPtr	fpInstanceP;
	Rect							bounds;
	
	HLock ((Handle) fontPalDataH);
	fontPalDataP = (FHI_PalDataPtr) (*fontPalDataH);

	GetPort (&savePort);
	SetPort (fontPalDataP->palettePort);

	ClearMenu (fontPalDataP->styleVarPopup, 1);
	
	fontIndex = FHIGetFontIDIndex (fontPalDataH, fontID, textface);
	if (fontIndex == -1)
		return;

	fpFontDescP = FP_FONT_DESC_P(fontPalDataH, fontIndex);

	if (fpFontDescP->instanceH && FP_INSTANCE_P(fontPalDataH, fontIndex)->numInstances)
	{
		HLock ((Handle) FP_FONT_DESC_P(fontPalDataH, fontIndex)->instanceH);
		fpInstanceP = FP_INSTANCE_P(fontPalDataH, fontIndex);
		
		for (i = 0; i < fpInstanceP->numInstances; i++)
		{
			AppendMenu (fontPalDataP->styleVarPopup, fpInstanceP->instances[i].name);
			
			if (fpInstanceP->numSelInstances == 1 && fpInstanceP->instances[i].selected)
				CheckItem (fontPalDataP->styleVarPopup, i+1, true);
			else if (fpInstanceP->instances[i].selected)
				SetItemMark (fontPalDataP->styleVarPopup, i+1, DashMark);
		}
		
		HUnlock ((Handle) FP_FONT_DESC_P(fontPalDataH, fontIndex)->instanceH);
	}
	else
	{
		AppendMenu (fontPalDataP->styleVarPopup, "\p(None in font");
		CheckItem (fontPalDataP->styleVarPopup, 1, true);
	}
	
	FHIGetWRect (fontPalDataP->paletteWitlH, StyleVarPopupItem, &bounds);
	InsetRect (&bounds, 2, 2);
	InvalRect (&bounds);
	
	SetPort (savePort);
	
	HUnlock ((Handle) fontPalDataH);
}

/*------------------------------------------------------- FHIMarkFeatPopup */
void FHIMarkFeatPopup (FHI_PalDataHan fontPalDataH, font fontID, Style textface)
{
	short				fontIndex, numMenuItems, numOfFeatures, i;
	MenuHandle			theMenu;
	FHI_FeatMenuInfoHan	featH;
	FHI_FeatMenuInfoPtr	featP;
	Boolean				check;

	fontIndex = FHIGetFontIDIndex (fontPalDataH, fontID, textface);
	if (fontIndex == -1)
		return;

	theMenu = (**fontPalDataH).featuresPopup;

	numMenuItems = CountMItems (theMenu);
	for (i = 1; i <= numMenuItems; i++)
		CheckItem (theMenu, i, false);

	featH = FP_FONT_DESC_P(fontPalDataH, fontIndex)->featH;
	HLock ((Handle) featH);
	featP = *featH;
	
	numOfFeatures = featP->numOfFeatures;
	
	for (i = 0; i < numOfFeatures; i++)
	{
		check = featP->features[i].selected ? true : false;
		CheckItem (theMenu, featP->features[i].menuItem, check);
	}
	
	HUnlock ((Handle) featH);
}


/*------------------------------------------------------- FHIMarkMultFeatPopup */
void FHIMarkMultFeatPopup (FHI_PalDataHan fontPalDataH, font fontID, Style textface)
{
	short				fontIndex, numMenuItems, numOfFeatures, i, j;
	MenuHandle			theMenu;
	FHI_FeatMenuInfoHan	featH;
	FHI_FeatMenuInfoPtr	featP;
	short				mutualCount;
	
	fontIndex = FHIGetFontIDIndex (fontPalDataH, fontID, textface);
	if (fontIndex == -1)
		return;

	theMenu = (**fontPalDataH).featuresPopup;

	numMenuItems = CountMItems (theMenu);
	for (i = 1; i <= numMenuItems; i++)
		CheckItem (theMenu, i, false);

	featH = FP_FONT_DESC_P(fontPalDataH, fontIndex)->featH;
	HLock ((Handle) featH);
	featP = *featH;
	
	numOfFeatures = featP->numOfFeatures;
	
	for (i = 0; i < numOfFeatures; i++) 
		if (featP->features[i].selected)
			if (featP->features[i].featureFlag & mutuallyExclusiveFeature)
			{
				mutualCount = 1;
				
				for ( j = i+1; j < numOfFeatures; j++ )
					if (featP->features[i].featureType == featP->features[j].featureType)
					{
						if (featP->features[j].featureFlag & mutuallyExclusiveFeature &&
							featP->features[j].selected)
						{
							mutualCount++;
							SetItemMark (theMenu, featP->features[j].menuItem, DashMark);
						}
					}
					else
						break;
				
				if ( mutualCount > 1 )
				{
					SetItemMark (theMenu, featP->features[i].menuItem, DashMark);
					i = j-1;	/* skips features which have been marked already */
				}
				else if ( mutualCount == 1 )
					CheckItem (theMenu, featP->features[i].menuItem, true);
			}
			else
				SetItemMark (theMenu, featP->features[i].menuItem, DashMark);
	
	HUnlock ((Handle) featH);
}

/*------------------------------------------------------- BuildFeaturePopup */
static void BuildFeaturePopup (FHI_PalDataHan fontPalDataH, font fontID, Style textface)
{	
	MenuHandle				theMenu;
	short					fontIndex;
	FHI_FeatMenuInfoHan		featH;
	FHI_FeatMenuInfoPtr		featP;
	short					numberOfFeatures, i;
	fontFeature				featureType;
	Str255					featureName;
	
	theMenu = (**fontPalDataH).featuresPopup;
	
	ClearMenu (theMenu, FirstFeaturesItem);

	fontIndex = FHIGetFontIDIndex (fontPalDataH, fontID, textface);
	if (fontIndex == -1)
		return;

	featH = FP_FONT_DESC_P(fontPalDataH, fontIndex)->featH;
	HLock ((Handle) featH);
	featP = *featH;
	
	numberOfFeatures = featP->numOfFeatures;

	featureType = -1;
	
	for ( i=0; i < numberOfFeatures; i++ )
	{
		/* draw the separator line - it gets drawn before each new feature type */
		if ( featureType > -1 && featureType != featP->features[i].featureType )
			AppendMenu (theMenu, "\p(-");

		/* get the real name from the font */
		FindFontPName ( fontID, featP->features[i].featureSetting.nameID, featureName );
		AppendMenu ( theMenu, featureName );
		featureType = featP->features[i].featureType;
	}
	
	HUnlock ((Handle) featH);

	if (CountMItems (theMenu) == 1)
		AppendMenu (theMenu, "\p(None in font");

	if ((**featH).numFeatGroups > 1)
		FHIMarkMultFeatPopup (fontPalDataH, fontID, textface);
	else
		FHIMarkFeatPopup (fontPalDataH, fontID, textface);
}

/*------------------------------------------------------- BuildPopups */
static void BuildPopups (FHI_PalDataHan fontPalDataH)
{
	font			fontID;
	Style			textface;
	Rect			bounds;
	Str255			familyName;
	short			curFontIndex = (**fontPalDataH).curFontIndex;
	
	BuildFontPopup (fontPalDataH);
	FHI_Pstrcpy (familyName, FP_FONT_DESC_P(fontPalDataH, curFontIndex)->familyName);
	BuildStylePopup (fontPalDataH, familyName);
	
	fontID = FP_FONT_DESC_P(fontPalDataH, curFontIndex)->fontID;
	textface = FP_FONT_DESC_P(fontPalDataH, curFontIndex)->textface;
	FHIBuildInstancePopup (fontPalDataH, fontID, textface);
	BuildFeaturePopup (fontPalDataH, fontID, textface);
}

/*----------------------------------------------------- UpdateTrackControl */
void UpdateTrackControl (FHI_PalDataHan fontPalDataH)
{
	GrafPtr			savePort;
	short			min, max, cur;
	short			curFontIndex = (**fontPalDataH).curFontIndex;
	ControlHandle	trackingSliderH = (**fontPalDataH).trackingSliderH;
	
	GetPort (&savePort);
	SetPort ((**fontPalDataH).palettePort);

	if ( FP_FONT_DESC_P(fontPalDataH, curFontIndex)->trackingH &&
		FP_TRACKING_P(fontPalDataH, curFontIndex)->numTracks)
	{
		min = FixedToInt (FloatToFixed (FP_TRACKING_P(fontPalDataH, curFontIndex)->minVal));
		max = FixedToInt (FloatToFixed (FP_TRACKING_P(fontPalDataH, curFontIndex)->maxVal));
		cur = FixedToInt (FloatToFixed (FP_TRACKING_P(fontPalDataH, curFontIndex)->curVal));

		SetCtlMin (trackingSliderH, 0);
		SetCtlMax (trackingSliderH, max - min);
		SetCtlValue (trackingSliderH, FHI_Abs(min));

		ShowControl (trackingSliderH);
	}
	else
		HideControl (trackingSliderH);

	SetPort (savePort);
}
 
/*------------------------------------------------------- RemoveRealMappings */
static void RemoveRealMappings (short *numMappings, FHI_FontMapPtr fontMappings)
{
	short			i;
	Size			blockSize;
	
	for (i = 0; i < *numMappings; i++)
		if (fontMappings[i].newFontID != fontMappings[i].oldFontID ||
			fontMappings[i].newTextface != fontMappings[i].oldTextface)
		{
			blockSize = (Size) ((*numMappings - i - 1) * sizeof (FHI_FontMapRec));
			BlockMove ((fontMappings + i+1), (fontMappings + i), blockSize);
			
			(*numMappings)--;
		}
}

/*------------------------------------------------------- GetSavedSelections */
static SelectionsHan GetSavedSelections (FHI_PalDataHan fontPalDataH,
										 short numMappings, FHI_FontMapPtr fontMappings)
{
	SelectionsHan					selectionsH = NULL;
	register SelectionsPtr			selectionsP;
	short							i, j, fontIndex;
	register FHI_InstMenuInfoPtr	fpInstanceP;
	register FHI_FeatMenuInfoPtr	fpFeatP;
	short							numInstances, numFeatures;
	short							numSelInstances, numSelFeatures;
	short							insertInd;

	/*	save instance and feature selections for fixed mappings
	 */
	if (numMappings)
	{
		selectionsH = (SelectionsHan) NewHandle (numMappings * sizeof (SelectionsRec));
		if (MemError () != noErr) goto errorReturn;
	}
	
	for (i = 0; i < numMappings; i++)
	{
		selectionsP = (SelectionsPtr) (*selectionsH + i);
		
		selectionsP->fontID = fontMappings[i].newFontID;
		selectionsP->textface = fontMappings[i].newTextface;
		selectionsP->selInstances = NULL;
		selectionsP->selFeatures = NULL;
		
		fontIndex = FHIGetFontIDIndex (fontPalDataH, selectionsP->fontID, selectionsP->textface);
		if (fontIndex != -1)
		{
			fpInstanceP = FP_INSTANCE_P(fontPalDataH, fontIndex);
			numInstances = fpInstanceP->numInstances;
			selectionsP->numSelInstances = numSelInstances = fpInstanceP->numSelInstances;
			
			fpFeatP = FP_FEAT_P(fontPalDataH, fontIndex);
			numFeatures = fpFeatP->numOfFeatures;
			numSelFeatures = 0;
			for (j = 0; j < numFeatures; j++)
				if (fpFeatP->features[j].selected)
					numSelFeatures++;
			selectionsP->numSelFeatures = numSelFeatures;
			
			(*selectionsH + i)->selInstances =
				(short *) NewPtr (numSelInstances * sizeof (short));
			if (MemError () != noErr) goto errorReturn;

			(*selectionsH + i)->selFeatures =
				(short *) NewPtr (numSelFeatures * sizeof (short));
			if (MemError () != noErr) goto errorReturn;
			
			fpInstanceP = FP_INSTANCE_P(fontPalDataH, fontIndex);
			fpFeatP = FP_FEAT_P(fontPalDataH, fontIndex);
			selectionsP = (SelectionsPtr) (*selectionsH + i);
			
			for (j = 0, insertInd = 0; j < numInstances && insertInd < numSelInstances; j++)
				if (fpInstanceP->instances[j].selected)
					selectionsP->selInstances[insertInd++] = j;

			for (j = 0, insertInd = 0; j < numFeatures && insertInd < numSelFeatures; j++)
				if (fpFeatP->features[j].selected)
					selectionsP->selFeatures[insertInd++] = j;
		}
		else
			goto errorReturn;
	}
	
	return (selectionsH);

errorReturn:
	if (selectionsH)
	{
		for (j = 0; j < i; j++)  /* i is the current selection's insertion index above */
		{
			if ((*selectionsH + j)->selInstances)
				DisposePtr ((Ptr) (*selectionsH + j)->selInstances);
			if ((*selectionsH + j)->selFeatures)
				DisposePtr ((Ptr) (*selectionsH + j)->selFeatures);
		}
		DisposeHandle ((Handle) selectionsH);
	}
	return (NULL);
}

/*------------------------------------------------------- SetSavedSelections */
static void SetSavedSelections (FHI_PalDataHan fontPalDataH,
								SelectionsHan selectionsH, short numFonts)
{
	register SelectionsPtr			selectionsP;
	short							i, j, fontIndex;
	register FHI_InstMenuInfoPtr	fpInstanceP;
	register FHI_FeatMenuInfoPtr	fpFeatP;
	short							numInstances, numFeatures;
	short							numSelInstances, numSelFeatures;

	if (!selectionsH) return;
	
	/*	set appropriate instances and features to be selected, based on saved data
	 */
	for (i = 0; i < numFonts; i++)
	{
		selectionsP = (SelectionsPtr) (*selectionsH + i);
		
		fontIndex = FHIGetFontIDIndex (fontPalDataH, selectionsP->fontID, selectionsP->textface);
		
		if (fontIndex != -1)
		{
			fpInstanceP = FP_INSTANCE_P(fontPalDataH, fontIndex);
			numInstances = fpInstanceP->numInstances;
			numSelInstances = selectionsP->numSelInstances;
			
			for (j = 0; j < numInstances; j++)
				fpInstanceP->instances[j].selected = false;
			fpInstanceP->numSelInstances = 0;
			
			for (j = 0; j < numSelInstances; j++)
				fpInstanceP->instances[selectionsP->selInstances[j]].selected = true;
			fpInstanceP->numSelInstances = numSelInstances;
	
			fpFeatP = FP_FEAT_P(fontPalDataH, fontIndex);
			numFeatures = fpFeatP->numOfFeatures;
			numSelFeatures = selectionsP->numSelFeatures;
			
			for (j = 0; j < numFeatures; j++)
				fpFeatP->features[j].selected = false;
			
			for (j = 0; j < numSelFeatures; j++)
				fpFeatP->features[selectionsP->selFeatures[j]].selected = true;
		}
	}
}

/*------------------------------------------------------- FHIPaletteFontChange */
/*
 *	always call after any kind of font change or text selection
 *
 *	invoke with mappings after calling FontHIFamilyMenuSetItem or FontHIStyleMenuSetItem,
 *	otherwise pass numMappings = 0, fontIDs = NULL
 */
short FHIPaletteFontChange (FHI_PalDataHan fontPalDataH, FHI_MenuDataHan fontMenuDataH,
							short numMappings, FHI_FontMapPtr fontMappings)
{
	short					err = noErr;
	GrafPtr					savePort;
	short					numFixedMappings = numMappings;
	FHI_FontMapPtr			fixedMappings = NULL;
	short					i;
	SelectionsHan			selectionsH = NULL;
	
	GetPort (&savePort);
	SetPort ((**fontPalDataH).palettePort);
	
	if (numMappings && fontMappings)
	{
		fixedMappings = (FHI_FontMapPtr) NewPtr (numFixedMappings * sizeof (FHI_FontMapRec));
		if ((err = MemError ()) != noErr) return (err);
		BlockMove (fontMappings, fixedMappings, GetPtrSize ((Ptr) fontMappings));
	
		RemoveRealMappings (&numFixedMappings, fixedMappings);
		
		selectionsH = GetSavedSelections (fontPalDataH, numFixedMappings, fixedMappings);
		
		/*	update font palette data with newly selected font(s)
		 */
		err = FHIUpdatePaletteFontInfo (fontPalDataH, fontMenuDataH);
		
		if (selectionsH)
			SetSavedSelections (fontPalDataH, selectionsH, numFixedMappings);
	}
	else
		err = FHIUpdatePaletteFontInfo (fontPalDataH, fontMenuDataH);
	
	if (!err)
	{
		BuildPopups (fontPalDataH);
		UpdateTrackControl (fontPalDataH);
	}
		
	if (selectionsH)
	{
		for (i = 0; i < numFixedMappings; i++)
		{
			if ((*selectionsH + i)->selInstances)
				DisposePtr ((Ptr) (*selectionsH + i)->selInstances);
			if ((*selectionsH + i)->selFeatures)
				DisposePtr ((Ptr) (*selectionsH + i)->selFeatures);
		}
		DisposeHandle ((Handle) selectionsH);
	}

	SetPort (savePort);
	
	return (err);
}

/*----------*/
FHI_PalDataHan FHIOpenPalette	(FHI_MenuDataHan fontMenuDataH)
{
	WindowPtr				newWindow;
	Rect					bounds;
	FHI_PalDataHan			fontPalDataH = NULL;
	register FHI_PalDataPtr	fontPalDataP;

	newWindow = FHIGetWindow (FontPaletteWindID);

	fontPalDataH = FHIInitPaletteData ();
	
	if (fontPalDataH)
	{
		HLock ((Handle) fontPalDataH);
		fontPalDataP = (FHI_PalDataPtr) (*fontPalDataH);
	
		fontPalDataP->palettePort = newWindow;
	
		fontPalDataP->paletteWitlH = GetResource ('Witl', FontPaletteWindID);
		fontPalDataP->paletteWictH = GetResource ('Wict', FontPaletteWindID);
	
		if (!fontPalDataP->selFontPopup)
			fontPalDataP->selFontPopup	= GetMenu (SelFontPopupID);
		if (!fontPalDataP->selStylePopup)
			fontPalDataP->selStylePopup	= GetMenu (SelStylePopupID);
		if (!fontPalDataP->styleVarPopup)
			fontPalDataP->styleVarPopup	= GetMenu (StyleVarPopupID);
		if (!fontPalDataP->featuresPopup)
			fontPalDataP->featuresPopup	= GetMenu (FeaturesPopupID);

		fontPalDataP->trackingSliderH = GetNewControl (SliderCntlID, newWindow);
	
		FHIUpdatePaletteFontInfo (fontPalDataH, fontMenuDataH);
		BuildPopups (fontPalDataH);
		UpdateTrackControl (fontPalDataH);
	
		fontPalDataP->paletteVisible = true;
		
		HUnlock ((Handle) fontPalDataH);
	}

/*	GhostWindow = newWindow; */

	return (fontPalDataH);

} /*FHIOpenPalette*/

/*------------------------------------------------------- FHIShowHidePalette */
void FHIShowHidePalette (FHI_PalDataHan fontPalDataH, Boolean show)
{
	if (show)
	{
		(**fontPalDataH).paletteVisible = true;
				
		ShowHide ((**fontPalDataH).palettePort, true);
	}
	else
	{
		(**fontPalDataH).paletteVisible = false;

		ShowHide ((**fontPalDataH).palettePort, false);
	}
}

/*----------*/
void FHIDisposePalette	(FHI_PalDataHan fontPalDataH)
{
	DisposeWindow ((**fontPalDataH).palettePort);
	
	if (fontPalDataH)
		FHIDisposePaletteData (fontPalDataH);

} /*FHIDisposePalette*/

/*----------*/
FHI_Changes FHIDoPaletteControl	(FHI_PalDataHan	fontPalDataH,
								 ControlHandle	whichControl, 
								 short			whichPart, 
						 		 Point			where)
{
	GrafPtr			savePort;
	FHI_Changes		paletteChange = FHI_NoChange;
	float			trVal;
	ControlHandle	trackingSliderH = (**fontPalDataH).trackingSliderH;
	short			curFontIndex = (**fontPalDataH).curFontIndex;
	
	GetPort (&savePort);
	SetPort ((**fontPalDataH).palettePort);

	if (whichControl == trackingSliderH)
	{
		FHITrackScroll (trackingSliderH, whichPart, where, nil);
		
		trVal = (float) GetCtlValue ( trackingSliderH );
	
		FP_TRACKING_P(fontPalDataH, curFontIndex)->curVal =
			FP_TRACKING_P(fontPalDataH, curFontIndex)->minVal + trVal;
		
		paletteChange = FHI_TrackingChange;
	}
	
	SetPort (savePort);
	
	return (paletteChange);
		
} /*FHIDoPaletteControl*/

/*------------------------------------------------------- DoClickInFontPopup */
static FHI_Changes DoClickInFontPopup (FHI_PalDataHan fontPalDataH, Rect *bounds)
{
	FHI_Changes				paletteChange = FHI_NoChange;
	register FHI_PalDataPtr	fontPalDataP;
	short					oldChoice, newChoice, newFontIndex;
	Str255					familyName, styleName;
	font					fontID;
	Style					textface;

	HLock ((Handle) fontPalDataH);
	fontPalDataP = (FHI_PalDataPtr) (*fontPalDataH);

	newChoice = oldChoice = fontPalDataP->selFontChoice;
	FHITrackPopup (*bounds, SelFontPopupID, &newChoice);

	if (newChoice != oldChoice)
	{
		fontPalDataP->selFontChoice = newChoice;

		GetItem (fontPalDataP->selFontPopup, newChoice, familyName);
		BuildStylePopup (fontPalDataH, familyName);
		GetItem (fontPalDataP->selStylePopup, 1, styleName);
		newFontIndex = FHIGetFontNameIndex (fontPalDataH, familyName, styleName);
		fontPalDataP->curFontIndex = newFontIndex;
		
		fontID = FP_FONT_DESC_P(fontPalDataH, newFontIndex)->fontID;
		textface = FP_FONT_DESC_P(fontPalDataH, newFontIndex)->textface;
		FHIBuildInstancePopup (fontPalDataH, fontID, textface);
		BuildFeaturePopup (fontPalDataH, fontID, textface);
		
		UpdateTrackControl (fontPalDataH);
		
		paletteChange = FHI_FontChange;
	}
	
	HUnlock ((Handle) fontPalDataH);
	
	return (paletteChange);
}

/*------------------------------------------------------- DoClickInStylePopup */
static FHI_Changes DoClickInStylePopup (FHI_PalDataHan fontPalDataH, Rect *bounds)
{
	FHI_Changes				paletteChange = FHI_NoChange;
	register FHI_PalDataPtr	fontPalDataP;
	short					oldChoice, newChoice, newFontIndex;
	Str255					familyName, styleName;
	font					fontID;
	Style					textface;

	HLock ((Handle) fontPalDataH);
	fontPalDataP = (FHI_PalDataPtr) (*fontPalDataH);

	newChoice = oldChoice = fontPalDataP->selStyleChoice;
	FHITrackPopup (*bounds, SelStylePopupID, &newChoice);
	
	if (newChoice != oldChoice)
	{
		fontPalDataP->selStyleChoice = newChoice;

		GetItem (fontPalDataP->selFontPopup, fontPalDataP->selFontChoice, familyName);
		GetItem (fontPalDataP->selStylePopup, newChoice, styleName);
		newFontIndex = FHIGetFontNameIndex (fontPalDataH, familyName, styleName);
		fontPalDataP->curFontIndex = newFontIndex;
					
		fontID = FP_FONT_DESC_P(fontPalDataH, newFontIndex)->fontID;
		textface = FP_FONT_DESC_P(fontPalDataH, newFontIndex)->textface;
		FHIBuildInstancePopup (fontPalDataH, fontID, textface);
		BuildFeaturePopup (fontPalDataH, fontID, textface);

		UpdateTrackControl (fontPalDataH);
		
		paletteChange = FHI_FontChange;
	}

	HUnlock ((Handle) fontPalDataH);
	
	return (paletteChange);
}
	
/*------------------------------------------------------- DoClickInInstancePopup */
static FHI_Changes DoClickInInstancePopup (FHI_PalDataHan fontPalDataH, Rect *bounds)
{
	FHI_Changes						paletteChange = FHI_NoChange;
	short							oldChoice, newChoice;
	short							curFontIndex = (**fontPalDataH).curFontIndex, i;
	Point							mouseLoc;
	register FHI_InstMenuInfoPtr	fpInstanceP;

	newChoice = oldChoice = FHIGetFirstMarkedInstance (fontPalDataH);
	if (FP_FONT_DESC_P(fontPalDataH, curFontIndex)->instanceH &&
		FP_INSTANCE_P(fontPalDataH, curFontIndex)->numSelInstances > 1)
		FHITrackMultSelPopup (*bounds, StyleVarPopupID, &newChoice);
	else
		FHITrackPopup (*bounds, StyleVarPopupID, &newChoice);

	GetMouse (&mouseLoc);
	
	if (newChoice != 0 &&
		(newChoice != oldChoice ||
		(FP_INSTANCE_P(fontPalDataH, curFontIndex)->numSelInstances > 1 && PtInRect (mouseLoc, bounds))))
	{
		fpInstanceP = FP_INSTANCE_P(fontPalDataH, curFontIndex);
		for (i = 0; i < fpInstanceP->numInstances; i++)
		{
			if (fpInstanceP->instances[i].menuItem == newChoice)
				fpInstanceP->instances[i].selected = true;
			else
			{
				CheckItem ((**fontPalDataH).styleVarPopup, fpInstanceP->instances[i].menuItem, false);
				fpInstanceP->instances[i].selected = false;
			}
		}
		fpInstanceP->numSelInstances = 1;

		paletteChange = FHI_InstanceChange;
	}

	return (paletteChange);
}

/*------------------------------------------------------- DoClickInFeaturePopup */
static FHI_Changes DoClickInFeaturePopup (FHI_PalDataHan fontPalDataH, Rect *bounds,
											fontRunFeature *feature, Boolean *mutExcl)
{
	FHI_Changes			paletteChange = FHI_NoChange;
	short				newChoice = 0, numOfFeatures, i;
	FHI_FeatMenuInfoHan	featH;
	FHI_FeatMenuInfoPtr	featP;
	fontFeature			featureType;
	font				fontID;
	short				theMark = 0;
	
	DelMenuItem ((**fontPalDataH).featuresPopup, 1);
	FHITrackMultSelPopRight (*bounds, FeaturesPopupID, &newChoice);
	InsMenuItem ((**fontPalDataH).featuresPopup, FirstFeaturesStr, 0);
		
	if (newChoice != 0)
	{
		newChoice++;  /* account for first item in menu, "Features", which is unselectable */
	
		feature->featureType = -1;
		feature->setting = -1;
		*mutExcl = false;
		
		featH = FP_FONT_DESC_P(fontPalDataH, (**fontPalDataH).curFontIndex)->featH;
		HLock ((Handle) featH);
		featP = *featH;
		
		numOfFeatures = featP->numOfFeatures;
	
		for (i = 0; i < numOfFeatures; i++)
		{	
			if ( featP->features[i].menuItem == newChoice )
			{	
				featureType = featP->features[i].featureType;
	
				feature->featureType = featureType;
				feature->setting = featP->features[i].featureSetting.setting;
				
				/* for mutually exclusive stuff */
				if ( featP->features[i].featureFlag & mutuallyExclusiveFeature )
				{
					featP->features[i].selected = true;
					*mutExcl = true;
					break;
				}
				
				/* for non-mutually-exclusive stuff */
				if ( featP->features[i].selected ) {
					
					GetItemMark ((**fontPalDataH).featuresPopup, newChoice, &theMark);
						
					if (theMark != DashMark)
					{
						featP->features[i].selected = false;
						feature->setting++;
					}
				} else
					featP->features[i].selected = true;
				break;
			}
		}
		
		/*	in the case of mutually exclusive stuff deselect
			everything else in the same feature type
		 */
		for (i = 0; i < numOfFeatures; i++)
			if (featP->features[i].featureFlag & mutuallyExclusiveFeature)
				if (featP->features[i].featureType == featureType &&
					featP->features[i].menuItem != newChoice) {
					featP->features[i].selected = false;
				}
		
		fontID = FHIGetPaletteFontID ( fontPalDataH );
		
		if (featP->numFeatGroups > 1)
		{
			FHIMarkMultFeatPopup (fontPalDataH, fontID, 0);
			
			/* dash turns into check when chosen */
			if (theMark == DashMark)
				CheckItem ((**fontPalDataH).featuresPopup, newChoice, true);
		}
		else
			FHIMarkFeatPopup (fontPalDataH, fontID, 0);
		
		HUnlock ((Handle) featH);

		paletteChange = FHI_FeatureChange;
	}
	
	return (paletteChange);
}

/*----------*/
FHI_Changes FHIDoPaletteClick	(FHI_PalDataHan	fontPalDataH,
								 Point			where,
								 short			modifiers,
								 fontRunFeature	*newFeature,
								 Boolean		*newFeatExcl)
{
	FHI_Changes				paletteChange = FHI_NoChange;
	register FHI_PalDataPtr	fontPalDataP;
	GrafPtr					savePort;
	Rect					popupBounds, labelBounds;
	
	newFeature->featureType = -1;
	newFeature->setting = -1;
	*newFeatExcl = false;

	HLock ((Handle) fontPalDataH);
	fontPalDataP = (FHI_PalDataPtr) (*fontPalDataH);

	GetPort (&savePort);
	SetPort (fontPalDataP->palettePort);
	
	FHIGetWRect (fontPalDataP->paletteWitlH, SelFontPopupItem, &popupBounds);
	if (PtInRect (where, &popupBounds)) {
		paletteChange = DoClickInFontPopup (fontPalDataH, &popupBounds);
	}
	
	FHIGetWRect (fontPalDataP->paletteWitlH, SelStylePopupItem, &popupBounds);
	if (PtInRect (where, &popupBounds)) {
		paletteChange = DoClickInStylePopup (fontPalDataH, &popupBounds);
	}
	
	FHIGetWRect (fontPalDataP->paletteWitlH, StyleVarPopupItem, &popupBounds);
	FHIGetWRect (fontPalDataP->paletteWitlH, VariationLabel, &labelBounds);
	if (PtInRect (where, &popupBounds) || PtInRect (where, &labelBounds)) {
		InvertRect (&labelBounds);
		paletteChange = DoClickInInstancePopup (fontPalDataH, &popupBounds);
		InvertRect (&labelBounds);
	}
	
	FHIGetWRect (fontPalDataP->paletteWitlH, FeaturesPopupItem, &popupBounds);
	popupBounds.right -= FeaturesPopupInset;
	FHIGetWRect (fontPalDataP->paletteWitlH, FeaturesLabel, &labelBounds);
	if (PtInRect (where, &popupBounds) || PtInRect (where, &labelBounds)) {
		InvertRect (&labelBounds);
		paletteChange = DoClickInFeaturePopup (fontPalDataH, &popupBounds, newFeature, newFeatExcl);
		InvertRect (&labelBounds);
	}

	SetPort (savePort);

	HUnlock ((Handle) fontPalDataH);

	return (paletteChange);
	
} /*FHIDoPaletteClick*/

/*----------*/
void FHIUpdatePalette (FHI_PalDataHan fontPalDataH)
{
	GrafPtr					savePort;
	Handle					paletteWictH, paletteWitlH;
	Rect					bounds;
	
	GetPort (&savePort);
	SetPort ((**fontPalDataH).palettePort);
	
	paletteWictH = (**fontPalDataH).paletteWictH;
	paletteWitlH = (**fontPalDataH).paletteWitlH;
	
	FHISetWFont (paletteWictH, SelectedTextLabel);
	FHIGetWRect (paletteWitlH, SelectedTextLabel, &bounds);
	FHITextIDBox (SelectedTextTextID, bounds);
	
	FHIGetWRect (paletteWitlH, Line2, &bounds);
	FHIDrawGrayLine (bounds);
	
	FHISetWFont (paletteWictH, LabelFontItem);
	FHIGetWRect (paletteWitlH, SelFontPopupItem, &bounds);
	FHIUpdatePopup (bounds, SelFontPopupID, (**fontPalDataH).selFontChoice);
	
	FHISetWFont (paletteWictH, LabelFontItem);
	FHIGetWRect (paletteWitlH, SelStylePopupItem, &bounds);
	FHIUpdatePopup (bounds, SelStylePopupID, (**fontPalDataH).selStyleChoice);
	
	FHIGetWRect (paletteWitlH, Line5, &bounds);
	FHIDrawGrayLine (bounds);
	
	FHISetWFont (paletteWictH, VariationLabel);
	FHIGetWRect (paletteWitlH, VariationLabel, &bounds);
	FHITextIDBox (VariationTextID, bounds);
	FHIGetWRect (paletteWitlH, StyleVarPopupItem, &bounds);
	FHIUpdatePopup (bounds, StyleVarPopupID, FHIGetFirstMarkedInstance (fontPalDataH));
	FHIGetWRect (paletteWitlH, StyleVarPict, &bounds);
	FHIDrawPictID (StyleVarPictID, bounds);
	
	FHISetWFont (paletteWictH, FeaturesLabel);
	FHIGetWRect (paletteWitlH, FeaturesLabel, &bounds);
	FHITextIDBox (FeaturesTextID, bounds);
	FHIGetWRect (paletteWitlH, FeaturesPopupItem, &bounds);
	bounds.right -= FeaturesPopupInset;
	FHIUpdatePopRight (bounds, FeaturesPopupID, 1);
	FHIGetWRect (paletteWitlH, DesignPict, &bounds);
	FHIDrawPictID (DesignPictID, bounds);

	FHISetWFont (paletteWictH, TightLabel);
	FHIGetWRect (paletteWitlH, TightLabel, &bounds);
	FHITextIDBox (TightTextID, bounds);
	FHISetWFont (paletteWictH, NormalLabel);
	FHIGetWRect (paletteWitlH, NormalLabel, &bounds);
	FHITextIDBox (NormalTextID, bounds);
	FHISetWFont (paletteWictH, LooseLabel);
	FHIGetWRect (paletteWitlH, LooseLabel, &bounds);
	FHITextIDBox (LooseTextID, bounds);
	FHIGetWRect (paletteWitlH, Rect14, &bounds);
	FrameRect (&bounds);
	FHISetWFont (paletteWictH, TrackingLabel);
	FHIGetWRect (paletteWitlH, TrackingLabel, &bounds);
	FHITextIDBox (TrackingTextID, bounds);
	FHIGetWRect (paletteWitlH, TrackingPict, &bounds);
	FHIDrawPictID (TrackingPictID, bounds);
	
	if ((**fontPalDataH).trackingSliderH)
		Draw1Control ((**fontPalDataH).trackingSliderH);

	HLock ((Handle) fontPalDataH);
	ValidRect (&(**fontPalDataH).palettePort->portRect);
	HUnlock ((Handle) fontPalDataH);
	
	SetPort (savePort);

} /*FHIUpdatePalette*/

/*----------*/
void FHIActivatePalette	(FHI_PalDataHan fontPalDataH, Boolean activate)
{
	GrafPtr					savePort;

	GetPort (&savePort);
	SetPort ((**fontPalDataH).palettePort);

	FHIEnableControl ((**fontPalDataH).trackingSliderH, activate);

	SetPort (savePort);

} /*FHIActivatePalette*/
