/*
	File:		FontHIValidation.c

	Contains:	validation routines for FontHI library

	Written by:	Dennis Buchheim

	Copyright:	 1993 by Apple Computer, Inc., All rights reserved.

	Change History (most recent first):


*/

#include <Traps.h>
#include <OSUtils.h>

#include "FontHIUtils.h"
#include "FontHIMenuData.h"
#include "FontHIPaletteData.h"
#include "FontHIValidation.h"


#define BoldItalicTextface	3


FHI_Errors FHIValidateMenuData (void *dataH)
{
	FHI_Errors					err = FontHI_NoErr;

#if FontHIMenuValidationOn

	FHI_MenuDataHan				fontMenuDataH;
	register FHI_MenuDataPtr	fontMenuDataP;
	short						i;
	register FHI_StylePtr		styleP;

	fontMenuDataH = (FHI_MenuDataHan) dataH;
	
	if (!fontMenuDataH)
		return (FontHI_NullStruct);
	
	if ((**fontMenuDataH).magicNum != FHI_MenuMagic)
		return (FontHI_BadMagic);
	
	HLock ((Handle) fontMenuDataH);
	fontMenuDataP = (FHI_MenuDataPtr) (*fontMenuDataH);
	
	for (i = 0; i < fontMenuDataP->numFamilies; i++)
	{
		HLock ((Handle) fontMenuDataP->families[i].styleH);	
		styleP = (FHI_StylePtr) (*(**fontMenuDataH).families[i].styleH);
		
		if (fontMenuDataP->numSelectedFonts < 0)
			err = FontHI_BadFields;
		else if (fontMenuDataP->totNumActualStyles < 0)
			err = FontHI_BadFields;
		else if (fontMenuDataP->numFamilies < 0)
			err = FontHI_BadFields;
		else if (fontMenuDataP->selFamilyMenuItem < 0 ||
				 fontMenuDataP->selFamilyMenuItem > fontMenuDataP->numFamilies)
			err = FontHI_BadFields;
		else if (fontMenuDataP->selStyleMenuItem < 0 ||
				 fontMenuDataP->selStyleMenuItem > fontMenuDataP->totNumActualStyles)
			err = FontHI_BadFields;
		else if (!FHIPstrcmp ((char *) fontMenuDataP->families[i].name, (char *) "\p"))
			err = FontHI_BadFields;
		else if (fontMenuDataP->families[i].familyMenuItem < 0 ||
				 fontMenuDataP->families[i].familyMenuItem > fontMenuDataP->numFamilies)
			err = FontHI_BadFields;
		else if (fontMenuDataP->families[i].numStyles < 0)
			err = FontHI_BadFields;
		else if (fontMenuDataP->families[i].defaultStyleIndex < 0 ||
				 fontMenuDataP->families[i].defaultStyleIndex >= fontMenuDataP->families[i].numStyles)
			err = FontHI_BadFields;
		else if (!fontMenuDataP->families[i].styleH)
			err = FontHI_BadFields;
		else if (!FHIPstrcmp ((char *) styleP->name, (char *) "\p"))
			err = FontHI_BadFields;
		else if (styleP->styleMenuItem < 0 || styleP->styleMenuItem > fontMenuDataP->totNumActualStyles)
			err = FontHI_BadFields;
		else if (styleP->fontID <= (font) 0 ||
				 styleP->textface < -1 || styleP->textface > BoldItalicTextface)
			err = FontHI_BadFields;
		else if (styleP->selected < 0 || styleP->selected > 1)
			err = FontHI_BadFields;
		
		HUnlock ((Handle) fontMenuDataP->families[i].styleH);
		
		if (err != FontHI_NoErr)
			break;
	}
	
	HUnlock ((Handle) fontMenuDataH);

#endif
	
	return (err);
}


FHI_Errors FHIValidatePaletteData (void *dataH)
{

#if FontHIPaletteValidationOn
	
	FHI_PalDataHan			fontPalDataH;
	short					numFontMItems, numStyleMItems;
	register FHI_PalDataPtr	fontPalDataP;

	fontPalDataH = (FHI_PalDataHan) dataH;

	if (!fontPalDataH)
		return (FontHI_NullStruct);
	
	numFontMItems = CountMItems ((**fontPalDataH).selFontPopup);
	numStyleMItems = CountMItems ((**fontPalDataH).selStylePopup);
	
	fontPalDataP = (FHI_PalDataPtr) (*fontPalDataH);
	
	if (fontPalDataP->magicNum != FHI_PaletteMagic)
		return (FontHI_BadMagic);
	
	if (!fontPalDataP->palettePort)
		return (FontHI_BadFields);
	if (!fontPalDataP->paletteWictH || !fontPalDataP->paletteWitlH)
		return (FontHI_BadFields);
	if (!fontPalDataP->selFontPopup || !fontPalDataP->selStylePopup ||
		!fontPalDataP->styleVarPopup || !fontPalDataP->featuresPopup)
		return (FontHI_BadFields);
	if (!fontPalDataP->trackingSliderH)
		return (FontHI_BadFields);
	if (!fontPalDataP->fontDescH)
		return (FontHI_BadFields);
	if (fontPalDataP->paletteVisible < 0 || fontPalDataP->paletteVisible > 1)
		return (FontHI_BadFields);
	if (fontPalDataP->selFontChoice < 1 || fontPalDataP->selFontChoice > numFontMItems)
		return (FontHI_BadFields);
	if (fontPalDataP->selStyleChoice < 1 || fontPalDataP->selStyleChoice > numStyleMItems)
		return (FontHI_BadFields);
	if (fontPalDataP->numFontsInPalette < 0)
		return (FontHI_BadFields);
	if (fontPalDataP->curFontIndex < 0 ||
		fontPalDataP->curFontIndex >= fontPalDataP->numFontsInPalette)
		return (FontHI_BadFields);
	
#endif

	return (FontHI_NoErr);
}


void FHIPostError (FHI_Errors err)
{
	long	trapAddr;
	long	unimplementedAddr;
	
	/* check to see if DebugStr() trap is installed */
	
	trapAddr = GetOSTrapAddress(_DebugStr);
	unimplementedAddr = GetOSTrapAddress(_Unimplemented);

	if(trapAddr != unimplementedAddr)
	{
		switch (err)
		{
			case FontHI_BadMagic:
				DebugStr ("\pFont HI Error: Bad magic number");
				break;
			case FontHI_NullStruct:
				DebugStr ("\pFont HI Error: Null data struct");
				break;
			case FontHI_BadFields:
				DebugStr ("\pFont HI Error: Invalid data struct");
				break;
			case FontHI_OtherArgsBad:
				DebugStr ("\pFont HI Error: Bad arguments passed");
				break;
		}
	}
}
