/*
	File:			font library additions.c
	
	Contains:
	
	Written by:		Barton R. House
	
	Copyright:		 1993 by Apple Computer, Inc., All rights reserved.
	
*/

#include <Types.h>
#include <Memory.h>
#include <OSUtils.h>

#include "font routines.h"
#include "font library additions.h"
#include "layout feature constants.h"

#define		mortTableTag	0x6d6f7274	/* 'mort' */

typedef	unsigned short	uint16;
typedef unsigned long	uint32;

typedef struct {
	uint16		featureType;			/* The type of feature. */
	uint16		featureSetting;			/* The feature's setting. */
} TextFeature;

typedef struct {
	uint16		length;					/* Length of subtable, including this
										   header, in bytes. */
	uint16		coverage;				/* Coverage flags and atomic transformation
										   type. */
	uint32		subFeatureFlags;		/* Flags for sub-features this subtable
										   describes. */
										   
} SubtableHdr, *SubtableHdrPtr, **SubtableHdrHan;

typedef struct {
	TextFeature	feature;				/* The feature type and setting. */
	uint32		enableFlags;			/* Flags for sub-features this feature
										   and setting enables. */
	uint32		disableFlags;			/* Complement of flags for sub-features
										   this feature and setting disables. */
										   
} FeatureTableEntry, *FeatureTableEntryP, **FeatureTableEntryHan;

typedef struct {
	uint16		length;					/* Length of subtable, including this
										   header, in bytes. */
	uint16		coverage;				/* Coverage flags and atomic transformation
										   type. */
	uint32		subFeatureFlags;		/* Flags for sub-features this subtable
										   describes. */
										   
} SubtableHeader, *SubtableHeaderP, **SubtableHeaderHan;

typedef struct {
	uint32		defaultFlags;			/* The default sub-feature flags for this
										   chain. */
	uint32		chainLength;			/* The length of the chain in bytes, including
										   this header. */
	uint16		nFeatureEntries;		/* The number of entries in the chain's
										   feature table. */
	uint16		nSubtables;				/* The number of subtables in the chain. */
	
} ChainHeader, *ChainHeaderP, **ChainHeaderHan;

typedef struct {
	Fixed		version;				/* Version number of the glyph metamorphosis
										   table (0x00010000 for the initial version. */
	uint32		nChains;				/* Number of metamorphosis chains which
										   follow. */
} MortHeader, *MortHeaderP, **MortHeaderHan;

static void GetDefaultFeatures(font fontID, long mortIndex,
									short * numFeatures, fontRunFeature ** features);
static void GetDefaultChainFeatures(ChainHeaderP chainHdr, FeatureTableEntryP chainEntries,
										short * numFeatures, fontRunFeature ** features);
static void AddFeatureToList(short * numFeatures, fontRunFeature ** featureList,
																fontRunFeature * feature);
static void RemoveFeatureFromList(short * numFeatures, fontRunFeature ** featureList,
																fontRunFeature * feature);
static fontRunFeature * FindFeatureInList(short numFeatures, fontRunFeature * featureList,
																fontRunFeature * feature);

void GetFontDefaultFeatures(font fontID, short * numFeatures, fontRunFeature *** features)
{
	long				mortSize;
	long				mortIndex;
	Ptr					mortData;
	
	*numFeatures = 0;
	*features = (fontRunFeature **) NewHandle(0);
	
	if(MemError() != noErr) {
		*features = nil;
		return;
	}
	
	mortSize = FindFontTable(fontID, mortTableTag, nil, &mortIndex);
	
	if(mortSize == 0)
		return;
	
	GetDefaultFeatures(fontID, mortIndex, numFeatures, *features);
	
}

static void GetDefaultFeatures(font fontID, long mortIndex, short * numFeatures, fontRunFeature ** features)
{
	MortHeader				mortHdr;
	ChainHeader				chainHdr;
	short					i;
	long					mortOffset;
	FeatureTableEntryP		chainEntries;
	long					size;
	long					chainIndex;
	fontRunFeature			enableFeature;
	
	mortOffset = 0;
	
	size = sizeof(MortHeader);
	GetFontTableParts(fontID, mortIndex, mortOffset, size, &mortHdr, nil);
	mortOffset += size;
	
	size = sizeof(ChainHeader);
	GetFontTableParts(fontID, mortIndex, mortOffset, size, &chainHdr, nil);
	mortOffset += size;
	
	chainIndex = 0;
	
	while(chainIndex++ < mortHdr.nChains) {
	
		if(chainHdr.nFeatureEntries) {
		
			size = sizeof(FeatureTableEntry) * chainHdr.nFeatureEntries;
			chainEntries = (FeatureTableEntryP) NewPtr(size);
			GetFontTableParts(fontID, mortIndex, mortOffset, size, chainEntries, nil);
			
			GetDefaultChainFeatures(&chainHdr, chainEntries, numFeatures, features);
		
			DisposePtr((Ptr) chainEntries);

		}
		
		mortOffset += chainHdr.chainLength;
	}
	
	if(*numFeatures) {
	
		enableFeature.featureType = typographicFeaturesType;
		enableFeature.setting = enableEffectsOnSelector;
	
		AddFeatureToList(numFeatures, features, &enableFeature);
	}
}

static void GetDefaultChainFeatures(ChainHeaderP chainHdr, FeatureTableEntryP chainEntries,
											short * numFeatures, fontRunFeature ** features)
{
	uint32					defaultFlags;
	FeatureTableEntryP		featureEntry;
	short					i;
	
	defaultFlags = chainHdr->defaultFlags;
	
	featureEntry = chainEntries;
	
	for(i=0; i<chainHdr->nFeatureEntries; i++, featureEntry++)
		if((featureEntry->enableFlags & defaultFlags) == featureEntry->enableFlags)
			AddFeatureToList(numFeatures, features, (fontRunFeature *) &featureEntry->feature);
		else
			RemoveFeatureFromList(numFeatures, features, (fontRunFeature *) &featureEntry->feature);
	
}

static void AddFeatureToList(short * numFeatures, fontRunFeature ** featureList, fontRunFeature * feature)
{
	if(FindFeatureInList(*numFeatures, *featureList, feature) != nil)
		return;
		
	PtrAndHand((Ptr) feature, (Handle) featureList, sizeof(fontRunFeature));
	
	if(MemError() == noErr)
		(*numFeatures)++;
			
}

static void RemoveFeatureFromList(short * numFeatures, fontRunFeature ** featureList, fontRunFeature * feature)
{
	fontRunFeature		* fp;
	
	if((fp = FindFeatureInList(*numFeatures, *featureList, feature)) == nil)
		return;
		
	BlockMove((Ptr) (fp+1), (Ptr) fp, (*featureList + *numFeatures) - fp);
	(*numFeatures)--;
	SetHandleSize((Handle) featureList, *numFeatures * sizeof(fontRunFeature));
	
}

static fontRunFeature * FindFeatureInList(short numFeatures, fontRunFeature * featureList, fontRunFeature * feature)
{
	short				i;
	fontRunFeature		* fp;
	
	fp = featureList;
	
	for(i=0; i<numFeatures; i++, fp++)
		if(fp->featureType == feature->featureType && fp->setting == feature->setting)
			return(fp);
			
	return(nil);
}

