/*
	File:			GXEditDoc.c
	
	Contains:
	
	Written by:	Barton R. House
	
	Copyright:	 1993 by Apple Computer, Inc., All rights reserved.
	
*/

/* GX includes */

#include "PrintingManager.h"
#include "graphics routines.h"
#include "graphics libraries.h"
#include "font routines.h"
#include "math routines.h"
#include "layout routines.h"

#include "GXEdit.h"
#include "GXEditDebug.h"
#include "GXEditDoc.h"
#include "GXEditParagraph.h"
#include "GXEditStyle.h"
#include "GXEditSelection.h"
#include "GXEditError.h"
#include "GXEditUtils.h"
#include "GXEditScrap.h"

unsigned char *		gDefaultFonts[] = {
						"\pGeneva",
						"\pTimes",
						"\pMonaco"
				};

static void		CalcStartsAndOffsets(DocPtr dp, short paragraphIndex);

static void		InsertNewParagraph(DocPtr dp, short paragraphIndex, short paragraphOffset);

static void		CalcPrintPages(DocPtr dp, Job printJob, short * numPages, long *** pageStartsPtr);
static shape 	GetAPage(DocPtr dp, long docOffset);

static void		DisposeScrap(ScrapPtr scrap);
static void 	InitRunControl(RunControls * runControl);

static void		BlinkCursor(DocPtr dp);
static Boolean	CheckReflow(DocPtr dp);

static void 	Insert(DocPtr dp, void * text, short startNumText, short textSize);
static void 	Clear(DocPtr dp);

static void		CalcPositionAndDraw(DocPtr dp, short startIndex);
static void		Draw(DocPtr dp, long start, long end);
static void		Position(DocPtr dp, long docOffset, Rect * rect);

static void		HitTest(DocPtr dp, Point where, long * docOffset, Boolean * endOfLine);

static void		CheckCaretVisibility(DocPtr dp);

void NewDoc(DocHan * docHanPtr, Rect * viewRect, viewPort docViewPort, short leftMargin, short rightMargin)
{
	DocPtr			dp, * doc;
	font			defaultFont;
	short			index;
	short			numDefaultFonts;
	viewPort		newViewPort;
	shape			clip;
	rectangle		rect;
	StyleRec		defaultStyle;
	
	doc = (DocHan) NewHandle(sizeof(DocRec));
	
	*docHanPtr = doc;
	
	HLock((Handle) doc);
	
	dp = *doc;
	
	dp->magic = kGXEditMagic;
	
	/* initialize error reporting */
	
	dp->error = gx_edit_no_error;
	dp->stickyError = gx_edit_no_error;
	dp->errorProc = nil;

	dp->length = 0;
	dp->lineWidth = viewRect->right - viewRect->left - leftMargin - rightMargin;
	dp->leftMargin = leftMargin;
	dp->rightMargin = rightMargin;
	
	dp->numText = 0;
	
	/* initialize the default style */
	
	dp->numStyles = 0;
	dp->styles = (StyleHan) NewHandle(0);
	
	defaultStyle.textStyle = NewStyle();
	
	SetStyleTextSize(defaultStyle.textStyle, ff(24));
	
	/* initialize the default layout options */
	
	dp->currentLayoutOptions.width = ff(dp->lineWidth);
	dp->currentLayoutOptions.flush = flushLeft;
	dp->currentLayoutOptions.just = noJustification;
	dp->currentLayoutOptions.flags = 0;					/* no flags set */
	dp->currentLayoutOptions.baselineRec = nil;
	
	/* find a good default font */
		
	defaultFont = (font) -1;

	numDefaultFonts = sizeof(gDefaultFonts) / sizeof(char *);
	
	for(index = 0; index < numDefaultFonts, defaultFont == (font) -1; index++)
		FindFonts(nil, familyFontName, 
							macintoshPlatform, romanScript, noLanguage,
							gDefaultFonts[index][0], &gDefaultFonts[index][1],
							1, 1, &defaultFont);
							
	if(defaultFont == (font) -1)
		FindFonts(nil, familyFontName,
							macintoshPlatform, romanScript, noLanguage,
							0, nil, 1, 1, &defaultFont);
							
	if(defaultFont != (font) -1)
		SetStyleFont(defaultStyle.textStyle, defaultFont);
		
	dp->numCurrentFonts = 1;
	dp->currentFonts = (font **) NewHandle(sizeof(font));
	(*dp->currentFonts)[0] = GetStyleFont(defaultStyle.textStyle);
	
	dp->numCurrentStyles = 1;
	dp->currentStyles = (short **) NewHandle(sizeof(short));
	(*dp->currentStyles)[0] = AddDocStyle(dp, &defaultStyle);
	
	DisposeStyle(defaultStyle.textStyle);

	/* set the view rect and viewPort */
	
	dp->viewRect = *viewRect;
	dp->top = 0;
	dp->maxTop = 0;
	
	newViewPort = CopyToViewPort(nil, docViewPort);
	SetViewPortParent(newViewPort, docViewPort);
	
	rect.top = ff(viewRect->top);
	rect.left = ff(viewRect->left);
	rect.right = ff(viewRect->right);
	rect.bottom = ff(viewRect->bottom);
	clip = NewRectangle(&rect);
	SetViewPortClip(newViewPort, clip);
	DisposeShape(clip);
	
	dp->docViewPort = newViewPort;

	/* the document has not been drawn */
	
	dp->drawn = false;
	dp->bottomClear = dp->viewRect.bottom;
	
	/* allocate offscreen bitmap and viewport */
	
	dp->offscreenBytes = nil;
	dp->offscreenPort = nil;
	dp->offscreenDevice = nil;
	dp->offscreenGroup = nil;
	dp->offscreenBitmap = nil;
	
	gxEditBuildOffscreen(dp);
	
	/* now initialize an empty paragraph */
	
	dp->numParagraphs = 1;
	dp->paragraphs = (ParaHan) NewHandle(sizeof(ParaRec));
	
	HLock((Handle) dp->paragraphs);
	
	gxEditNewParagraph(dp, *dp->paragraphs, (*dp->currentStyles)[0], sizeof(char));
	
	CalcStartsAndOffsets(dp, 0);
		
	/* selection at the start of the document */
	
	dp->emptyStyle = (*dp->currentStyles)[0];
	SetSelection(dp, 0, 0, false, false);
		
	/* mark the document as unactive */
	
	dp->active = false;
	
	/* setup the scrap to be empty */
	
	dp->scrap.numRuns = 0;
	dp->scrap.runs = (NewRunHan) NewHandle(0);
	
	
	HUnlock((Handle) dp->paragraphs);
	HUnlock((Handle) doc);
		
}

void DisposeDoc(DocHan doc)
{
	DocPtr			dp;
	ParaPtr			pp;
	short			paragraphIndex;
	StylePtr			sp;
	short			styleIndex;
	
	HLock((Handle) doc);
	
	dp = *doc;
	
	HLock((Handle) dp->styles);
	
	sp = *dp->styles;
	
	for(styleIndex=0; styleIndex < dp->numStyles; styleIndex++, sp++)
		DisposeStyle(sp->textStyle);
		
	HUnlock((Handle) dp->styles);
	
	DisposeHandle((Handle) dp->styles);
	
	HLock((Handle) dp->paragraphs);
	
	pp = *dp->paragraphs;
	
	for(paragraphIndex=0; paragraphIndex < dp->numParagraphs; paragraphIndex++, pp++)
		gxEditDisposeParagraph(pp);
		
	HUnlock((Handle) dp->paragraphs);
	
	DisposeHandle((Handle) dp->paragraphs);

	DisposeHandle((Handle) dp->currentStyles);
	
	DisposeScrap(&dp->scrap);
	
	DisposeViewPort(dp->docViewPort);
	
	gxEditDisposeOffscreen(dp);
		
	HUnlock((Handle) doc);
	
	DisposeHandle((Handle) doc);
}

void DocSetSelection(DocPtr dp, long selStart, long selEnd)
{
	SetSelection(dp, selStart, selEnd, true, false);
	
	Draw(dp, dp->viewRect.top, dp->viewRect.bottom);
}

void DocInsert(DocPtr dp, void *text, short startNumText, short textSize)
{
	Insert(dp, text, startNumText, textSize);
	CheckCaretVisibility(dp);
	Draw(dp, dp->viewRect.top, dp->viewRect.bottom);
}

void DocClear(DocPtr dp)
{
	if(dp->selection.startDocOffset == dp->selection.endDocOffset)
		return;		/* nothing to clear */
		
	Clear(dp);
	
	/* reflow the document */
	
	while(CheckReflow(dp)) ;
	
	CheckCaretVisibility(dp);

	Draw(dp, dp->viewRect.top, dp->viewRect.bottom);
		
	
}

long DocHitTest(DocPtr dp, Point where)
{
	long				docOffset;
	Boolean				endOfLine;
	
	HitTest(dp, where, &docOffset, &endOfLine);
	
	return(docOffset);
	
}

void DocUpdate(DocPtr dp, long start, long end)
{
	ParaPtr			pp;
	short			paragraphIndex;
	Rect			rect;
	rectangle		aRect;
	shape			eraser;
	
	if(start > end)
		gxEditPostFatalError(dp, gx_edit_internal_fatal_error);
		
	if(start < dp->viewRect.top || start > dp->viewRect.bottom)
		gxEditPostFatalError(dp, gx_edit_internal_fatal_error);

	if(end < dp->viewRect.top || end > dp->viewRect.bottom)
		gxEditPostFatalError(dp, gx_edit_internal_fatal_error);
		
	HLock((Handle) dp->paragraphs);
	
	/* map the view area coords to doc coords */
	
	start = start - dp->viewRect.top + dp->top;
	end = end - dp->viewRect.top + dp->top;
	
	/* loop through all of the paragraphs, draw paragraphs that fall in the range */
	
	/* this loop should be optimized in the future */
	
	pp = *dp->paragraphs;
	
	for(paragraphIndex=0; paragraphIndex < dp->numParagraphs; pp++, paragraphIndex++)
		if((pp->start + pp->height) >= start && (pp->start < end)) {
			
			pp->drawn = false;
			DrawParagraph(dp, pp, start, end);
			
		}
			
	/* erase from the last paragraph to end of view if necessary */
	
	if(paragraphIndex == dp->numParagraphs) {
	
		pp = *dp->paragraphs + dp->numParagraphs - 1;
		
		rect.top = pp->start + pp->height + dp->viewRect.top - dp->top;
		rect.bottom = dp->bottomClear;
		rect.left = dp->viewRect.left;
		rect.right = dp->viewRect.right;
		
		if(rect.top < rect.bottom) {
		
			aRect.top = ff(rect.top);
			aRect.left = ff(rect.left);
			aRect.bottom = ff(rect.bottom);
			aRect.right = ff(rect.right);
			
			eraser = NewRectangle(&aRect);
			
			SetViewPortClip(dp->docViewPort, eraser);
	
			SetShapeCommonColor(eraser, gxWhite);
			SetShapeViewPorts(eraser, 1, &dp->docViewPort);
			DrawShape(eraser);
			
			DisposeShape(eraser);
			
			dp->bottomClear = rect.top;
			
		}
				
	}
	
	dp->drawn = true;
		
	HUnlock((Handle) dp->paragraphs);
		
}

void DocClick(DocPtr dp, Point where, Boolean extend)
{
	long		docOffset;
	long		startDocOffset;
	long		endDocOffset;
	Boolean		endOfLine;
	Boolean		done;
	long		oldDocOffset;
	
	HitTest(dp, where, &docOffset, &endOfLine);

	if(extend) {
	
		startDocOffset = dp->selection.startDocOffset;
		endDocOffset = dp->selection.endDocOffset;
		
	} else
		startDocOffset = docOffset;
	
	done = false;
	
	while(!done) {
		
		if(extend) {
		
			if(docOffset > startDocOffset)
				SetSelection(dp, startDocOffset, docOffset, true, endOfLine);
			else
				SetSelection(dp, docOffset, endDocOffset, true, endOfLine);
				
		} else {
		
			if(docOffset > startDocOffset)
				SetSelection(dp, startDocOffset, docOffset, true, endOfLine);
			else
				SetSelection(dp, docOffset, startDocOffset, true, endOfLine);
				
		}
		
		Draw(dp, dp->viewRect.top, dp->viewRect.bottom);
		
		oldDocOffset = docOffset;
		
		do {
		
			if((done = !StillDown()))
				break;
			
			GetMouse(&where);
			
			HitTest(dp, where, &docOffset, &endOfLine);
			
		} while(docOffset == oldDocOffset);
		
	}
	
}

void DocScroll(DocPtr dp, short dv)
{
	RgnHandle		updateRgn;
	long			top;
	long			bottom;
	ParaPtr			pp;
	LinePtr			lp;
	short			paraIndex;
	short			lineIndex;
	
	updateRgn = NewRgn();
	
	ScrollRect(&dp->viewRect, 0, dv, updateRgn);
	
	DisposeRgn(updateRgn);
	
	if(dv < 0) {
		top = dp->viewRect.bottom + dv;
		bottom = dp->viewRect.bottom;
	} else {
		top = dp->viewRect.top;
		bottom = dp->viewRect.top + dv;
	}
	
	dp->top -= dv;

	/* mark any line that falls in this range not drawn */
	
	top -= dp->viewRect.top;
	bottom -= dp->viewRect.top;
	
	top += dp->top;
	bottom += dp->top;
	
	pp = *dp->paragraphs;
	for(paraIndex=0; paraIndex < dp->numParagraphs; paraIndex++, pp++)
		if(pp->start < bottom && (pp->start + pp->height) > top)
			for(lp = *pp->lines, lineIndex=0; lineIndex < pp->numLines; lineIndex++, lp++)
				if((pp->start + lp->start) < bottom && (pp->start + lp->start + lp->height) > top)
					lp->drawn = false;
		
	
	Draw(dp, dp->viewRect.top, dp->viewRect.bottom);
	
}

long DocHeight(DocPtr dp)
{
	return(dp->length);
}

void DocSize(DocPtr dp, Rect * viewRect)
{
	shape			clip;
	rectangle		rect;
	short			paraIndex;
	ParaPtr			pp;
	
	dp->viewRect = *viewRect;
	
	rect.top = ff(viewRect->top);
	rect.left = ff(viewRect->left);
	rect.right = ff(viewRect->right);
	rect.bottom = ff(viewRect->bottom);

	clip = NewRectangle(&rect);
	SetViewPortClip(dp->docViewPort, clip);
	DisposeShape(clip);

	DocSetMaxTop(dp);
	
	/* update the offscreen bitmap */
	
	gxEditBuildOffscreen(dp);

	/* change the line width */
	
	dp->lineWidth = viewRect->right - viewRect->left - dp->leftMargin - dp->rightMargin;
	
	/* change the line width for all of the paragraphs */
	
	pp = *dp->paragraphs;
	
	for(paraIndex = 0; paraIndex < dp->numParagraphs; paraIndex++, pp++) {
	
		if(pp->width != dp->lineWidth) {
			
			pp->width = dp->lineWidth;
			pp->layoutOptions.width = ff(pp->width);
			
			if(pp->numLines) {
			
				pp->reflow = true;
				(*pp->lines)[0].reflow = true;
				
			}
			
		}
		
	}
	
	/* mark the entire document as not drawn */
	
	dp->drawn = false;
		
}	
	
long DocSetTop(DocPtr dp, long top)
{
	short	dv;
	
	if(top > dp->maxTop)
		top = dp->maxTop;
	
	dv = dp->top - top;
	
	if(dv != 0)
		DocScroll(dp, dv);
	
	return(dp->top);
}

long DocGetTop(DocPtr dp)
{
	return(dp->top);
}

long DocGetMaxTop(DocPtr dp)
{
	return(dp->maxTop);
}

void DocKey(DocPtr dp, char key)
{

	switch(key) {
	case '\b':
	
		if(dp->selection.startDocOffset == dp->selection.endDocOffset &&
		   dp->selection.startDocOffset != 0)
			SetSelection(dp, dp->selection.startDocOffset - 1, dp->selection.endDocOffset, false, false);			
		
		DocClear(dp);
		
		break;
		
	default:
		DocInsert(dp, (void *) &key, 1, sizeof(char));
		break;
	}
	
}

void DocSetTextFonts(DocPtr dp, short numFonts, font * srcFonts, font * dstFonts)
{

	/* modify the current selection */
	
	if(SetSelectionTextFonts(dp, numFonts, srcFonts, dstFonts))
		CalcPositionAndDraw(dp, dp->selection.startParaIndex);
			
	SetSelection(dp, dp->selection.startDocOffset, dp->selection.endDocOffset, false, dp->selection.endOfLine);

}

void DocGetTextFonts(DocPtr dp, short * numFonts, font * fonts)
{

	if(numFonts != nil)
		*numFonts = dp->numCurrentFonts;
	
	if(fonts != nil)
		BlockMove((Ptr) *dp->currentFonts, (Ptr) fonts, sizeof(font) * dp->numCurrentFonts);
}

void DocSetTextSize(DocPtr dp, short size)
{
	if(SetSelectionTextSize(dp, size))
		CalcPositionAndDraw(dp, dp->selection.startParaIndex);
				
	SetSelection(dp, dp->selection.startDocOffset, dp->selection.endDocOffset, false, dp->selection.endOfLine);
}

short DocGetTextSize(DocPtr dp)
{
	return(dp->currentSize);
}

void DocSetFeatures(DocPtr dp, font fontId, short numFeatureGroups,
									short * srcFeatureGroupSizes,
									fontRunFeature ** srcFeatureGroups,
									short * dstFeatureGroupSizes,
									fontRunFeature ** dstFeatureGroups)
{
	short			** newStyles;
	
	if(dp->numCurrentStyles != 0) {
	
		newStyles = (short **) NewHandle(dp->numCurrentStyles * sizeof(short));
		
		HLock((Handle) dp->currentStyles);
		HLock((Handle) newStyles);
		
		ModifyStyleFeatures(dp, fontId, numFeatureGroups,
										srcFeatureGroupSizes, srcFeatureGroups,
										dstFeatureGroupSizes, dstFeatureGroups,
										dp->numCurrentStyles, *dp->currentStyles, *newStyles);
		
		if(SetSelectionStyles(dp, dp->numCurrentStyles, *dp->currentStyles, *newStyles))
			CalcPositionAndDraw(dp, dp->selection.startParaIndex);
				
		HUnlock((Handle) dp->currentStyles);
		HUnlock((Handle) newStyles);
		
		DisposeHandle((Handle) newStyles);
			
		SetSelection(dp, dp->selection.startDocOffset, dp->selection.endDocOffset, false, dp->selection.endOfLine);
		
	} else
		gxEditPostFatalError(dp, gx_edit_internal_fatal_error);
	
}

void DocGetFeatures(DocPtr dp, font fontId, short * numFeatureGroupsPtr,
									short *** featureGroupSizesPtr,
									fontRunFeature **** featureGroupsPtr)
{

	if(dp->numCurrentStyles != 0) {
	
		HLock((Handle) dp->currentStyles);
		
		UnionStyleFeatures(dp, dp->numCurrentStyles, *dp->currentStyles, 
								fontId, numFeatureGroupsPtr, featureGroupSizesPtr, featureGroupsPtr);
		
		HUnlock((Handle) dp->currentStyles);
	
		
	} else {
		
		*numFeatureGroupsPtr = 0;
		*featureGroupSizesPtr = nil;
		*featureGroupsPtr = nil;
	}

}

void DocSetVariations(DocPtr dp, font fontId, short numVariations, fontVariation * variations)
{
	short			** newStyles;
	
	if(dp->numCurrentStyles != 0) {
	
		newStyles = (short **) NewHandle(dp->numCurrentStyles * sizeof(short));
		
		HLock((Handle) dp->currentStyles);
		HLock((Handle) newStyles);
		
		ModifyStyleVariations(dp, fontId, numVariations, variations, dp->numCurrentStyles, *dp->currentStyles, *newStyles);
		
		if(SetSelectionStyles(dp, dp->numCurrentStyles, *dp->currentStyles, *newStyles))
			CalcPositionAndDraw(dp, dp->selection.startParaIndex);

		HUnlock((Handle) dp->currentStyles);
		HUnlock((Handle) newStyles);
		
		DisposeHandle((Handle) newStyles);
			
		SetSelection(dp, dp->selection.startDocOffset, dp->selection.endDocOffset, false, dp->selection.endOfLine);
		
	} else
		gxEditPostFatalError(dp, gx_edit_internal_fatal_error);

}

void DocGetVariations(DocPtr dp, font fontId, short * numInstancesPtr,
						short *** instanceSizesPtr, fontVariation **** instancesPtr)
{
	if(dp->numCurrentStyles != 0) {
	
		HLock((Handle) dp->currentStyles);
		
		UnionStyleVariations(dp, dp->numCurrentStyles, *dp->currentStyles,
								fontId, numInstancesPtr, instanceSizesPtr, instancesPtr);
		
		HUnlock((Handle) dp->currentStyles);
	
		
	} else {
		
		*numInstancesPtr = 0;
		*instanceSizesPtr = nil;
		*instancesPtr = nil;
	
	}
	
}

void DocGetRunControls(DocPtr dp, RunControls * runControls, RunControls * mask)
{
	if(dp->numCurrentStyles != 0) {
	
		HLock((Handle) dp->currentStyles);
		
		UnionStyleRunControls(dp, dp->numCurrentStyles, *dp->currentStyles, runControls, mask);
	
		HUnlock((Handle) dp->currentStyles);
				
	} else
		gxEditPostFatalError(dp, gx_edit_internal_fatal_error);

}

void DocSetRunControls(DocPtr dp, RunControls * runControls, RunControls * mask)
{
	short			** newStyles;

	if(dp->numCurrentStyles != 0) {
	
		newStyles = (short **) NewHandle(dp->numCurrentStyles * sizeof(short));
		
		HLock((Handle) dp->currentStyles);
		HLock((Handle) newStyles);
		
		ModifyStyleRunControls(dp, runControls, mask, dp->numCurrentStyles, *dp->currentStyles, *newStyles);
		
		if(SetSelectionStyles(dp, dp->numCurrentStyles, *dp->currentStyles, *newStyles))
			CalcPositionAndDraw(dp, dp->selection.startParaIndex);

		HUnlock((Handle) dp->currentStyles);
		HUnlock((Handle) newStyles);
		
		DisposeHandle((Handle) newStyles);
			
		SetSelection(dp, dp->selection.startDocOffset, dp->selection.endDocOffset, false, dp->selection.endOfLine);
		
	} else
		gxEditPostFatalError(dp, gx_edit_internal_fatal_error);
	
}

void DocGetLayoutOptions(DocPtr dp, LayoutOptions * options, LayoutOptions * optionsMask)
{
	ParaPtr			pp;
	short			startParagraphIndex;
	short			endParagraphIndex;
	short			paragraphIndex;
	short			start;
	short			end;
	Boolean			firstTime;
			
	/* find the current selected paragraph */
	
	startParagraphIndex = dp->selection.startParaIndex;
	endParagraphIndex = dp->selection.endParaIndex;
	
	/* modify the selected paragraphs */
	
	HLock((Handle) dp->paragraphs);
	
	pp = *dp->paragraphs + startParagraphIndex;
	paragraphIndex = startParagraphIndex;
	
	start = dp->top;
	end = dp->top + (dp->viewRect.bottom - dp->viewRect.top);
				
	for(;paragraphIndex <= endParagraphIndex; paragraphIndex++, pp++) {
	
		firstTime = (paragraphIndex == startParagraphIndex);
		
		gxEditGetBitMaskOp(&pp->layoutOptions, options, optionsMask, sizeof(LayoutOptions), firstTime);		
			
	}
	
	/* fix up the mask */
	
	if(~optionsMask->width)
		optionsMask->width = 0;
	
	if(~optionsMask->flush)
		optionsMask->flush = 0;
	
	if(~optionsMask->just)
		optionsMask->just = 0;
	
	if(~((long) optionsMask->baselineRec))
		optionsMask->baselineRec = nil;
		
	HUnlock((Handle) dp->paragraphs);
	
}

void DocSetLayoutOptions(DocPtr dp, LayoutOptions *options, LayoutOptions * optionsMask)
{
	ParaPtr			pp;
	LinePtr			lp;
	short			startParagraphIndex;
	short			endParagraphIndex;
	short			paragraphIndex;
	short			lineIndex;
	short			start;
	short			end;
		
	/* find the current selected paragraph */
	
	startParagraphIndex = dp->selection.startParaIndex;
	endParagraphIndex = dp->selection.endParaIndex;
	
	/* modify the selected paragraphs */
	
	HLock((Handle) dp->paragraphs);
	
	pp = *dp->paragraphs + startParagraphIndex;
	paragraphIndex = startParagraphIndex;
	
	start = dp->top;
	end = dp->top + (dp->viewRect.bottom - dp->viewRect.top);
				
	for(;paragraphIndex <= endParagraphIndex; paragraphIndex++, pp++) {
	
		/* set the paragraph options */
				
		if(gxEditSetBitMaskOp(options, &pp->layoutOptions, optionsMask, sizeof(LayoutOptions))) {
		
			pp->width = FixedToInt(pp->layoutOptions.width);
		
			/* change all of the layout options of the lines */
			
			HLock((Handle) pp->lines);
			
			lp = *pp->lines;
			
			for(lineIndex=0; lineIndex < pp->numLines-1; lineIndex++, lp++) {
				SetLayout(lp->layout, 0, nil, nil, 0, nil, nil, 0, nil, nil, &pp->layoutOptions, nil);
				lp->drawn = false;
			}
			
			HUnlock((Handle) pp->lines);
		
			/* now redraw the paragraph if it is displayed on the screen */
			
			if(pp->start + pp->height >= start || pp->start < end)
				DrawParagraph(dp, pp, start, end);
		}
		
	}
	
	gxEditSetBitMaskOp(options, &dp->currentLayoutOptions, optionsMask, sizeof(LayoutOptions));
	
	HUnlock((Handle) dp->paragraphs);
	
}

void DocPrint(DocPtr dp, Job printJob)
{
	short			numPages;
	long			** pageStarts;
	short			page;
	OSErr			err;
	viewPort		aViewPort;
	shape			picture;

	aViewPort = dp->docViewPort;
	
	CalcPrintPages(dp, printJob, &numPages, &pageStarts);
	
	/* document title should be used !!! */
	
	StartJob(printJob, "\pGX Edit Doc", numPages);
	
	if((err = GetJobError(printJob)) != noErr)
		gxEditPostFatalError(dp, gx_edit_internal_fatal_error);
	
	for(page=0; page<numPages; page++) {
		
		picture = GetAPage(dp, (*pageStarts)[page]);
		
		PrintPage(printJob,  page+1, GetJobFormat((Job) printJob, 1), picture);
				
		if((err = GetJobError((Job) printJob)) != noErr)
			gxEditPostFatalError(dp, gx_edit_internal_fatal_error);
			
		DisposeShape(picture);
		
	}
	
	FinishJob(printJob);
	
	if((err = GetJobError(printJob)) != noErr)
		gxEditPostFatalError(dp, gx_edit_internal_fatal_error);
		
	DisposeHandle((Handle) pageStarts);
}

void DocCopy(DocPtr dp)
{
	ParaPtr				pp;
	NewRunPtr			rp;
	short				paragraphIndex;
	short				runIndex;
	long				start;
	long				numText;
	NewRunRec			run;
	long				size;
	ScrapPtr			scrap;
	short				numAppendText;
	
	start = dp->selection.startDocOffset;
	numText = dp->selection.endDocOffset - start;

	if(!numText)
		return;			/* nothing to cut */
		
	scrap = &dp->scrap;
	
	/* clear out the current scrap */
	
	scrap->numRuns = 0;
	scrap->numText = numText;
	SetHandleSize((Handle) scrap->runs, 0);
		
	/* get to the start of the selection and make a copy of the first partial run,
	   all intermediate runs and the partial last run */
	   
	HLock((Handle) dp->paragraphs);
	
	pp = *dp->paragraphs;
	
	for(paragraphIndex=0; paragraphIndex < dp->numParagraphs; paragraphIndex++, pp++)
		if(start < (pp->docOffset + pp->numText))
			break;
			
	if(paragraphIndex == dp->numParagraphs)
		gxEditPostFatalError(dp, gx_edit_internal_fatal_error);
		
	start -= pp->docOffset;
		
	HLock((Handle) pp->runs);
	
	rp = *pp->runs;
	
	for(runIndex = 0; runIndex < pp->numRuns; runIndex++, rp++)
		if(start < (rp->paraOffset + rp->numText))
			break;
			
	if(runIndex == pp->numRuns)
		gxEditPostFatalError(dp, gx_edit_internal_fatal_error);
		
	start -= rp->paraOffset;
		
	/* grab the text */
		
	run.styleIndex = rp->styleIndex;
	run.numText = rp->numText - start;
	run.textSize = rp->textSize;
	
	if(run.numText > numText)
		run.numText = numText;
	
	size = run.numText * run.textSize;
	run.text = (void **) NewHandle(size);
	
	BlockMove((Ptr) ((long) *rp->text + (start*rp->textSize)), (Ptr) *run.text, size);
	
	PtrAndHand((Ptr) &run, (Handle) scrap->runs, sizeof(NewRunRec));
	scrap->numRuns++;
	
	numText -= run.numText;
	
	runIndex++;
	rp++;
		
	while(numText) {
	
		if(runIndex == pp->numRuns) {
		
			HUnlock((Handle) pp->runs);
			
			paragraphIndex++;
			pp++;
				
			if(paragraphIndex == dp->numParagraphs)
				gxEditPostFatalError(dp, gx_edit_internal_fatal_error);
					
			HLock((Handle) pp->runs);
				
			rp = *pp->runs;
			runIndex = 0;
					
		}
		
		if(rp->styleIndex == run.styleIndex) {
		
			/* just add the text to the last scrap run */

			numAppendText = rp->numText;
						
			if(numAppendText > numText)
				numAppendText = numText;
			
			size = (run.numText + numAppendText) * run.textSize;
			SetHandleSize((Handle) run.text, size);
			
			size = numAppendText * run.textSize;
			BlockMove((Ptr) *rp->text, (Ptr) ((long) *run.text + (run.numText * run.textSize)), size);
			
			run.numText += numAppendText;

			BlockMove((Ptr) &run, (Ptr) (*scrap->runs + scrap->numRuns - 1), sizeof(NewRunRec));
			
			numText -= numAppendText;
			
		} else {
		
			/* add a new run to the scrap */

			run.styleIndex = rp->styleIndex;
			run.numText = rp->numText;
			
			if(run.numText > numText)
				run.numText = numText;
			
			size = run.numText * run.textSize;
			run.text = (void **) NewHandle(size);
			
			BlockMove((Ptr) *rp->text, (Ptr) *run.text, size);
			
			PtrAndHand((Ptr) &run, (Handle) scrap->runs, sizeof(NewRunRec));
			scrap->numRuns++;
			
			numText -= run.numText;
			
		}
		
		if(numText == 0)
			break;
			
		runIndex++;
		rp++;
		
	}
	
	gxEditScrapToDesk(dp);
	
	HUnlock((Handle) pp->runs);
	HUnlock((Handle) dp->paragraphs);
	
}


void DocPaste(DocPtr dp)
{
	NewRunPtr		rp;
	short			runIndex;
	
	gxEditScrapFromDesk(dp);
	
	if(dp->scrap.numRuns == 0)
		return;		/* no scrap text to paste */
		
	HLock((Handle) dp->scrap.runs);
	
	rp = *(dp->scrap.runs);
	
	for(runIndex=0; runIndex < dp->scrap.numRuns; runIndex++, rp++) {
	
		HLock((Handle) rp->text);
		
		(*dp->currentStyles)[0] = rp->styleIndex;
		
		Insert(dp, *rp->text, rp->numText, rp->textSize);

		HUnlock((Handle) rp->text);
	}
	
	HUnlock((Handle) dp->scrap.runs);
	
	/* reflow the document */
	
	while(CheckReflow(dp)) ;
	
	CheckCaretVisibility(dp);
	
	Draw(dp, dp->viewRect.top, dp->viewRect.bottom);
		
}

long DocGetScrapLen(DocPtr dp)
{
	return(dp->scrap.numText);
}

long DocGetSelectionLen(DocPtr dp)
{
	return(dp->selection.endDocOffset - dp->selection.startDocOffset);
}

long DocGetTextLen(DocPtr dp)
{
	return(dp->numText);
}

Boolean DocIdle(DocPtr dp)
{
	Boolean			modified;
	
	dp->currentTime = TickCount();
	
	BlinkCursor(dp);
	
	if(modified = CheckReflow(dp)) {

		CheckCaretVisibility(dp);
	
		Draw(dp, dp->viewRect.top, dp->viewRect.bottom);
		
	}
	
	return(modified);
		
}

static void CalcPrintPages(DocPtr dp, Job printJob, short * numPagesPtr, long *** pageStartsPtr)
{
	short				numPages;
	long				** pageStarts;
	ParaPtr				pp;
	LinePtr				lp;
	short				paragraphIndex;
	short				lineIndex;
	short				top;
	short				bottom;
	long				docOffset;
	Format				format;
	rectangle			pageSize;
	rectangle			paperSize;

	numPages = 0;
	pageStarts = (long **) NewHandle(0);
	
	/* calculate the page dimensions */
	
	format = GetJobFormat((Job) printJob, 1);
	GetFormatDimensions(format, &pageSize, &paperSize);
	
	dp->pageTop = FixedToInt(paperSize.top) + kPrintTopMargin;
	dp->pageLeft = FixedToInt(paperSize.left) + kPrintLeftMargin;
	dp->pageBottom = FixedToInt(paperSize.bottom) - kPrintBottomMargin;
	dp->pageRight = FixedToInt(paperSize.right) - kPrintRightMargin;
		
	HLock((Handle) dp->paragraphs);
	
	pp = *dp->paragraphs;
	
	top = dp->pageTop;
	bottom = dp->pageBottom;
	docOffset = 0;
	
	for(paragraphIndex = 0; paragraphIndex < dp->numParagraphs; paragraphIndex++, pp++) {
	
		HLock((Handle) pp->lines);
		
		lp = *pp->lines;
		
		for(lineIndex = 0; lineIndex < pp->numLines; lineIndex++, lp++) {
		
			if(top + lp->height > bottom) {
			
				PtrAndHand((Ptr) &docOffset, (Handle) pageStarts, sizeof(long));
				docOffset = lp->paraOffset + pp->docOffset;
				top = dp->pageTop;
				numPages++;
			
			}
			
			top += lp->height;
			
		}
		
		HUnlock((Handle) pp->lines);
		
	}
	
	HUnlock((Handle) dp->paragraphs);

	PtrAndHand((Ptr) &docOffset, (Handle) pageStarts, sizeof(long));
	numPages++;
	
	*pageStartsPtr = pageStarts;
	*numPagesPtr = numPages;
}

static shape GetAPage(DocPtr dp, long docOffset)
{
	ParaPtr				pp;
	LinePtr				lp;
	shape				picture;
	short				top;
	fixed				fixedLeft;
	fixed				fixedTop;
	short				paragraphIndex;
	short				lineIndex;
	short				paraOffset;
	short				bottom;
	long				numLayouts;
	void				** layouts;
	
	picture = NewPicture(0, nil, nil, nil, nil);
	layouts = (void **) NewHandle(0);
	numLayouts = 0;
	
	HLock((Handle) dp->paragraphs);
	
	pp = *dp->paragraphs;
	
	for(paragraphIndex = 0; paragraphIndex < dp->numParagraphs; paragraphIndex++, pp++)
		if(pp->docOffset + pp->numText > docOffset)
			break;
	
	if(paragraphIndex == dp->numParagraphs)
		gxEditPostFatalError(dp, gx_edit_internal_fatal_error);
		
	HLock((Handle) pp->lines);
	
	lp = *pp->lines;
	paraOffset = docOffset - pp->docOffset;
	
	for(lineIndex = 0; lineIndex < pp->numLines; lineIndex++, lp++)
		if(lp->paraOffset + lp->numText > paraOffset)
			break;
			
	if(lineIndex == pp->numLines)
		gxEditPostFatalError(dp, gx_edit_internal_fatal_error);
		
	top = dp->pageTop;
	bottom = dp->pageBottom;
	
	fixedLeft = ff(dp->pageLeft);

	while(top + lp->height < bottom) {
	
		fixedTop = ff(top + lp->ascent);
		
		MoveShapeTo((shape) lp->layout, fixedLeft, fixedTop);
		lp->layoutPos.x = fixedLeft;
		lp->layoutPos.y = fixedTop;
		
		PtrAndHand((Ptr) &lp->layout, (Handle) layouts, sizeof(void *));
		
		top += lp->height;
		
		lineIndex++;
		lp++;
		
		if(lineIndex == pp->numLines) {
		
			paragraphIndex++;
			
			if(paragraphIndex == dp->numParagraphs)
				break;

			HUnlock((Handle) pp->lines);
			
			pp++;
			
			HLock((Handle) pp->lines);
			
			lp = *pp->lines;
			lineIndex = 0;
			
		}
		
	}
	
	HUnlock((Handle) pp->lines);
	
	HUnlock((Handle) dp->paragraphs);
	
	numLayouts = GetHandleSize((Handle) layouts) / sizeof(void *);
	
	HLock((Handle) layouts);
	
	picture = NewPicture(numLayouts, *layouts, nil, nil, nil);
	
	HUnlock((Handle) layouts);
	
	DisposeHandle((Handle) layouts);
		
	return(picture);

}

static void CalcStartsAndOffsets(DocPtr dp, short paragraphIndex)
{
	ParaPtr				pp;
	long				start;
	short				docOffset;
	
	if(paragraphIndex < 0 || paragraphIndex > dp->numParagraphs)
		gxEditPostFatalError(dp, gx_edit_internal_fatal_error);
		
	pp = *dp->paragraphs + paragraphIndex;
	
	if(paragraphIndex == 0) {
	
		if(pp->start != 0)
			pp->drawn = false;
			
		pp->start = 0;
		pp->docOffset = 0;
		
	} else {
	
		paragraphIndex--;
		pp--;
	}

	start = pp->start + pp->height;
	docOffset = pp->docOffset + pp->numText;
	
	paragraphIndex++;
	pp++;
	
	for(;paragraphIndex < dp->numParagraphs; paragraphIndex++, pp++) {
	
		if(pp->start != start)
			pp->drawn = false;		/* the paragraph has moved -- redraw */
		
		pp->start = start;
		pp->docOffset = docOffset;

		start += pp->height;
		docOffset += pp->numText;

		
	}
		
	dp->length = start;
	dp->numText = docOffset;
	
	DocSetMaxTop(dp);
	
}

static void InsertNewParagraph(DocPtr dp, short paragraphIndex, short paragraphOffset)
{
	ParaRec		newPara;
	ParaPtr		pp;
	long		size;
	
	HLock((Handle) dp->paragraphs);
	
	pp = *dp->paragraphs + paragraphIndex;
		
	SplitParagraph(dp, pp, paragraphOffset, &newPara);
	
	HUnlock((Handle) dp->paragraphs);
	
	dp->numParagraphs++;
	
	SetHandleSize((Handle) dp->paragraphs, dp->numParagraphs * sizeof(ParaRec));
	
	size = sizeof(ParaRec) * (dp->numParagraphs - paragraphIndex - 2);
	
	BlockMove((Ptr) (*dp->paragraphs + paragraphIndex + 1),
			  (Ptr) (*dp->paragraphs + paragraphIndex + 2), size);
			  
	BlockMove((Ptr) &newPara, (Ptr) (*dp->paragraphs + paragraphIndex + 1), sizeof(ParaRec));
		
}

void DocSetMaxTop(DocPtr dp)
{
	short		dv;
	
	dp->maxTop = dp->length - (dp->viewRect.bottom - dp->viewRect.top);
	
	if(dp->maxTop < 0)
		dp->maxTop = 0;
		
	if(dp->top > dp->maxTop) {
	
		dv = dp->top - dp->maxTop;
		
		DocScroll(dp, dv);
			
	}
}

static void DisposeScrap(ScrapPtr scrap)
{
	DisposeHandle((Handle) scrap->runs);
}

static void InitRunControl(RunControls * runControl)
{

	runControl->flags = 0;		/* all off id the default state */
	runControl->beforeWithStreamShift = ff(0);
	runControl->afterWithStreamShift = ff(0);
	runControl->crossStreamShift = ff(0);
	runControl->imposedWidth = ff(0);
	runControl->track = noTracking;
	runControl->hangingInhibitFactor = frac1;
	runControl->kerningInhibitFactor = frac1;
	runControl->decompositionAdjustmentFactor = ff(0);
	runControl->baselineType = 0;		/* roman baseline */

}

static void BlinkCursor(DocPtr dp)
{
	if(dp->selection.startDocOffset != dp->selection.endDocOffset)
		return;		/* selection range is not a cursor */
		
	/* code to blink cursor */	
}

static Boolean CheckReflow(DocPtr dp)
{
	ParaPtr		pp;
	short		paragraphIndex;
	
	/* try to find a paragraph that needs to be reflowed */
	
	HLock((Handle) dp->paragraphs);
	
	pp = *dp->paragraphs;
	
	for(paragraphIndex=0; paragraphIndex < dp->numParagraphs; paragraphIndex++, pp++)
		if(pp->reflow) {
		
			ReflowParagraph(dp, pp);
			
			HUnlock((Handle) dp->paragraphs);
			
			CalcStartsAndOffsets(dp, paragraphIndex);

			return(true);
		}
		
	HUnlock((Handle) dp->paragraphs);
}

static void CalcPositionAndDraw(DocPtr dp,short startIndex)		
 {
	/* calculate starts and offsets -- set document height */
	CalcStartsAndOffsets(dp, startIndex);
	CheckCaretVisibility(dp);
	Draw(dp, dp->viewRect.top, dp->viewRect.bottom);
}

static void Position(DocPtr dp, long docOffset, Rect * rect)
{
	ParaPtr		pp;
	short		paragraphIndex;
	short		paragraphOffset;
		
	GetParagraphIndexAndOffset(dp, docOffset, &paragraphIndex, &paragraphOffset, false);
	
	HLock((Handle) dp->paragraphs);
	
	pp = *dp->paragraphs + paragraphIndex;
	
	ParagraphPosition(dp, pp, paragraphOffset, rect);
	
	rect->top += pp->start;
	rect->bottom += pp->start;
	
	HUnlock((Handle) dp->paragraphs);
	
}

static void Draw(DocPtr dp, long start, long end)
{
	ParaPtr			pp;
	short			paragraphIndex;
	Rect			rect;
	rectangle		aRect;
	shape			eraser;
		
	if(start > end)
		gxEditPostFatalError(dp, gx_edit_internal_fatal_error);
		
	if(start < dp->viewRect.top || start > dp->viewRect.bottom)
		gxEditPostFatalError(dp, gx_edit_internal_fatal_error);

	if(end < dp->viewRect.top || end > dp->viewRect.bottom)
		gxEditPostFatalError(dp, gx_edit_internal_fatal_error);
		
	HLock((Handle) dp->paragraphs);
	
	/* map the view area coords to doc coords */
	
	start = start - dp->viewRect.top + dp->top;
	end = end - dp->viewRect.top + dp->top;
				
	/* loop through all of the paragraphs, draw paragraphs that fall in the range */
	
	/* this loop should be optimized in the future */
	
	pp = *dp->paragraphs;
	
	for(paragraphIndex=0; paragraphIndex < dp->numParagraphs; pp++, paragraphIndex++) {
					
		if((pp->start + pp->height) >= start && (pp->start < end)) {
					   
			DrawParagraph(dp, pp, start, end);
			
		} else {
		
			pp->drawn = false;

		}
		
	}
	
	/* erase from the last paragraph to end of view if necessary */
	
	if(paragraphIndex == dp->numParagraphs) {
	
		pp = *dp->paragraphs + dp->numParagraphs - 1;
		
		rect.top = pp->start + pp->height + dp->viewRect.top - dp->top;
		rect.bottom = dp->bottomClear;
		rect.left = dp->viewRect.left;
		rect.right = dp->viewRect.right;
		
		if(rect.top < rect.bottom) {
		
			aRect.top = ff(rect.top);
			aRect.left = ff(rect.left);
			aRect.bottom = ff(rect.bottom);
			aRect.right = ff(rect.right);
			
			eraser = NewRectangle(&aRect);
			
			SetViewPortClip(dp->docViewPort, eraser);
	
			SetShapeCommonColor(eraser, gxWhite);
			SetShapeViewPorts(eraser, 1, &dp->docViewPort);
			DrawShape(eraser);
			
			DisposeShape(eraser);
			
			dp->bottomClear = rect.top;
			
		}
				
	}
		
	dp->drawn = true;
		
	HUnlock((Handle) dp->paragraphs);
		
}

static void HitTest(DocPtr dp, Point where, long * docOffsetPtr, Boolean * endOfLinePtr)
{
	ParaPtr				pp;
	long				docOffset;
	short				paraIndex;
	short				paraOffset;
	Boolean				endOfLine;
	
	if(where.v < dp->viewRect.top)
		where.v = dp->viewRect.top;
	
	if(where.v > dp->viewRect.bottom)
		where.v = dp->viewRect.bottom;
	
	where.v -= dp->viewRect.top;
	where.h -= dp->viewRect.left + dp->leftMargin;
		
 	where.v += dp->top;
 	
	HLock((Handle) dp->paragraphs);
	
	/* find the paragraph and hit test paragraph */
	
	pp = *dp->paragraphs;
	docOffset = 0;
	
	for(paraIndex=0; paraIndex < dp->numParagraphs; paraIndex++,pp++)
		if(where.v >= pp->start && where.v < (pp->start + pp->height)) {
		
			where.v -= pp->start;
			
			HitTestParagraph(dp, pp, where, &paraOffset, &endOfLine);
			
			docOffset =  paraOffset + pp->docOffset;
			break;
			
		}
	
	if(paraIndex == dp->numParagraphs && dp->numParagraphs != 0)
		docOffset = (pp-1)->docOffset + (pp-1)->numText;
	
	HUnlock((Handle) dp->paragraphs);
	
	*docOffsetPtr = docOffset;
	*endOfLinePtr = endOfLine;
	
}

static void Insert(DocPtr dp, void * text, short startNumText, short textSize)
{
	short		paragraphIndex, paragraphOffset;
	ParaPtr		pp;
	short		numParaText;
	char		* ptr;
	short		startIndex;
	short		startOffset;
	short		newParagraphs;
	short		startHeight;
	long		docOffset;
	short		numText;
	
	Clear(dp);
		
	if(dp->numCurrentStyles == 0)
		gxEditPostFatalError(dp, gx_edit_internal_fatal_error);

	/* find the paragraph index and offset from the current selection start */
	
	paragraphIndex = dp->selection.startParaIndex;
	paragraphOffset = dp->selection.startParaOffset;
	
	numText = startNumText;
	
	if(numText == 0)
		return;
	
	HLock((Handle) dp->paragraphs);
	
	/* insert all characters up to the newline character */
	
	startIndex = paragraphIndex;
	startOffset = paragraphOffset;
	newParagraphs = 0;
	startHeight = 0;
	
	if(textSize == 1) {
		for(numParaText=0, ptr=(char *) text; numParaText<numText; numParaText++, ptr++)
			if(*ptr == '\r')
				break;			
	} else
		numParaText = numText;
	
	if(numParaText) {
	
		pp = *dp->paragraphs + paragraphIndex;
		
		startHeight = pp->height;
	
		InsertParagraphText(dp, pp, paragraphOffset, text, numParaText, textSize, (*dp->currentStyles)[0]);
		
		numText -= numParaText;
		text = (void *) ((long) text + (numParaText * textSize));
		paragraphOffset += numParaText;
		
	}
	
	/* now add new paragraphs as appropriate */
	
	while(numText) {
	
		text = (void *) ((long) text + textSize);
		numText--;			/* skip past newline */
	
		if(textSize == 1) {
			for(numParaText=0, ptr=(char *) text; numParaText<numText; numParaText++, ptr++)
				if(*ptr == '\r')
					break;			
		} else
			gxEditPostFatalError(dp, gx_edit_internal_fatal_error);
				
		HUnlock((Handle) dp->paragraphs);
		
		InsertNewParagraph(dp, paragraphIndex, paragraphOffset);
		
		newParagraphs++;
		
		paragraphIndex++;
		paragraphOffset = 0;
		
		HLock((Handle) dp->paragraphs);
		
		if(numParaText) {
		
			pp = *dp->paragraphs + paragraphIndex;
			
			InsertParagraphText(dp, pp, 0, text, numParaText, textSize, (*dp->currentStyles)[0]);
			
			numText -= numParaText;
			text = (void *) ((long) text + (numParaText * textSize));
			paragraphOffset += numParaText;
			
		}
	
		
	}
	
	HUnlock((Handle) dp->paragraphs);
	
	/* calculate starts and offsets -- set document height */
	
	CalcStartsAndOffsets(dp, startIndex);
	
	/* update the current selection */
	
	docOffset = dp->selection.startDocOffset + startNumText;
	
	SetSelection(dp, docOffset, docOffset, false, false);
	
}

static void Clear(DocPtr dp)
{
	ParaPtr			pp;
	long			start;
	long			end;
	short			paragraphIndex;
	short			paragraphOffset;
	short			numText;
	long			size;
	Boolean			needToJoin;
	short			paraText;
	short			index;
	short			startIndex;
	
	if(dp->selection.startDocOffset == dp->selection.endDocOffset)
		return;		/* nothing to clear */
		
	/* make sure we don't clear the entire docuemnt -- leave at least one empty paragraph */
	
	if(dp->selection.startDocOffset == 0 && dp->selection.endDocOffset == dp->numText)
		dp->selection.endDocOffset--;
		
	start = dp->selection.startDocOffset;
	end = dp->selection.endDocOffset;
			
	numText = end - start;
		
	/* find out where to start */
	
	HLock((Handle) dp->paragraphs);
	
	GetParagraphIndexAndOffset(dp, start, &paragraphIndex, &paragraphOffset, false);
	
	startIndex = index = paragraphIndex;
	
	needToJoin = false;
	
	if(paragraphOffset != 0) {
	
		pp = *dp->paragraphs + index;
		
		paraText = pp->numText - paragraphOffset;
		
		if(numText < paraText)
			paraText = numText;
		else {
			paraText--;		/* don't delete '\r' */
			needToJoin = true;
		}
					
		ParagraphClear(dp, pp, paragraphOffset, paraText);
		
		numText -= paraText;
		
		if(needToJoin)
			numText--;		/* the '\r' will be removed when the paragraphs are joined */
				
		index++;
		
	}
	
	while(numText) {
	
		pp = *dp->paragraphs + index;
		
		if(numText >= pp->numText) {
		
			numText -= pp->numText;
			
			gxEditDisposeParagraph(pp);
			
			dp->numParagraphs--;
			
			size = (dp->numParagraphs - index) * sizeof(ParaRec);
			BlockMove((Ptr) (pp+1), (Ptr) pp, size);
			
			HUnlock((Handle) dp->paragraphs);
			
			size = dp->numParagraphs * sizeof(ParaRec);
			SetHandleSize((Handle) dp->paragraphs, size);
			
			HLock((Handle) dp->paragraphs);
			
		} else {
		
			ParagraphClear(dp, pp, 0, numText);
			
			numText = 0;
			
			
		
		}
		
	}
	
	if(needToJoin && index != dp->numParagraphs) {
	
		pp = *dp->paragraphs + index;
		
		JoinParagraphs(dp, pp-1, pp);
		
		dp->numParagraphs--;
		
		size = (dp->numParagraphs - index) * sizeof(ParaRec);
		BlockMove((Ptr) (pp+1), (Ptr) pp, size);
		
		HUnlock((Handle) dp->paragraphs);
		
		size = dp->numParagraphs * sizeof(ParaRec);
		SetHandleSize((Handle) dp->paragraphs, size);
		
		HLock((Handle) dp->paragraphs);
	
	}
	
	HUnlock((Handle) dp->paragraphs);
	
	SetSelection(dp, dp->selection.startDocOffset, dp->selection.startDocOffset, false, false);
	
	CalcStartsAndOffsets(dp, paragraphIndex);
	
}

static void CheckCaretVisibility(DocPtr dp)
{
	Rect			rect;
	short			dv;
	
	/* scroll the screen if the caret is not in the visible area */
	
	Position(dp, dp->selection.startDocOffset, &rect);
	
	if(rect.bottom > dp->top + (dp->viewRect.bottom - dp->viewRect.top)) {
	
		dv = dp->top - (rect.bottom - (dp->viewRect.bottom - dp->viewRect.top));
	
		DocScroll(dp, dv);
		
	}
	
}
