/*
	File:				GXEditFileIO.c
	
	Contains:
	
	Written by:		Barton R. House
	
	Copyright:		 1993 by Apple Computer, Inc., All rights reserved.
	
*/

#include "GXEdit.h"
#include "GXEditDoc.h"
#include "GXEditDebug.h"
#include "GXEditError.h"
#include "GXEditUtils.h"
#include "GXEditSelection.h"

/* GX includes */

#include "graphics toolbox.h"
#include "graphics routines.h"
#include "math routines.h"

#define kBufferSize		2056

static void WriteParagraph(DocPtr dp, short fRefNum, ParaPtr pp);
static void WriteLine(DocPtr dp, short fRefNum, LinePtr lp);
static void WriteNewRun(DocPtr dp, short fRefNum, NewRunPtr rp);
static void WriteStyle(DocPtr dp, short fRefNum, StylePtr sp);
static void WriteData(short fRefNum, Ptr p, long size);
static void WriteLong(short fRefNum, long l);
static void WriteShort(short fRefNum, short s);
static void WriteBoolean(short fRefNum, Boolean b);


static void ReadParagraph(DocPtr dp, short fRefNum, ParaPtr pp);
static void ReadLine(DocPtr dp, short fRefNum, LinePtr lp);
static void ReadNewRun(DocPtr dp, short fRefNum, NewRunPtr rp);
static void ReadStyle(DocPtr dp, short fRefNum, StylePtr sp);
static void ReadLayout(DocPtr dp, short fRefNum, shape * layoutPtr);
static void ReadData(short fRefNum, Ptr p, long size);
static void ReadLong(short fRefNum, long * l);
static void ReadShort(short fRefNum, short * s);
static void ReadBoolean(short fRefNum, Boolean * b);


/* FILE FORMAT - Version 0

	magic								long
	version								long
	
	doc
	
		length							long
		numStyles						short
		numText							long
		numParagraphs					short
		top								long
		
		
	selection
	
		startDocOffset					long
		endDocOffset					long
		endOfLine						Boolean
	
	styles[numStyles]
	
		styleSize						long
		
		flatten data					char[styleSize]
		
	paragraphs[numPara]
	
		start							short
		width							short
		height							short
		docOffset						long
		numText							short
		numLines						short
		numRuns							short
		options							LayoutOptions		(this should be a private data structure)
		
		lines[numLines]
		
			start						short
			paraOffset					short
			numText						short
		
		runs[numRuns]
		
			styleIndex					short
			paraOffset					short
			numText						short
			textSize					short
			
			text						char[numText * textSize]
	
*/

typedef struct {
	spoolBlock		sb;			/* this must be first */
	short			fRefNum;
	long			start;
	long			numBytes;
	DocPtr			dp;
} MySpoolBlock;

static long MySpoolProcedure(spoolCommand command, spoolBlock * sb);

void DocSave(DocPtr dp, short fRefNum)
{
	ParaPtr			pp;
	StylePtr		sp;
	short			paragraphIndex;
	short			styleIndex;
	
	/* write out magic */
	
	WriteLong(fRefNum, kGXEditMagic);
	
	/* write out file format version */
	
	WriteLong(fRefNum, kGXEditFileVersion);
	
	/* write out the document information */
	
	WriteLong(fRefNum, dp->length);
	WriteShort(fRefNum, dp->numStyles);
	WriteLong(fRefNum, dp->numText);
	WriteShort(fRefNum, dp->numParagraphs);
	WriteLong(fRefNum, dp->top);
	
	/* write out selection information */
	
	WriteLong(fRefNum, dp->selection.startDocOffset);
	WriteLong(fRefNum, dp->selection.endDocOffset);
	WriteBoolean(fRefNum, dp->selection.endOfLine);
	
	/* write out style information */
	
	HLock((Handle) dp->styles);
	
	for(sp = *dp->styles, styleIndex = 0; styleIndex < dp->numStyles; styleIndex++, sp++)
		WriteStyle(dp, fRefNum, sp);
		
	HUnlock((Handle) dp->styles);
	
	/* write out the paragraphs */
	
	HLock((Handle) dp->paragraphs);
	
	for(paragraphIndex=0, pp = *dp->paragraphs; paragraphIndex < dp->numParagraphs; paragraphIndex++,pp++)
		WriteParagraph(dp, fRefNum, pp);
	
	HUnlock((Handle) dp->paragraphs);
		
}

static void WriteParagraph(DocPtr dp, short fRefNum, ParaPtr pp)
{
	LinePtr			lp;
	short			lineIndex;
	NewRunPtr		rp;
	short			runIndex;
	
	WriteShort(fRefNum, pp->start);
	WriteShort(fRefNum, pp->width);
	WriteShort(fRefNum, pp->height);
	WriteLong(fRefNum, pp->docOffset);
	WriteShort(fRefNum, pp->numText);
	WriteShort(fRefNum, pp->numLines);
	WriteShort(fRefNum, pp->numRuns);
	
	/* this should really use a private structure */
	
	WriteData(fRefNum, (Ptr) &pp->layoutOptions, sizeof(pp->layoutOptions));
	
	/* write out lines */
		
	HLock((Handle) pp->lines);
	
	for(lineIndex=0, lp = *pp->lines; lineIndex < pp->numLines; lineIndex++, lp++)
		WriteLine(dp, fRefNum, lp);
	
	HUnlock((Handle) pp->lines);
	
	/* write out runs */
	
	HLock((Handle) pp->runs);
	
	for(runIndex = 0, rp = *pp->runs; runIndex < pp->numRuns; runIndex++, rp++)
		WriteNewRun(dp, fRefNum, rp);
		
	HUnlock((Handle) pp->runs);
	
}

static void WriteLine(DocPtr dp, short fRefNum, LinePtr lp)
{
	#pragma unused(dp)
	WriteShort(fRefNum, lp->start);
	WriteShort(fRefNum, lp->paraOffset);
	WriteShort(fRefNum, lp->numText);	
		
}

static void WriteNewRun(DocPtr dp, short fRefNum, NewRunPtr rp)
{
	#pragma unused(dp)

	WriteShort(fRefNum, rp->styleIndex);
	WriteShort(fRefNum, rp->paraOffset);
	WriteShort(fRefNum, rp->numText);
	WriteShort(fRefNum, rp->textSize);
	
	/* save out text */
	
	HLock((Handle) rp->text);
	
	WriteData(fRefNum, *rp->text, rp->numText * rp->textSize);
		
	HUnlock((Handle) rp->text);
	
}

static void WriteStyle(DocPtr dp, short fRefNum, StylePtr sp)
{
	MySpoolBlock		msb;
	shape				emptyShape;
	long				startOfStyle;
	long				endOfStyle;
	long				styleSize;
	OSErr				err;
	
	/* save the file position of start of style */
	
	GetFPos(fRefNum, &startOfStyle);
	
	/* save out the size of the flatten style */
	
	styleSize = 0;	/* we don't know how big the flatten style is yet */
	
	WriteLong(fRefNum, styleSize);

	/* save out the flatten style information */
	
	msb.sb.buffer = NewPtr(kBufferSize);
	msb.sb.bufferSize = kBufferSize;
	msb.sb.spoolProcedure = MySpoolProcedure;
	
	GetFPos(fRefNum, &msb.start);
	
	msb.fRefNum = fRefNum;
	msb.numBytes = 0;
	msb.dp = dp;
	
	/* create an empty shape and set the style to the style we want to save */
	
	emptyShape = NewShape(emptyType);
	SetShapeStyle(emptyShape, sp->textStyle);
	
	/* look into setting the flags so that a list of fonts that are used is generated */
	
	FlattenShape(emptyShape, (flattenFlags) 0, (spoolBlock *) &msb);
	
	/* dispose of the temporary shape */
	
	DisposeShape(emptyShape);

	/* now record the file position for the end of the style */
	
	GetFPos(fRefNum, &endOfStyle);
	
	/* double check the file position with the number of bytes written */
	
	if(msb.numBytes + msb.start != endOfStyle)
		gxEditPostFatalError(dp, gx_edit_internal_fatal_error);
		
	/* finish up recording the size of the flatten style */
	
	styleSize = endOfStyle - msb.start;
	
	if((err = SetFPos(fRefNum, fsFromStart, startOfStyle)) != noErr)
		gxEditPostFatalError(dp, gx_edit_internal_fatal_error);

	WriteLong(fRefNum, styleSize);
			
	if((err = SetFPos(fRefNum, fsFromStart, endOfStyle)) != noErr)
		gxEditPostFatalError(dp, gx_edit_internal_fatal_error);
	
	/* free up the buffer */
	
	DisposePtr(msb.sb.buffer);
	
}


static void WriteData(short fRefNum, Ptr p, long size)
{
	long		count;
	OSErr		err;
	
	count = size;
		
	if((err = FSWrite(fRefNum, &count, p)) != noErr)
		gxEditPostFatalError(nil, gx_edit_internal_fatal_error);
		
	if(count != size)
		gxEditPostFatalError(nil, gx_edit_internal_fatal_error);
		
}

static void WriteLong(short fRefNum, long l)
{
	WriteData(fRefNum, (Ptr) &l, sizeof(long));
}

static void WriteShort(short fRefNum, short s)
{
	WriteData(fRefNum, (Ptr) &s, sizeof(short));
}

static void WriteBoolean(short fRefNum, Boolean b)
{
	WriteData(fRefNum, (Ptr) &b, sizeof(Boolean));
}
		
void DocOpen(DocHan * docHanPtr, Rect * viewRect, viewPort docViewPort, short fRefNum, short leftMargin, short rightMargin)
{
	DocHan				doc;
	DocPtr				dp;
	ParaPtr				pp;
	StylePtr			sp;
	short				paraIndex;
	short				styleIndex;
	viewPort			newViewPort;
	shape				clip;
	rectangle			rect;
	long				magic;
	long				version;
	
	doc = (DocHan) NewHandle(sizeof(DocRec));
	
	HLock((Handle) doc);
	
	dp = *doc;
	
	/* read in magic */
	
	ReadLong(fRefNum, &magic);
	
	if(magic != kGXEditMagic)
		gxEditPostFatalError(nil, gx_edit_internal_fatal_error);

	/* read the file format version */
	
	ReadLong(fRefNum, &version);
	
	if(version != kGXEditFileVersion)
		gxEditPostFatalError(nil, gx_edit_internal_fatal_error);

	/* read the the document information */
	
	ReadLong(fRefNum, &dp->length);
	ReadShort(fRefNum, &dp->numStyles);
	ReadLong(fRefNum, &dp->numText);
	ReadShort(fRefNum, &dp->numParagraphs);
	ReadLong(fRefNum, &dp->top);
	
	/* read the selection information */
	
	ReadLong(fRefNum, &dp->selection.startDocOffset);
	ReadLong(fRefNum, &dp->selection.endDocOffset);
	ReadBoolean(fRefNum, &dp->selection.endOfLine);
			
	/* set generic doc information */
	
	dp->magic = magic;
	dp->active = false;
	
	/* set the error control values */
	
	dp->errorProc = nil;
	dp->error = 0;
	dp->stickyError = 0;
	
	/* set the correct view port and view rect */
	
	dp->lineWidth = viewRect->right - viewRect->left - leftMargin - rightMargin;
	dp->leftMargin = leftMargin;
	dp->rightMargin = rightMargin;
	
	dp->viewRect = *viewRect;
	
	newViewPort = CopyToViewPort(nil, docViewPort);
	SetViewPortParent(newViewPort, docViewPort);
	rect.top = ff(viewRect->top);
	rect.left = ff(viewRect->left);
	rect.right = ff(viewRect->right);
	rect.bottom = ff(viewRect->bottom);
	clip = NewRectangle(&rect);
	SetViewPortClip(newViewPort, clip);
	DisposeShape(clip);
	
	dp->docViewPort = newViewPort;
	
	/* allocate offscreen bitmap and viewport */
	
	dp->offscreenBytes = nil;
	dp->offscreenPort = nil;
	dp->offscreenDevice = nil;
	dp->offscreenGroup = nil;
	dp->offscreenBitmap = nil;
		
	gxEditBuildOffscreen(dp);
	
	/* read in the styles */
	
	dp->styles = (StyleHan) NewHandle(sizeof(StyleRec) * dp->numStyles);
	
	HLock((Handle) dp->styles);
	
	for(styleIndex=0, sp = *dp->styles; styleIndex<dp->numStyles; styleIndex++, sp++)
		ReadStyle(dp, fRefNum, sp);
	
	HUnlock((Handle) dp->styles);
	
	/* read in the paragraphs */
	
	dp->paragraphs = (ParaHan) NewHandle(sizeof(ParaRec) * dp->numParagraphs);
	
	HLock((Handle) dp->paragraphs);
	
	for(paraIndex=0, pp = *dp->paragraphs; paraIndex<dp->numParagraphs; paraIndex++, pp++)
		ReadParagraph(dp, fRefNum, pp);
	
	HUnlock((Handle) dp->paragraphs);
		
	/* initialize empty scrap */
	
	dp->scrap.numText = 0;
	dp->scrap.numRuns = 0;
	dp->scrap.runs = (NewRunHan) NewHandle(0);
	
	/* invalidate the entire document for drawing purposes */
	
	dp->drawn = false;
	
	/* set the paragraph widths and reflow if the paragraph width has changed  */
	
	pp = *dp->paragraphs;
	
	for(paraIndex = 0; paraIndex < dp->numParagraphs; paraIndex++, pp++)
		if(pp->width != dp->lineWidth) {
			
			pp->width = dp->lineWidth;
			pp->layoutOptions.width = ff(pp->width);
			
			if(pp->numLines) {
			
				pp->reflow = true;
				(*pp->lines)[0].reflow = true;
				
			}
			
		}
		
	/* setup the current selection */
	
	dp->numCurrentFonts = 0;
	dp->currentFonts = (font **) NewHandle(0);
	
	dp->numCurrentStyles = 0;
	dp->currentStyles = (short **) NewHandle(0);
	
	dp->currentSize = 0;
	dp->emptyStyle = 0;
	
	SetSelection(dp, dp->selection.startDocOffset, dp->selection.endDocOffset, true, dp->selection.endOfLine); 
	
	DocSetMaxTop(dp);
	
	HUnlock((Handle) doc);
	
	*docHanPtr = doc;
}

static void ReadParagraph(DocPtr dp, short fRefNum, ParaPtr pp)
{
	LinePtr				lp;
	short				lineIndex;
	NewRunPtr			rp;
	short				runIndex;
	
	ReadShort(fRefNum, &pp->start);
	ReadShort(fRefNum, &pp->width);
	ReadShort(fRefNum, &pp->height);
	ReadLong(fRefNum, &pp->docOffset);
	ReadShort(fRefNum, &pp->numText);
	ReadShort(fRefNum, &pp->numLines);
	ReadShort(fRefNum, &pp->numRuns);
	
	pp->drawn = false;
	pp->reflow = false;
	
	/* this should really use a private structure */
	
	ReadData(fRefNum, (Ptr) &pp->layoutOptions, sizeof(pp->layoutOptions));
	
	/* adjust the layout options */
	
	pp->layoutOptions.baselineRec = nil;		/* FUTURE ENHANCEMENT -- THIS SHOULD BE RESTORED */
	
	/* read in the lines */
	
	pp->lines = (LineHan) NewHandle(pp->numLines * sizeof(LineRec));
	
	HLock((Handle) pp->lines);
	
	for(lineIndex = 0, lp = *pp->lines; lineIndex < pp->numLines; lineIndex++, lp++)
		ReadLine(dp, fRefNum, lp);
		
	HUnlock((Handle) pp->lines);
	
	/* read in runs */
	
	pp->runs = (NewRunHan) NewHandle(pp->numRuns * sizeof(NewRunRec));
	
	HLock((Handle) pp->runs);
	
	for(runIndex = 0, rp = *pp->runs; runIndex < pp->numRuns; runIndex++, rp++)
		ReadNewRun(dp, fRefNum, rp);
		
	HUnlock((Handle) pp->runs);
	
}


static void ReadLine(DocPtr dp, short fRefNum, LinePtr lp)
{
	#pragma unused(dp)

	ReadShort(fRefNum, &lp->start);
	ReadShort(fRefNum, &lp->paraOffset);
	ReadShort(fRefNum, &lp->numText);
	
	lp->layout = nil;
	lp->layoutPos.x = 0; lp->layoutPos.y = 0;
	lp->dirty = true;
	lp->reflow = false;
	lp->drawn = false;
	lp->selected = false;
	
	lp->ascent = 0;
	lp->descent = 0;
	
}

static void ReadNewRun(DocPtr dp, short fRefNum, NewRunPtr rp)
{
	#pragma unused(dp)
	ReadShort(fRefNum, &rp->styleIndex);
	ReadShort(fRefNum, &rp->paraOffset);
	ReadShort(fRefNum, &rp->numText);
	ReadShort(fRefNum, &rp->textSize);
	
	/* read in text */
	
	rp->text = (void **) NewHandle(rp->numText * rp->textSize);

	HLock((Handle) rp->text);
	
	ReadData(fRefNum, *rp->text, rp->numText * rp->textSize);
		
	HUnlock((Handle) rp->text);
		
}

static void ReadStyle(DocPtr dp, short fRefNum, StylePtr sp)
{
	MySpoolBlock		msb;
	shape				emptyShape;
	long				styleSize;
	
	ReadLong(fRefNum, &styleSize);

	/* readin the GX style information */
	
	msb.sb.buffer = NewPtr(kBufferSize);
	msb.sb.bufferSize = kBufferSize;
	msb.sb.spoolProcedure = MySpoolProcedure;
	
	GetFPos(fRefNum, &msb.start);
	
	msb.numBytes = styleSize;
	
	msb.fRefNum = fRefNum;
	msb.dp = dp;
	
	emptyShape = UnflattenShape((spoolBlock *) &msb, 1, &dp->docViewPort );
	
	sp->textStyle = GetShapeStyle(emptyShape);		/* this does not increment the owner count */
	sp->textStyle = CopyToStyle(nil, sp->textStyle);
	
	DisposeShape(emptyShape);
	
	/* check that the correct number of bytes was read */
	
	if(msb.numBytes != 0)
		gxEditPostFatalError(dp, gx_edit_internal_fatal_error);
	
	DisposePtr(msb.sb.buffer);
	
	/* initialize the redundant data */
	
	sp->textFont = GetStyleFont(sp->textStyle);
	sp->textSize = GetStyleTextSize(sp->textStyle);
		
}

static void ReadData(short fRefNum, Ptr p, long size)
{
	long				count;
	OSErr				err;
	
	count = size;
	
	if((err = FSRead(fRefNum, &count, p)) != noErr)
		gxEditPostFatalError(nil, gx_edit_internal_fatal_error);
		
	if(count != size)
		gxEditPostFatalError(nil, gx_edit_internal_fatal_error);
		
}

static void ReadLong(short fRefNum, long * l)
{
	ReadData(fRefNum, (Ptr) l, sizeof(long));
}

static void ReadShort(short fRefNum, short * s)
{
	ReadData(fRefNum, (Ptr) s, sizeof(short));
}

static void ReadBoolean(short fRefNum, Boolean * b)
{
	ReadData(fRefNum, (Ptr) b, sizeof(Boolean));
}
		
static long MySpoolProcedure(spoolCommand command, spoolBlock * sb)
{
	long				count;
	MySpoolBlock		* msb;
	OSErr				err;
	
	msb = (MySpoolBlock *) sb;
	
	switch(command) {
	
	case openReadSpool:
	case openWriteSpool:
	
		SetFPos(msb->fRefNum, fsFromStart, msb->start);
		break;
		
	case readSpool:
	
		count = sb->count;
		
		if(count > msb->numBytes) {
		
			count = msb->numBytes;
			sb->bufferSize = msb->numBytes;
			
		}
		
		if((err = FSRead(msb->fRefNum, &count, sb->buffer)) != noErr)
			gxEditPostFatalError(msb->dp, gx_edit_internal_fatal_error);
		
		if((sb->count > msb->numBytes && count != msb->numBytes) ||
			(sb->count <= msb->numBytes && count != sb->count))
			gxEditPostFatalError(msb->dp, gx_edit_internal_fatal_error);
		
		msb->numBytes -= count;
		
		break;
		
	case writeSpool:

		count = sb->count;
		
		if((err = FSWrite(msb->fRefNum, &count, sb->buffer)) != noErr)
			gxEditPostFatalError(msb->dp, gx_edit_internal_fatal_error);

		if(sb->count != count)
			gxEditPostFatalError(msb->dp, gx_edit_internal_fatal_error);
		
		msb->numBytes += count;
		
		break;
		
	case closeSpool:
		break;
	}
	
	return(noErr);
	
}
