/*
	File:				GXEditLine.c
	
	Contains:
	
	Written by:		Barton R. House
	
	Copyright:		 1993 by Apple Computer, Inc., All rights reserved.
	
*/

#include <Memory.h>

#include "GXEdit.h"
#include "GXEditDebug.h"
#include "GXEditDoc.h"
#include "GXEditLine.h"
#include "GXEditNewRun.h"
#include "GXEditStyle.h"
#include "GXEditSelection.h"
#include "GXEditError.h"

#include "graphics routines.h"
#include "math routines.h"
#include "graphics libraries.h"
#include "layout routines.h"

#define MIN(x,y) ((x) < (y) ? (x) : (y))
#define MAX(x,y) ((x) > (y) ? (x) : (y))

static void		CalcLayout(DocPtr dp, ParaPtr pp, LinePtr lp);														  
static void		CalcOffsets(DocPtr dp, LinePtr lp, short runIndex);
static void		CalcHeight(LinePtr lp);
static void 	DrawHighlight(DocPtr dp, shape layout, short startOffset, short endOffset);
static void		DrawCaret(DocPtr dp, shape layout, short offset);

/* insert text into the line */

void InsertDocLineText(DocPtr dp, ParaPtr pp, LinePtr lp, short lineOffset, short numText)
{
	#pragma unused(pp)
	if(lineOffset < 0 || lineOffset > lp->numText)
		gxEditPostFatalError(dp, gx_edit_internal_fatal_error);
			
	lp->numText += numText;	
	lp->dirty = true;
				
}


/* join the two lines disposing of the right line data structures */

void JoinDocLines(DocPtr dp, ParaPtr pp, LinePtr left, LinePtr right)
{
	#pragma unused(dp)
	#pragma unused(pp)
	left->numText += right->numText;
	left->dirty = true;
	
	if(right->layout != nil)
		DisposeShape(right->layout);
	
}

void SplitDocLine(DocPtr dp, ParaPtr pp, LinePtr lp, short lineOffset, LinePtr newLine)
{
	#pragma unused(pp)
	if(lineOffset < 0 || lineOffset > lp->numText)
		gxEditPostFatalError(dp, gx_edit_internal_fatal_error);
	
	lp->dirty = true;
	
	newLine->dirty = true;
	newLine->layout = nil;
		
	newLine->numText = lp->numText - lineOffset;
	lp->numText = lineOffset;
	
}

/* the line is our basic drawing unit */
								 
void DrawDocLine(DocPtr dp, ParaPtr pp, LinePtr lp)
{
	Rect			rect;
	long			docOffset;
	long			startOffset;
	long			endOffset;
	point			textPosition;
	rectangle		aRect;
	shape			clip;
	Boolean			drawLine;
	Boolean			drawSelection;
	Boolean			clearSelection;
	SelPtr			sp;
	Boolean			startAtEnd;
	Boolean			endAtStart;
	Boolean			isCaret;
	
	drawLine = !(dp->drawn & pp->drawn & lp->drawn);
	
	/* see if this line is selected and needs to be drawn */
	
	sp = &dp->selection;
	
	docOffset = pp->docOffset + lp->paraOffset;
	endOffset = docOffset + lp->numText;
	
	startAtEnd = (sp->startDocOffset == endOffset);
	endAtStart = (sp->endDocOffset == docOffset);
	isCaret = (sp->endDocOffset == sp->startDocOffset);
		
	if(docOffset <= sp->endDocOffset && endOffset >= sp->startDocOffset &&
	  !(isCaret && ((sp->endOfLine && endAtStart) || (!sp->endOfLine && startAtEnd))) &&
	  !(!isCaret && (endAtStart || startAtEnd))) {
	
	   	/* this line falls in the selection range */
	   
	   	if(docOffset > dp->selection.startDocOffset)
	   		startOffset = docOffset;
	   	else
	   		startOffset =  dp->selection.startDocOffset;
	   
	   	if(docOffset + lp->numText < dp->selection.endDocOffset)
	   		endOffset = docOffset + lp->numText;
	   	else
	   		endOffset =  dp->selection.endDocOffset;
	   		
	   	startOffset -= docOffset;
	   	endOffset -= docOffset;
	   	
	   	/* NOTE!!!! This is a quick hack for parry.  It should be removed when the
	   	   real end of paragraph code is put in */
	   	   
		if(endOffset && endOffset > startOffset && endOffset + lp->paraOffset == pp->numText)
			endOffset--;
	   	   
	   	/* end of hack */
	   		   		   		
	   	if(lp->selected)
	   		if(lp->startOffset != startOffset || lp->endOffset != endOffset)
	   			drawSelection = true;
	   		else
	   			drawSelection = false;
	   	else
	   		drawSelection = true;
	   		   	
		drawSelection |= drawLine;
		
		clearSelection = false;
		
	} else {
	
		if(lp->selected && !drawLine)
			clearSelection = true;
		else
			clearSelection = false;
	
		drawSelection = false;
			
	}
	
	if(!(drawLine || drawSelection || clearSelection))
		return;
	
	if(lp->dirty)
		CalcLayout(dp, pp, lp);

	/* set the correct clip */
	
	rect.top = dp->viewRect.top + pp->start + lp->start - dp->top;
	
	if(rect.top < dp->viewRect.top)
		rect.top = dp->viewRect.top;
		
	rect.left = dp->viewRect.left;
	
	rect.bottom = rect.top + lp->height;
	
	if(rect.bottom > dp->viewRect.bottom)
		rect.bottom = dp->viewRect.bottom;
	
	rect.right = dp->viewRect.right;
	
	aRect.top = ff(rect.top);
	aRect.left = ff(rect.left);
	aRect.bottom = ff(rect.bottom);
	aRect.right = ff(rect.right);
	
	clip = NewRectangle(&aRect);

	/* This ignores is needed because we do not keep track of the current
	      clip of the view port.  Thus, we will often set the clip to the same rect
	      which generates a notice*/
	      
	IgnoreGraphicsNotice(port_clip_already_set);
	SetViewPortClip(dp->docViewPort, clip);
	PopGraphicsNotice();
	
	/* make sure the layout is at the right position */
	
	textPosition.x = ff(dp->viewRect.left + dp->leftMargin);
	textPosition.y = ff(dp->viewRect.top + pp->start + lp->start - dp->top + lp->ascent);

	if(textPosition.x != lp->layoutPos.x || textPosition.y != lp->layoutPos.y) {
		MoveShapeTo(lp->layout, textPosition.x, textPosition.y);
		lp->layoutPos = textPosition;
	}

	if(drawLine) {

		
		/* erase the line */
	
		SetShapeCommonColor(clip, gxWhite);
		SetShapeViewPorts(clip, 1, &dp->offscreenPort);
		DrawShape(clip);
		
		if(lp->numText == 0)
			return;				/* nothing to draw or select */

		/* now draw the layout */
		
		DrawShape(lp->layout);

		/* now draw the bitmap to the screen */
		
		SetShapeViewPorts(dp->offscreenBitmap, 1, &dp->docViewPort);
		DrawShape(dp->offscreenBitmap);
		SetShapeViewPorts(dp->offscreenBitmap, 1, &dp->offscreenPort);
		
		lp->drawn = true;
		lp->selected = false;
		
		/* check to see if we drawn in the bottom cleared area */
		
		if(dp->bottomClear < rect.bottom)
			dp->bottomClear = rect.bottom;
		
	}
		
	if(drawSelection) {
	    		
		if(lp->selected && (lp->startOffset == lp->endOffset)) {
		
			DrawCaret(dp, lp->layout, lp->startOffset);
			lp->selected = false;
			
		}
				
	    if(startOffset == endOffset) {
	    
	    	if(lp->selected)
	    		DrawHighlight(dp, lp->layout, lp->startOffset, lp->endOffset);
	
			DrawCaret(dp, lp->layout, startOffset);
			
	    	
	    } else {
	    
	    	if(lp->selected) {
	    	
	    	
	    		/* 					|--------------|								*/
	    		/*		|----|										case 0			*/
	    		/*										|---|		case 1			*/
	    		/*				|-----|								case 2			*/
	    		/*								|-----|				case 3			*/
	    		/*						|----|						case 4			*/
	    		/*				|----------------------|			case 5			*/
	    		
	    	
	    		if(startOffset < lp->startOffset) {
	    		
	    			if(endOffset > lp->endOffset) {
	    			
	    				/* case 5 */
	    				
	    				DrawHighlight(dp, lp->layout, startOffset, lp->startOffset);
	    				DrawHighlight(dp, lp->layout, lp->endOffset, endOffset);
	    				
	    			} else if(endOffset > lp->startOffset) {
	    			
	    				/* case 2 */
	    			
	    				DrawHighlight(dp, lp->layout, startOffset, lp->startOffset);
	    				DrawHighlight(dp, lp->layout, endOffset, lp->endOffset);
	    				
	    			} else {
	    			
	    				/* case 0 */
	    			
	    				DrawHighlight(dp, lp->layout, lp->startOffset, lp->endOffset);
	    				DrawHighlight(dp, lp->layout, startOffset, endOffset);
	    				
	    			}
	    		
	    		} else {
	    		
	    			if(startOffset > lp->endOffset) {
	    			
	    				/* case 1 */
	    				
	    				DrawHighlight(dp, lp->layout, lp->startOffset, lp->endOffset);
	    				DrawHighlight(dp, lp->layout, startOffset, lp->endOffset);
	    				
	    			} else if(endOffset < lp->endOffset) {
	    			
	    				/* case 4 */
	    				
	    				DrawHighlight(dp, lp->layout, lp->startOffset, startOffset);
	    				DrawHighlight(dp, lp->layout, endOffset, lp->endOffset);
	    				
	    			} else {
	    			
	    				/* case 3 */
	    			
	    				DrawHighlight(dp, lp->layout, lp->startOffset, startOffset);
	    				DrawHighlight(dp, lp->layout, lp->endOffset, endOffset);
	    			}
	    			
	    		}

	    	} else
	    		DrawHighlight(dp, lp->layout, startOffset, endOffset);
	    
	    }
	    
	    lp->selected = true;
	    lp->startOffset = startOffset;
	    lp->endOffset = endOffset;
		
	} else if(clearSelection) {
				
			if(lp->startOffset == lp->endOffset)
				DrawCaret(dp, lp->layout, lp->startOffset);
			else
				DrawHighlight(dp, lp->layout, lp->startOffset, lp->endOffset);
	
			lp->selected = false;
	}
	
	DisposeShape(clip);
		
}

void NewDocLine(DocPtr dp, ParaPtr pp, LinePtr lp)
{
	#pragma unused(dp)
	#pragma unused(pp)

	lp->layout = nil;
	lp->paraOffset = 0;
	lp->numText = 0;
	lp->start = 0;
	lp->reflow = false;
	lp->drawn = false;
	lp->selected = false;
	lp->dirty = true;
		
}

short HitTestDocLine(DocPtr dp, ParaPtr pp, LinePtr lp, Point where)
{
	short			lineOffset;
	LayoutHitInfo	hitInfo;
	point			hitDown;
	
	if(where.h <= 0)
		return(0);
		
	if(lp->dirty)
		CalcLayout(dp, pp, lp);
	
	hitDown.x = ff(where.h) + lp->layoutPos.x;
	hitDown.y = ff(where.v) + lp->layoutPos.y;
	
	lineOffset = HitTestLayout((shape) lp->layout, &hitDown, highlightStraight, &hitInfo,
									nil);
			
	return(lineOffset);
	
}

void DocLinePosition(LinePtr lp, short lineOffset, Rect * rect)
{
	/* currently not fully supported */
	#pragma unused(lineOffset)

	rect->top = 0;
	rect->bottom = lp->height;
	rect->left = 0;
	rect->right = 0;
			
}


void DocLineClear(DocPtr dp, ParaPtr pp, LinePtr lp, short lineOffset, short numText)
{
	#pragma unused(pp)

	if(numText == 0)
		return;

	if(lineOffset < 0 || lineOffset > lp->numText)
		gxEditPostFatalError(dp, gx_edit_internal_fatal_error);
		
	if((lineOffset + numText) > lp->numText)
		gxEditPostFatalError(dp, gx_edit_internal_fatal_error);
		
	lp->numText -= numText;
	lp->dirty = true;
		
}

void DisposeDocLine(LinePtr lp)
{
	
	DisposeShape(lp->layout);
	
}

short GetLineHeight(DocPtr dp, ParaPtr pp, LinePtr lp)
{
	if(lp->dirty)
		CalcLayout(dp, pp, lp);
		
	return(lp->height);
}

static void CalcLayout(DocPtr dp, ParaPtr pp, LinePtr lp)
{
	long			textRunCount;
	short			* textRunLengths;
	void			** textPtrs;
	long			styleRunCount;
	short			* styleRunLengths;
	style			* styles;
	NewRunPtr		rp;
	short			runIndex;
	LayoutOptions	options;
	short			lineText;
	short			runText;
	short			maxRunCount;
	short			startRunIndex;
	short			runOffset;
	short			i;
	
	if(lp->numText == 0) {
	
		/* really should create an empty shape and set line values accordingly */
		/* but for speed -- we will just return for now */
		/* this should not be a problem */
		
		return;
	}
	
	HLock((Handle) pp->runs);

	if(lp->layout != nil)
		DisposeShape(lp->layout);
	
	maxRunCount = pp->numRuns;
	
	textRunLengths = (short *) NewPtr(maxRunCount * sizeof(short));
	textPtrs = (void **) NewPtr(maxRunCount * sizeof(void *));
	styles = (style *) NewPtr(maxRunCount * sizeof(style));
	
	GetNewRunIndexAndOffset(dp, pp, lp->paraOffset, &startRunIndex, &runOffset);
	lineText = lp->numText;
	textRunCount = 0;
	
	runIndex = startRunIndex;
	rp = *pp->runs + runIndex;
	for(; runIndex < maxRunCount && lineText > 0; runIndex++, rp++, lineText -= runText) {
		
		runText = (rp->numText - runOffset > lineText ? lineText : rp->numText - runOffset);
		textRunLengths[textRunCount] = runText * rp->textSize;

		/* NOTE!!!!! this is a quick hack for parry.  It should be removed when the real end of
		   paragraph handling is put in -- this hack causes the end of paragraph '\n' not to
		   be placed in the layout */
		   
		if(runText + runOffset == rp->numText && pp->numRuns == runIndex + 1 && (runText > 1 || pp->numRuns > 1)) {

			textRunLengths[textRunCount] -= rp->textSize;
		
			/* GX graphics flattening breaks when we have zero length runs */
			
			if(textRunLengths[textRunCount] == 0)
				continue;
				
		}
		   
		/* end of hack */

		styles[textRunCount] = GetDocStyle(dp, rp->styleIndex)->textStyle;
		
		HLock((Handle) rp->text);
		
		textPtrs[textRunCount++] = (Ptr) *rp->text + (runOffset * rp->textSize);
		
		runOffset = 0;
		
	}
	
	if(lineText > 0)
		gxEditPostFatalError(dp, gx_edit_internal_fatal_error);
	
	styleRunCount = textRunCount;
	styleRunLengths = textRunLengths;
	
	/* setup layout options */
	
	options = pp->layoutOptions;
	
	/* turn off justification if last line in paragraph */
	
	if(options.just != noJustification &&
	   StripAddress(lp) == StripAddress((*pp->lines + pp->numLines - 1)))
		options.just = noJustification;
		
	/* these values may be totally wrong, but they might be right.  If they are right, it
	   will save us from having to move it later. And thus, save us from throwing away caches */
	   
	lp->layoutPos.x = ff(dp->viewRect.left + dp->leftMargin);
	lp->layoutPos.y = ff(dp->viewRect.top + pp->start + lp->start - dp->top + lp->ascent);

	if (textRunCount)
		lp->layout = (void *) NewLayout(textRunCount, textRunLengths, (void *) textPtrs,
								styleRunCount, styleRunLengths, styles, 0, nil, nil,
								&options, nil);
	else
		lp->layout = (void *) NewLayout(0, nil, nil, 0, nil, nil, 0, nil, nil, &options, nil);
								
	SetShapeViewPorts(lp->layout, 1, &dp->offscreenPort);
								
	MoveShapeTo(lp->layout, lp->layoutPos.x, lp->layoutPos.y);

	rp = *pp->runs + startRunIndex;
	for(i = 0; i < textRunCount; i++, rp++)
		HUnlock((Handle) rp->text);
		
	DisposePtr((Ptr) textRunLengths);
	DisposePtr((Ptr) textPtrs);
	DisposePtr((Ptr) styles);
		
	CalcHeight(lp);
	
	lp->drawn = false;		/* this line needs be redrawn */
	lp->dirty = false;		/* no longer dirty */
	
	HUnlock((Handle) pp->runs);
}

static void CalcHeight(LinePtr lp)
{
	fixed		lineAscent;
	fixed		lineDescent;
	fixed		boundsAscent;
	fixed		boundsDescent;
	rectangle	bounds;
	
	if(lp->numText == 0) {
	
		lp->ascent = lp->descent = lp->height = 0;
		return;
		
	}
	
	GetLayoutSpan((shape) lp->layout, &lineAscent, &lineDescent);
	GetShapeLocalBounds((shape) lp->layout, &bounds);
	
	boundsAscent = lp->layoutPos.y - bounds.top;
	boundsDescent = bounds.bottom - lp->layoutPos.y;
	
	if(boundsAscent > lineAscent || boundsDescent > lineDescent) {
	
		lineAscent = lineAscent > boundsAscent ? lineAscent : boundsAscent;
		lineDescent = lineDescent > boundsDescent ? lineDescent : boundsDescent;
		
		SetLayoutSpan(lp->layout, lineAscent, lineDescent);
	}
	
	lp->ascent = FixedToInt(lineAscent);
	lp->descent = FixedToInt(lineDescent);

	lp->height = lp->ascent + lp->descent;	
	
}

static void DrawHighlight(DocPtr dp, shape layout, short startOffset, short endOffset)
{
	shape		highlight;
	
	if(startOffset == endOffset)
		return;		/* no highlight */
	
	 highlight = GetLayoutHighlight(layout, startOffset, endOffset, highlightAverageAngle, nil);
    										
	SetShapeCommonTransfer(highlight, highlightMode);
	SetShapeCommonColor(highlight, gxWhite);
	SetShapeViewPorts(highlight, 1, &dp->docViewPort);
   										
	DrawShape(highlight);
    	
	 DisposeShape(highlight);
	    	
}

static void DrawCaret(DocPtr dp, shape layout, short offset)
{
	shape		caret;
	
	caret = GetLayoutCaret(layout, offset, highlightAverageAngle, splitCaretType, nil);

	SetShapeCommonTransfer(caret, xorMode);
	SetShapeCommonColor(caret, gxWhite);
	SetShapeViewPorts(caret, 1, &dp->docViewPort);
 	
 	DrawShape(caret);
    	
	DisposeShape(caret);
    
}

