/*
	File:				GXEditStyle.c
	
	Contains:
	
	Written by:		Barton R. House
	
	Copyright:		 1993 by Apple Computer, Inc., All rights reserved.
	
*/

#include "GXEdit.h"
#include "GXEditDoc.h"
#include "GXEditStyle.h"
#include "GXEditDebug.h"
#include "GXEditError.h"
#include "GXEditUtils.h"

#include "math routines.h"
#include "graphics routines.h"
#include "graphics libraries.h"
#include "font routines.h"
#include "layout routines.h"


StylePtr GetDocStyle(DocPtr dp, short styleIndex)
{
	if(styleIndex < 0 || styleIndex > dp->numStyles)
		gxEditPostFatalError(dp, gx_edit_internal_fatal_error);
		
	return(&(*dp->styles)[styleIndex]);
}

short FindDocStyle(DocPtr dp, StylePtr matchSp)
{
	short		styleIndex;
	StylePtr	sp;
	
	sp = *dp->styles;
	
	for(styleIndex = 0; styleIndex < dp->numStyles; styleIndex++, sp++)
		if(matchSp->textFont == sp->textFont &&
		   matchSp->textSize == sp->textSize &&
		   EqualStyle(matchSp->textStyle, sp->textStyle))
			return(styleIndex);
	
	return(-1);
	
}

short AddDocStyle(DocPtr dp, StylePtr sp)
{
	StyleRec			newStyle;
	
	newStyle = *sp;

	newStyle.textStyle = CopyToStyle(nil, newStyle.textStyle);
	
	newStyle.textSize = FixedToInt(GetStyleTextSize(newStyle.textStyle));
	newStyle.textFont = GetStyleFont(newStyle.textStyle);
		
	PtrAndHand((Ptr) &newStyle, (Handle) dp->styles, sizeof(StyleRec));
	
	return(dp->numStyles++);

}

short SetDocStyleTextFont(DocPtr dp, short styleIndex, font fontId)
{
	StyleRec	newStyle;
	
	newStyle = *GetDocStyle(dp, styleIndex);
	 
	newStyle.textStyle = CopyToStyle(nil, newStyle.textStyle);
	
	SetStyleFont(newStyle.textStyle, fontId);
	newStyle.textFont = fontId;
	
	styleIndex = FindDocStyle(dp, &newStyle);
	
	if(styleIndex == -1)
		styleIndex = AddDocStyle(dp, &newStyle);
		
	DisposeStyle(newStyle.textStyle);
	
	return(styleIndex);

}

font GetDocStyleTextFont(DocPtr dp, short styleIndex)
{
	StylePtr		sp;
	
	if(styleIndex < 0 || styleIndex >= dp->numStyles)
		gxEditPostFatalError(dp, gx_edit_internal_fatal_error);
		
	sp = *dp->styles + styleIndex;
	
	return(sp->textFont);
}

short SetDocStyleTextSize(DocPtr dp, short styleIndex, short size)
{
	StyleRec	newStyle;
	
	newStyle = *GetDocStyle(dp, styleIndex);
	 
	newStyle.textStyle = CopyToStyle(nil, newStyle.textStyle);
	
	SetStyleTextSize(newStyle.textStyle, ff(size));
	
	styleIndex = FindDocStyle(dp, &newStyle);
	
	if(styleIndex == -1)
		styleIndex = AddDocStyle(dp, &newStyle);
		
	DisposeStyle(newStyle.textStyle);
	
	return(styleIndex);

}

short GetDocStyleTextSize(DocPtr dp, short styleIndex)
{
	StylePtr		sp;
	
	if(styleIndex < 0 || styleIndex >= dp->numStyles)
		gxEditPostFatalError(dp, gx_edit_internal_fatal_error);
		
	sp = *dp->styles + styleIndex;
	
	return(sp->textSize);
}

short SetDocStyleFeatures(DocPtr dp, short styleIndex, short numFeatures, fontRunFeature * features)
{
	StyleRec	newStyle;
	
	newStyle = *GetDocStyle(dp, styleIndex);
	
	newStyle.textStyle = CopyToStyle(nil, newStyle.textStyle);	
	
	SetStyleRunFeatures(newStyle.textStyle, numFeatures, (RunFeature *) features);
	
	styleIndex = FindDocStyle(dp, &newStyle);
	
	if(styleIndex == -1)
		styleIndex = AddDocStyle(dp, &newStyle);		
		
	DisposeStyle(newStyle.textStyle);
		
	return(styleIndex);	
}

void GetDocStyleFeatures(DocPtr dp, short styleIndex, short * numFeatures, fontRunFeature * features)
{
	StylePtr		sp;
	long			num;
	
	if(styleIndex < 0 || styleIndex >= dp->numStyles)
		gxEditPostFatalError(dp, gx_edit_internal_fatal_error);
		
	sp = *dp->styles + styleIndex;
	
	num = GetStyleRunFeatures(sp->textStyle, (RunFeature *) features);
	
	if(numFeatures != nil)
		*numFeatures = num;
	
}

void ModifyStyleFeatures(DocPtr dp, font fontId, short numFeatureGroups,
			short * srcFeatureGroupSizes,
			fontRunFeature ** srcFeatureGroups,
			short * dstFeatureGroupSizes,
			fontRunFeature ** dstFeatureGroups,
			short numStyles, short * oldStyles, short * newStyles)
{
	short				i, j, k, l;
	short				numFeatures;
	fontRunFeature		* features;
	
	if(numStyles == 0)
		return;
		
	/* now loop through all of the given styles */

	for(i = 0; i < numStyles; i++, oldStyles++, newStyles++) {
	
		*newStyles = *oldStyles;		/* in case we continue */
	
		if(GetDocStyleTextFont(dp, *oldStyles) != fontId)
			continue;
			
		GetDocStyleFeatures(dp, *oldStyles, &numFeatures, nil);
		
		features = (fontRunFeature *) NewPtr(sizeof(fontRunFeature) * numFeatures);
		
		GetDocStyleFeatures(dp, *oldStyles, nil, features);
		
		/* find feature group that matches */
		
		for(j=0; j<numFeatureGroups; j++) {
			
			if(srcFeatureGroupSizes[j] != numFeatures)
				continue;
		
			for(k=0; k<numFeatures; k++) {
			
				for(l=0; l<numFeatures; l++)
					if((srcFeatureGroups[j])[k].featureType == features[l].featureType &&
					   (srcFeatureGroups[j])[k].setting == features[l].setting)
						break;
						
				if(l == numFeatures)
					break;
			
			}
			
			if(k == numFeatures)
				break;
		}
		
		if(j != numFeatureGroups)
			*newStyles = SetDocStyleFeatures(dp, *oldStyles,
									dstFeatureGroupSizes[j], dstFeatureGroups[j]);
									
		DisposePtr((Ptr) features);
				
	}
	
}

void ModifyStyleVariations(DocPtr dp, font fontId,
								short numVariations, fontVariation * variations,
								short numStyles, short * oldStyles, short * newStyles)
{
	StyleRec			newStyle;
	short				i,j,k;
	short				numFontVariations;
	fontVariation		** newVariations;
	fontVariation		* vp;
	fontVariation		* np;
	
	if(numStyles == 0)
		return;
		
	/* now loop through all of the given styles */

	for(i = 0; i < numStyles; i++, oldStyles++, newStyles++) {
	
		*newStyles = *oldStyles;		/* in case we continue */
	
		/* get the soon to be old variations */
	
		newStyle = *GetDocStyle(dp, *oldStyles);
		
		if(newStyle.textFont != fontId)
			continue;
	 
		newStyle.textStyle = CopyToStyle(nil,newStyle.textStyle);
		
		numFontVariations = GetStyleFontVariations(newStyle.textStyle, nil);
		newVariations = (fontVariation **) NewHandle(sizeof(fontVariation) * numFontVariations);
		
		HLock((Handle) newVariations);
		
		GetStyleFontVariations(newStyle.textStyle, *newVariations);
		
		vp = variations;
		
		for(j=0; j<numVariations; j++, vp++) {
		
			np = *newVariations;
			
			for(k=0; k<numFontVariations; k++, np++)
				if(np->name == vp->name) {
					np->value = vp->value;
					break;
				}
				
			if(k == numFontVariations) {
			
				/* we need to add it */
				
				HUnlock((Handle) newVariations);
				
				numFontVariations++;
				SetHandleSize((Handle) newVariations, sizeof(fontVariation) * numFontVariations);
				
				HLock((Handle) newVariations);
				
				(*newVariations)[numFontVariations-1] = *vp;
				
			}
				
		}
		
		SetStyleFontVariations(newStyle.textStyle, numFontVariations, *newVariations);
		
		*newStyles = FindDocStyle(dp, &newStyle);
		
		if(*newStyles == -1)
			*newStyles = AddDocStyle(dp, &newStyle);
			
		HUnlock((Handle) newVariations);
		
		DisposeStyle(newStyle.textStyle);
		DisposeHandle((Handle) newVariations);
		
	}
	
}

void ModifyStyleRunControls(DocPtr dp, RunControls * newRunControls, RunControls * runControlsMask,
								short numStyles, short * oldStyles, short * newStyles)
{
	RunControls			runControls;
	StyleRec			newStyle;
	short				i;
	char				* dst;
	long				size;
	
	if(numStyles == 0)
		return;
		
	/* now loop through all of the given styles */

	for(i = 0; i < numStyles; i++, oldStyles++, newStyles++) {
	
		*newStyles = *oldStyles;		/* in case we don't match */
	
		/* get the soon to be old run features */
	
		newStyle = *GetDocStyle(dp, *oldStyles);
		
		newStyle.textStyle = CopyToStyle(nil, newStyle.textStyle);
		
		if(!GetStyleRunControls(newStyle.textStyle, &runControls)) {
		
			/* initialize style run control to default values -- all zeros */

			size = sizeof(runControls);
			dst = (char *) &runControls;
			
			while(size--)
				*dst++ = 0;
			
		}
		
		if(gxEditSetBitMaskOp(newRunControls, &runControls, runControlsMask,
																sizeof(runControls))) {
						
			SetStyleRunControls(newStyle.textStyle, &runControls);
			
			*newStyles = FindDocStyle(dp, &newStyle);
			
			if(*newStyles == -1)
				*newStyles = AddDocStyle(dp, &newStyle);
				
		}
		
		DisposeStyle(newStyle.textStyle);
		
	}
		
}

void UnionStyleFeatures(DocPtr dp, short numStyles, short *styles,
							font fontId, short * numFeatureGroupsPtr,
							short *** featureGroupSizesPtr,
							fontRunFeature **** featureGroupsPtr)
{
	short			numFeatureGroups;
	fontRunFeature	*** featureGroups;
	short			** featureGroupSizes;
	short			numFeatures;
	fontRunFeature	* features;
	short			i, j, k, l;
	
	numFeatureGroups = 0;
	featureGroups = (fontRunFeature ***) NewHandle(0);
	featureGroupSizes = (short **) NewHandle(0);
	
	for(i=0; i<numStyles; i++, styles++) {
	
		if(GetDocStyleTextFont(dp, *styles) != fontId)
			continue;
			
		/* get the feature group for the style */
		
		GetDocStyleFeatures(dp, *styles, &numFeatures, nil);
		
		features = (fontRunFeature *) NewPtr(sizeof(fontRunFeature) * numFeatures);
		
		GetDocStyleFeatures(dp, *styles, nil, features);
		
		/* try to find the feature set in our list of feature sets */
		
		for(j=0; j<numFeatureGroups; j++) {
		
			if((*featureGroupSizes)[j] != numFeatures)
				continue;
				
			for(k=0; k<numFeatures; k++) {
			
				for(l=0; l<numFeatures; l++)
					if(((*featureGroups)[j])[k].featureType == features[l].featureType &&
					    ((*featureGroups)[j])[k].setting == features[l].setting)
						break;
				
				if(l == numFeatures)
					break;
			}
			
			if(k == numFeatures)
				break;
			
		}
		
		/* if it was not found, then add to our group of feature sets */
		
		if(j == numFeatureGroups) {
		
			PtrAndHand((Ptr) &features, (Handle) featureGroups, sizeof(fontRunFeature *));
			PtrAndHand((Ptr) &numFeatures, (Handle) featureGroupSizes, sizeof(short));
			numFeatureGroups++;
		
		} else {
		
			DisposePtr((Ptr) features);
			
		}
		
	}

	*numFeatureGroupsPtr = numFeatureGroups;
	*featureGroupSizesPtr = featureGroupSizes;
	*featureGroupsPtr = featureGroups;

}

void UnionStyleVariations(DocPtr dp, short numStyles, short *styles,
						font fontId, short * numInstancesPtr,
						short *** instanceSizesPtr, fontVariation **** instancesPtr)
{
	fontVariation	* variation;
	short			numStyleVariations;
	fontVariation	* styleVariations;
	fontVariation	* styleVariation;
	short			i, j, k;
	short			* stylePtr;
	StyleRec		aStyle;
	short			numInstances;
	fontVariation	*** instances, ** instance;
	short			** instanceSizes;
		
	numInstances = 0;
	instances = (fontVariation ***) NewHandle(0);
	instanceSizes = (short **) NewHandle(0);
	
	stylePtr = styles;
	
	for(i=0; i<numStyles; i++, stylePtr++) {
	
		aStyle = *GetDocStyle(dp, *stylePtr);
		
		if(aStyle.textFont != fontId)
			continue;
		
		numStyleVariations = GetStyleFontVariations(aStyle.textStyle, nil);

		styleVariations = (fontVariation *) NewPtr((numStyleVariations) * sizeof(fontVariation));
		
		GetStyleFontVariations(aStyle.textStyle, styleVariations);
		
		
		/* try to find this instance among the list of instances */
		
		instance = *instances;
		
		for(j=0; j<numInstances; j++, instance++) {
		
			if((*instanceSizes)[j] != numStyleVariations)
				continue;
			
			variation = *instance;
			styleVariation = styleVariations;
			
			for(k=0; k<numStyleVariations; k++, variation++, styleVariation++)
				if(variation->name != styleVariation->name ||
				   variation->value != styleVariation->value)
				   	break;
				   	
			if(k != numStyleVariations)
				continue;
				
			break;
			
		}
		
		if(j == numInstances) {
		
			PtrAndHand((Ptr) &styleVariations, (Handle) instances, sizeof(fontVariation *));
			PtrAndHand((Ptr) &numStyleVariations, (Handle) instanceSizes, sizeof(short));
			numInstances++;
			
		} else {
				
			DisposePtr((Ptr) styleVariations);
			
		}
		
	}
	
	*numInstancesPtr = numInstances;
	*instanceSizesPtr = instanceSizes;
	*instancesPtr = instances;
	
}

void UnionStyleRunControls(DocPtr dp, short numStyles, short *styles,
										RunControls * runControls, RunControls * runControlsMask)
{
	short			i;
	short			* stylePtr;
	char 			* dst;
	long			size;
	RunControls		styleRunControls;
	Boolean			firstTime;
	StyleRec		aStyle;
			
	stylePtr = styles;
	
	for(i=0; i<numStyles; i++, stylePtr++) {
	
		aStyle = *GetDocStyle(dp, *stylePtr);
		
		if(!GetStyleRunControls(aStyle.textStyle, &styleRunControls)) {
		
			/* set the default values -- all zeros */
			
			size = sizeof(styleRunControls);
			dst = (char *) &styleRunControls;
			
			while(size--)
				*dst++ = 0;
		
		}
		
		firstTime = (i == 0);
		
		gxEditGetBitMaskOp(&styleRunControls, runControls, runControlsMask, 
								sizeof(styleRunControls), firstTime);
										
	}
	
	/* fix up the mask */
	
	if(~runControlsMask->beforeWithStreamShift)
		runControlsMask->beforeWithStreamShift = 0;

	if(~runControlsMask->afterWithStreamShift)
		runControlsMask->afterWithStreamShift = 0;
	
	if(~runControlsMask->crossStreamShift)
		runControlsMask->crossStreamShift = 0;
		
	if(~runControlsMask->imposedWidth)
		runControlsMask->imposedWidth = 0;
		
	if(~runControlsMask->track)
		runControlsMask->track = 0;
		
	if(~runControlsMask->hangingInhibitFactor)
		runControlsMask->hangingInhibitFactor = 0;
		
	if(~runControlsMask->kerningInhibitFactor)
		runControlsMask->kerningInhibitFactor = 0;
	
	if(~runControlsMask->decompositionAdjustmentFactor)
		runControlsMask->decompositionAdjustmentFactor = 0;
		
	if(~runControlsMask->baselineType)
		runControlsMask->baselineType = 0;
		
}
