/* graphics libraries:
	offscreen support library
	by Cary Clark, Georgiann Delaney, Michael Fairman, Dave Good, Robert Johnson, Keith McGreggor, Oliver Steele, David Van Brink, Chris Yerga
	Copyright 1987 - 1991 Apple Computer, Inc.  All rights reserved.	*/

	#include <Memory.h>
#include "graphics libraries.h"
#include "offscreen library.h"
#include "memory routines.h"

typedef struct viewPortBufferRecord {
	viewPort		masterPort;
	viewPort		slavePort;
	shape		area;
	shape		draw;
	viewGroup	group;
	long			deviceCount;
	viewDevice	devices[1];
} viewPortBufferRecord;

/*
	Given a bitmap shape, this routine fills out the offscreen structure with the objects necessary to draw into
	the offscreen and transfer the offscreen onscreen.  See the header file for details about the offscreen structure.
*/
void CreateOffscreen(register offscreen *os, shape bitShape)
{
	NilParamReturn(os);
	FillBytes(os, sizeof(offscreen), 0);
	NilShapeReturn(bitShape);
	os->group = NewViewGroup();
	os->device = NewViewDevice(os->group, bitShape);
	os->port = NewViewPort(os->group);
	os->draw = GetViewDeviceBitmap(os->device);			/* draw this shape to copy the offscreen to the screen */
	os->xform = NewTransform();						/* assign this to shapes and they will draw offscreen */
	SetTransformViewPorts(os->xform, 1, &os->port);
}


/*
	When you're done with the offscreen, call this routine.
*/
void DisposeOffscreen(offscreen *os)
{
	NilParamReturn(os);
	DisposeShape(os->draw);			/* dispense with bitmap shape */
	DisposeTransform(os->xform);		/* and transform */
	DisposeViewGroup(os->group);		/* and viewGroup (which disposes viewPort and viewDevice) */
}


void CopyToBitmaps(shape dest, shape source)
{
	offscreen destOff;
	
	NilShapeReturn(dest);
	NilShapeReturn(source);
	CreateOffscreen(&destOff, dest);
	{	transform saved = GetShapeTransform(source);
		
		CloneTransform(saved);
		SetShapeTransform(source, destOff.xform);
		DrawShape(source);
		SetShapeTransform(source, saved);
		DisposeTransform(saved);
	}
	DisposeOffscreen(&destOff);
}

typedef struct {
	bitmap		bits;
	point			offset;
} areaCharacteristics;

static areaCharacteristics GetAreaCharacteristics(shape area, viewDevice device, viewPort port)
{
	areaCharacteristics x;
	rectangle bounds;
	shape bitShape;
	mapping map;

	FillBytes(&x, sizeof(x), 0);
	FillBytes(&bounds, sizeof(bounds), 0);
	FillBytes(&map, sizeof(map), 0);
	bitShape = GetViewDeviceBitmap(device);
	GetShapeGlobalBounds(area, port, nil, &bounds);
	GetBitmap(bitShape, &x.bits, nil);
	if (x.bits.space == indexedSpace)
		CloneColorSet(x.bits.set);
	if (x.bits.profile)
		CloneColorProfile(x.bits.profile);
	DisposeShape(bitShape);
	x.offset.x = bounds.left;
	x.offset.y = bounds.top;
	x.bits.width = FixedRound(bounds.right) - FixedRound(bounds.left);
	x.bits.height = FixedRound(bounds.bottom) - FixedRound(bounds.top);
	InvertMapping(&map, GetViewPortGlobalMapping(port, &map));
	MapPoints(&map, 1, &x.offset);
	return x;
}

static void DisposeAreaCharacteristics(areaCharacteristics *x)
{
	if (x->bits.space == indexedSpace)
		DisposeColorSet(x->bits.set);
	if (x->bits.profile)
		DisposeColorProfile(x->bits.profile);
}

static viewDevice NewDeviceWithAreaCharacteristics(viewGroup group, areaCharacteristics *x)
{
	viewDevice device;
	shape bitShape;
	mapping map;

	NilParamReturnNil(x);
	bitShape = NewBitmap(&x->bits, nil);
	device = NewViewDevice(group, bitShape);
	DisposeShape(bitShape);
	ResetMapping(&map);
	MoveMapping(&map, -x->offset.x, -x->offset.y);
	SetViewDeviceMapping(device, &map);
	return device;
}

static areaCharacteristics GetDeviceAreaCharacteristics(viewDevice device)
{
	areaCharacteristics x;
	shape bitShape;
	mapping map;

	FillBytes(&x, sizeof(x), 0);
	FillBytes(&map, sizeof(map), 0);
	bitShape = GetViewDeviceBitmap(device);
	GetBitmap(bitShape, &x.bits, nil);
	if (x.bits.space == indexedSpace)
		CloneColorSet(x.bits.set);
	if (x.bits.profile)
		CloneColorProfile(x.bits.profile);
	DisposeShape(bitShape);
	GetViewDeviceMapping(device, &map);
	x.offset.x = -map.map[2][0];
	x.offset.y = -map.map[2][1];
	return x;
}

static void SetDeviceAreaCharacteristics(viewDevice device, areaCharacteristics *x)
{
	shape bitShape;
	bitmap bits;
	mapping map;

	NilParamReturn(x);
	bits.width = 1;
	bits.height = 1;
	bits.pixelSize = 1;
	bits.rowBytes = 0;
	bits.image = nil;
	bits.space = indexedSpace;
	bits.set = nil;
	bits.profile = nil;
	bitShape = NewBitmap(&bits, nil);
	SetViewDeviceBitmap(device, bitShape);
	DisposeShape(bitShape);

	bitShape = NewBitmap(&x->bits, nil);
	SetViewDeviceBitmap(device, bitShape);
	DisposeShape(bitShape);
	ResetMapping(&map);
	MoveMapping(&map, -x->offset.x, -x->offset.y);
	SetViewDeviceMapping(device, &map);
}

static boolean EqualAreaCharacteristics(areaCharacteristics *x, areaCharacteristics *y)
{
	NilParamReturnNil(x);
	NilParamReturnNil(y);
	return	x->bits.space == y->bits.space
		&&	x->bits.set == y->bits.set
		&&	x->bits.profile == y->bits.profile
		&&	x->offset.x == y->offset.x
		&&	x->offset.y == y->offset.y
		&&	x->bits.width == y->bits.width
		&&	x->bits.height == y->bits.height
		&&	x->bits.pixelSize == y->bits.pixelSize;
}

static void AddBitmapShapes(viewPortBufferRecord *buffers)
{
	short i;

	NilParamReturn(buffers);
	for (i = 0; i < buffers->deviceCount; i++) 
	{	viewDevice device = buffers->devices[i];
		shape bitShape = GetViewDeviceBitmap(device);
		mapping map;

		FillBytes(&map, sizeof(map), 0);
		GetViewDeviceMapping(device, &map);
		SetShapeAttributes(bitShape, GetShapeAttributes(bitShape) | mapTransformShape);
		MoveShape(bitShape, -map.map[2][0], -map.map[2][1]);
		SetShapeParts(buffers->draw, 0, 0, bitShape, 0);
		DisposeShape(bitShape);
	}
}

viewPortBuffer NewViewPortBuffer(viewPort port)
{
	viewPortBuffer buffersHandle;
	viewPortBufferRecord *buffers;
	transform xform;
	shape area;
	long deviceCount;
	short i;

	NilParamReturnNil(port);
	area = NewShape(fullType);
	xform = NewTransform();
	SetTransformViewPorts(xform, 1, &port);
	SetShapeTransform(area, xform);
	DisposeTransform(xform);
	deviceCount = GetShapeGlobalViewDevices(area, port, nil);
	buffersHandle = (viewPortBuffer) NewHandle(sizeof(viewPortBufferRecord) + (deviceCount - 1) * sizeof(viewDevice));
	NilParamReturnNil(buffersHandle);
	HLock((Handle) buffersHandle);
	buffers = *buffersHandle;
	buffers->group = NewViewGroup();
	buffers->masterPort = port;
	buffers->slavePort = NewViewPort(buffers->group);
	buffers->area = area;
	buffers->draw = NewShape(pictureType);
	buffers->deviceCount = deviceCount;
	SetViewPortDither(buffers->slavePort, GetViewPortDither(port));
	GetShapeGlobalViewDevices(area, port, buffers->devices);
	for (i = 0; i < deviceCount; i++) 
	{	viewDevice device = buffers->devices[i];
		areaCharacteristics x = GetAreaCharacteristics(area, device, port);
		
		buffers->devices[i] = NewDeviceWithAreaCharacteristics(buffers->group, &x);
		DisposeAreaCharacteristics(&x);
	}
	AddBitmapShapes(buffers);
	HUnlock((Handle) buffersHandle);
	return buffersHandle;
}

void DisposeViewPortBuffer(viewPortBuffer buffersHandle)
{
	NilParamReturn(buffersHandle);
	DisposeShape((*buffersHandle)->area);
	DisposeShape((*buffersHandle)->draw);
	DisposeViewGroup((*buffersHandle)->group);
	DisposHandle((Handle) buffersHandle);
}

boolean ValidViewPortBuffer(viewPortBuffer buffersHandle)
{
	long deviceCount;
	viewDevice *devices;
	boolean valid = true;
	short i;

	NilParamReturnNil(buffersHandle);
	deviceCount = GetShapeGlobalViewDevices((*buffersHandle)->area, (*buffersHandle)->masterPort, nil);
	if (deviceCount != (*buffersHandle)->deviceCount)
		return false;
	devices = (viewDevice *) NewPtr(deviceCount * sizeof(viewDevice));
	NilParamReturnNil(devices);
	for (i = 0; i < deviceCount; i++) 
	{	areaCharacteristics x = GetAreaCharacteristics((*buffersHandle)->area, devices[i], (*buffersHandle)->masterPort);
		areaCharacteristics y = GetDeviceAreaCharacteristics((*buffersHandle)->devices[i]);
		
		valid = EqualAreaCharacteristics(&x, &y);
		DisposeAreaCharacteristics(&x);
		DisposeAreaCharacteristics(&y);
		if (!valid) break;
	}
	DisposPtr((Ptr) devices);
	return valid;
}

boolean UpdateViewPortBuffer(viewPortBuffer buffersHandle)
{
	viewPortBufferRecord *buffers;
	short state;
	long deviceCount;
	viewDevice *devices;
	boolean valid = true;
	short i;

	NilParamReturnNil(buffersHandle);
	state = HGetState((Handle) buffersHandle);
	HLock((Handle) buffersHandle);
	buffers = *buffersHandle;
	deviceCount = GetShapeGlobalViewDevices(buffers->area, buffers->masterPort, nil);
	if (deviceCount != buffers->deviceCount) {
		valid = false;
		SetPicture(buffers->draw, 0, nil, nil, nil, nil);
		if (deviceCount < buffers->deviceCount) {
			for (i = deviceCount; i < buffers->deviceCount; i++)
				DisposeViewDevice(buffers->devices[i]);
			SetHandleSize((Handle) buffersHandle, sizeof(viewPortBufferRecord) + (deviceCount - 1) * sizeof(viewDevice));
			buffers = *buffersHandle;
		} else {
			HUnlock((Handle) buffersHandle);
			SetHandleSize((Handle) buffersHandle, sizeof(viewPortBufferRecord) + (deviceCount - 1) * sizeof(viewDevice));
			HLock((Handle) buffersHandle);
			buffers = *buffersHandle;
			for (i = buffers->deviceCount; i < deviceCount; i++)
				buffers->devices[i] = nil;
		}
		buffers->deviceCount = deviceCount;
	}
	devices = (viewDevice *) NewPtr(deviceCount * sizeof(viewDevice));
	NilParamReturnNil(devices);
	deviceCount = GetShapeGlobalViewDevices(buffers->area, buffers->masterPort, devices);
	for (i = 0; i < deviceCount; i++) 
	{	viewDevice device = buffers->devices[i];
		areaCharacteristics x = GetAreaCharacteristics(buffers->area, devices[i], buffers->masterPort);
		
		if (device) 
		{	areaCharacteristics y = GetDeviceAreaCharacteristics(device);
		
			if (EqualAreaCharacteristics(&x, &y) == false) {
				valid = false;
				SetPicture(buffers->draw, 0, nil, nil, nil, nil);
				SetDeviceAreaCharacteristics(device, &x);
			}
			DisposeAreaCharacteristics(&y);
		} else
			buffers->devices[i] = NewDeviceWithAreaCharacteristics(buffers->group, &x);
		DisposeAreaCharacteristics(&x);
	}
	DisposPtr((Ptr) devices);
	if (valid == false)
		AddBitmapShapes(buffers);
	HSetState((Handle) buffersHandle, state);
	return valid;
}

viewPort GetViewPortBufferViewPort(viewPortBuffer buffersHandle)
{
	NilParamReturnNil(buffersHandle);
	return (*buffersHandle)->slavePort;
}

shape GetViewPortBufferShape(viewPortBuffer buffersHandle)
{
	NilParamReturnNil(buffersHandle);
	return (*buffersHandle)->draw;
}
