/* shape controls library:
	library to allow the user to manipulate (transform) a set of shapes
	by Dave Good and Jeff Kreegar
	Copyright 1991 - 1992 Apple Computer, Inc.  All rights reserved.	*/

#include <Memory.h>
#include <Events.h>
#include <OSUtils.h>
#include <ToolUtils.h>

#include "graphics debugging.h"
#include "graphics errors.h"
#include "graphics toolbox.h"
#include "memory routines.h"
#include "graphics libraries.h"
#include "shape controls library.h"


/*----------------------------------------------------------------------------------------------------*/
/* private constants and structures */
/*----------------------------------------------------------------------------------------------------*/


/*** this is not used or completed; however it is a placeholder for the future */
typedef struct portBuffer {
	viewGroup	group;
} portBuffer;


typedef struct shapeControlRecord {
	/* ----- can be accessed in a public way ----- */
	shape		items;
	shape		foreground;
	shape		background;
	shape		selection;
	/* ----- private fields follow ----- */
	viewPort		**originalPorts;
	shape		controls;
	point			savedCenter;			/* this is either (positiveInfinity, positiveInfinity) or the saved rotate center */
	long			buildHiddenLevel;
	rectangle		**invalidAreas;
} shapeControlRecord, *shapeControlPtr;


#define	controlHandleGap		ff(6)		/* this is the distance between the bounds of the shape and the bounds of the control handles */
#define	minimumObjectSizeH		ff(2)		/* this is the smallest horizontal size that an object may be scaled to */
#define	minimumObjectSizeV		ff(2)		/* this is the smallest vertical size that an object may be scaled to */
#define	rotateControlRadius		ff(7)
#define	rotateCenterSize		ff(2)


typedef enum {
	/* ----- special commands ----- */
	firstSpecialCommand = -1,
	commandSelect = firstSpecialCommand,
	commandMove,
	lastSpecialCommand = commandMove,
	/* ----- indexes into the shapeControl->controls picture that correspond to different commands ----- */
	firstIndexCommand = 1,
	commandRotate = firstIndexCommand,
	commandMoveRotateCenter,
	commandScale,
	commandSkew = commandScale + 4,
	lastIndexCommand = commandSkew + 3,
	/* ----- modifiers for commandScale follow ----- */
	leftTop = 0,
	rightTop,
	rightBottom,
	leftBottom,
	/* ----- modifiers for commandSkew follow ----- */
	centerTop = 0,
	centerRight,
	centerBottom,
	centerLeft
} shapeControlCommands;


/*----------------------------------------------------------------------------------------------------*/
/* private generic routines */
/*----------------------------------------------------------------------------------------------------*/


static fixed FixedAbs(fixed value)
{
	if( value < 0 )
		return -value;
	return value;
}

static viewPort *DereferenceViewPortList(viewPort **list)
{
	HLock((Handle)list);
	return *list;
}

static void ReleaseViewPortList(viewPort **list)
{
	HUnlock((Handle)list);
}

static viewPort **NewViewPortListFromShape(shape source)
{
	viewPort **list = (viewPort **)NewHandleClear((GetShapeViewPorts(source, nil) + 1) * sizeof(viewPort));

	IfErrorReturnNil(list == nil, out_of_memory);
	GetShapeViewPorts(source, DereferenceViewPortList(list));
	ReleaseViewPortList(list);
	return list;
}

static long CountViewPortList(viewPort **list)
{
	return (GetHandleSize((Handle)list) / sizeof(viewPort)) - 1;
}

static void DisposeViewPortList(viewPort **list)
{
	NilParamReturn(list);
	DisposeHandle((Handle)list);
}

static boolean ShapeInPicture(shape test, shape picture)
{
	long count;

	NilShapeReturnNil(test);
	NilShapeReturnNil(picture);

	count = GetPicture(picture, nil, nil, nil, nil);
	while( count ) {
		shape item;

		GetPictureParts(picture, count, 1, &item, nil, nil, nil);
		if( item == test )
			return true;
		--count;
	}
	return false;
}

static void GetPictureBounds(shape picture, rectangle *bounds, mapping *concatMatrix)
{
	long count = GetPicture(picture, nil, nil, nil, nil);

	/* if there are no items in the picture, then there is nothing to do */
	if( count == 0 )
		return;

	do {	shape item;
		transform itemXform;
		mapping matrix;

		GetPictureParts(picture, count, 1, &item, nil, nil, &itemXform);

		/* if there was no overriding transform, then use the items transform. calculate the concatenated mapping */
		if( itemXform == nil )
			itemXform = GetShapeTransform(item);
		MapMapping(GetTransformMapping(itemXform, &matrix), concatMatrix);

		if( GetShapeType(item) == pictureType ) {
			GetPictureBounds(item, bounds, &matrix);
		} else {
			rectangle itemBounds;
			mapping tempMatrix;
			shape boundsShape = NewRectangle( GetShapeLocalBounds(item, &itemBounds) );

			/* map our bounds rectangle through the concatenated mapping. note that since GetShapeLocalBounds already
			maps the bounds by the items transform, we need to undo this effect. */
			IgnoreGraphicsNotice(transform_mapping_unaffected);
				MapMapping(&matrix, InvertMapping(&tempMatrix, GetTransformMapping(GetShapeTransform(item), &tempMatrix)));
				MapShape(boundsShape, &matrix);
			PopGraphicsNotice();

			/* get the bounds of our transformed bounds rectangle and then dispose of the temporary bounds shape */
			GetShapeBounds(boundsShape, 0L, &itemBounds);
			DisposeShape(boundsShape);

			/* union this items bounds into the main picture bounds (work around a problem with UnionRectangle) */
			if( bounds->right < bounds->left || bounds->bottom < bounds->top )
				*bounds = itemBounds;
			else
				UnionRectangle(bounds, bounds, &itemBounds);
		}
	} while( --count );
}

static rectangle *GetShapeTransformedBounds(shape source, rectangle *bounds)
{
	mapping matrix;

	/* since graphics currently cant get the correct bounds for a picture shape, we need special code */
	if( GetShapeType(source) == pictureType ) {
		bounds->left = bounds->top = positiveInfinity;
		bounds->right = bounds->bottom = negativeInfinity;
		GetPictureBounds( source, bounds, GetShapeMapping(source, &matrix) );
		return bounds;
	}

	/* otherwise, just use GetShapeLocalBounds */
	return GetShapeLocalBounds(source, bounds);
}

static point *GetShapeTransformedCenter(shape source, point *center)
{
	mapping matrix;

	/* since graphics currently cant calculate the center of picture shapes, we need special code */
	if( GetShapeType(source) == pictureType ) {
		rectangle bounds;

		GetShapeTransformedBounds(source, &bounds);
		center->x = (bounds.right >> 1) + (bounds.left >> 1);
		center->y = (bounds.top >> 1) + (bounds.bottom >> 1);
	} else
		MapPoints( GetShapeMapping(source, &matrix), 1, GetShapeCenter(source, 0L, center) );
	return center;
}


/*----------------------------------------------------------------------------------------------------*/
/* private internal routines */
/*----------------------------------------------------------------------------------------------------*/


#if 0
static boolean AllocateOffscreen(shapeControl source)
{
	return false;
	bitmap deepestBits;
	shape area;

	/* this ensures that the first device that we find will be put in deepestBits */
	deepestBits.pixelSize = -1;

	area = NewShape(fullType);
	SetShapeViewPorts(area, portCount, ports);
	while( portCount-- ) {
		viewDevice devices[maximumViewThings];
		long deviceCount;

		if( (deviceCount = GetShapeGlobalViewDevices(area, ports[portCount], devices)) > maximumViewThings )
			DebugStr("\pexceded the maximum number of viewDevices for a shape control");
		while( deviceCount-- ) {
			shape bitsShape;
			bitmap bits;

			bitsShape = GetViewDeviceBitmap(devices[deviceCount]);
			GetBitmap(bitsShape, &bits, nil);
			if( bits.pixelSize > deepestBits.pixelSize )
				deepestBits = bits;
			DisposeShape(bitsShape);
		}
	}
	DisposeShape(area);

	if( deepestBits.pixelSize > 0 ) {
		shape bitsShape;

		deepestBits.image = nil;	/* make graphics allocate our offscreen bit image */
		bitsShape = NewBitmap(&deepestBits, nil);
		CreateOffscreen(&source->buffer, bitsShape);
		DisposeShape(bitsShape);
	} else
		FillBytes(&source->buffer, sizeof(source->buffer), 0);

	/*** needs work */
}
#endif

static void InvalidateRectangle(shapeControlPtr targetPtr, rectangle *newArea)
{
	if( targetPtr->invalidAreas ) {
		rectangle *listPtr = *targetPtr->invalidAreas;
		rectangle *maxPtr = (rectangle *)((char *)listPtr + GetHandleSize((Handle)targetPtr->invalidAreas));

		while( listPtr != maxPtr &&
			(listPtr->top < newArea->top || listPtr->top == newArea->top && listPtr->left < newArea->left) ) {
			++listPtr;
		}
		Munger((Handle)targetPtr->invalidAreas, (char *)listPtr - (char *)*targetPtr->invalidAreas, nil, 0,
			(Ptr)newArea, sizeof(rectangle));
	} else {
		targetPtr->invalidAreas = (rectangle **)NewHandle(sizeof(rectangle));
		**targetPtr->invalidAreas = *newArea;

		/*** reduce the rectangle list here */
	}

#if 0  /* re-enable this to show which areas will be updated */
	{	shape invalidShape = NewRectangle(newArea);
		IgnoreGraphicsNotice(color_already_set);
			SetShapeCommonColor(invalidShape, gxBlack);
		PopGraphicsNotice();
		DrawShape(invalidShape);
		DisposeShape(invalidShape);
	}
#endif
}

static void InvalidatePicture(shapeControlPtr targetPtr, shape control)
{
	/*** later, we need to fix the picture code below and then only invalidate the bounds of the sub-items */
	if( false && GetShapeType(control) == pictureType ) {
		long count = GetPicture(control, nil, nil, nil, nil);
		while( count > 0 ) {
			shape item;
			transform itemXform, originalXform;
			rectangle bounds;

			GetPictureParts(control, count, 1, &item, nil, nil, &itemXform);
			if( itemXform ) {
				originalXform = GetShapeTransform(item);
				SetShapeTransform(item, itemXform);
			}
			InvalidateRectangle(targetPtr, GetShapeTransformedBounds(item, &bounds));
			if( itemXform )
				SetShapeTransform(item, originalXform);
			--count;
		}
	} else {
		rectangle bounds;
		InvalidateRectangle(targetPtr, GetShapeTransformedBounds(control, &bounds));
	}
}

static void RedrawShapeControl(shapeControlPtr targetPtr)
{
	if( targetPtr->invalidAreas ) {
		rectangle *listPtr, *maxPtr;

		/* lock down our handle of rectangular invalid areas */
		HLock((Handle)targetPtr->invalidAreas);
		listPtr = *targetPtr->invalidAreas;
		maxPtr = (rectangle *)((char *)listPtr + GetHandleSize((Handle)targetPtr->invalidAreas));

		do {	shape clip;
			viewPort *portList, *childList;
			viewPort **childPorts;

			/* make a viewPort list for all of our child ports (these will hold our invalid clip) */
			childPorts = (viewPort **)NewHandleClear( GetHandleSize((Handle)targetPtr->originalPorts) );
			childList = DereferenceViewPortList(childPorts);

			/* clip all of the viewPorts that we are drawing into to our current invalid rectangle */
			clip = NewRectangle(listPtr);
			portList = DereferenceViewPortList(targetPtr->originalPorts);
			while( *portList ) {
				*childList = NewViewPort(GetViewPortViewGroup(*portList));
				SetViewPortParent(*childList, *portList);
				SetViewPortClip(*childList, clip);
				++portList;
				++childList;
			}
			ReleaseViewPortList(targetPtr->originalPorts);
			DisposeShape(clip);

			/* re-direct all of our drawing into the child viewPorts */
			IgnoreGraphicsNotice(transform_viewPorts_already_set); {
				long portCount = CountViewPortList(childPorts);

				childList = *childPorts;
				SetShapeViewPorts(targetPtr->background, portCount, childList);
				if( targetPtr->foreground )
					SetShapeViewPorts(targetPtr->foreground, portCount, childList);
				if( targetPtr->controls )
					SetShapeViewPorts(targetPtr->controls, portCount, childList);
			} PopGraphicsNotice();

			DrawShape(targetPtr->background);

			/* draw all the items in the picture, one at a time in the proper order using our clip */
		#if 1
			{	long count = GetPicture(targetPtr->items, nil, nil, nil, nil);
				long index;

				for(index = 1; index <= count; ++index ) {
					viewPort **itemList;
					shape item;

					GetPictureParts(targetPtr->items, index, 1, &item, nil, nil, nil);

					/* save the items viewPort list and re-direct the item into our childPorts */
					itemList = NewViewPortListFromShape(item);
					SetShapeViewPorts(item, CountViewPortList(childPorts), *childPorts);

					DrawShape(item);

					/* restore the items viewPort list and throw away our saved list memory */
					DereferenceViewPortList(itemList);
					SetShapeViewPorts(item, CountViewPortList(itemList), *itemList);
					DisposeViewPortList(itemList);
				}
			}
		#else
			DrawShape(targetPtr->items);
		#endif

			if( targetPtr->foreground )
				DrawShape(targetPtr->foreground);
			if( targetPtr->controls )
				DrawShape(targetPtr->controls);

			/* re-direct all of our drawing back into the main viewPorts */
			IgnoreGraphicsNotice(transform_viewPorts_already_set); {
				long portCount = CountViewPortList(targetPtr->originalPorts);

				portList = DereferenceViewPortList(targetPtr->originalPorts);
				SetShapeViewPorts(targetPtr->background, portCount, portList);
				if( targetPtr->foreground )
					SetShapeViewPorts(targetPtr->foreground, portCount, portList);
				if( targetPtr->controls )
					SetShapeViewPorts(targetPtr->controls, portCount, portList);
				ReleaseViewPortList(targetPtr->originalPorts);
			} PopGraphicsNotice();

			/* dispose of all the viewPorts in the child list and then dispose of the child list itself */
			childList = DereferenceViewPortList(childPorts);
			while( *childList )
				DisposeViewPort(*childList++);
			ReleaseViewPortList(childPorts);
			DisposeViewPortList(childPorts);
		} while( ++listPtr < maxPtr );

		/* throw away all the invalid areas since we have just updated them */
		DisposeHandle((Handle)targetPtr->invalidAreas);
		targetPtr->invalidAreas = nil;
	}
}

static void BuildSelectionControls(shapeControlPtr targetPtr)
{
	rectangle bounds, controlBounds;
	shape controls, workShape;

	/* dont re-build anything if building is hidden */
	if( targetPtr->buildHiddenLevel )
		return;

	/* throw away the old controls and see if there is a selection; if not, then we are done */
	DisposeShapeAt(&targetPtr->controls);
	if( targetPtr->selection == nil )
		return;

	/* create a new controls picture in all of our viewPorts */
	controls = targetPtr->controls = NewShape(pictureType);
	{ viewPort *portList = DereferenceViewPortList(targetPtr->originalPorts);
	SetShapeViewPorts(targetPtr->controls, CountViewPortList(targetPtr->originalPorts), portList);
	ReleaseViewPortList(targetPtr->originalPorts); }

	/* get the bounds of the entire selection and calculate the controlBounds */
	GetShapeTransformedBounds(targetPtr->selection, &bounds);
	IfDebug(bounds.right < bounds.left || bounds.bottom < bounds.top, "\pthe selection should always have a bounds here");
	controlBounds.left = bounds.left - controlHandleGap;
	controlBounds.top = bounds.top - controlHandleGap;
	controlBounds.right = bounds.right + controlHandleGap;
	controlBounds.bottom = bounds.bottom + controlHandleGap;

	/* create the scale controls (these are simply rotated images of each other) */
	{	static long scalePoly[] =	{ 1, 4,
								ff(0), ff(0),
								ff(5) + 0x5000, ff(0),
								ff(0), ff(5) + 0x5000,
								ff(0), ff(0)
							};
		shape tempShape;

		workShape = NewPolygons((polygons *)scalePoly);
		SetShapeFill(workShape, windingFill);
		IgnoreGraphicsNotice(color_already_set);
			SetShapeCommonColor(workShape, gxBlack);
		PopGraphicsNotice();

		tempShape = CopyToShape(nil, workShape);
		MoveShapeTo(tempShape, controlBounds.left, controlBounds.top);
		SetPictureParts(controls, 0, 0, 1, &tempShape, nil, nil, nil);
		DisposeShape(tempShape);

		RotateShape(workShape, ff(90), ff(0), ff(0));
		tempShape = CopyToShape(nil, workShape);
		MoveShapeTo(tempShape, controlBounds.right, controlBounds.top);
		SetPictureParts(controls, 0, 0, 1, &tempShape, nil, nil, nil);
		DisposeShape(tempShape);

		RotateShape(workShape, ff(90), ff(0), ff(0));
		tempShape = CopyToShape(nil, workShape);
		MoveShapeTo(tempShape, controlBounds.right, controlBounds.bottom);
		SetPictureParts(controls, 0, 0, 1, &tempShape, nil, nil, nil);
		DisposeShape(tempShape);

		RotateShape(workShape, ff(90), ff(0), ff(0));
		tempShape = CopyToShape(nil, workShape);
		MoveShapeTo(tempShape, controlBounds.left, controlBounds.bottom);
		SetPictureParts(controls, 0, 0, 1, &tempShape, nil, nil, nil);
		DisposeShape(tempShape);

		DisposeShape(workShape);
	}

	/* create the skew controls in (top, right, bottom, left) order */
	{	static long skewPoly[] =	{ 1, 5,
								-ff(3), ff(0),
								ff(0), -ff(3),
								ff(3), ff(0),
								ff(0), ff(3),
								-ff(3), ff(0)
							};
		shape tempShape;
		point center;

		center.x = (bounds.right >> 1) + (bounds.left >> 1);
		center.y = (bounds.top >> 1) + (bounds.bottom >> 1);

		workShape = NewPolygons((polygons *)skewPoly);
		SetShapeFill(workShape, windingFill);
		IgnoreGraphicsNotice(color_already_set);
			SetShapeCommonColor(workShape, gxBlack);
		PopGraphicsNotice();

		tempShape = CopyToShape(nil, workShape);
		MoveShape(tempShape, center.x, controlBounds.top);
		SetPictureParts(controls, 0, 0, 1, &tempShape, nil, nil, nil);
		DisposeShape(tempShape);

		tempShape = CopyToShape(nil, workShape);
		MoveShape(tempShape, controlBounds.right + ff(1), center.y);
		SetPictureParts(controls, 0, 0, 1, &tempShape, nil, nil, nil);
		DisposeShape(tempShape);

		tempShape = CopyToShape(nil, workShape);
		MoveShape(tempShape, center.x, controlBounds.bottom + ff(3));
		SetPictureParts(controls, 0, 0, 1, &tempShape, nil, nil, nil);
		DisposeShape(tempShape);

		tempShape = CopyToShape(nil, workShape);
		MoveShape(tempShape, controlBounds.left, center.y);
		SetPictureParts(controls, 0, 0, 1, &tempShape, nil, nil, nil);
		DisposeShape(tempShape);

		DisposeShape(workShape);
	}

	/* create the rotate control and its center point */
	{	rectangle tempRect;
		point center;

		/* get the center of the selection (or the individual shape) for the rotate control and its center point */
		if( targetPtr->savedCenter.x != positiveInfinity ) {
			center = targetPtr->savedCenter;
		} else if( GetPicture(targetPtr->selection, nil, nil, nil, nil) == 1 ) {
			shape item;

			GetPicture(targetPtr->selection, &item, nil, nil, nil);
			GetShapeTransformedCenter(item, &center);
		} else {
			center.x = (bounds.right >> 1) + (bounds.left >> 1);
			center.y = (bounds.top >> 1) + (bounds.bottom >> 1);
		}

		/* create the rotate control */
		tempRect.left		= center.x - rotateControlRadius;
		tempRect.top		= center.y - rotateControlRadius;
		tempRect.right		= center.x + rotateControlRadius;
		tempRect.bottom	= center.y + rotateControlRadius;
		workShape = NewOval(&tempRect);
		SetShapeFill(workShape, closedFrameFill);
		IgnoreGraphicsNotice(color_already_set);
			SetShapeCommonColor(workShape, gxBlack);
		PopGraphicsNotice();
	#if useFilledOval
		SetShapePen(workShape, ff(2));
		PrimitiveShape(workShape);
		SetShapeFill(workShape, windingFill);
	#endif
		SetPictureParts(controls, 1, 0, 1, &workShape, nil, nil, nil);
		DisposeShape(workShape);

		/* create the rotate center */
		tempRect.left		= center.x - rotateCenterSize;
		tempRect.top		= center.y - rotateCenterSize;
		tempRect.right		= center.x + rotateCenterSize;
		tempRect.bottom	= center.y + rotateCenterSize;
		workShape = NewOval(&tempRect);
		SetShapeFill(workShape, windingFill);
		IgnoreGraphicsNotice(color_already_set);
			SetShapeCommonColor(workShape, gxBlack);
		PopGraphicsNotice();
		SetPictureParts(controls, 2, 0, 1, &workShape, nil, nil, nil);
		DisposeShape(workShape);
	}

	/* reset the savedCenter to the no center saved value */
	targetPtr->savedCenter.x = targetPtr->savedCenter.y = positiveInfinity;
}

static void TrackShapeControl(shapeControlPtr targetPtr, long command, point originalClick)
{
	boolean hiddenControls = false;
	boolean rebuildControls = false;
	polar originalPolar;
	point originalSize;
	point pivot;
	point adjust;
	mapping oldMatrix;
	point oldMouse = originalClick;
	fixed polySrc[10];
	fixed polyDst[10];

	/* if we are rotating, then calculate the pivot point for the rotation and save the original angle */
	if( command == commandRotate ) {
		shape centerShape;
		point tempPoint;

		GetPictureParts(targetPtr->controls, commandMoveRotateCenter, 1, &centerShape, nil, nil, nil);
		GetShapeTransformedCenter(centerShape, &pivot);
		tempPoint.x = originalClick.x - pivot.x;
		tempPoint.y = originalClick.y - pivot.y;
		PointToPolar(&tempPoint, &originalPolar);
		targetPtr->savedCenter = pivot;
	} else if( command >= commandScale && command <= commandSkew + 3 ) {
		rectangle bounds;

		/* get the bounds of the entire selection for all the commands that require it */
		GetShapeTransformedBounds(targetPtr->selection, &bounds);
		IfDebug(bounds.right < bounds.left || bounds.bottom < bounds.top, "\pthe selection should always have a bounds here");

		/* calculate the originalSize of the selection and set the original scale mapping to identity */
		originalSize.x = bounds.right - bounds.left;
		originalSize.y = bounds.bottom - bounds.top;
		ResetMapping(&oldMatrix);

		/* calculate the pivot (or anchor) point for the scale */
		switch( command ) {
			case commandScale + leftTop:
				pivot.x = bounds.right;
				pivot.y = bounds.bottom;
				break;
			case commandScale + rightTop:
				pivot.x = bounds.left;
				pivot.y = bounds.bottom;
				break;
			case commandScale + rightBottom:
				pivot.x = bounds.left;
				pivot.y = bounds.top;
				break;
			case commandScale + leftBottom:
				pivot.x = bounds.right;
				pivot.y = bounds.top;
				break;
			default:
				polySrc[0] = 4;
				polySrc[1] = bounds.left;
				polySrc[2] = bounds.top;
				polySrc[3] = bounds.right;
				polySrc[4] = bounds.top;
				polySrc[5] = bounds.right;
				polySrc[6] = bounds.bottom;
				polySrc[7] = bounds.left;
				polySrc[8] = bounds.bottom;
				MoveBytes(&polySrc[0], &polyDst[0], sizeof(polySrc));
		}

		/* this is the amount that we must adjust the mouse position by in order to have the scaling come out correct */
		adjust.x = FixedAbs(originalClick.x - pivot.x) - originalSize.x;
		adjust.y = FixedAbs(originalClick.y - pivot.y) - originalSize.y;
	}

	while( StillDown() ) {
		point newMouse;

		GetViewPortMouse(0, &newMouse);
		if( newMouse.x != oldMouse.x || newMouse.y != oldMouse.y ) {
			fixed deltaX = newMouse.x - oldMouse.x;
			fixed deltaY = newMouse.y - oldMouse.y;
			mapping newMatrix;
			rectangle bounds;
			long count;
			shape item;

			/* if the controls are not hidden yet, then do so now, unless we are moving the rotate center */
			if( command != commandMoveRotateCenter && ! hiddenControls ) {
				/* set our flag so that we wont try to hide the controls again */
				hiddenControls = true;

				/* mark the area covered by the controls as invalid and then temporarily hide the controls picture */
				InvalidatePicture(targetPtr, targetPtr->controls);
				SetShapeFill(targetPtr->controls, noFill);
				RedrawShapeControl(targetPtr);
			}

			switch( command ) {

				case commandSelect:
					/*** needs work */
					break;

				case commandMove:
					/* move all the items in the selection, marking their old and new bounds as redraw areas */
					count = GetPicture(targetPtr->selection, nil, nil, nil, nil);
					IfDebug(count == 0, "\pthe selection should always have items in it here");
					do {
						GetPictureParts(targetPtr->selection, count, 1, &item, nil, nil, nil);
						InvalidateRectangle(targetPtr, GetShapeTransformedBounds(item, &bounds));
			 			MoveShape(item, deltaX,  deltaY);
						InvalidateRectangle(targetPtr, GetShapeTransformedBounds(item, &bounds));
					} while( --count );

					/* move the controls for all the selected items and update the screen to reflect the new position */
					MoveShape(targetPtr->controls, deltaX, deltaY);
					RedrawShapeControl(targetPtr);
					break;

				case commandRotate:
				{	point tempPoint;
					polar newPolar;

					tempPoint.x = newMouse.x - pivot.x;
					tempPoint.y = newMouse.y - pivot.y;
					PointToPolar(&tempPoint, &newPolar);

					/* rotate all the items in the selection, marking their old and new bounds as redraw areas */
					count = GetPicture(targetPtr->selection, nil, nil, nil, nil);
					IfDebug(count == 0, "\pthe selection should always have items in it here");
					do {
						GetPictureParts(targetPtr->selection, count, 1, &item, nil, nil, nil);
						InvalidateRectangle(targetPtr, GetShapeTransformedBounds(item, &bounds));
			 			RotateShape(item, newPolar.angle - originalPolar.angle, pivot.x, pivot.y);
						InvalidateRectangle(targetPtr, GetShapeTransformedBounds(item, &bounds));
					} while( --count );

					/* copy the newPolar into the originalPolar for our next pass through the loop */
					originalPolar = newPolar;

					/* invalidate the controls for all the selected items and update the screen to reflect the new position */
					rebuildControls = true;
					RedrawShapeControl(targetPtr);
					break;
				}

				case commandMoveRotateCenter:
				{	long index;

					/* move the rotate center control part and the rotate control part */
					for(index = commandRotate; index <= commandMoveRotateCenter; ++index) {
						shape control;
						transform xform;
						long pass;

						GetPictureParts(targetPtr->controls, index, 1, &control, nil, nil, &xform);
						/* invalidate the position before the move and the position after the move */
						for( pass = 0; pass <= 1; ++pass ) {
							shape boundsShape;
							transform originalXform;
							mapping matrix;

							if( xform ) {
								originalXform = GetShapeTransform(control);
								SetShapeTransform(control, xform);
							}
							boundsShape = NewRectangle(GetShapeTransformedBounds(control, &bounds));
							if( xform )
								SetShapeTransform(control, originalXform);
							MapShape(boundsShape, GetShapeMapping(targetPtr->controls, &matrix));
							GetShapeBounds(boundsShape, 0L, &bounds);
							DisposeShape(boundsShape);
							InvalidateRectangle(targetPtr, &bounds);

							/* if we are on our first pass, then move the control */
							if( pass == 0 )
								MoveShape(control, deltaX, deltaY);
						}
					}

					/* update the screen to reflect the new positions of these controls */
					RedrawShapeControl(targetPtr);
					break;
				}

				/* scale the shape around the pivot point */
				case commandScale + leftTop:
				case commandScale + rightTop:
				case commandScale + rightBottom:
				case commandScale + leftBottom:
				{	fixed scaleX, scaleY;

					/* calculate the scale values for the selection */
					scaleX = FixedDivide(FixedAbs(newMouse.x - pivot.x) - adjust.x, originalSize.x);
					scaleY = FixedDivide(FixedAbs(newMouse.y - pivot.y) - adjust.y, originalSize.y);

					/* if we are on the other side of the pivot point from where we started, then negate the scale */
					if( ((newMouse.x - pivot.x) ^ (originalClick.x - pivot.x)) < 0 )
						scaleX = -scaleX;
					if( ((newMouse.y - pivot.y) ^ (originalClick.y - pivot.y)) < 0 )
						scaleY = -scaleY;

					/* save our new true scale mapping and calculate the delta mapping to apply to all the shapes */
					ResetMapping(&newMatrix);
					ScaleMapping(&newMatrix, scaleX, scaleY, pivot.x, pivot.y);
					goto mapSelection;
				}

				case commandSkew + centerTop:
					polyDst[1] += deltaX;
					polyDst[2] += deltaY;
					polyDst[3] = polyDst[1] + originalSize.x;
					polyDst[4] = polyDst[2];
					goto skewSelection;

				case commandSkew + centerRight:
					polyDst[3] += deltaX;
					polyDst[4] += deltaY;
					polyDst[5] = polyDst[3];
					polyDst[6] = polyDst[4] + originalSize.y;
					goto skewSelection;

				case commandSkew + centerBottom:
					polyDst[7] += deltaX;
					polyDst[8] += deltaY;
					polyDst[5] = polyDst[7] + originalSize.x;
					polyDst[6] = polyDst[8];
					goto skewSelection;

				case commandSkew + centerLeft:
					polyDst[1] += deltaX;
					polyDst[2] += deltaY;
					polyDst[7] = polyDst[1];
					polyDst[8] = polyDst[2] + originalSize.y;
					/* fall through into skewSelection */

skewSelection:			PolyToPolyMap((polygon *)&polySrc, (polygon *)&polyDst, &newMatrix);
					/* fall through into mapSelection */

mapSelection:		{	mapping deltaMatrix, tempMatrix;

					/* calculate the delta mapping */
					CopyToMapping(&tempMatrix, &newMatrix);
					InvertMapping(&deltaMatrix, &oldMatrix);
					MapMapping(&deltaMatrix, &tempMatrix);

					/* map all the shapes in the selection (and invalidate them) */
					count = GetPicture(targetPtr->selection, nil, nil, nil, nil);
					IfDebug(count == 0, "\pthe selection should always have items in it here");
					do {	shape itemCopy;
	
						GetPictureParts(targetPtr->selection, count, 1, &item, nil, nil, nil);
						itemCopy = CopyToShape(nil, item);
						MapShape(itemCopy, &deltaMatrix);
						GetShapeTransformedBounds(itemCopy, &bounds);
						DisposeShape(itemCopy);
						if( bounds.right >= bounds.left + minimumObjectSizeH &&
							bounds.bottom >= bounds.top + minimumObjectSizeV ) {
								InvalidateRectangle(targetPtr, GetShapeTransformedBounds(item, &bounds));
								MapShape(item, &deltaMatrix);
								InvalidateRectangle(targetPtr, GetShapeTransformedBounds(item, &bounds));
						} else {
							/* if the object would have disappeared, then keep the objects old full-delta mapping */
							CopyToMapping(&newMatrix, &oldMatrix);
						}
					} while( --count );

					/* save the new full-delta mapping in our original mapping variable */
					CopyToMapping(&oldMatrix, &newMatrix);

					/* update the screen to reflect changes to the shape and invalidate the controls */
					RedrawShapeControl(targetPtr);
					rebuildControls = true;
					break;
				}
			}

			oldMouse = newMouse;
		}
	}

	/* if we changed the geometry of the selection, then re-build the controls for it */
	if( rebuildControls )
		BuildSelectionControls(targetPtr);

	if( hiddenControls ) {
		SetShapeFill(targetPtr->controls, evenOddFill);
		DrawShape(targetPtr->controls);
	}
}

static void DeselectShape(shapeControlPtr targetPtr, shape item)
{
	if( targetPtr->selection ) {
		long count = GetPicture(targetPtr->selection, nil, nil, nil, nil);
		shape temp;

		do {
			GetPictureParts(targetPtr->selection, count, 1, &temp, nil, nil, nil);
		} while( temp != item && --count );

		/* if we actually found the item in the selection, then de-select it */
		if( count ) {
			/* invalidate the controls for the selection, since we are changing it */
			InvalidatePicture(targetPtr, targetPtr->controls);

			/* remove the item from the selection picture and dispose of the picture item was the only selected shape */
			SetPictureParts(targetPtr->selection, count, 1, 0, nil, nil, nil, nil);
			if( GetPicture(targetPtr->selection, nil, nil, nil, nil) == 0 )
				DisposeShapeAt(&targetPtr->selection);

			/* rebuild the controls for the selection */
			BuildSelectionControls(targetPtr);
		}
	}
}

static void SelectShape(shapeControlPtr targetPtr, shape item, boolean replaceSelection, boolean bringToFront)
{
	/* move the passed item shape to the front of the picture if we are supposed to do so */
	if( bringToFront ) {
		long index = GetPicture(targetPtr->items, nil, nil, nil, nil);
		shape tempShape;

		IfDebug(index == 0, "\ptargetPtr->items must always have items");
		do {
			GetPictureParts(targetPtr->items, index, 1, &tempShape, nil, nil, nil);
			if( tempShape == item ) {
				rectangle bounds;

				/* add the shape to the end first, then remove it so that our references are always valid */
				SetPictureParts(targetPtr->items, 0, 0, 1, &item, nil, nil, nil);
				SetPictureParts(targetPtr->items, index, 1, 0, nil, nil, nil, nil);
				InvalidateRectangle(targetPtr, GetShapeTransformedBounds(item, &bounds));
			}
		} while( --index );
	}

	/* if we are replacing the selection, then un-select all the selected shapes */
	if( replaceSelection && targetPtr->selection ) {
		InvalidatePicture(targetPtr, targetPtr->controls);
		DisposeShapeAt(&targetPtr->selection);
		DisposeShapeAt(&targetPtr->controls);
	}

	/* if there is an old selection that we are adding to, then invalidate the area for its controls */
	if( targetPtr->controls )
		InvalidatePicture(targetPtr, targetPtr->controls);

	if( item == nil ) {
		DisposeShapeAt(&targetPtr->controls);
		return;
	}

	if( targetPtr->selection == nil ) {
		targetPtr->selection = NewPicture(1, &item, nil, nil, nil);
	} else {
		long count = GetPicture(targetPtr->selection, nil, nil, nil, nil);
		long index = 0;
		shape selectionItem;
		shape mainItem;

		/* look through the selection for the first shape AFTER item in the main items picture; insert item before this index */
		IfDebug(count == 0, "\pselection must have items in it here");
		do {	long mainIndex = 0;

			GetPictureParts(targetPtr->selection, ++index, 1, &selectionItem, nil, nil, nil);
			do {
				IfDebug(mainIndex == GetPicture(targetPtr->items, nil, nil, nil, nil),
					"\pselected item not in main items picture");
				GetPictureParts(targetPtr->items, ++mainIndex, 1, &mainItem, nil, nil, nil);
			} while( mainItem != selectionItem && mainItem != item );
		} while( mainItem == selectionItem && --count );

		/* if we hit the end of the selection without finding a shape after item, then set index to zero to insert at the end */
		if( count == 0 )	index = 0;

		/* insert our newly selected item in the correct place in the selection */
		SetPictureParts(targetPtr->selection, index, 0, 1, &item, nil, nil, nil);
	}

	/* calculate the geometry of the controls for the selection and invalidate this new area */
	BuildSelectionControls(targetPtr);
	InvalidatePicture(targetPtr, targetPtr->controls);
}


/*----------------------------------------------------------------------------------------------------*/
/* public routines */
/*----------------------------------------------------------------------------------------------------*/


shapeControl NewShapeControl(shape items, shape background, shape foreground)
{
	shapeControl target;

	NilShapeReturnNil(items);
	IfWarningReturnNil(GetShapeType(items) != pictureType, picture_expected);
	IfDebug(firstIndexCommand <= lastSpecialCommand, "\pinvalid shape control commands - bug in shape control library");

	if( target = (shapeControl)NewHandleClear(sizeof(shapeControlRecord)) ) {
		viewPort **originalPorts = NewViewPortListFromShape(items);
		shapeControlPtr targetPtr;

		HLock((Handle)target);
		targetPtr = *target;

		targetPtr->items = CloneShape(items);

		if( background )
			targetPtr->background = CloneShape(background);
		else {
			color whiteColor;

			/* if there is no background, then create a white full shape (we always need some sort of background) */
			targetPtr->background = NewShape(fullType);
			whiteColor.space = rgbSpace;
			whiteColor.profile = nil;
			whiteColor.element.rgb.red = 0xFFFF;
			whiteColor.element.rgb.green = 0xFFFF;
			whiteColor.element.rgb.blue = 0xFFFF;
			SetShapeColor(targetPtr->background, &whiteColor);
		}
		/* ensure that the background draws to the same ports as the items */
		DereferenceViewPortList(originalPorts);
		IgnoreGraphicsNotice(transform_viewPorts_already_set);
			SetShapeViewPorts(targetPtr->background, CountViewPortList(originalPorts), *originalPorts);
		PopGraphicsNotice();

		if( foreground ) {
			targetPtr->foreground = CloneShape(foreground);
			SetShapeViewPorts(targetPtr->foreground, CountViewPortList(originalPorts), *originalPorts);
		}

		/* set the savedCenter to the no center saved value */
		targetPtr->savedCenter.x = targetPtr->savedCenter.y = positiveInfinity;

		ReleaseViewPortList(originalPorts);
		targetPtr->originalPorts = originalPorts;
		HUnlock((Handle)target);
	}

	return target;
}


void DisposeShapeControl(shapeControl target)
{
	shapeControlPtr targetPtr;

	NilParamReturn(target);

	HLock((Handle)target);
	targetPtr = *target;
	DisposeShape(targetPtr->items);
	DisposeShape(targetPtr->background);
	DisposeShapeAt(&targetPtr->foreground);
	DisposeShapeAt(&targetPtr->selection);
	DisposeShapeAt(&targetPtr->controls);
	DisposeViewPortList(targetPtr->originalPorts);
	HUnlock((Handle)target);

	/* dispose of the shape control memory itself */
	DisposeHandle((Handle)target);
}


shape GetShapeControlSelection(const shapeControl source)
{
	NilParamReturn(source);

	if( (*source)->selection )
		return CopyToShape(nil, (*source)->selection);
	else
		return nil;
}


void SetShapeControlsSelection(shapeControl target, shape shapesToSelect, boolean bringToFront, boolean replaceSelection)
{
	shapeControlPtr targetPtr;
	long count;
	boolean updateControls = false;

	NilParamReturn(target);
	HLock((Handle)target);
	targetPtr = *target;

	/* prevent the selections controls from getting re-build by the various intermediate steps */
	++targetPtr->buildHiddenLevel;

	/* if we are replacing the old selection, then de-select all the old shapes that will not be in the new selection */
	if( replaceSelection && targetPtr->selection ) {
		count = GetPicture(targetPtr->selection, nil, nil, nil, nil);

		do {	/* scan through all the items in the selection, de-selecting all the ones not in shapesToSelect */
			shape item;
			boolean itemInNewSelection;

			GetPictureParts(targetPtr->selection, count, 1, &item, nil, nil, nil);
			itemInNewSelection = false;
			if( shapesToSelect == nil || (shapesToSelect != item && GetShapeType(shapesToSelect) == pictureType
				&& ShapeInPicture(item, shapesToSelect) == false) ) {
					updateControls = true;
					SetPictureParts(targetPtr->selection, count, 1, 0, nil, nil, nil, nil);
			}
		} while( --count );

		if( GetPicture(targetPtr->selection, nil, nil, nil, nil) == 0 ) {
			DisposeShapeAt(&targetPtr->selection);
			DisposeShapeAt(&targetPtr->controls);
		}
	}

	/* if there is a new selection, then remove any already selected shapes from it */
	if( shapesToSelect ) {
		/* is the new selection a non-empty list of items? */
		if( GetShapeType(shapesToSelect) == pictureType && (count = GetPicture(shapesToSelect, nil, nil, nil, nil)) ) {
			/*** bring all the new items to the front in the order that they are in shapesToSelect */

			/* make a copy of the new item list to be selected so that we can munge the list */
			shapesToSelect = CopyToShape(nil, shapesToSelect);

			do {	/* remove all the already selected items from our new list to select */
				shape item;

				/* if the item in our new selection is already selected, then remove it from our new selection */
				GetPictureParts(shapesToSelect, count, 1, &item, nil, nil, nil);
				if( targetPtr->selection && ShapeInPicture(item, targetPtr->selection) ||
					ShapeInPicture(item, targetPtr->items) == false )
						SetPictureParts(shapesToSelect, count, 1, 0, nil, nil, nil, nil);
			} while( --count );

			/* if there are any items left in the new list to be selected, then add them to the current selection */
			if( count = GetPicture(shapesToSelect, nil, nil, nil, nil) ) {
				do {	/* note that we dont bring these items to the front, since we have already done this */
					shape item;

					GetPictureParts(shapesToSelect, count, 1, &item, nil, nil, nil);
					SelectShape(targetPtr, item, false, false);
					updateControls = true;
				} while( --count );
			}

			/* dispose of our munged copy of the new items that have been selected */
			DisposeShape(shapesToSelect);
		} else if( ShapeInPicture(shapesToSelect, targetPtr->selection) == false ) {
			if( ShapeInPicture(shapesToSelect, targetPtr->items) ) {
				SelectShape(targetPtr, shapesToSelect, bringToFront, false);
				updateControls = true;
			}
		}
	}

	/* allow the selections controls to get re-built and re-build them if they have changed */
	--targetPtr->buildHiddenLevel;
	if( updateControls )
		BuildSelectionControls(targetPtr);

	HUnlock((Handle)target);
}


shape GetShapeControlSelectionHandles(const shapeControl source)
{
	NilParamReturn(source);

	if( (*source)->controls )
		return CopyDeepToShape(nil, (*source)->controls);
	else
		return nil;
}


boolean SendEventToShapeControl(shapeControl target, EventRecord *event)
{
	boolean handledEvent = false;

	NilParamReturn(target);
	NilParamReturn(event);

	HLock((Handle)target);
	switch( event->what ) {
		case keyDown:
			break;	/*** change this to handle the key and return true if it is ours */
		case mouseDown:
		{	point originalClick;
			hitTestInfo result;
			shape itemShape;
			long command;
			shapeControlPtr targetPtr = *target;

			QDLocalToFixedGlobal(&event->where, 0, &originalClick);

			/* if there is a selection, then see if any of the selections controls have been hit and track them if they have */
			if( targetPtr->selection ) {
				if( HitTestPicture(targetPtr->controls, &originalClick, &result, 1, 1) != noPart ) {
					TrackShapeControl(targetPtr, result.containerIndex, originalClick);
					handledEvent = true;
					break;
				}
			}

			/* see if an item shape was hit; if so then add it to the selection and track it */
			if( HitTestPicture(targetPtr->items, &originalClick, &result, 1, 1) != noPart ) {
				command = commandMove;
				itemShape = result.which;

				if( targetPtr->selection && ShapeInPicture(itemShape, targetPtr->selection) ) {
					if( event->modifiers & shiftKey ) {
						/* the shape was already selected and we shift-clicked it, so deselect it and return */
						DeselectShape(targetPtr, itemShape);
						RedrawShapeControl(targetPtr);
						handledEvent = true;
						break;
					}
				} else {
					/* select the new shape - the last two parameters are (boolean replaceSelection, bringToFront) */
					SelectShape(targetPtr, itemShape, (event->modifiers & shiftKey) == 0,
						(event->modifiers & optionKey) != 0);
					RedrawShapeControl(targetPtr);
				}

				/* allow the user to move the entire selection around */
				TrackShapeControl(targetPtr, commandMove, originalClick);
				handledEvent = true;
				break;
			}

			/* see if we are within the bounds of any of the original viewPorts of the items */
			{	shape testShape = NewPoint(&originalClick);
				viewPort *portList = DereferenceViewPortList(targetPtr->originalPorts);

				IgnoreGraphicsNotice(transform_viewPorts_already_set);
					SetShapeViewPorts(testShape, CountViewPortList(targetPtr->originalPorts), portList);
				PopGraphicsNotice();
				/* scan through all the original viewPorts of the items */
				while( *portList ) {
					/* if a point at the original click location would draw to any devices in this port, then we are
					within the bounds of the port. we should now drag out a gray rectangle to select all the items
					contained within it */
					if( GetShapeGlobalViewDevices(testShape, *portList, nil) ) {
						DisposeShape(testShape);

					#if 0	 /*** needs work */
						TrackShapeControl(targetPtr, commandSelect, originalClick);
					#else
						/* for now, just de-select all the items */
						SelectShape(targetPtr, nil, true, false);
					#endif

						RedrawShapeControl(targetPtr);
						ReleaseViewPortList(targetPtr->originalPorts);
						HUnlock((Handle)target);
						return true;
					}
					++portList;
				}
				DisposeShape(testShape);
				ReleaseViewPortList(targetPtr->originalPorts);
			}

			/* nothing was hit, so break and return that we didnt handle the event */
			break;
		}
	}

	HUnlock((Handle)target);
	return handledEvent;
}


void InvalidateShapeControlShape(shapeControl target, shape invalidShape)
{
	NilParamReturn(target);
	NilShapeReturn(invalidShape);

	HLock((Handle)target);
	InvalidatePicture(*target, invalidShape);
	HUnlock((Handle)target);
}


void InvalidateShapeControlRectangle(shapeControl target, rectangle *bounds)
{
	NilParamReturn(target);
	NilParamReturn(bounds);

	HLock((Handle)target);
	InvalidateRectangle(*target, bounds);
	HUnlock((Handle)target);
}


void UpdateShapeControl(shapeControl target)
{
	NilParamReturn(target);
	HLock((Handle)target);
	RedrawShapeControl(*target);
	HUnlock((Handle)target);
}
