/* graphics libraries:	
	primitive flattening library routines
	by Cary Clark, Georgiann Delaney, Michael Fairman, Dave Good, Robert Johnson, Keith McGreggor, Oliver Steele, David Van Brink, Chris Yerga
	Copyright 1987 - 1991 Apple Computer, Inc.  All rights reserved.	*/


	#include <Memory.h>
	#include <Errors.h>
	#include <Files.h>
#include "font routines.h"
#include "graphics libraries.h"
#include "scaler types.h"
#include "storage library.h"
/*
	The following routine is an example of how to write a spooling procedure to
	work with FlattenShape and UnflattenShape.  For writing, it stores the data in
	a Handle which must be disposed of after the FlattenShape call.  When reading,
	it expects the 'data' field of the spoolBlock to have a handle to the data to be
	unflattened.
*/

#define allocationIncrement	1024		/* the storage handle is grown by this amount */

long HandleSpoolProc(spoolCommand command,  userSpool *block)
{
	switch (command)
	{
		case openReadSpool:
			block->size = 0;
			block->position = 0;
		break;
		case openWriteSpool:
			block->data = NewHandle(allocationIncrement);
			block->size = allocationIncrement;
			block->position = 0;
		break;
		
		case readSpool:
			BlockMove((*(char **) block->data) + block->position, block->spool.buffer, block->spool.count);
			block->position += block->spool.count;
		break;

		case writeSpool:
		{	register long oldPosition;

			oldPosition = block->position;
			block->position += block->spool.count;

			/* make sure there is at least enough room for one buffer size past current pointer */
			if (block->position + block->spool.bufferSize > block->size)		
			{
				block->size += block->spool.bufferSize;
				HUnlock(block->data);
				SetHandleSize(block->data, block->size);
				HLock(block->data);
			}
			BlockMove(block->spool.buffer, (*(char **) block->data + oldPosition), block->spool.count);
		}
		break;
		
		case closeSpool:
			SetHandleSize(block->data, block->position);
		break;
	}
	return 0;
}


/*
	This routine spools to and from a file.  The file must have previously been created and opened,
	and its refNum is expected to be in the refnum field of the block.
*/
static long FileSpoolProc(spoolCommand command, userSpool *block)
{
	short err;
	
	switch (command) {
		case openReadSpool:
		break;
		case openWriteSpool:
		break;
		case readSpool:
		{
			long count = block->spool.count;
			err = FSRead(block->reference, &count, block->spool.buffer);
			IfDebug(err && err != eofErr, "\pFSRead failed");
		} break;
		case writeSpool:
		{
			long count = block->spool.count;
			err = FSWrite(block->reference, &count, block->spool.buffer);
			IfDebug(err, "\pFSWrite failed");
		} break;
		case closeSpool:
		break;
		default:
			IfDebug(true, "\punexpected spool command");
	}
	return 0;
}

Handle ShapeToHandle(shape source)
{
	userSpool block;
	
	NilShapeReturnNil(source);
	block.spool.spoolProcedure = (long (*)(spoolCommand, struct spoolBlock *)) HandleSpoolProc;
	block.spool.buffer = nil;
	block.spool.bufferSize = 0;
	FlattenShape(source, fontListFlatten | fontGlyphsFlatten, &block.spool);
	return block.data;
}

shape HandleToShape(Handle source, long count, const viewPort portList[])
{
	userSpool block;
	
	block.spool.spoolProcedure = (long (*)(spoolCommand, struct spoolBlock *)) HandleSpoolProc;
	block.spool.buffer = nil;
	block.spool.bufferSize = 0;
	block.data = source;
	return UnflattenShape(&block.spool, count, portList);
}

void ShapeToFile(shape source, Str255 fileName, short vRefNum, OSType creator, OSType fileType)
{
	userSpool block;
	short refNum;
	short err;
	
	NilShapeReturn(source);
	NilParamReturn(fileName);
	
	if (creator == 0) creator = 'sLib';
	if (fileType == 0) fileType = 'flat';
	err = Create(fileName, vRefNum, creator, fileType);
	if (err == dupFNErr) {
		FSDelete(fileName, vRefNum);
		err = Create(fileName, vRefNum, creator, fileType);
	}
	IfDebug(err, "\pCreate failed");
	err = FSOpen(fileName, vRefNum, &refNum);
	IfDebug(err, "\pFSOpen failed");
	
	block.spool.spoolProcedure = (long (*)(spoolCommand, struct spoolBlock *)) FileSpoolProc;
	block.reference = refNum;
	block.spool.buffer = nil;
	block.spool.bufferSize = 0;
	FlattenShape(source,  fontListFlatten | fontGlyphsFlatten, &block.spool);
	err = FSClose(refNum);
	IfDebug(err, "\pFSClose failed");
}

shape FileToShape(Str255 fileName, short vRefNum, long count, const viewPort portList[])
{
	userSpool block;
	short refNum;
	short err;
	shape dest;
	
	err = FSOpen(fileName, vRefNum, &refNum);
	IfDebug(err, "\pFSOpen failed");
	block.spool.spoolProcedure = (long (*)(spoolCommand, struct spoolBlock *)) FileSpoolProc;
	block.reference = refNum;
	block.spool.buffer = nil;
	block.spool.bufferSize = 0;
	dest = UnflattenShape(&block.spool, count, portList);
	err = FSClose(refNum);
	IfDebug(err, "\pFSClose failed");
	return dest;
}

Handle FontToHandle(font fontID, long glyphBits[])
{
	userSpool block;
	struct scalerStream stream;
	
	block.spool.spoolProcedure = (long (*)(spoolCommand, struct spoolBlock *)) HandleSpoolProc;
	block.spool.buffer = nil;
	block.spool.bufferSize = 0;

	stream.streamRefCon = fontID;
	stream.type = flattenedStream;
	stream.flags = downloadFlag;
	stream.memorySize = 0;
	stream.variationCount = selectAllVariations;
	stream.variations = nil;
	stream.encoding = 0;
	stream.glyphBits = glyphBits;
	stream.fontName = nil;

	FlattenFont(fontID, &stream, &block.spool);
	return block.data;
}

font HandleToFont(Handle source)
{
	/*
	 *	NewFont does not copy the data in source, so you can't dispose
	 *	of it until you have called DisposeFont().
	*/
	return NewFont(handleFontStorage, source, 0);
}

