/* graphics libraries:
	transfer mode library
	by Cary Clark, Georgiann Delaney, Michael Fairman, Dave Good, Robert Johnson, Keith McGreggor, Oliver Steele, David Van Brink, Chris Yerga
	Copyright 1987 - 1991 Apple Computer, Inc.  All rights reserved.	*/

	#include <Memory.h>
#include "graphics libraries.h"

/*
  * Routines
  */

static void InitColorMatrix(fixed m[5][4])
{
	register fixed *x;
	register short i;

	x = &m[0][0];
	for(i = 19; i>=0; i--)
		*x++ = 0;
	m[0][0] = m[1][1] = m[2][2] = m[3][3] = fix1;			/* Identity matrix, for cleanliness	*/
}

/*
  * Initialize all fields of a transfermode record
  * to a very reasonable rgb copy mode.
  */
void InitTransferMode(transferMode *mode)
{
	NilParamReturn(mode);
	SetCommonTransfer(mode,copyMode,0,nil);
}	

transferMode *SetCommonTransfer(register transferMode *mode,  commonTransferMode type, unsigned short opValue, const color *opColor)
{
	register transferComponent *component;
	register short i;
	color oColor;

	NilParamReturnNil(mode);

	if(opColor)
		{
		oColor = *opColor;
		ConvertColor(&oColor,rgbaSpace, nil, nil);
		}
	else
		{
		oColor.space = rgbaSpace;
		oColor.element.rgba.red =
		oColor.element.rgba.green =
		oColor.element.rgba.blue = opValue;
		oColor.element.rgba.alpha = 65535;
		}

	if(type >= overMode && type <= fadeMode)
		mode->space = rgbaSpace;
	else
		mode->space = rgbSpace;

	mode->flags = 0;
	mode->set = nil;
	mode->profile = nil;

	component = &mode->component[0];
	InitColorMatrix(mode->sourceMatrix);
	InitColorMatrix(mode->deviceMatrix);
	InitColorMatrix(mode->resultMatrix);

	component->mode = type;
	component->flags = 0;

	component->sourceMinimum = component->deviceMinimum = component->clampMinimum = 0;
	component->sourceMaximum = component->deviceMaximum = component->clampMaximum = 65535;

	for(i = 3; i>=0; i--)
		{
		component = &mode->component[i];

		*component = mode->component[0];
		component->operand = (&oColor.element.rgb.red)[i];

		switch(type)
			{
			case highlightMode:
				if(!opColor)					/* if no opColor passed, use lomem RGBHilite	*/
					component->operand = ((short *)0xDA0)[i];
				break;
			case overMode:						/* use it like a real alpha mode				*/
				if(i==3)
					component->mode = rampOrMode;
				break;
			case atopMode:						/* use it like a real alpha mode				*/
				if(i==3)
					{
					component->flags |= reverseComponent;
					component->mode = copyMode;	/* result alpha = device alpha		*/
					}
				break;
			case excludeMode:					/* use it like a real alpha mode				*/
				if(i==3)
					component->mode = rampXorMode;
				break;
			case fadeMode:
				if(i==3)
					component->mode = addMode;
				break;
			case commonAddOverMode:
				component->mode = addMode;
				component->flags |= overResultComponent;
				break;
			case commonSubtractOverMode:
				component->mode = addMode;
				component->flags |= overResultComponent;
				mode->sourceMatrix[i][i] = -fix1;
				break;
			case commonSubtractPinMode:
				component->mode = addMode;
				mode->sourceMatrix[i][i] = -fix1;
				break;
			case commonTransparentMode:			/* the opColor specifies which color to avoid from source	*/
				component->mode = copyMode;
				mode->flags |= rejectSourceTransfer;
				if(opColor)
					{
					component->sourceMinimum =
					component->sourceMaximum = (&oColor.element.rgb.red)[i];
					}
				else							/* default is avoid drawing white */
					{
					component->sourceMinimum =
					component->sourceMaximum = 65535;
					}
				break;
			case commonInMode:
				if(i==3)
					component->mode = rampAndMode;
				else
					component->mode = copyMode;
				break;
			case commonOutMode:
				if(i==3)
					{
					component->mode = rampAndMode;
					mode->deviceMatrix[3][3] = -fix1;
					mode->deviceMatrix[4][3] = 65535;
					}
				else
					component->mode = copyMode;
				break;


			}

		}

	return mode;
}

void SetInkCommonTransfer(ink source, commonTransferMode type)
{
	transferMode mode;
	
	NilInkReturn(source);
	SetCommonTransfer(&mode, type,32768,nil);
	SetInkTransfer(source, &mode);
}

void SetShapeCommonTransfer(shape source, commonTransferMode type)
{
	transferMode mode;
	
	NilShapeReturn(source);
	SetCommonTransfer(&mode, type,32768,nil);
	SetShapeTransfer(source, &mode);
}

commonTransferMode GetShapeCommonTransfer(shape source)
{
	transferMode mode;
	
	NilShapeReturnNil(source);
	GetShapeTransfer(source, &mode);
	return mode.component[0].mode;
}

commonTransferMode GetInkCommonTransfer(ink source)
{
	transferMode mode;
	
	NilInkReturnNil(source);
	GetInkTransfer(source, &mode);
	return mode.component[0].mode;
}

color *TransmogrifyColor(color *dstColor, const color *srcColor, const transferMode *mode)
{
	ink tempInk = NewInk();
	
	SetInkColor(tempInk,srcColor);
	SetInkTransfer(tempInk,mode);
	CombineColor(dstColor, tempInk);
	DisposeInk(tempInk);
	return dstColor;
}

/* Given an ink and a device, set's up the ink's mode so that drawing over the background color leaves
  * the result color. Also sets up the transferMode colorSpace etc. so that the fast case is taken.
  */
void SetInkFastXorTransfer(ink inky, viewDevice destDevice, color *background, color *result)
{
	shape deviceBitsShape;
	bitmap deviceBits;
	transferMode fastMode;
	color inkColor;

	color localBackground;
	color localResult;

	localBackground.space = rgbSpace;
	localBackground.profile = nil;
	localBackground.element.rgb.red = 0xFFFF;
	localBackground.element.rgb.green = 0xFFFF;
	localBackground.element.rgb.blue = 0xFFFF;
	
	localResult.space = rgbSpace;
	localResult.profile = nil;
	localResult.element.rgb.red = 0;
	localResult.element.rgb.green = 0;
	localResult.element.rgb.blue = 0;
	
	if (background)
		localBackground = *background;
	if (result)
		localResult = *result;
		
	deviceBitsShape = GetViewDeviceBitmap(destDevice);
	GetBitmap(deviceBitsShape, &deviceBits, nil);
	
	InitTransferMode(&fastMode);
	fastMode.flags |= singleComponentTransfer;
	fastMode.component[0].mode = xorMode;
	if ((fastMode.space = deviceBits.space) == indexedSpace) {
		fastMode.set = deviceBits.set;
		ConvertColor(&localBackground, indexedSpace, deviceBits.set, deviceBits.profile);
		ConvertColor(&localResult, indexedSpace, deviceBits.set, deviceBits.profile);
		if (localBackground.element.indexed.index == localResult.element.indexed.index) {
			if (localResult.element.indexed.index > 1)
				--localResult.element.indexed.index;
			else
				++localResult.element.indexed.index;	 /* This assumes that colorSet has at least 2 colors */
		}
		inkColor.space = indexedSpace;
		inkColor.profile = deviceBits.profile;
		inkColor.element.indexed.set = deviceBits.set;
		inkColor.element.indexed.index = ((localBackground.element.indexed.index - 1) ^
								(localResult.element.indexed.index - 1)) + 1;
	} else {
		fastMode.space = rgbSpace;
		ConvertColor(&localBackground, rgbSpace, nil, deviceBits.profile);
		ConvertColor(&localResult, rgbSpace, nil, deviceBits.profile);
		inkColor.space = rgbSpace;
		inkColor.profile = deviceBits.profile;
		inkColor.element.rgb.red = localBackground.element.rgb.red ^ localResult.element.rgb.red;
		inkColor.element.rgb.green = localBackground.element.rgb.green ^ localResult.element.rgb.green;
		inkColor.element.rgb.blue = localBackground.element.rgb.blue ^ localResult.element.rgb.blue;
	}
	SetInkColor(inky, &inkColor);
	SetInkTransfer(inky, &fastMode);
	DisposeShape(deviceBitsShape);
}

void SetShapeFastXorTransfer(shape source, color *background, color *result)
{
	long viewPortCount, viewDeviceCount;
	void *buffer;
	viewPort vp;
	viewDevice vd;
	ink inky;
	
	viewPortCount = GetTransformViewPorts(GetShapeTransform(source), nil);
	buffer = NewPtr(sizeof(viewPort) * viewPortCount);
	if (MemError()) Debugger();
	GetTransformViewPorts(GetShapeTransform(source), (viewPort *)buffer);
	
	viewDeviceCount = GetViewPortViewDevices(vp = *(viewPort *)buffer, nil);
	SetPtrSize(buffer, viewDeviceCount * sizeof(viewDevice));
	if (MemError())
	{	DisposPtr(buffer);
		buffer = NewPtr(viewDeviceCount * sizeof(viewDevice));
		if (MemError()) Debugger();
	}
	GetViewPortViewDevices(vp, (viewDevice *)buffer);
	vd = *(viewDevice *)buffer;
	DisposPtr(buffer);

	if (GetInkOwners(inky = GetShapeInk(source)) > 1) {
		SetShapeInk(source, inky = NewInk());
		DisposeInk(inky);
	}
	SetInkFastXorTransfer(inky, vd, background, result);
	SetShapeInkAttributes(source, GetShapeInkAttributes(source) | suppressDitherInk);
}

