/* ------------------------------------------------------------------------------

FILENAME
	PrintingManager.h

DESCRIPTION
		This file defines all the public level data types
		and interface calls for the Printing Manager.
	
COPYRIGHT
	  Apple Computer, Inc. 1990, 1991, 1992, 1993
	 All rights reserved. 

-------------------------------------------------------------------------------- */

#ifndef __PRINTINGMANAGER__
#define __PRINTINGMANAGER__


#ifndef __PRINTING__
	#include <PrintTraps.h>
#endif

#ifndef __FILES__
	#include <Files.h>
#endif

#ifndef __COLLECTIONMANAGER__
	#include <Collections.h>
#endif

#ifndef __PRINTINGERRORS__
	#include <PrintingErrors.h>
#endif

#ifndef graphicsTypesIncludes
	#include <graphics types.h>
#endif

#ifndef fontTypesIncludes
	#include <font types.h>
#endif

#ifndef scalerTypeIncludes
	#include <scaler types.h>
#endif

/* ------------------------------------------------------------------------------

							Printing Manager API Contants and Types

-------------------------------------------------------------------------------- */



/* Printing Manager Gestalt Selector... */


enum {
	gestaltPrintingMgrVersion = 'pmgr'
};


/*--------------------------*/
/* abstract data types...   */
/*--------------------------*/



typedef struct PrivatePrinterRecord *Printer;
typedef struct PrivateJobRecord *Job;
typedef struct PrivateFormatRecord *Format;
typedef struct PrivatePaperTypeRecord *PaperType;
typedef struct PrivatePrintFileRecord *PrintFile;



/*----------------------------------------*/
/* new function constants and types...    */
/*----------------------------------------*/



typedef Boolean LoopStatus;

enum {						// possible values for LoopStatus
	stopLooping = false,
	keepLooping = true
};

typedef pascal LoopStatus (*ViewDeviceProc) (viewDevice, void *refCon);
typedef pascal LoopStatus (*FormatProc) (Format, void *refCon);
typedef pascal LoopStatus (*PaperTypeProc) (PaperType, void *refCon);

typedef pascal OSErr (*PrintingFlattenProc) (long size, void *data, void *refCon);



/*----------------------------------------------*/
/* printing collection constants and types...	*/
/*----------------------------------------------*/


// The following constants are used to set collection item flags in printing collections.
// The Printing Manager purges certain items whenever a driver switch occurs.
// If the formatting driver changes, all items marked as volatileFormattingDriverCategory
// will be purged.  If the output driver changes, all items marked as
// volatileOutputDriverCategory will be purged.
// Note that to prevent items from being flattened when FlattenJob is called, you should
// unset the collectionPersistenceBit (defined in Collections.h), which is on by default.


typedef short CollectionCategory;		// stored in collection items' user attribute bits

enum {
	noCollectionCategory			= (CollectionCategory) 0x0000,
	outputDriverCategory			= (CollectionCategory) 0x0001,
	formattingDriverCategory	= (CollectionCategory) 0x0002,
	driverVolatileCategory		= (CollectionCategory) 0x0004,
	
	volatileOutputDriverCategory = outputDriverCategory + driverVolatileCategory,
	volatileFormattingDriverCategory = formattingDriverCategory + driverVolatileCategory
};



/*-----------------------------*/
/* printing collection tags... */
/*-----------------------------*/



enum { printingTagID = -28672 };			// all pre-defined tags have this ID



/*--------------------------*/
/* job collection tags...   */
/*--------------------------*/



enum { jobTag = 'job ' };

	enum { 										// priorities for jobs
		printJobUrgent	= 0x00000001,
		printJobAtTime	= 0x00000002, 
		printJobASAP	= 0x00000003
	};
	
	enum { printJobHoldingBit = 0x00001000 };	// reserved bit in the priority field means the job is on hold
	
	enum {
		printJobHolding			= (printJobHoldingBit + printJobASAP),
		printJobHoldingAtTime	= (printJobHoldingBit + printJobAtTime),
		printJobHoldingUrgent	= (printJobHoldingBit + printJobUrgent)
	};
	
	enum {										// ===== Job Alert User constants (jobAlert field) =====
		noPrintTimeAlert	=	0,					// don't alert user when we print
		alertBefore			=	1,					// alert user before we print
		alertAfter			=	2,					//	alert user after we print
		alertBothTimes		=	3					//	alert before and after we print
	};
		
	enum {										// ===== Job Timeout constants (jobTimeout field) =====
		thirtySeconds		=	1800,				// 30 seconds in ticks
		twoMinutes			=	7200				// 2 minutes in ticks
	};	

	typedef struct {
		long	numPages;
		long	priority;
		long	timeToPrint;
		long	jobTimeout;				// in ticks
		long	firstPageToPrint;		// start printing from this page
		short	jobAlert;
		Str31	appName;
		Str31 documentName;
		Str31 userName;
	} JobInfo;

enum { collationTag = 'sort' };

	typedef struct {
		Boolean collation;					// true if copies are to be collated
	} CollationInfo;

enum { copiesTag = 'copy' };

	typedef struct {
		long copies;							// number of copies of document to print
	} CopiesInfo;

enum { pageRangeTag = 'rang' };

	enum {										// for optionChosen field in SimplePageRangeInfo
		defaultPageRange		= (char) 0,
		replacePageRange		= (char) 1,
		customizePageRange	= (char) 2
	};

	typedef struct {
		char			optionChosen;					// from options listed above
		Boolean		printAll;						// true if user wants to print all pages
		long			fromPage;						// for defaultPageRange, current value
		long			toPage;							// for defaultPageRange, current value
	} SimplePageRangeInfo;

	typedef struct {
		SimplePageRangeInfo	simpleRange;		// info which will be returned for GetJobPageRange		
		Str31						fromString;			// for customizePageRange, current value
		Str31						toString;			// for customizePageRange, current value
		long						minFromPage;		// for defaultPageRange, we parse with this, ignored if nil
		long						maxToPage;			// for defaultPageRange, we parse with this, ignored if nil
		char						replaceString[1];	// for replacePageRange, string to display
	} PageRangeInfo;

enum { qualityTag = 'qual' };

	typedef struct {
		Boolean	disableQuality;			// true to disable standard quality controls
		short		defaultQuality;
		short		currentQuality;
		short		qualityCount;				// number of quality menu items in popup menu
		char		qualityNames[1];			// array of packed pascal strings for popup menu titles
	} QualityInfo;

enum { fileDestinationTag = 'dest' };

	typedef struct {
		Boolean toFile;						// true if destination is a file
	} FileDestinationInfo;

enum { fileLocationTag = 'floc' };

	typedef struct {
		FSSpec fileSpec;						// location to put file, if destination is file
	} FileLocationInfo;

enum { fileFormatTag = 'ffmt' };

	typedef struct {
		Str31 fileFormatName;				// name of file format if destination is file
	} FileFormatInfo;

enum { fileFontsTag = 'incf' };

	typedef struct {
		Boolean includeFonts;				// true if including fonts, if destination is file
	} FileFontsInfo;

enum { paperFeedTag = 'feed' };

	typedef struct {
		Boolean	autoFeed;					// true if automatic feed, false if manual
	} PaperFeedInfo;

typedef long TrayIndex;						// type of tray index reference

enum { trayFeedTag = 'tray' };			// public Printing Manager tray setting

	typedef struct {
		TrayIndex feedTrayIndex;			// tray to feed paper from
	} TrayFeedInfo;

enum { manualFeedTag = 'manf' };

	typedef struct {
		long	numPaperTypeNames;			// number of paperTypes to manually feed
		Str31	paperTypeNames[1];			// array of names of paperTypes to manually feed
	} ManualFeedInfo;

enum { normalMappingTag = 'nmap' };

	typedef struct {
		Boolean normalPaperMapping;		// true if not overriding normal paper mapping
	} NormalMappingInfo;

enum { specialMappingTag = 'smap' };

	enum {										// for paperMapping field in SpecialMappingInfo
		redirectPages			= (char) 1,
		scalePages				= (char) 2,
		tilePages				= (char) 3
	};

	typedef struct {
		char specialMapping;					// enumerated redirect, scale or tile setting
	} SpecialMappingInfo;

enum { trayMappingTag = 'tmap' };

	typedef struct {
		TrayIndex mapPaperToTray;			// tray to map all paper to
	} TrayMappingInfo;

enum { paperMappingTag = 'pmap' };

	// This collection item contains a flattened paper type resource

enum { printPanelTag = 'ppan' };

	typedef struct {
		Str31 startPanelName;				// name of starting panel in job print dialog
	} PrintPanelInfo;

enum { formatPanelTag = 'fpan' };

	typedef struct {
		Str31 startPanelName;				// name of starting panel in format dialog
	} FormatPanelInfo;



/*--------------------------*/
/* format collection tags...*/
/*--------------------------*/



enum { paperTypeLockTag = 'ptlk' };		// paperType lock tag

	typedef struct {
		Boolean paperTypeLocked;			// true if format's paperType is locked
	} PaperTypeLockInfo;

enum { orientationTag = 'layo' };

	enum {										// for orientation field in orientationTag
		portraitLayout 			= (char) 0,
		landscapeLayout 			= (char) 1,
		rotatedPortraitLayout 	= (char) 2,
		rotatedLandscapeLayout 	= (char) 3
	};
	
	typedef struct {
		char orientation;						// an enumerated orientation value
	} OrientationInfo;

enum { scalingTag = 'scal' };

	typedef struct {
		Fixed horizontalScaleFactor;		// current horizontal scaling factor
		Fixed verticalScaleFactor;			// current vertical scaling factor
		short minScaling;						// minimum current scaling
		short maxScaling;						// maximum current scaling
	} ScalingInfo;

enum { directModeTag = 'dirm' };

	typedef struct {
		Boolean directModeOn;				// true if direct mode is enabled
	} DirectModeInfo;

enum { formatHalftoneTag = 'half' };

	typedef struct {
		long numHalftones;					// how many halftones
		halftone	halftones[1]; 				// any number of halftones
	} FormatHalftoneInfo;

enum { invertPageTag = 'invp' };

	typedef struct	{
		Boolean		invert;					// true -> invert the page
	} InvertPageInfo;							// missing | false -> don't invert the page

enum { flipPageHorizontalTag = 'flph' };

	typedef struct	{
		Boolean		flipHorizontal;		// true -> flip x coordinates on the page
	} FlipPageHorizontalInfo;				// missing | false -> don't flip horizontally
	
enum { flipPageVerticalTag = 'flpv' };

	typedef struct	{
		Boolean		flipVertical;			// true -> flip y coordinates on the page
	} FlipPageVerticalInfo;					// missing | false -> don't flip vertically
	
enum { preciseBitmapsTag = 'pbmp' };
	
	typedef struct {
		Boolean		preciseBitmaps;		// true -> scale the page by 96%
	} PreciseBitmapInfo;						// missing | false -> don't scale the page by 96%



/*-----------------------------*/
/* paperType collection tags...*/
/*-----------------------------*/



enum { baseTag = 'base' };

	enum {										// for baseType field in baseTag
		unknownBase		=	0,						// base paper type from which this paper type is
		usLetterBase	=	1,						// derived.  This is not a complete set.
		usLegalBase		=	2,
		a4LetterBase	=	3,
		b5LetterBase	=	4,
		tabloidBase		=	5
	};
	
	typedef struct {
		long baseType;							// paperType's base type
	} BaseInfo;

enum { creatorTag = 'crea' };

	enum {										// for creator field in creatorTag
		sysPaperType			=	'sypt',		// system paper type creator
		userPaperType			=	'uspt'		// user paper type creator
														// if driver paperTypes then = driver's creator type
	};
	
	typedef struct {
		OSType creator;						// paperType's creator
	} CreatorInfo;

enum { unitsTag = 'unit' };

	enum {										// for units field in unitsTag
		picas						=	(char) 0,	// pica measurement
		mms						=	(char) 1,	// millimeter measurement
		inches					=	(char) 2		// inches measurement
	};
	
	typedef struct {
		char units;								// paperType's units
	} UnitsInfo;

enum { flagsTag = 'flag' };

	enum {										// paper type flags (bit positions)
		oldPaperTypeFlag				=	0x00800000,	// indicates a paper type with 7.0 settings
		newPaperTypeFlag				=	0x00400000,	// indicates a paper type with post 7.0 settings
		oldAndNewPaperTypeFlag		=	0x00C00000,	// indicates a paper type that's both old and new
		fixedPrintArea					=	0x00200000,	//	indicates the margins cannot change when format for a different device
		defaultPaperTypeFlag			=	0x00100000,	//	indicates the default paper type in the group	
		defaultBasePT					=	0x00080000	//	indicates the default paper type for a specific base type (e.g. US Letter)
	};
	
	typedef struct {
		long flags;								// paperType's flags
	} FlagsInfo;

enum { commentTag = 'cmnt' };

	typedef struct {
		Str255 comment;						// paperType's comment
	} CommentInfo;



/*---------------------------*/
/* printer viewDevice tags...*/
/*---------------------------*/



enum { penTableTag = 'pent' };

	enum {					// possible values for the flags field in the PenTableEntry structure
		deviceUnits = 0,
		mmUnits = 1,
		inchesUnits = 2
	};
	
	enum {    
		  penNotLoaded	 = -1
	};
	
	typedef struct { 
		Str31      penName;			// name of the pen
		color      penColor;			// color that is part of the color set
		fixed      penThickness;	// size of the pen
		short      penUnits;			// specifies units in which pen thickness is defined
		short      penPosition;		// pen position in the carousel, -1 (kPenNotLoaded) if not loaded
	} PenTableEntry;
		 
		 
	typedef struct {
		long   		 	numPens;		// number of pen entries in the following array
		PenTableEntry	pens[1];		// array of pen entries
	} PenTable, *PenTablePtr, **PenTableHdl;



/*----------------------------------------*/
/* dialog-related constants and types...  */
/*----------------------------------------*/



typedef long DialogResult;

enum {
	cancelSelected		= (DialogResult) 0,
	okSelected 			= (DialogResult) 1,
	revertSelected		= (DialogResult) 2
};


typedef struct {
	short	editMenuID;
	short	cutItem;
	short	copyItem;
	short	pasteItem;
	short	clearItem;
	short	undoItem;
} EditMenuRecord;


/*-----------------------------------------*/
/* job format mode constants and types...  */
/*-----------------------------------------*/



typedef long QueryType;

enum {
	getJobFormatLineConstraintQuery		= (QueryType)  0,
	getJobFormatFontsQuery					= (QueryType)  1,
	getJobFormatFontCommonStylesQuery	= (QueryType)  2,
	getJobFormatFontConstraintQuery		= (QueryType)  3,
	setStyleJobFormatCommonStyleQuery	= (QueryType)  4
};


typedef OSType JobFormatMode;

enum {
	graphicsJobFormatMode	= (JobFormatMode) 'grph',
	textJobFormatMode			= (JobFormatMode) 'text',
	postscriptJobFormatMode = (JobFormatMode) 'post'
};


typedef struct {
	long				numModes;
	JobFormatMode	modes[1];		// any number of modes
} JobFormatModeTable, *JobFormatModeTablePtr, **JobFormatModeTableHdl;


enum { constraintRange = -1 };	// in numSizes field of PositionConstraintTable to indicate a range in sizes array


typedef struct {						// for text mode	
	point		phase;
	point		offset;
	long		numSizes;
	Fixed		sizes[1];				// any number of sizes
} PositionConstraintTable, *PositionConstraintTablePtr, **PositionConstraintTableHdl;

typedef struct {						// for text mode
	long			numStyleNames;			// number of style names
	Str255		styleNames[1];			// any number of style names
} StyleNameTable, *StyleNameTablePtr, **StyleNameTableHdl;

typedef struct {						// for text mode
	long			numFonts;				// number of font references
	font			fonts[1];				// any number of font references
} FontTable, *FontTablePtr, **FontTableHdl;



/* ------------------------------------------------------------------------------

								Printing Manager API Functions

-------------------------------------------------------------------------------- */


#define PUBLICAPIGLUE(selector) {0x203C, 0x0000, selector, 0xABFE}


#ifdef __cplusplus
extern "C" {
#endif


// Global Routines (Basic)


pascal OSErr InitPrinting (void)
    = PUBLICAPIGLUE(0);

pascal OSErr ExitPrinting (void)
    = PUBLICAPIGLUE(1);


// Job Routines (Basic)


pascal OSErr NewJob (Job *)
    = PUBLICAPIGLUE(2);
	
pascal OSErr DisposeJob (Job)
    = PUBLICAPIGLUE(3);
	 
pascal void FlattenJob (Job, PrintingFlattenProc, void *)
    = PUBLICAPIGLUE(4);
	
pascal Job UnflattenJob (Job, PrintingFlattenProc, void *)
    = PUBLICAPIGLUE(5);
	 
pascal Handle FlattenJobToHdl (Job, Handle)
    = PUBLICAPIGLUE(6);
	
pascal Job UnflattenJobFromHdl (Job, Handle)
    = PUBLICAPIGLUE(7);

pascal void InstallApplicationOverride (Job, short messageID, void *override)
	= PUBLICAPIGLUE(8);


// Format Routines (Basic)


pascal Format NewFormat (Job)
    = PUBLICAPIGLUE(9);
	
pascal void DisposeFormat (Format)
    = PUBLICAPIGLUE(10);
	
								
// PaperType Routines (Basic)


pascal PaperType NewPaperType (Job, Str31 name, rectangle *pageSize, rectangle *paperSize)
    = PUBLICAPIGLUE(11);
	
pascal void DisposePaperType (PaperType)
    = PUBLICAPIGLUE(12);

pascal PaperType GetNewPaperType (Job, short resID)
    = PUBLICAPIGLUE(13);


// Error-Handling Routines (Basic)
					
								
pascal OSErr GetJobError (Job)
	= PUBLICAPIGLUE(14);

pascal void SetJobError (Job, OSErr)
	= PUBLICAPIGLUE(15);


// Dialog Routines (Basic)
					
								
pascal DialogResult JobDefaultFormatDialog (Job, EditMenuRecord *)
   = PUBLICAPIGLUE(16);

pascal DialogResult JobPrintDialog (Job, EditMenuRecord *)
   = PUBLICAPIGLUE(17);

pascal DialogResult FormatDialog (Format, EditMenuRecord *, StringPtr title)
	= PUBLICAPIGLUE(18);

								
// Format Routines (Basic)
					
								
pascal Format GetJobFormat (Job, long whichFormat)
    = PUBLICAPIGLUE(19);
	
pascal Job GetFormatJob (Format)
	= PUBLICAPIGLUE(20);

pascal PaperType GetFormatPaperType (Format)
    = PUBLICAPIGLUE(21);
	
pascal void GetFormatDimensions (Format, rectangle *pageSize, rectangle *paperSize)
    = PUBLICAPIGLUE(22);
	

// Spooling Routines (Basic)
					
								
pascal void GetJobPageRange (Job theJob, long *firstPage, long *lastPage)
	= PUBLICAPIGLUE(23);
	
pascal void StartJob (Job, StringPtr docName, long pageCount)
    = PUBLICAPIGLUE(24);
	
pascal void PrintPage (Job, long pageNumber, Format, shape thePage)
   = PUBLICAPIGLUE(25);
		
pascal Boolean StartPage (Job, long pageNumber, Format, long numViewPorts, viewPort *viewPortList)
	= PUBLICAPIGLUE(26);

pascal void FinishPage (Job)
	= PUBLICAPIGLUE(27);

pascal void FinishJob (Job)
    = PUBLICAPIGLUE(28);
	

// Job Routines (Advanced)
											

pascal Collection GetJobCollection (Job)
	= PUBLICAPIGLUE(29);

pascal void *GetJobRefCon (Job)
	= PUBLICAPIGLUE(30);

pascal void SetJobRefCon (Job, void *refCon)
	= PUBLICAPIGLUE(31);
	
pascal Job CopyJob (Job srcJob, Job dstJob)
	= PUBLICAPIGLUE(32);

pascal void SelectJobFormattingPrinter (Job, Str31 printerName)
	= PUBLICAPIGLUE(33);
	
pascal void SelectJobOutputPrinter (Job, Str31 printerName)
	= PUBLICAPIGLUE(34);

pascal void ForEachJobFormatDo (Job, FormatProc, void *refCon)
	= PUBLICAPIGLUE(35);

pascal long CountJobFormats (Job)
	= PUBLICAPIGLUE(36);

pascal Boolean UpdateJob (Job)
	= PUBLICAPIGLUE(37);
	
pascal void ConvertPrintRecord (Job, THPrint)
	= PUBLICAPIGLUE(38);


// Printer Routines (Advanced)


pascal Printer GetJobFormattingPrinter (Job)
	= PUBLICAPIGLUE(39);
	
pascal Printer GetJobOutputPrinter (Job)
	= PUBLICAPIGLUE(40);
	
pascal Printer GetJobPrinter (Job)
	= PUBLICAPIGLUE(41);
	
pascal Job GetPrinterJob (Printer)
	= PUBLICAPIGLUE(42);

pascal void ForEachPrinterViewDeviceDo (Printer, ViewDeviceProc, void *refCon)
	= PUBLICAPIGLUE(43);

pascal long CountPrinterViewDevices (Printer)
	= PUBLICAPIGLUE(44);

pascal viewDevice GetPrinterViewDevice (Printer, long	whichViewDevice)
	= PUBLICAPIGLUE(45);
	
pascal void SelectPrinterViewDevice (Printer, long	whichViewDevice)
	= PUBLICAPIGLUE(46);
	
pascal void GetPrinterName (Printer, Str31)
	= PUBLICAPIGLUE(47);

pascal OSType GetPrinterType (Printer)
	= PUBLICAPIGLUE(48);

pascal void GetPrinterDriverName (Printer, Str31)
	= PUBLICAPIGLUE(49);

pascal OSType GetPrinterDriverType (Printer)
	= PUBLICAPIGLUE(50);


// Format Routines (Advanced)


pascal Collection GetFormatCollection (Format)
	= PUBLICAPIGLUE(51);

pascal void ChangedFormat (Format)
	= PUBLICAPIGLUE(52);

pascal Format CopyFormat (Format srcFormat, Format dstFormat)
	= PUBLICAPIGLUE(53);
	
pascal Format CloneFormat (Format)
	= PUBLICAPIGLUE(54);
		
pascal long CountFormatOwners (Format)
	= PUBLICAPIGLUE(55);
		
pascal void GetFormatMapping (Format, mapping *)
	= PUBLICAPIGLUE(56);

pascal shape GetFormatForm (Format, shape *mask)
	= PUBLICAPIGLUE(57);
	
pascal void SetFormatForm (Format, shape form, shape mask)
	= PUBLICAPIGLUE(58);


// Job Format Modes Routines (Advanced)


pascal void SetAvailableJobFormatModes (Job, JobFormatModeTableHdl)
	= PUBLICAPIGLUE(59);

pascal JobFormatMode GetPreferredJobFormatMode (Job, Boolean *directOnly)
	= PUBLICAPIGLUE(60);

pascal JobFormatMode GetJobFormatMode (Job)
	= PUBLICAPIGLUE(61);

pascal void SetJobFormatMode (Job, JobFormatMode)
	= PUBLICAPIGLUE(62);

pascal void JobFormatModeQuery (Job, QueryType, void *srcData, void *dstData)
	= PUBLICAPIGLUE(63);


// Dialog Routines  (Advanced)


pascal void EnableJobScalingPanel (Job, Boolean enabled)
	= PUBLICAPIGLUE(64);

pascal void GetJobPanelDimensions (Job, Rect *)
	= PUBLICAPIGLUE(65);


// PaperType Routines (Advanced)


pascal long CountJobPaperTypes (Job, Boolean forFormatDevice)
	= PUBLICAPIGLUE(66);

pascal PaperType GetJobPaperType (Job, long whichPaperType, Boolean forFormatDevice, PaperType)
	= PUBLICAPIGLUE(67);

pascal void ForEachJobPaperTypeDo (Job, PaperTypeProc, void *refCon, Boolean forFormattingPrinter)
	= PUBLICAPIGLUE(68);

pascal PaperType CopyPaperType (PaperType srcPaperType, PaperType dstPaperType)
	= PUBLICAPIGLUE(69);

pascal void GetPaperTypeName (PaperType, Str31)
	= PUBLICAPIGLUE(70);

pascal void GetPaperTypeDimensions (PaperType, rectangle *pageSize, rectangle *paperSize)
	= PUBLICAPIGLUE(71);

pascal Job GetPaperTypeJob (PaperType)
	= PUBLICAPIGLUE(72);

pascal Collection GetPaperTypeCollection (PaperType)
	= PUBLICAPIGLUE(73);


// PrintViewer Routines (Advanced)


pascal PrintFile OpenPrintFile (Job, FSSpecPtr, char permission)
	= PUBLICAPIGLUE(74);

pascal void ClosePrintFile (PrintFile)
	= PUBLICAPIGLUE(75);

pascal Job GetPrintFileJob (PrintFile)
	= PUBLICAPIGLUE(76);

pascal long CountPrintFilePages (PrintFile)
	= PUBLICAPIGLUE(77);

pascal void ReadPrintFilePage (PrintFile, long pageNumber, long numViewPorts, viewPort *viewPortList, Format *, shape *)
	= PUBLICAPIGLUE(78);

pascal void ReplacePrintFilePage (PrintFile, long pageNumber, Format, shape)
	= PUBLICAPIGLUE(79);

pascal void InsertPrintFilePage (PrintFile, long atPageNumber, Format, shape)
	= PUBLICAPIGLUE(80);

pascal void DeletePrintFilePageRange (PrintFile, long fromPageNumber, long toPageNumber)
	= PUBLICAPIGLUE(81);

pascal void SavePrintFile (PrintFile, FSSpec *)
	= PUBLICAPIGLUE(82);


#ifdef __cplusplus
}
#endif


#endif __PRINTINGMANAGER__
