/* ------------------------------------------------------------------------------

	FILENAME
		PrintingMessages.h

	DESCRIPTION
		This file defines all the universal printing messages,
		and Printing Manager API functions callable only from
		within a message override.

	COPYRIGHT
		Copyright  Apple Computer, Inc. 1989-1993
		All rights reserved. 
	
------------------------------------------------------------------------------- */


#ifndef __PRINTINGMESSAGES__
#define __PRINTINGMESSAGES__


#ifndef __PRINTINGMANAGER__
	#include <PrintingManager.h>
#endif

#ifndef __PRINTINGRESTYPES__
	#include <PrintingResTypes.h>
#endif

#ifndef __MESSAGEMANAGER__
	#include <Messages.h>
#endif

#ifndef __MENUS__
	#include <Menus.h>
#endif

#ifndef __LISTS__
	#include <Lists.h>
#endif


/* ------------------------------------------------------------------------------

									Constants and Types

-------------------------------------------------------------------------------- */



/*--------------------------*/
/* abstract data types...   */
/*--------------------------*/



typedef struct PrivateFileRecord *SpoolFile;



/*---------------------------------------*/
/* dialog panel constants and types...   */
/*---------------------------------------*/



typedef long PanelEvent;

enum  {													// dialog panel events
	panelNoEvt 					= (PanelEvent) 0,
	panelOpenEvt 				= (PanelEvent) 1,			// initialize and draw
	panelCloseEvt 				= (PanelEvent) 2,			// your panel is going away ( panel switch, confirm or cancel )
	panelHitEvt					= (PanelEvent) 3,			// there's a hit in your panel
	panelActivateEvt			= (PanelEvent) 4,			// the dialog window has just been activated
	panelDeactivateEvt 		= (PanelEvent) 5,			// the dialog window is about to be deactivated
	panelIconFocusEvt 		= (PanelEvent) 6,			// the focus changes from the panel to the icon list
	panelPanelFocusEvt 		= (PanelEvent) 7,			// the focus changes from the icon list to the panel
	panelFilterEvt				= (PanelEvent) 8,			// this is called to filter every event
	panelCancelEvt				= (PanelEvent) 9,			// the user has cancelled the dialog
	panelConfirmEvt			= (PanelEvent) 10,		// the user has confirmed the dialog
	panelDialogEvt				= (PanelEvent) 11,		// event to be handle by dialoghandler ( you get to see all events )
	panelOtherEvt				= (PanelEvent) 12,		// osEvts, etc.
	panelUserWillConfirmEvt	= (PanelEvent) 13			// user has selected confirm, time to parse panel interdependencies
};


typedef long PanelResult;

enum {										// possible panel responses to dialog handler calls
	panelNoResult			 	= 0,
	panelCancelConfirmation = 1		// only valid from panelUserWillConfirmEvt - used to keep the dialog from going away
};


enum {										// values for the evtAction field in PanelInfoRecord
	otherAction 				= 0,				// current item will not change
	closePanelAction			= 1,				// panel will be closed
	cancelDialogAction		= 2,				// dialog will be cancelled
	confirmDialogAction		= 3				// dialog will be confirmed
};


typedef struct {							// for FilterPanelEvent and HandlePanelEvent messages
	PanelEvent 		panelEvt;					// why we're calling
	short 			panelResId;					// 'panl' resource id of current panel
	DialogPtr 		pDlg;							// pointer to dialog
	EventRecord 	*theEvent;					// pointer to event
	short 			itemHit;						// actual item number as dialog mgr thinks
	short 			itemCount;					// number of items before your items
	short				evtAction;					// once this event is processed, the action that will result
															// (only meaningful in filtering, used for parsing)
	short				errorStringId;				// STR id of string to put in error alert (0 means no string)
	Format			theFormat;					// the current format (only meaningful in a format dialog)
	void				*refCon;						// refCon passed in PanelSetupRecord
} PanelInfoRecord;
					

typedef long PrintingPanelKind;

enum {
	applicationPanel 	= (PrintingPanelKind) 0,
	extensionPanel		= (PrintingPanelKind) 1,
	driverPanel			= (PrintingPanelKind) 2
};
	

// The PanelSetupInfo structure is passed to the dialog handler when the client calls
// to setup panels for the dialog


typedef struct {
	PrintingPanelKind			panelKind;
	short 						panelResId;
	short 						resourceRefNum;
	void							*refCon;
} PanelSetupRecord;


typedef long ParsePageRangeResult;		// returned by ParsePageRange message

enum {
	rangeNotParsed		= (ParsePageRangeResult) 0,		// default initial value
	rangeParsed			= (ParsePageRangeResult) 1,
	rangeBadFromValue	= (ParsePageRangeResult) 2,
	rangeBadToValue	= (ParsePageRangeResult) 3
};



/*----------------------------------------*/
/* status-related constants and types...  */
/*----------------------------------------*/



// status type ids


enum {
	nonFatalError = 1,				// effects icon on spooling dialog
	fatalError = 2,					// sends up cool alert on spooling dialog
	printerReady = 3,					// signals PFE to leave alert mode
	userAttention = 4,				// signals initiation of a modal alert
	userAlert = 5,						// signals initiation of a moveable modal alert
	pageTransmission = 6,			// signals page sent to printer, increments page count in strings to user
	openConnectionStatus = 7,		// signals PFE to begin animation on printer icon
	informationalStatus = 8,		// default status type, no side effects
	spoolingPageStatus = 9,			// signals page spooled, increments page count in spooling dialog
	endStatus = 10,					// signals end of spooling
	percentageStatus = 11			// signals the PFE as to the amount of the job which is currently complete
};


typedef struct {							// for status messages
	unsigned short		statusType;			// one of the ids listed above ( nonFatalError, etc. )
	unsigned short		statusId;			// specific status ( out of paper, etc. ) - To be defined
	unsigned short		statusAlertId;		//	cool alert id (if any) for status
	Signature			statusOwner;		// creator type of status owner
	short					statResId;			// id for 'stat' resource
	short					statResIndex;		// index into 'stat' resource for this status
	short					dialogResult;		// id of button string selected on dismissal of cool alert
	unsigned short		bufferLen;			// number of bytes in status buffer - total record size must be <= 512
	char					statusBuffer[1];	// user response from alert
} StatusRecord;




typedef struct {							// for WriteStatusToDTPWindow message
	Boolean			useText;
	Handle			hPicture;
	Str255			theText;
} DisplayRecord;



/*-----------------------------------------------*/
/* paper mapping-related constants and types...  */
/*-----------------------------------------------*/


typedef long TrayMapping;

enum {
	defaultTrayMapping		= (TrayMapping) 0,
	configuredTrayMapping 	= (TrayMapping) 1
};



/* ------------------------------------------------------------------------------

				API Functions callable only from within message overrides

-------------------------------------------------------------------------------- */



#define PROTECTEDAPIGLUE(selector) {0x203C, 0x0001, selector, 0xABFE}

#define PRINTINGDISPATCH(segID, selector) {0x203C, 0x0001, 0, 0x223C, (segID & 0x0FFF), selector << 2, 0xABFE}


#ifdef __cplusplus
extern "C" {
#endif


#ifdef applec

	#pragma parameter __D0 PrintingDispatch(__D1)
	OSErr PrintingDispatch (long selector, ...)
  	  = PROTECTEDAPIGLUE(0);

#else

	OSErr PrintingDispatch (long selector, ...)
  	  = {0x221F, 0x203C, 0x0001, 0x0000, 0xABFE, 0x598F};

#endif applec

/*
	How to use the PRINTINGDISPATCH macro...
	
	If your driver or extension is large, you may want to segment it
	across smaller boundaries than is permitted by the messaging system.
	Without using the Printing Manager's segmentation manager directly,
	the smallest segment you can create consists of the code to override
	a single message.  If you are overriding workhorse messages such as
	RenderPage, you may want to divide up the work among many functions
	distributed across several segments.  Here's how...
	
	The Printing Manager segment scheme involves the construction of a
	single 32-bit dispatch selector, which contains all the information
	necessary for the dispatcher to find a single routine.  It contains the
	segment's resource ID, and the offset within the segment which contains
	the start of the routine.  The PRINTINGDISPATCH macro will construct the
	dispatch selector for you, as well as the code to do the dispatch.
	
	Usually, it is convenient to start your segment with a long aligned jump table,
	beginning after the 4 byte header required by the Printing Manager.  The
	macro assumes this is the case and takes a 1-based routine selector from
	which it conmstructs the offset.
	
	For example, if your code is in resource 'pdvr' (print driver), ID=2
	at offset=12 (third routine in segment), you would declare your
	routine as follows:
	
	OSErr MyRenderingRoutine (long param1, Ptr param2)
		= PRINTINGDISPATCH(2, 3);
		
	Remember, ALL segment dispatches must return OSErr.  If your routine
	does not generate errors, you must still declare it to return OSErr
	and have the routine itself return noErr.
	
	An alternative way to call across segments is to call the PrintingDispatch
	function directly.  You must construct the 32-bit selector yourself and pass
	it as the first parameter.  This is usually not preferable since you don't get
	type-checking unless you declare a prototype as shown above, and your code
	isn't as easy to read.
	
	So given the above prototype, there are two ways to call the function:
	
		anErr = MyRenderingRoutine(p1, p2);			// free type checking!
		
	or:
	
		#define kMyRenderRoutineSelector 0x0002000C
		anErr = PrintingDispatch(kMyRenderRoutineSelector, p1, p2);		// no type-checking!
	
	
	Both have the same effect.
*/


Job GetJob (void)
    = PROTECTEDAPIGLUE(1);

short GetMyResFile (void)
	= PROTECTEDAPIGLUE(2);
	
Boolean SpoolingAborted (void)
	= PROTECTEDAPIGLUE(3);
	
OSErr JobIdle (void)
	= PROTECTEDAPIGLUE(4);

OSErr ReportStatus (short statusID, unsigned short statusIndex)
	= PROTECTEDAPIGLUE(5);

OSErr AlertTheUser (StatusRecord *)
	= PROTECTEDAPIGLUE(6);

OSErr SetupDialogPanel (PanelSetupRecord *)
	= PROTECTEDAPIGLUE(7);

OSErr CountTrays (TrayIndex *numTrays)
	= PROTECTEDAPIGLUE(8);
	
OSErr GetTrayName (TrayIndex trayNumber, Str31 trayName)
	= PROTECTEDAPIGLUE(9);

OSErr SetTrayPaperType (TrayIndex whichTray, PaperType)
	= PROTECTEDAPIGLUE(10);

OSErr GetTrayPaperType (TrayIndex whichTray, PaperType)
	= PROTECTEDAPIGLUE(11);

OSErr GetTrayMapping (TrayMapping *trayMapping)
	= PROTECTEDAPIGLUE(12);

void CleanupStartJob (void)
	= PROTECTEDAPIGLUE(13);

void CleanupStartPage (void)
	= PROTECTEDAPIGLUE(14);

void CleanupOpenConnection (void)
	= PROTECTEDAPIGLUE(15);

void CleanupStartSendPage (void)
	= PROTECTEDAPIGLUE(16);



/* ------------------------------------------------------------------------------

					Constants and types for Universal Printing Messages

-------------------------------------------------------------------------------- */



enum {										// options for CreateSpoolFile message
	noCreateOptions	= 0x00000000,	// just create the file
	inhibitAlias		= 0x00000001,	// do not create an alias in the PMD folder
	inhibitUniqueName	= 0x00000002	// do not append to the filename to make it unique
};


enum {										// options for CloseSpoolFile message
	noCloseOptions	= 0x00000000,		// just close the file
	deleteOnClose	= 0x00000001,		// delete the file rather than closing it
	updateJobData	= 0x00000002,		// write current job information into file prior to closing
	makeRemoteFile	= 0x00000004		// mark job as a remote file
};


enum { 										// options for CreateImageFile message
	noImageFile 	= 0x00000000,				// don't create image file
	makeImageFile 	= 0x00000001,				// create an image file
	eachPlane		= 0x00000002,				// only save up planes before rewinding
	eachPage			= 0x00000004,				// save up entire pages before rewinding
	entireFile		= eachPlane + eachPage	// save up the entire file before rewinding
};


enum {										// options for BufferData message
	noBufferOptions	= 0x00000000, 
	makeBufferHex		= 0x00000001, 
	dontSplitBuffer	= 0x00000002
};


typedef struct {							// for DumpBuffer and FreeBuffer messages
	long			size;							//	size of buffer in bytes
	long			userData;					//	client assigned id for the buffer
	char			data[1];						// array of size bytes
} PrintingBuffer;


typedef struct {							// for RenderPage message
	long		docPageNum;						// number of page being printed
	long		copyNum;							// copy number being printed
	Boolean	formatChanged;					// true if format changed from last page
	Boolean	pageChanged;					// true if page contents changed from last page
	long		internalUse;					// private
} PageInfoRecord;



/* ------------------------------------------------------------------------------

								Universal Printing Messages

-------------------------------------------------------------------------------- */


#define UNIVSENDGLUE(selector) SendMessageGlue(0x0000, selector)
#define FORWARDGLUE ForwardThisMessageGlue


#define Send_FetchTaggedDriverData(tag, id, pHandle) Send_FetchTaggedData(tag, id, pHandle, 'drvr')
#define Forward_FetchTaggedDriverData(tag, id, pHandle) Forward_FetchTaggedData(tag, id, pHandle, 'drvr')


OSErr Send_JobIdle (void)
    = UNIVSENDGLUE(jobIdle);
OSErr Forward_JobIdle (void)
    = FORWARDGLUE;

OSErr Send_JobStatus (StatusRecord *)
    = UNIVSENDGLUE(jobStatus);
OSErr Forward_JobStatus (StatusRecord *)
    = FORWARDGLUE;

OSErr Send_PrintingEvent (EventRecord *, Boolean filterEvent)
    = UNIVSENDGLUE(printingEvent);
OSErr Forward_PrintingEvent (EventRecord *, Boolean filterEvent)
    = FORWARDGLUE;


OSErr Send_JobDefaultFormatDialog (DialogResult *)
    = UNIVSENDGLUE(jobDefaultFormatDialog);
OSErr Forward_JobDefaultFormatDialog (DialogResult *)
    = FORWARDGLUE;
	
OSErr Send_FormatDialog (Format, StringPtr title, DialogResult *)
    = UNIVSENDGLUE(formatDialog);
OSErr Forward_FormatDialog (Format, StringPtr title, DialogResult *)
    = FORWARDGLUE;
	
OSErr Send_JobPrintDialog (DialogResult *)
    = UNIVSENDGLUE(jobPrintDialog);
OSErr Forward_JobPrintDialog (DialogResult *)
    = FORWARDGLUE;

OSErr Send_FilterPanelEvent (PanelInfoRecord *, Boolean *)
	= UNIVSENDGLUE(filterPanelEvent);

OSErr Send_HandlePanelEvent (PanelInfoRecord *, PanelResult *)
	= UNIVSENDGLUE(handlePanelEvent);

OSErr Send_ParsePageRange (StringPtr fromString, StringPtr toString, ParsePageRangeResult *result)
	= UNIVSENDGLUE(parsePageRange);
OSErr Forward_ParsePageRange (StringPtr fromString, StringPtr toString, ParsePageRangeResult *result)
	= FORWARDGLUE;
	
	
OSErr Send_DefaultJob (void)
    = UNIVSENDGLUE(defaultJob);
OSErr Forward_DefaultJob (void)
    = FORWARDGLUE;
	
OSErr Send_DefaultFormat (Format)
    = UNIVSENDGLUE(defaultFormat);
OSErr Forward_DefaultFormat (Format)
    = FORWARDGLUE;

OSErr Send_DefaultPaperType (PaperType)
    = UNIVSENDGLUE(defaultPaperType);
OSErr Forward_DefaultPaperType (PaperType)
    = FORWARDGLUE;
	
OSErr Send_DefaultPrinter (Printer)
    = UNIVSENDGLUE(defaultPrinter);
OSErr Forward_DefaultPrinter (Printer)
    = FORWARDGLUE;
	

OSErr Send_CreateSpoolFile (FSSpecPtr, long createOptions, SpoolFile *)
    = UNIVSENDGLUE(createSpoolFile);
OSErr Forward_CreateSpoolFile (FSSpecPtr, long createOptions, SpoolFile *)
    = FORWARDGLUE;
	
OSErr Send_SpoolPage (SpoolFile, Format, shape)
    = UNIVSENDGLUE(spoolPage);
OSErr Forward_SpoolPage (SpoolFile, Format, shape)
    = FORWARDGLUE;

OSErr Send_SpoolData (SpoolFile, Ptr data, long *length)
    = UNIVSENDGLUE(spoolData);
OSErr Forward_SpoolData (SpoolFile, Ptr data, long *length)
    = FORWARDGLUE;

OSErr Send_SpoolResource (SpoolFile, Handle, ResType, short id)
    = UNIVSENDGLUE(spoolResource);
OSErr Forward_SpoolResource (SpoolFile, Handle, ResType, short id)
    = FORWARDGLUE;

OSErr Send_CompleteSpoolFile (SpoolFile)
    = UNIVSENDGLUE(completeSpoolFile);
OSErr Forward_CompleteSpoolFile (SpoolFile)
    = FORWARDGLUE;


OSErr Send_CountPages (SpoolFile, long *numPages)
    = UNIVSENDGLUE(countPages);
OSErr Forward_CountPages (SpoolFile, long *numPages)
    = FORWARDGLUE;

OSErr Send_DespoolPage (SpoolFile, long, Format, shape *,  Boolean *formatChanged)
    = UNIVSENDGLUE(despoolPage);
OSErr Forward_DespoolPage (SpoolFile, long, Format, shape *,  Boolean *formatChanged)
    = FORWARDGLUE;

OSErr Send_DespoolData (SpoolFile, Ptr data, long *length)
    = UNIVSENDGLUE(despoolData);
OSErr Forward_DespoolData (SpoolFile, Ptr data, long *length)
    = FORWARDGLUE;

OSErr Send_DespoolResource (SpoolFile, ResType, short id, Handle *)
    = UNIVSENDGLUE(despoolResource);
OSErr Forward_DespoolResource (SpoolFile, ResType, short id, Handle *)
    = FORWARDGLUE;

OSErr Send_CloseSpoolFile (SpoolFile, long closeOptions)
    = UNIVSENDGLUE(closeSpoolFile);
OSErr Forward_CloseSpoolFile (SpoolFile, long closeOptions)
    = FORWARDGLUE;
	
	
OSErr Send_StartJob (StringPtr docName, long pageCount)
    = UNIVSENDGLUE(startJob);
OSErr Forward_StartJob (StringPtr docName, long pageCount)
    = FORWARDGLUE;
	
OSErr Send_FinishJob (void)
    = UNIVSENDGLUE(finishJob);
OSErr Forward_FinishJob (void)
    = FORWARDGLUE;
	
OSErr Send_StartPage (Format, long numViewPorts, viewPort *viewPortList)
	= UNIVSENDGLUE(startPage);
OSErr Forward_StartPage (Format, long numViewPorts, viewPort *viewPortList)
	= FORWARDGLUE;

OSErr Send_FinishPage (void)
	= UNIVSENDGLUE(finishPage);
OSErr Forward_FinishPage (void)
	= FORWARDGLUE;
	
OSErr Send_PrintPage (Format, shape)
	= UNIVSENDGLUE(printPage);
OSErr Forward_PrintPage (Format, shape)
	= FORWARDGLUE;
	

OSErr Send_SetupImageData (void *imageData)
    = UNIVSENDGLUE(setupImageData);
OSErr Forward_SetupImageData (void *imageData)
    = FORWARDGLUE;

OSErr Send_ImageJob(SpoolFile, long *closeOptions)
    = UNIVSENDGLUE(imageJob);
OSErr Forward_ImageJob (SpoolFile, long *closeOptions)
    = FORWARDGLUE;

OSErr Send_ImageDocument (SpoolFile, void *imageData)
    = UNIVSENDGLUE(imageDocument);
OSErr Forward_ImageDocument (SpoolFile, void *imageData)
    = FORWARDGLUE;

OSErr Send_ImagePage (SpoolFile, long pageNumber, Format, void *imageData)
    = UNIVSENDGLUE(imagePage);
OSErr Forward_ImagePage (SpoolFile, long pageNumber, Format, void *imageData)
    = FORWARDGLUE;

OSErr Send_RenderPage (Format, shape, PageInfoRecord *, void *imageData)
    = UNIVSENDGLUE(renderPage);
OSErr Forward_RenderPage (Format, shape, PageInfoRecord *, void *imageData)
    = FORWARDGLUE;


OSErr Send_CreateImageFile (FSSpecPtr, long imageFileOptions, long *)
    = UNIVSENDGLUE(createImageFile);
OSErr Forward_CreateImageFile (FSSpecPtr, long imageFileOptions, long *)
    = FORWARDGLUE;


OSErr Send_OpenConnection (void)
    = UNIVSENDGLUE(openConnection);
OSErr Forward_OpenConnection (void)
    = FORWARDGLUE;

OSErr Send_CloseConnection (void)
    = UNIVSENDGLUE(closeConnection);
OSErr Forward_CloseConnection (void)
    = FORWARDGLUE;

OSErr Send_StartSendPage (Format)
    = UNIVSENDGLUE(startSendPage);
OSErr Forward_StartSendPage (Format)
    = FORWARDGLUE;

OSErr Send_FinishSendPage (void)
    = UNIVSENDGLUE(finishSendPage);
OSErr Forward_FinishSendPage (void)
    = FORWARDGLUE;


OSErr Send_WriteData (Ptr data, long length)
    = UNIVSENDGLUE(writeData);
OSErr Forward_WriteData (Ptr data, long length)
    = FORWARDGLUE;

OSErr Send_BufferData (Ptr data, long length, long bufferOptions)
    = UNIVSENDGLUE(bufferData);
OSErr Forward_BufferData (Ptr data, long length, long bufferOptions)
    = FORWARDGLUE;

OSErr Send_DumpBuffer (PrintingBuffer *)
    = UNIVSENDGLUE(dumpBuffer);
OSErr Forward_DumpBuffer (PrintingBuffer *)
    = FORWARDGLUE;

OSErr Send_FreeBuffer (PrintingBuffer *)
    = UNIVSENDGLUE(freeBuffer);
OSErr Forward_FreeBuffer (PrintingBuffer *)
    = FORWARDGLUE;


OSErr Send_CheckStatus (Ptr data, long length, short statusType, Signature owner)
    = UNIVSENDGLUE(checkStatus);
OSErr Forward_CheckStatus (Ptr data, long length, short statusType, Signature owner)
    = FORWARDGLUE;

OSErr Send_GetDeviceStatus (Ptr cmdData, long cmdSize, Ptr responseData, long *responseSize, Str255 termination)
    = UNIVSENDGLUE(getDeviceStatus);
OSErr Forward_GetDeviceStatus (Ptr cmdData, long cmdSize, Ptr responseData, long *responseSize, Str255 termination)
    = FORWARDGLUE;


OSErr Send_FetchTaggedData (ResType, short id, Handle *, Signature owner)
	= UNIVSENDGLUE(fetchTaggedData);
OSErr Forward_FetchTaggedData (ResType, short id, Handle *, Signature owner)
	= FORWARDGLUE;


OSErr	Send_GetDTPMenuList (MenuHandle)
	= UNIVSENDGLUE(getDTPMenuList);
OSErr	Forward_GetDTPMenuList (MenuHandle)
	= FORWARDGLUE;

OSErr	Send_DTPMenuSelect (short id, Boolean *done)
	= UNIVSENDGLUE(dtpMenuSelect);
OSErr	Forward_DTPMenuSelect (short id, Boolean *done)
	= FORWARDGLUE;

OSErr	Send_DTPMenuEventFilter (short id, Boolean *done, Boolean *handledEvent, EventRecord *)
	= UNIVSENDGLUE(dtpMenuEventFilter);
OSErr	Forward_DTPMenuEventFilter (short id, Boolean *done, Boolean *handledEvent, EventRecord *)
	= FORWARDGLUE;
	

OSErr Send_JobFormatModeQuery (QueryType, void *srcData, void *dstData)
	= UNIVSENDGLUE(jobFormatModeQuery);
OSErr Forward_JobFormatModeQuery (QueryType, void *srcData, void *dstData)
	= FORWARDGLUE;


OSErr Send_WriteStatusToDTPWindow (StatusRecord *, DisplayRecord *)
	= UNIVSENDGLUE(writeStatusToDTPWindow);
OSErr Forward_WriteStatusToDTPWindow (StatusRecord *, DisplayRecord *)
	= FORWARDGLUE;

OSErr Send_InitializeStatusAlert (StatusRecord *, DialogPtr *)
	= UNIVSENDGLUE(initializeStatusAlert);
OSErr Forward_InitializeStatusAlert (StatusRecord *, DialogPtr *)
	= FORWARDGLUE;

OSErr Send_HandleAlertStatus (StatusRecord *)
	= UNIVSENDGLUE(handleAlertStatus);
OSErr Forward_HandleAlertStatus (StatusRecord *)
	= FORWARDGLUE;
	
OSErr Send_HandleAlertEvent (StatusRecord *, DialogPtr, EventRecord *, short *response)
	= UNIVSENDGLUE(handleAlertEvent);
OSErr Forward_HandleAlertEvent (StatusRecord *, DialogPtr, EventRecord *, short *response)
	= FORWARDGLUE;
		

void Send_CleanupStartJob (void)
	= UNIVSENDGLUE(cleanupStartJob);
void Forward_CleanupStartJob (void)
	= FORWARDGLUE;

void Send_CleanupStartPage (void)
	= UNIVSENDGLUE(cleanupStartPage);
void Forward_CleanupStartPage (void)
	= FORWARDGLUE;

void Send_CleanupOpenConnection (void)
	= UNIVSENDGLUE(cleanupOpenConnection);
void Forward_CleanupOpenConnection (void)
	= FORWARDGLUE;

void Send_CleanupStartSendPage (void)
	= UNIVSENDGLUE(cleanupStartSendPage);
void Forward_CleanupStartSendPage (void)
	= FORWARDGLUE;


OSErr Send_DefaultDesktopPrinter (Str31)
	= UNIVSENDGLUE(defaultDesktopPrinter);
OSErr Forward_DefaultDesktopPrinter (Str31)
	= FORWARDGLUE;
	
OSErr	Send_CaptureOutputDevice (Boolean capture)
	= UNIVSENDGLUE(captureOutputDevice);
OSErr	Forward_CaptureOutputDevice (Boolean capture)
	= FORWARDGLUE;


OSErr Send_OpenConnectionRetry (ResType, void *, Boolean *, OSErr)
	= UNIVSENDGLUE(openConnectionRetry);
OSErr Forward_OpenConnectionRetry (ResType, void *, Boolean *, OSErr)
	= FORWARDGLUE;
	
OSErr Send_ExamineSpoolFile (SpoolFile)
	= UNIVSENDGLUE(examineSpoolFile);
OSErr Forward_ExamineSpoolFile (SpoolFile)
	= FORWARDGLUE;

OSErr Send_FinishSendPlane (void)
	= UNIVSENDGLUE(finishSendPlane);
OSErr Forward_FinishSendPlane (void)
	= FORWARDGLUE;

OSErr Send_DoesPaperFit( PaperType paper, Boolean *fits )
	= UNIVSENDGLUE(doesPaperFit);
OSErr Forward_DoesPaperFit( PaperType paper, Boolean *fits )
	= FORWARDGLUE;
	
OSErr Send_ChooserMessage (short message, short caller, StringPtr objName,
									StringPtr zoneName, ListHandle theList, long p2)
	= UNIVSENDGLUE(chooserMessage);
OSErr Forward_ChooserMessage (short message, short caller, StringPtr objName,
										StringPtr zoneName, ListHandle theList, long p2)
	= FORWARDGLUE;
	

#ifdef __cplusplus
}
#endif


#endif __PRINTINGMESSAGES__
