/* layout types.h -- Line Layout types

	by Dave Opstad and Eric Mader,
	with special thanks to the following folks:
		 Mark Davis, for starting it all off;
		 Russ Sonnenschein, for many good ideas;
		 Cary Clark, Dave Good, Rob Johnson, Keith McGreggor, and Chris Yerga for the graphics system;
		 Richard Becker, Lynn Bekkala and Mike Reed for fonts and the font system;
		 Robert Dierkes and the SQE team for testing and good ideas;
		 Parry Kejriwal and Richard Klein, for font tools;
		 Diane Patterson, for great patience in writing documentation;
		 Hugo Ayala, Tom Dowdy, Dan Lipton, Sean Parent and the rest of the PrintShop;
		 Mitch Allen for managing to manage us; and
		 many other good folk at Apple!
*/

/* Copyright 1988-1992 Apple Computer, Inc.  All rights reserved. */

#ifndef layoutTypesIncludes
#define layoutTypesIncludes

#ifndef mathTypesIncludes
#include "math types.h"
#endif mathTypesIncludes

#ifndef graphicsTypesIncludes
#include "graphics types.h"
#endif graphicsTypesIncludes

typedef unsigned short Glyphcode;   /* single glyph in a font */
typedef long ByteOffset;            /* byte offset within backing store */

/* The actual constants for feature types and selectors have been moved to a library. */

typedef unsigned short RunFeatureType;
typedef unsigned short RunFeatureSelector;

/* If tracking is not desired, specify the following value in the track field in the
    RunControls record (note that a track of 0 does *not* mean to turn tracking off;
    rather, it means to use normal tracking). */

#define noTracking negativeInfinity

/* The special "noStake" value is returned by the GetLayoutBreakOffset call to indicate
    the absence of a character offset that is stable with respect to metamorphosis and
    contextual kerning. */

#define noStake -1

/* A glyph's behavior with respect to other glyphs on its line is defined in part by its
    BaselineType. These types allow correct alignment of the baselines of all glyphs on
    the line. */

enum {
  romanBaseline = 0,
  ideographicCenterBaseline,
  ideographicLowBaseline,
  hangingBaseline,
  mathBaseline,
  lastBaseline = 31,
  numberOfBaselineTypes = lastBaseline + 1,
  noOverrideBaseline = 255
  };
typedef unsigned long BaselineType;

typedef fixed BaselineDeltas[numberOfBaselineTypes];

/* JustificationPriority defines the point during the justification process at which a
    glyph will begin to receive deltas before and after itself. */

enum {
  kashidaPriority           = 0,
  whiteSpacePriority        = 1,
  interCharPriority         = 2,
  nullJustificationPriority = 3,
  numberOfJustificationPriorities
  };
typedef unsigned char JustificationPriority;

/* JustificationFlags are used to control which fields of a WidthDeltaRecord are to be
    overridden and which are not if a PriorityJustificationOverride or
    GlyphJustificationOverride (qq.v.) is specified. */

enum {
  overridePriority          = 0x8000, /* use priority value from override */
  overrideLimits            = 0x4000, /* use limits values from override */
  overrideUnlimited         = 0x2000, /* use unlimited flag from override */
  unlimitedGapAbsorption    = 0x1000, /* glyph can take unlimited gap */
  justificationPriorityMask = 0x000F, /* justification priority */

	allJustificationFlags			= overridePriority|overrideLimits|overrideUnlimited|
															unlimitedGapAbsorption|justificationPriorityMask
  };
typedef unsigned short JustificationFlags;

/* The directional behavior of a glyph can be overridden using a DirectionOverride. */

enum {
  noDirectionOverride = 0,
  imposeLeftToRight   = 1,
  imposeRightToLeft   = 2,
  imposeArabic        = 3
  };
typedef unsigned short DirectionOverride;

/* RunControlFlags describe the nonparametric layout options contained in a style. */

#define noLigatureSplits        0x80000000
#define noCaretAngle            0x40000000
#define imposeWidth             0x20000000
#define noCrossKerning          0x10000000
#define noOpticalAlignment      0x08000000
#define forceHanging            0x04000000
#define noSpecialJustification  0x02000000
#define directionOverrideMask   0x00000003

#define allRunControlFlags (noLigatureSplits|noCaretAngle|imposeWidth|noCrossKerning|noOpticalAlignment|forceHanging|noSpecialJustification|directionOverrideMask)

typedef unsigned long RunControlFlags;

/* HighlightType is used to distinguish various highlighting methods, both in terms of
    character offset based vs. visual based, and in terms of angled sides vs. non-angled
    sides. */

enum {
  highlightStraight     = 0,  /* straight-edged simple highlighting */
  highlightAverageAngle = 1   /* takes average of two edge angles */
  };
typedef unsigned long HighlightType;

/* CaretType is used to control whether the caret that is returned from GetLayoutCaret
    is a split caret or a (keyboard-syncronized) single caret. */

enum {
  splitCaretType         = 0,  /* returns Mac-style split caret (default) */
  leftRightKeyboardCaret = 1,  /* single caret in left-right position */
  rightLeftKeyboardCaret = 2   /* single caret in right-left position */
  };
typedef unsigned long CaretType;

/* LayoutOffsetState describes the characteristics of a given ByteOffset in some
    layout. It is returned by the GetOffsetGlyphs call. Note that the offsetInsideLigature
    value is returned in addition to the 8/16 (or invalid) indication. */

enum {
  offset8_8            = 0,
  offset8_16           = 1,
  offset16_8           = 2,
  offset16_16          = 3,
  offsetInvalid        = 4,
  offsetInsideLigature = 0x8000
  };
typedef unsigned short LayoutOffsetState;

/* LayoutOptionsFlags are single-bit flags contained in a LayoutOptions record. We also
    define here some utility constants that are useful in setting various fields in the
    LayoutOptions record. */

#define noLayoutOptions 0

#define lineIsDisplayOnly   0x00000001

#define allLayoutOptionsFlags (lineIsDisplayOnly)

#define maxRunLevel 15
#define flushLeft 0
#define flushCenter (frac1/2)
#define flushRight frac1

#define noJustification 0
#define fullJustification frac1

typedef unsigned long LayoutOptionsFlags;

/* A RunFeature describes a feature and a level for that feature. */

typedef struct {
  RunFeatureType      featureType;
  RunFeatureSelector  featureSelector;
  } RunFeature;

/* A WidthDeltaRecord contains all of the information needed to describe the behavior of one
    class of glyphs during the justification process. */

typedef struct {
  fixed               beforeGrowLimit;    /* ems AW can grow by at most on LT */
  fixed               beforeShrinkLimit;  /* ems AW can shrink by at most on LT */
  fixed               afterGrowLimit;     /* ems AW can grow by at most on RB */
  fixed               afterShrinkLimit;   /* ems AW can shrink by at most on RB */
  JustificationFlags  growFlags;          /* flags controlling grow case */
  JustificationFlags  shrinkFlags;        /* flags controlling shrink case */
  } WidthDeltaRecord;

/* PriorityJustificationOverride contains an array of WidthDeltaRecords, one for each
    JustificationPriority. */

typedef struct {
  WidthDeltaRecord deltas[numberOfJustificationPriorities]; /* overrides for each of the priorities */
  } PriorityJustificationOverride;

/* GlyphJustificationOverride contains a WidthDeltaRecord that is to be used for a
    specific glyph in a specific run (this limitation is because glyphcodes vary by font). */

typedef struct {
  Glyphcode        glyph;
  WidthDeltaRecord override;
  } GlyphJustificationOverride;

/* RunControls contains flags, shifts, imposed widths and overrides for a run. */
/* NOTE: a value of "noTracking" in track disables tracking */

typedef struct {
  RunControlFlags flags;
  fixed           beforeWithStreamShift;
  fixed           afterWithStreamShift;
  fixed           crossStreamShift;
  fixed           imposedWidth;
  fixed           track;
  fract           hangingInhibitFactor;
  fract           kerningInhibitFactor;
  fixed           decompositionAdjustmentFactor;
  BaselineType    baselineType;
  } RunControls;

/* A GlyphSubstitution describes one client-provided substitution that occurs after all
    other automatic glyph changes have happened. */

typedef struct {
  Glyphcode originalGlyph;    /* Whenever you see this glyph... */
  Glyphcode substituteGlyph;  /* ...change it to this one. */
  } GlyphSubstitution;

/* KerningAdjustmentFactors specify an adjustment to automatic kerning.
    The adjustment is ax + b where x is the automatic kerning value, a
    is scaleFactor, and b is adjustmentPointSizeFactor times the run's
    point size. */
    
typedef struct {
  fract scaleFactor;
  fixed adjustmentPointSizeFactor;
} KerningAdjustmentFactors;

/* A KerningAdjustment identifies with- and cross-stream kerning adjustments
    for specific glyph pairs. */

typedef struct {
  Glyphcode                 firstGlyph;
  Glyphcode                 secondGlyph;
  KerningAdjustmentFactors  withStreamFactors;
  KerningAdjustmentFactors  crossStreamFactors;
  } KerningAdjustment;

/* a value of resetCrossStreamFactor in crossStreamFactors.adjustmentPointSizeFactor
    will reset the cross-stream kerning to the baseline. */
    
#define resetCrossStreamFactor negativeInfinity

/* LayoutHitInfo contains the output from the HitTestLayout call. */

typedef struct {
  fixed      firstPartialDist;
  fixed      lastPartialDist;
  ByteOffset hitSideOffset;
  ByteOffset nonHitSideOffset;
  boolean    leadingEdge;
  boolean    inLoose;
  } LayoutHitInfo;

/* A LineBaselineRecord contains the deltas from 0 to all the different baselines for
    the layout. It can be filled via a call to GetBaselineDeltas (q.v.). */

typedef struct {
  BaselineDeltas  deltas;
  } LineBaselineRecord;

/* The LayoutOptions type contains information about the layout characteristics of the
    whole line. */

typedef struct {
  fixed               width;
  fract               flush;
  fract               just;
  LayoutOptionsFlags  flags;
  LineBaselineRecord  *baselineRec;
  } LayoutOptions;

#endif layoutTypesIncludes
