/*	scaler types.h
**
**		These are the data types relevent to the API of a scaler.  Included are the
**	types for managing memory blocks, and for communicating with scalers in the input
**	and output directions.
**
**	Copyright:	 1992 by Apple Computer, Inc., all rights reserved.
**
**	Change History:
**
**	<1>		03/15/92 ABB	Created
*/

#ifndef	scalerTypeIncludes
#define	scalerTypeIncludes

#ifndef mathTypesIncludes
	#include "math types.h"
#endif

#ifndef sfntTypesIncludes
	#include "sfnt types.h"
#endif

#define	truetypeFontFormat		0x74727565	/* 'true' */
#define	type1FontFormat		0x74797031	/* 'typ1' */
#define	nfntFontFormat			0x6e666e74	/* 'nfnt' */

typedef enum {
	scaler_no_problem = 0,					/* Everything went OK */
	
	scaler_first_error,						/* Generally fatal errors */
	scaler_null_context = scaler_first_error,	/* Client passed a null context pointer */
	scaler_null_input,						/* Client passed a null input pointer */
	scaler_invalid_context,					/* There was a problem with the context */
	scaler_invalid_input,					/* There was a problem with an input */
	scaler_invalid_font_data,					/* A portion of the font was corrupt */
	scaler_new_block_failed,					/* A call to NewBlock() failed */
	scaler_bitmap_alloc_failed,				/* Call to alloc bitmap perm. block failed */
	scaler_outline_alloc_failed,				/* Call to alloc outline perm. block failed */
	scaler_required_table_missing,			/* A needed font table was not found */
	scaler_unsupported_outline_format,			/* Couldn't create an outline of the desired format */
	scaler_unsupported_stream_format,			/* ScalerStreamFont() call can't supply any requested format */
	scaler_unsupported_font_format,			/* No scaler supports the font format */
	scaler_hinting_error,					/* An error occurred during hinting */
	scaler_scan_error,						/* An error in scan conversion */
	scaler_internal_error,					/* Scaler has a bug */
	scaler_invalid_matrix,					/* The transform matrix was unusable */
	scaler_fixed_overflow,					/* An overflow ocurred during matrix operations */
	scaler_API_version_mismatch,				/* Scaler requires a newer/older version of the scaler API */
	scaler_no_kerning_applied,				/* ScalerKernGlyphs() found no kerning data, or it had no effect */
	scaler_last_error = scaler_no_kerning_applied,
	
	scaler_first_warning,					/* Informative warnings */
	scaler_no_output = scaler_first_warning,		/* Couldn't fulfill any glyph request. */
	scaler_fake_metrics,					/* Returned metrics aren't based on info in the font */
	scaler_fake_linespacing,					/* Linespacing metrics not based on info in the font */
	scaler_glyph_substitution,				/* Requested glyph out of range, a substitute was used */
	scaler_last_warning = scaler_glyph_substitution
} scalerErrors;

typedef long scalerError;

/*** ScalerOpen output type ***/

typedef struct {
	fontFormatTag	format;			/* Font format supported by this scaler */
	fixed			scalerVersion;		/* Version number of the scaler */
	fixed			APIVersion;		/* Version of API implemented (compare with version in scalerContext) */
} scalerInfo;

/*** ScalerNewFont output type ***/

typedef enum {
	requiresLayoutFont = 1,
	hasNormalLayoutFont = 2,
	canReorderFont = 4,
	canRearrangeFont = 8,
	hasOutlinesFont = 16
} scalerFontFlags;

typedef unsigned long scalerFontFlag;

typedef struct {
	unsigned long		unitsPerEm;
	scalerFontFlag		flags;
	unsigned long		numGlyphs;
} scalerFontInfo;

/*** ScalerNewTransform input types ***/

typedef enum {
	applyHintsTransform = 1,			/* Execute hinting instructions (grid fit) */
	exactBitmapTransform = 2,		/* Use embedded bitmap iff exact size */
	useThresholdTransform = 4,		/* Use scaled bitmap (if any) if below outline threshold */
	verticalTransform = 8			/* Glyphs will be in vertical orientation */
} scalerTransformFlags;

typedef unsigned long scalerTransformFlag;

typedef struct {
	scalerTransformFlag		flags;			/* Hint, embedded bitmap control, etc. */
	fixed					pointSize;			/* The desired pointsize */
	mapping				*fontMatrix;		/* The 3x3 matrix to apply to glyphs */
	point					resolution;		/* 2D device resolution */
	point					spotSize;			/* 2D pixel size */
} scalerTransform;

/*** ScalerNewTransform output type ***/

typedef struct {
	point		before;			/* Spacing of the line before */
	point		after;			/* Spacing of the line after */
	point		caretAngle;		/* Rise (y) and run (x) of the insertion caret */
	point		caretOffset;		/* adjustment to caret for variants like italic */
} scalerTransformInfo;

/*** ScalerNewGlyph input types ***/

typedef enum {
	noImageGlyph = 1				/* Don't return the bit image for this glyph */
} scalerGlyphFlags;

typedef unsigned long scalerGlyphFlag;

#define pathOutline		0x70617468	/* 'path' QuickDraw GX outline */

typedef long outlineFormat;

typedef struct {
	long				glyphIndex;	/* Index of the glyph to be considered */
	long				bandingTop;	/* Banding controls (scanline numbers) top==bottom==0 means no banding */
	long				bandingBottom;
	outlineFormat		format;		/* Format of outline to return, ignored if no outline desired */
	scalerGlyphFlag		flags;
} scalerGlyph;

/*** ScalerNewGlyph output types ***/

typedef struct {
	point		advance, sideBearing, otherSideBearing;
} scalerMetrics;

typedef struct {
	long	xMin;
	long	yMin;
	long	xMax;
	long	yMax;
} scalerRectangle;

typedef struct {
	char				*image;		/* Pointer to pixels */
	point				topLeft;		/* Bitmap positioning relative to client's origin */
	scalerRectangle		bounds;		/* Bounding box of bitmap */
	long				rowBytes;	/* Width in bytes */
#ifdef cpuPrinter
	long				actualBitmapSize;
	long				scanConvertScratchSize;
	long				dropoutScratchSize;
#endif
} scalerBitmap;

/*** ScalerKernGlyphs input/output types ***/

enum {
	lineStartKern		= 1,			/* Array of glyphs starts a line */
	lineEndKern		= 2,			/* Array of glyphs ends a line */
	noCrossKern		= 4,			/* Prohibit cross kerning */
	deviceValuesKern	= 8			/* Return device (vs. fractional) values */
};
typedef long scalerKerningFlags;

enum {
	noStakeNote = 1,
	crossStreamResetNote = 2
};
typedef unsigned short scalerKerningNotes;

typedef struct {
	long				numGlyphs;		/* Number of glyphs in the glyphs array */
	fract				scaleFactor;		/* Amount of kerning to apply (0 == none, frac1 == all) */
	scalerKerningFlags	flags;			/* Various control flags */
	unsigned short		*glyphs;			/* Pointer to the array of glyphs to be kerned */
}  scalerKerning;

/*** ScalerStream input/output types ***/

typedef enum {
	truetypeStream	= 0x0001,
	type1Stream		= 0x0002,
	type3Stream		= 0x0004,
	type42Stream		= 0x0008,
	type42GXStream	= 0x0010,
	portableStream		= 0x0020,
	flattenedStream	= 0x0040
} streamTypes;						/* Possible streamed font formats */

typedef unsigned long streamType;

typedef enum {
	downloadFlag = 0,
	asciiDownloadFlag,
	fontSizeQueryFlag,
	encodingOnlyFlag
} scalerStreamFlags;

typedef long scalerStreamFlag;

enum {
	selectAllVariations = -1
};

typedef struct scalerStream {
	void				*streamRefCon;	/* <-	private reference for client */
	streamType		type;			/* <->	data stream format */
	scalerStreamFlag	flags;			/* <-	what action to take */
	unsigned long		memorySize;		/* ->	worst case vm in printer */
	long				variationCount;		/* <-	the number of variations, or selectAllVariations */
	fontVariation		*variations;		/* <-   	a pointer to an array the variations */
	short			*encoding;		/* <-	Intention is * short[256] */
	long				*glyphBits;		/* <->	sizeof(glyphcount)/bitsperlong */
	char				*fontName;		/* <->	the printer font name to use/used */
} scalerStream;

typedef struct {
	long		hexFlag;
	long		byteCount;
	void	   	*data;
} scalerStreamData;

typedef enum {
	scalerScratchBlock = -1,		/* Scaler alloced/freed temporary memory */

	scalerOpenBlock,			/* Five permanent input/state block types */
	scalerFontBlock,
	scalerVariationBlock,
	scalerTransformBlock,
	scalerGlyphBlock,
	scalerBlockCount,			/* Number of permanent block types */
	
	scalerOutlineBlock = scalerBlockCount,			/* Two output block types */
	scalerBitmapBlock
	
} scalerBlockTypes;

typedef long scalerBlockType;

#define sfntDirectoryTag			0x64697220	/* 'dir ' */

typedef struct scalerContext {
	fixed			version;
	void			*theFont;
	fontFormatTag	format;
	
	long		(*GetFontTable)(struct scalerContext *context, fontTableTag tableTag, long offset, long length, void *data);
	void		(*ReleaseFontTable)(struct scalerContext *context, void* fontData);
	void*	(*NewBlock)(struct scalerContext *context, long size, scalerBlockType, void* oldBlock);
	void		(*DisposeBlock)(struct scalerContext *context, void* scratchData, scalerBlockType theType);
	void		(*StreamFunction)(struct scalerContext *context,  scalerStream* streamInfo, scalerStreamData *dataInfo);
	void		(*ScanLineFunction)(struct scalerContext *context,  scalerBitmap* scanLine);
	void		(*PostErrorFunction)(struct scalerContext *context, scalerError theProblem);
#ifdef cpuPrinter
	long		(*FindExtrema)( void *, void * );
	long		(*ScanChar2)( void *, void *, void *, short, short, long, struct scalerContext * );
#endif
	void*	scalerBlocks[scalerBlockCount];
	void		(*ScalerFunction)(struct scalerContext *context, void* data);
} scalerContext;

#endif