/* shape controls library:
	library to allow the user to manipulate (transform) a set of shapes
	by Dave Good and Jeff Kreegar
	Copyright 1991 - 1992 Apple Computer, Inc.  All rights reserved.	*/

#ifndef shapeControlsLibraryIncludes
	#define shapeControlsLibraryIncludes

	#include "graphics types.h"

/*
|	General Information:
|
|	 The items shape passed in is used directly by the shape controls library; the application may modify or
|	   look at it at any time. Note that if the application removes or adds a sub-shape (or changes a sub-shape
|	   in any way) then it needs to call InvalidShapeControlShape on the sub-shape.
|
|	 The foreground and background shapes are also used directly by the shape controls library and therefore
|	   can be modified at any time (and also must be invalidated). The background shape appears underneath all
|	   the items and the foreground shape appears above all the items, but below the control handles for the
|	   current selection.
|
|	 The foreground and background shapes will both be modified slightly to draw to the same list of viewPorts
|	   that the items picture goes to. Other than this, they will not be changed.
|
|	 GetShapeControlSelection and SetShapeControlSelection do NOT work directly with the selection picture
|	   inside the shape control; instead GetShapeControlSelection returns a completely new picture that contains
|	   all the sub-shapes of the item picture that are selected. The application must then dispose of this returned
|	   picture when it is through with it. SetShapeControlSelection will select all the sub-shapes of the items
|	   picture contained by shapesToSelect. The application may then dispose of this shapesToSelect picture. The
|	   shapesToSelect parameter may be a single sub-shape instead of a picture, but it must always be contained
|	   by the items picture.
|
|	 GetShapeControlSelectionHandles returns a copy of the shape that is drawn as the selections handles.
|
|	 SendEventToShapeControl returns true if it handled the event and false if it didnt. The events handled are:
|		* keyDown: delete removes the item
|		* mouseDown: in a destination viewPort for the items picture, the selection will be tracked; any other
|		   mouse event will return false.
|
|	 InvalidateShapeControlShape does NOT require that invalidShape be a member of the items picture for the
|	   shape control. However, if invalidShape is a member and it is selected, then the control handles will be
|	   added to the invalid area also. An application should use this call when invalidating a picture containly
|	   several widely separated disjoint areas; it should probably use InvalidateShapeControlRectangle to invalidate
|	   other types of shapes.
|
|	 UpdateShapeControl redraws all the invalid areas that have been recorded by calls to InvalidateShapeControlShape
|	   and InvalidateShapeControlRectangle.
|
|	 The shape controls library will try to allocate an offscreen buffer for each of the viewPorts that the items
|	   picture draws into. If this buffer is allocated, then the manipulation of the sub-objects will be flicker-free.
|	   If, on the other hand, there is not enough memory to allocate this buffer, then the shape controls library
|	   will still work, but the objects will flicker quite a bit as they are manipulated.
*/

	typedef struct shapeControlRecord **shapeControl;

	shapeControl NewShapeControl(shape items, shape background, shape foreground);
	void DisposeShapeControl(shapeControl target);

	shape GetShapeControlSelection(const shapeControl source);
	void SetShapeControlsSelection(shapeControl target, shape shapesToSelect, boolean bringToFront, boolean replaceSelection);
	shape GetShapeControlSelectionHandles(const shapeControl source);

	boolean SendEventToShapeControl(shapeControl target, EventRecord *event);

	void InvalidateShapeControlShape(shapeControl target, shape invalidShape);
	void InvalidateShapeControlRectangle(shapeControl target, rectangle *bounds);

	void UpdateShapeControl(shapeControl target);

#endif	/* shapeControlsLibraryIncludes */
