/* ------------------------------------------------------------------------------

	FILENAME
		PaperTypeLibrary.c
		
	DESCRIPTION
		Handy paperType functions.
		
	COPYRIGHT
		  Apple Computer, Inc. 1990-1991
		 All rights reserved. 
	
-------------------------------------------------------------------------------- */


#include <PrintingManager.h>


/****************************************************************************************

										GetPaperTypeBaseType

	function :
				GetPaperTypeBaseType returns the specified paper type's base paper type value.

	parameters :
				thePaperType			(in)	reference to target paper type
				theBasePaperType		(out)	paper type's base value
				
	returns :
				error code

****************************************************************************************/
long GetPaperTypeBaseType(
	PaperType	thePaperType)
{
	BaseInfo info;
	
	GetCollectionItem(GetPaperTypeCollection(thePaperType), baseTag, printingTagID, nil, &info);
	return info.baseType;
}
/* GetPaperTypeBaseType */


//<FF>
/****************************************************************************************

										SetPaperTypeBaseType

	function :
				SetPaperTypeBaseType saves the specified base paper type in the target paper type.

	parameters :
				thePaperType		(in)	reference to target paper type
				theBasePaperType	(in)	new base paper type value
				
	returns :
				error code

****************************************************************************************/
void SetPaperTypeBaseType(
	PaperType	thePaperType,
	long		theBasePaperType)
{
	BaseInfo info;
	
	info.baseType = theBasePaperType;
	AddCollectionItem(GetPaperTypeCollection(thePaperType), baseTag, printingTagID, sizeof(BaseInfo), &info);
}
/* SetPaperTypeBaseType */


//<FF>
/****************************************************************************************

									GetPaperTypeCreator

	function :
				GetPaperTypeCreator returns the specified paper type's creator.

	parameters :
				thePaperType	(in)	reference to target paper type
				theCreator		(out)	paper type's creator value
				
	returns :
				error code

****************************************************************************************/
OSType GetPaperTypeCreator(
	PaperType	thePaperType)
{
	CreatorInfo info;
	
	GetCollectionItem(GetPaperTypeCollection(thePaperType), creatorTag, printingTagID, nil, &info);
	return info.creator;
}
/* GetPaperTypeCreator */


//<FF>
/****************************************************************************************

									SetPaperTypeCreator

	function :
				SetPaperTypeCreator saves the specified creator in the target paper type.

	parameters :
				thePaperType	(in)	reference to target paper type
				theCreator		(in)	new paper type creator value
				
	returns :
				error code

****************************************************************************************/
void SetPaperTypeCreator(
	PaperType	thePaperType,
	OSType		theCreator)
{
	CreatorInfo info;
	
	info.creator = theCreator;
	AddCollectionItem(GetPaperTypeCollection(thePaperType), creatorTag, printingTagID, sizeof(CreatorInfo), &info);
}
/* SetPaperTypeCreator */


//<FF>
/****************************************************************************************

									GetPaperTypeUnits

	function :
				GetPaperTypeUnits returns the specified paper type's units.

	parameters :
				thePaperType	(in)	reference to target paper type
				theUnits		(out)	paper type's units
				
	returns :
				error code

****************************************************************************************/
unsigned char GetPaperTypeUnits(
	PaperType		thePaperType)
{
	UnitsInfo info;
	
	GetCollectionItem(GetPaperTypeCollection(thePaperType), unitsTag, printingTagID, nil, &info);
	return info.units;
}
/* GetPaperTypeUnits */


//<FF>
/****************************************************************************************

									SetPaperTypeUnits

	function :
				SetPaperTypeUnits saves the specified units in the target paper type.

				
	parameters :
				thePaperType	(in)	reference to target paper type
				theUnits		(in)	new paper type's units setting
				
	returns :
				error code

****************************************************************************************/
void SetPaperTypeUnits(
	PaperType		thePaperType,
	unsigned char	theUnits)
{
	UnitsInfo info;
	
	info.units = theUnits;
	AddCollectionItem(GetPaperTypeCollection(thePaperType), unitsTag, printingTagID, sizeof(UnitsInfo), &info);
}
/* SetPaperTypeUnits */


//<FF>
/****************************************************************************************

										GetPaperTypeFlags

	function :
				GetPaperTypeFlags returns the specified paper type's flags.

	parameters :
				thePaperType	(in)	reference to target paper type
				theFlags		(out)	paper type's flags
				
	returns :
				error code

****************************************************************************************/
unsigned long GetPaperTypeFlags(
	PaperType			thePaperType)
{
	FlagsInfo info;
	
	GetCollectionItem(GetPaperTypeCollection(thePaperType), flagsTag, printingTagID, nil, &info);
	return info.flags & 0x00FFFFFF;
}
/* GetPaperTypeFlags */


//<FF>
/****************************************************************************************

									SetPaperTypeFlags

	function :
				SetPaperTypeFlags saves the specified flags in the target paper type.

	parameters :
				thePaperType	(in)	reference to target paper type
				theFlags		(in)	new paper type's flags setting
				
	returns :
				error code

****************************************************************************************/
void SetPaperTypeFlags(
	PaperType		thePaperType,
	unsigned long	theFlags)
{
	FlagsInfo info;
	
	info.flags = theFlags & 0x00FFFFFF;
	AddCollectionItem(GetPaperTypeCollection(thePaperType), flagsTag, printingTagID, sizeof(FlagsInfo), &info);
}
/* SetPaperTypeFlags */


//<FF>
/****************************************************************************************

									GetPaperTypeComment

	function :
				GetPaperTypeComment returns the specified paper type's comment.

	parameters :
				thePaperType	(in)	reference to target paper type
				theComment		(out)	the comment returned (Pascal string)
				
	returns :
				error code

****************************************************************************************/
void GetPaperTypeComment(
	PaperType	thePaperType,
	Str255		theComment)
{
	long itemSize = sizeof(Str255);
	OSErr anErr = GetCollectionItem(GetPaperTypeCollection(thePaperType), commentTag, printingTagID,
											  &itemSize, theComment);
	if (anErr)
		theComment[0] = '\0';
}
/* GetPaperTypeComment */


//<FF>
/****************************************************************************************

										SetPaperTypeComment

	function :
				SetPaperTypeComment saves the specified comment in the target paper type.

	parameters :
				thePaperType	(in)	reference to target paper type
				theComment		(in)	the comment to save (Pascal string)
				
	returns :
				error code

****************************************************************************************/
OSErr SetPaperTypeComment (
	PaperType	thePaperType,
	Str255		theComment)
{
	return AddCollectionItem(GetPaperTypeCollection(thePaperType), commentTag, printingTagID,
									 theComment[0] + 1, theComment);
}
/* SetPaperTypeComment */


//<FF>
/****************************************************************************************

									LockFormatPaperType

	function :
				LockFormatPaperType sets the lock state of the paper type associated with the
				target format.  To lock the format's paper type, pass true for lockPaperType;
				otherwise pass false.  A special collection item is attached to the format's
				collection which contains a single Boolean indicating the current lock state
				(true if locked; false if not).

	parameters :
				theJob				(in)	reference to the target job
				theFormat			(in)	reference to the target format
				lockPaperType		(in)	true => lock the paper type; false => unlock it
				
	returns :
				nothing

****************************************************************************************/
void LockFormatPaperType(
	Format		theFormat,
	Boolean		lockPaperType)
{
	OSErr			anErr;
	Collection	formatCollection;
	
	formatCollection = GetFormatCollection(theFormat);

	anErr = AddCollectionItem(	formatCollection, 
										paperTypeLockTag, 
										printingTagID,
										sizeof(Boolean), 
										&lockPaperType);

	if (anErr != noErr)
		SetJobError(GetFormatJob(theFormat), anErr);
}
/* LockFormatPaperType */

