/* graphics libraries:
	QuickDraw region routines
	by Cary Clark, Georgiann Delaney, Michael Fairman, Dave Good, Robert Johnson, Keith McGreggor, Oliver Steele, David Van Brink, Chris Yerga
	Copyright 1987 - 1991 Apple Computer, Inc.  All rights reserved.	*/

#include "qd globals.h"
#include "qd emulation.h"
#include "qd utilities.h"
#include "qd library.h"
#include "graphics libraries.h"


pascal void StdRgn(GrafVerb verb, RgnHandle regionShape)
{
	NilShapeReturn(regionShape);
	DrawQDShape(regionShape, verb, true, false);
}


pascal void FrameRgn(RgnHandle regionShape)
{
	(*stdRgnProc)(frame, regionShape);
}


pascal void FastFrameRgn(RgnHandle regionShape)
{
	NilShapeReturn(regionShape);

	SetShapeFill(regionShape, closedFrameFill);
	FastPaintRgn(regionShape);
	SetShapeFill(regionShape, evenOddFill);
}



pascal void PaintRgn(RgnHandle regionShape)
{
	(*stdRgnProc)(paint, regionShape);
}


pascal void FastPaintRgn(RgnHandle regionShape)
{
	NilShapeReturn(regionShape);

	SetShapeStyle(regionShape, GetShapeStyle(GetDefaultShape(rectangleType)));
	DrawShape(regionShape);
}


pascal void EraseRgn(RgnHandle regionShape)
{
	(*stdRgnProc)(erase, regionShape);
}


pascal void FastEraseRgn(RgnHandle regionShape)
{
	NilShapeReturn(regionShape);
	SetShapeStyle(regionShape, qdEraseStyle);
	DrawShape(regionShape);
}


pascal void InvertRgn(RgnHandle regionShape)
{
	(*stdRgnProc)(invert, regionShape);
}


pascal void FastInvertRgn(RgnHandle regionShape)
{
	SetShapeInk(regionShape, qdInvertInk);
	DrawShape(regionShape);
}


pascal void FillRgn(RgnHandle regionShape, const Pattern pat)
{
	SaveQDFillStyleInk
	SetStyleQDPat(qdFillStyle, qdFillInk, pat);
	(*stdRgnProc)(fill, regionShape);
	RestoreQDFillStyleInk
}


pascal RgnHandle NewRgn(void)
{
	shape regionShape = NewShape(emptyType);
	SetShapeFill(regionShape, noFill);
	return regionShape;
}


pascal void OpenRgn(void)
{
	IfErrorReturn(openRgn, region_already_open);
	HidePen();
	openRgn = NewShape(emptyType);
	SetShapeFill(openRgn, noFill);
}


pascal void CloseRgn(RgnHandle rgn)
{
	NilShapeReturn(rgn);
	IfErrorReturn(openRgn == nil, no_open_region);
	SetShapeFill(openRgn,evenOddFill);
	ShowPen();
	CopyToShape(rgn, openRgn);
	DisposeShape(openRgn);
	openRgn = nil;
}


pascal void DisposeRgn(RgnHandle rgn)
{
	NilShapeReturn(rgn);
	DisposeShape(rgn);
}


pascal void CopyRgn(RgnHandle srcRgn, RgnHandle dstRgn)
{
	NilShapeReturn(srcRgn);
	NilShapeReturn(dstRgn);
	CopyToShape(dstRgn, srcRgn);
}


pascal void SetEmptyRgn(RgnHandle rgn)
{
	NilShapeReturn(rgn);
	if (GetShapeType(rgn) != emptyType)
		SetShapeType(rgn, emptyType);
}


pascal void SetFullRgn(RgnHandle rgn)
{
	NilShapeReturn(rgn);
	if (GetShapeType(rgn) != fullType)
		SetShapeType(rgn, fullType);
}


pascal void SetRectRgn(RgnHandle rgn, short left, short top, short right, short bottom)
{
	NilShapeReturn(rgn);
	if (left == -32768 && top == -32768 && right == 32767 && bottom == 32767)
		SetFullRgn(rgn);
	else if (left == right && top == bottom)
		SetEmptyRgn(rgn);
	else {
		SetShapeFill(rgn,evenOddFill);
		if (GetShapeType(rgn) != rectangleType) {
			IgnoreGraphicsWarning(new_shape_contains_invalid_data);
			SetShapeType(rgn, rectangleType);
			PopGraphicsWarning();
		}
		SetShape4(rgn, IntToFixed(left), IntToFixed(top), IntToFixed(right), IntToFixed(bottom));
	}
}


pascal void RectRgn(RgnHandle rgn, const Rect *rect)
{
	NilShapeReturn(rgn);
	NilParamReturn(rect);
	SetRectRgn(rgn, rect->left, rect->top, rect->right, rect->bottom);
}


pascal void OffsetRgn(RgnHandle rgn, short dh, short dv)
{
	NilShapeReturn(rgn);
	IgnoreGraphicsNotice(transform_mapping_unaffected);
	MoveShape(rgn, IntToFixed(dh), IntToFixed(dv));
	PopGraphicsNotice();
}


pascal void InsetRgn(RgnHandle rgn, short dh, short dv)
{
	fixed averageInset = (IntToFixed(dh + dv)) >> 1;

	NilShapeReturn(rgn);
	IgnoreGraphicsNotice(transform_mapping_unaffected);
	InsetShape(rgn, averageInset);
	PopGraphicsNotice();
}


pascal void SectRgn(RgnHandle a, RgnHandle b, RgnHandle dst)
{
	NilShapeReturn(a);
	NilShapeReturn(b);
	if (a == dst)
		a = b;
	else if (b != dst)
		CopyToShape(dst, b);
	IntersectShape(dst, a);
}


pascal void UnionRgn(RgnHandle a, RgnHandle b, RgnHandle dst)
{
	NilShapeReturn(a);
	NilShapeReturn(b);
	if (a == dst)
		a = b;
	else if (b != dst)
		CopyToShape(dst, b);
	UnionShape(dst, a);
}


pascal void DiffRgn(RgnHandle a, RgnHandle b, RgnHandle dst)
{
	NilShapeReturn(a);
	NilShapeReturn(b);
	if (a == dst)
		a = b;
	else if (b != dst)
		CopyToShape(dst, b);
	DifferenceShape(dst, a);
}


pascal void XorRgn(RgnHandle a, RgnHandle b, RgnHandle dst)
{
	NilShapeReturn(a);
	NilShapeReturn(b);
	if (a == dst)
		a = b;
	else if (b != dst)
		CopyToShape(dst, b);
	ExcludeShape(dst, a);
}


/* could hit test */
pascal boolean PtInRgn(Point pt, RgnHandle rgn)
{
	shape ptShape;
	boolean result;

	NilShapeReturnNil(rgn);
	ptShape = NewShape2(pointType, IntToFixed(pt.h), IntToFixed(pt.v));
	IntersectShape(ptShape, rgn);
	result = GetShapeType(ptShape) == pointType;
	DisposeShape(ptShape);
	return result;
}


pascal boolean RectInRgn(const Rect *rect, RgnHandle rgn)
{
	shape rectShape;
	boolean result;

	NilParamReturnNil(rect);
	NilShapeReturnNil(rgn);
	rectShape = NewShape4(rectangleType, IntToFixed(rect->left), IntToFixed(rect->top), IntToFixed(rect->right), IntToFixed(rect->bottom));
	IntersectShape(rectShape, rgn);
	result = GetShapeType(rectShape) != emptyType;
	DisposeShape(rectShape);
	return result;
}


pascal boolean EqualRgn(RgnHandle rgnA, RgnHandle rgnB)
{
	NilShapeReturnNil(rgnA);
	NilShapeReturnNil(rgnB);
	return EqualShape(rgnA, rgnB);
}


pascal boolean EmptyRgn(RgnHandle rgn)
{
	NilShapeReturnNil(rgn);
	return GetShapeType(rgn) == emptyType;
}

pascal void GetRgnBounds(RgnHandle rgn, Rect *bounds)
{
	shapeType rgnType;

	NilShapeReturn(rgn);
	NilParamReturn(bounds);
	rgnType = GetShapeType(rgn);
	if (rgnType == emptyType)
		SetRect(bounds, 0, 0, 0, 0);
	else if (rgnType == fullType)
		SetRect(bounds, -32768, -32768, 32767, 32767);
	else {
		rectangle fixedBounds;

		GetShapeBounds(rgn, 0, &fixedBounds);
		FixedRectToShort(&fixedBounds, bounds);
	}
}
